/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata.security;

import com.trifork.stamdata.SDMException;
import com.trifork.stamdata.SDMFaultMessage;

import org.apache.log4j.Logger;

import javax.inject.Inject;
import javax.inject.Named;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.stream.Collectors;

public class AccessControl {

    private static final Logger logger = Logger.getLogger(AccessControl.class);

    public static final String SECURITY_ALLOWED_ACTORS = "dk.nsi.stamdata.security.allowed.actors";

    private List<UserType> allowedActors;

    private WhitelistService whitelistService;

    @Inject
    public AccessControl(WhitelistServiceProvider whitelistServiceProvider, @Named(SECURITY_ALLOWED_ACTORS) String allowedActors) {
        this.allowedActors = splitActorList(allowedActors);
        this.whitelistService = whitelistServiceProvider.get();
    }

    public void checkAccess(RequestContextVO requestContext, String whitelistServiceName) throws SDMException {
        checkAccess(requestContext, true, whitelistServiceName);
    }

    public void checkAccess(RequestContextVO requestContext, boolean useWhitelisting, String whitelistServiceName) throws SDMException {
        validUserTypeForService(requestContext);
        if (UserType.Citizen.equals(requestContext.getServiceActor().getUserType())) {
        	if (requestContext.getPersonIdentifier() != null) {
        		citizenCallIsOnOwnCpr(requestContext);	
        	}
        } else {
            if(useWhitelisting) {
                checkCVRWhitelisted(requestContext.getServiceActor().getOrganizationId(), whitelistServiceName);
            }
        }
    }
    
    public String getCprForFilteringBasedOnAccess(RequestContextVO requestContext) throws SDMException {
    	if (UserType.Citizen.equals(requestContext.getServiceActor().getUserType())) {
    		if (requestContext.getServiceActor().getPersonIdentifier() != null) {
    			return requestContext.getServiceActor().getPersonIdentifier();
    		} else {
    			throw new SDMException(SDMFaultMessage.AUTHORIZATION_ERROR_CPR_ACCESS_DENIED);
    		}
    	}
    	return null;
    }

    private void checkCVRWhitelisted(String cvr, String serviceName) throws SDMException {
        if (!whitelistService.isCvrWhitelisted(cvr, serviceName)) {
            logger.error("Whitelist check failed. Servicename " + serviceName + " does not have access to CVR " + cvr);
            throw new SDMException("Dette CVR-nummer (" + cvr + ") er ikke whitelistet til at hente de ønskede oplysniger");
        }
    }

    private void validUserTypeForService(RequestContextVO requestContext) throws SDMException {
        if (!allowedActors.contains(requestContext.getServiceActor().getUserType())) {
            logger.error("Invalid user type: " + requestContext.getServiceActor().getUserType());
            throw new SDMException(SDMFaultMessage.AUTHORIZATION_ERROR_CPR_ACCESS_DENIED);
        }
    }

    private void citizenCallIsOnOwnCpr(RequestContextVO requestContext) throws SDMException {
        if (!requestContext.getPersonIdentifier().equals(requestContext.getServiceActor().getPersonIdentifier())) {
            logger.error("Citizen calls are only allowed on own cpr.");
            throw new SDMException(SDMFaultMessage.AUTHORIZATION_ERROR_CPR_ACCESS_DENIED);
        }
    }

    private List<UserType> splitActorList(String actors) {
        List<UserType> actorList = new ArrayList<>();

        List<String> trimmedActors = Arrays.stream(actors.split(","))
                .map(a -> a.trim())
                .filter(a -> !a.isEmpty())
                .collect(Collectors.toList());
        for (String actor : trimmedActors) {
            actorList.add(UserType.valueOf(actor));
        }

        return actorList;
    }
}
