/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata.security;

import com.trifork.stamdata.SDMException;
import com.trifork.stamdata.SDMFaultMessage;
import dk.sds.nsp.security.SecurityContext;
import dk.sds.nsp.security.SecurityContext.Organisation.OrganisationIdentifierFormat;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

public class ServiceActorProvider {

	private static final Logger logger = LogManager.getLogger(ServiceActorProvider.class);

	private String allowedAudience;


	public ServiceActorProvider(String allowedAudience) {
		this.allowedAudience = allowedAudience;
	}
	

	public ServiceActor getServiceActor(SecurityContext securityContext) throws SDMException {
		validateTicket(securityContext);
		return mapSecurityContextToValidatedServiceActor(securityContext);
	}

	private ServiceActor mapSecurityContextToValidatedServiceActor(SecurityContext securityContext) throws SDMException {


		// Gennemløb de understøttede aktørtyper for denne service og valider og map
		
		// Tjekke for sundhedsfagligbruger
		if (securityContext.getActingUser().filter(u -> u.getUserType().equals(SecurityContext.User.UserType.HealthcareProfessional)).isPresent()) {
			// Ingen delegering
			if (securityContext.getPrincipalUser().isPresent()) {
				throw new SDMException(SDMFaultMessage.AUTHORIZATION_ERROR_INGEN_DELEGERING);
			}

			ServiceActor result = mapToServiceActor(securityContext, UserType.HealthCareProfessional, true);
			return result;
		}
		// Borgerkald
		if (securityContext.getActingUser().filter(u -> u.getUserType().equals(SecurityContext.User.UserType.Citizen)).isPresent()) {

			// Ingen fuldmagt for borgere
			if (securityContext.getPrincipalUser().isPresent()) {
				throw new SDMException(SDMFaultMessage.AUTHORIZATION_ERROR_INGEN_DELEGERING);
			}
			// Der skal være audience tilstede
			if (!securityContext.getTicket().get().getAudience().isPresent()) {
				throw new SDMException(SDMFaultMessage.AUTHORIZATION_ERROR_INVALID_HEADER);
			}

	        // Validering af audience
	        if (!securityContext.getTicket().get().getAudience().get().equals(allowedAudience)) {
				throw new SDMException(SDMFaultMessage.AUTHORIZATION_ERROR_WRONG_AUDIENCE);
	        }
			
			ServiceActor result = mapToServiceActor(securityContext, UserType.Citizen, false);
			return result;
		}
		// Systembruger
		if (!securityContext.getActingUser().isPresent() && securityContext.getOrganisation().isPresent()) {

			ServiceActor result = mapToServiceActor(securityContext, UserType.SystemUser, true);

			return result;
		}

		throw new SDMException(SDMFaultMessage.AUTHORIZATION_ERROR_INGEN_DELEGERING);
	}


	private ServiceActor mapToServiceActor(SecurityContext securityContext, UserType userType, boolean requireCvr) throws SDMException {
		ServiceActor result = new ServiceActor(userType);

		if (securityContext.getOrganisation().isPresent()) {
			mapDataFromOrganisation(result, securityContext.getOrganisation().get());
		} else if (requireCvr) {
			throw new SDMException(SDMFaultMessage.AUTHORIZATION_ERROR_INGEN_CVR);
		}
		if (securityContext.getActingUser().isPresent()) {
			mapAndValidateDataFromUser(result, securityContext.getActingUser().get());
		}
		
		return result;
	}


	private void mapDataFromOrganisation(ServiceActor serviceActor, SecurityContext.Organisation organisation) throws SDMException {
		if (organisation.getIdentifierFormat().equals(OrganisationIdentifierFormat.CVR)) {
			serviceActor.setOrganizationId(organisation.getIdentifier());			
			serviceActor.setOrganizationIdSource(OrganisationIdentifierFormat.CVR);
			serviceActor.setOrganizationName(organisation.getName().orElse(null));			
		} else {
			throw new SDMException(SDMFaultMessage.AUTHORIZATION_ERROR_INGEN_CVR);
		}
	}
/*
	private void mapDataFromClient(ServiceActor ServiceActor, SecurityContext.Client client) {
		ServiceActor.setItSystemName(client.getName().orElse(null));
	}
*/
	private void mapAndValidateDataFromUser(ServiceActor aerviceActor, SecurityContext.User user) throws SDMException {
		if (user.getIdentifierFormat().equals(SecurityContext.User.PersonIdentifierFormat.CPR)) {
			aerviceActor.setPersonIdentifier(user.getIdentifier());
		} else {
			throw new SDMException(SDMFaultMessage.AUTHORIZATION_ERROR_USER_NOT_CPR);
		}
		aerviceActor.setFirstName(user.getGivenName().orElse(null));
		aerviceActor.setLastName(user.getSurname().orElse(null));
		if (user.getCredentials().isPresent()) {
			SecurityContext.User.Credentials credentials = user.getCredentials().get();
			aerviceActor.setAuthorisationNumber(credentials.getAuthorizationCode().orElse(null));
			aerviceActor.setEducationCode(credentials.getEducationCode().orElse(null));
		}
	}

	private void validateTicket(SecurityContext securityContext) throws SDMException {
		if (securityContext == null || !securityContext.getTicket().isPresent() || !securityContext.getTicket().get().isValid()) {
			throw new SDMException(SDMFaultMessage.AUTHORIZATION_ERROR_INVALID_HEADER);
		}
	}
}
