/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata.testing;
import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.Stream;

public class DatabaseTestInitializer {
    private static boolean initialized = false;
    private final Connection connection;

    public DatabaseTestInitializer(Connection connection) {
        this.connection = connection;
    }

    public Connection getConnection() {
        return connection;
    }

    public void initialize() {
        if(initialized) {
            return;
        }
        String[] scriptPaths = getScripts();
        List<String> sqlScripts = new ArrayList<>();

        for (String path : scriptPaths) {
            try {
                StringBuilder sql = new StringBuilder();

                List<String> lines = Files.readAllLines(new File(path).toPath());
                for (String line : Files.readAllLines(new File(path).toPath())) {
                    // The H2 database was added after the SQL scripts were written, and we would like not to change the
                    // scripts. Preprocess the SQL scripts to ignore collate-directives (not supported by H2), and change the
                    // type of one column, to avoid a casting error.

                    line = line
                            .replace("USE sdm_warehouse;", "")
                            .replace("COLLATE=utf8_bin", "")
                            .replace("CPR_AUT_ID", "")
                            .replace("SET @OLD_UNIQUE_CHECKS=@@UNIQUE_CHECKS, UNIQUE_CHECKS=0;", "")
                            .replace("SET @OLD_FOREIGN_KEY_CHECKS=@@FOREIGN_KEY_CHECKS, FOREIGN_KEY_CHECKS=0;", "")
                            .replace("SET @OLD_SQL_MODE=@@SQL_MODE, SQL_MODE='TRADITIONAL,ALLOW_INVALID_DATES';", "")
                            .replace("SET SQL_MODE=@OLD_SQL_MODE;", "")
                            .replace("SET FOREIGN_KEY_CHECKS=@OLD_FOREIGN_KEY_CHECKS;", "")
                            .replace("SET UNIQUE_CHECKS=@OLD_UNIQUE_CHECKS;", "")
                            .replace("COLLATE=utf8_danish_ci", "")
                            .replace("COLLATE utf8_danish_ci", "")
                            .replace("AUTO_INCREMENT=22", "")
                            .replace("`isPID` TINYINT", "`isPID` INT")
                            .replace("\"", "'");

                    sql.append(line + "\n");
                    if(line.contains(";")) {
                        sqlScripts.add(sql.toString());
                        sql.setLength(0);
                    }
                }
                sqlScripts.add(sql.toString());
            } catch (IOException e) {
                e.printStackTrace();
            }

        }
        for (String sqlScript: sqlScripts) {
            try {
                connection.createStatement().executeUpdate(sqlScript);
                connection.commit();
            } catch (SQLException e) {
                throw new RuntimeException(e);
            }
        }
        initialized = true;
    }


    private String[] getScripts() {
        List<File> directories = new ArrayList<>();
        File dir = new File("../../compose/database");
        if(!dir.isDirectory()) {
            dir = new File("../compose/database");
        }
        directories.add(dir);

        List<String> scripts = new ArrayList<>();
        for(File directory : directories) {
            try (Stream<Path> fs = Files.list(directory.toPath())) {
                scripts.addAll(fs.filter(Files::isRegularFile)
                        .filter(f -> (!f.endsWith("01-dropAndCreateDB.sql") && !f.toString().endsWith(".swp")))
                        .map(f -> f.toString())
                        .sorted()
                        .collect(Collectors.toList()));
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        }
        return scripts.toArray(new String[scripts.size()]);
    }
}
