/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata.testing;

import dk.sds.nsp.security.SecurityContext;
import org.w3c.dom.Element;

import java.time.ZonedDateTime;
import java.time.temporal.ChronoUnit;
import java.util.HashSet;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.atomic.AtomicReference;

public class SDMSecurityContextHelperCitizen extends SDMTestSecurityContext implements SecurityContext {

    private static AtomicReference<Ticket> globalTicket = new AtomicReference<>(new DevelopmentTicket());

    private static AtomicReference<User> globalActingUser = new AtomicReference<>(new DevelopmentUser());

    private static AtomicReference<User> globalPrincipalUser = new AtomicReference<>();

    private static AtomicReference<Organisation> globalOrganisation = new AtomicReference(new DevelopmentOrganisation());

    private static AtomicReference<Client> globalClient = new AtomicReference<>(new DevelopmentClient());

    private static AtomicReference<Message> globalMessage = new AtomicReference<>(new DevelopmentMessage());

    private static ThreadLocal<Ticket> threadTicket = new ThreadLocal<>();

    private static ThreadLocal<User> threadActingUser = new ThreadLocal<>();

    private static ThreadLocal<User> threadPrincipalUser = new ThreadLocal<>();

    private static ThreadLocal<Organisation> threadOrganisation = new ThreadLocal<>();

    private static ThreadLocal<Client> threadClient = new ThreadLocal<>();

    private static ThreadLocal<Message> threadMessage = new ThreadLocal<>();

    private static final class DevelopmentTicket implements Ticket {

        private final ZonedDateTime now;

        private DevelopmentTicket() {
            this.now = ZonedDateTime.now();
        }

        @Override
        public boolean isValid() {
            return true;
        }

        @Override
        public Federation getFederation() {
            return Federation.Development;
        }

        @Override
        public Optional<String> getAudience() {
            return Optional.of("https://fsk");
        }

        @Override
        public Optional<ZonedDateTime> getCreated() {
            return Optional.of(now.minus(1, ChronoUnit.MINUTES));
        }

        @Override
        public Optional<ZonedDateTime> getValidFrom() {
            return Optional.of(now.minus(1, ChronoUnit.MINUTES));
        }

        @Override
        public Optional<ZonedDateTime> getValidTo() {
            return Optional.of(now.plus(1, ChronoUnit.MINUTES));
        }

		@Override
		public Optional<String> createRequestWithContext(Element arg0, Element... arg1) {
            return Optional.empty();
		}
    }

    private static final class DevelopmentOrganisation implements Organisation {
        private DevelopmentOrganisation() {
        }

        public OrganisationIdentifierFormat getIdentifierFormat() {
            return OrganisationIdentifierFormat.CVR;
        }

        public String getIdentifier() {
            return "33257872";
        }

        public Optional<String> getName() {
            return Optional.of("Statens Serum Institut");
        }
    }

    private static final class DevelopmentUser implements User {

        private static final class DevelopmentCredentials implements Credentials {

            @Override
            public Optional<String> getAuthorizationCode() {
                return Optional.of("AABB123");
            }

            @Override
            public Optional<String> getEducationCode() {
                return Optional.of("7170");
            }

            @Override
            public Optional<String> getNationalRole() {
                return null;
            }

            @Override
            public Optional<String> getUnverifiedRole() {
                return null;
            }

            @Override
            public Set<String> getPowerOfAttorneyPrivileges() {
                Set<String> privs = new HashSet<String>();
                privs.add("test");
                return privs;
            }

        }


        @Override
        public UserType getUserType() {
            return UserType.Citizen;
        }

        @Override
        public PersonIdentifierFormat getIdentifierFormat() {
            return PersonIdentifierFormat.CPR;
        }

        @Override
        public String getIdentifier() { return "2222222222"; }

        @Override
        public Optional<String> getGivenName() {
            return Optional.of("Hans Helmuth");
        }

        @Override
        public Optional<String> getSurname() {
            return Optional.of("Hansen");
        }

        @Override
        public Optional<Credentials> getCredentials() {
            return Optional.of(new DevelopmentCredentials());
        }

		@Override
		public Optional<PersistentUniqueKey> getPersistentUniqueKey() {
            return Optional.empty();
		}
    }

    private static final class DevelopmentClient implements Client {

        @Override
        public Optional<String> getName() {
            return Optional.of("Development Systems ltd.");
        }

		@Override
		public Optional<PersistentUniqueKey> getPersistentUniqueKey() {
            return Optional.empty();
		}

    }

    private static final class DevelopmentMessage implements Message {

        @Override
        public Optional<String> getIdentifier() { return Optional.of("1234"); }

        @Override
        public Optional<String> getConversationIdentifier() {
            return Optional.of("5678");
        }

        @Override
        public Optional<String> getAction() {
            return Optional.empty();
        }
    }

    @Override
    public SecurityContext getSecurityContext() {
        System.err.println("WARNING: This NSP Security Provider must not be used in test or production environments. It is only for development");
        return this;
    }

    @Override
    public Optional<Ticket> getTicket() {
        Ticket ticket = threadTicket.get();
        if (ticket == null) {
            ticket = globalTicket.get();
        }
        return Optional.ofNullable(ticket);
    }

    public void setGlobalTicket(Ticket ticket) {
        globalTicket.set(ticket);
    }

    public void setThreadTicket(Ticket ticket) {
        threadTicket.set(ticket);
    }


    @Override
    public Optional<Message> getMessage() {
        Message message = threadMessage.get();
        if (message == null) {
            message = globalMessage.get();
        }
        return Optional.ofNullable(message);
    }

    public void setGlobalMessage(Message message) {
        globalMessage.set(message);
    }

    public void setThreadMessage(Message message) { threadMessage.set(message); }


    @Override
    public Optional<User> getActingUser() {
        User actingUser = threadActingUser.get();
        if (actingUser == null) {
            actingUser = globalActingUser.get();
        }
        return Optional.ofNullable(actingUser);
    }

    public void setGlobalActingUser(User user) {
        globalActingUser.set(user);
    }

    public void setThreadActingUser() {
        threadActingUser.remove();
    }

    @Override
    public Optional<User> getPrincipalUser() {
        return Optional.empty();
    }

    public void setGlobalPrincipalUser(User user) {
        globalPrincipalUser.set(user);
    }

    public void setThreadPrincipalUser(User user) {
        threadPrincipalUser.set(user);
    }


    public Optional<Organisation> getOrganisation() {
        Organisation organisation = (Organisation)threadOrganisation.get();
        if (organisation == null) {
            organisation = (Organisation)globalOrganisation.get();
        }

        return Optional.of(organisation);
    }

    public void setGlobalOrganisation(Organisation organisation) {
        globalOrganisation.set(organisation);
    }

    public void setThreadOrganisation(Organisation organisation) {
        threadOrganisation.set(organisation);
    }


    @Override
    public Optional<Client> getClient() {
        Client client = threadClient.get();
        if (client == null) {
            client = globalClient.get();
        }
        return Optional.ofNullable(client);
    }

    public void setGlobalClient(Client client) {
        globalClient.set(client);
    }

    public void setThreadClient(Client client) {
        threadClient.set(client);
    }

}