/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata.testing;

import com.trifork.stamdata.sla.SDMSLALoggingFilter;
import io.undertow.Undertow;
import io.undertow.server.HttpHandler;
import io.undertow.server.handlers.resource.FileResourceManager;
import io.undertow.servlet.Servlets;
import io.undertow.servlet.api.DeploymentInfo;
import io.undertow.servlet.api.DeploymentManager;
import io.undertow.servlet.api.FilterInfo;

import javax.servlet.DispatcherType;
import javax.servlet.ServletContextListener;
import javax.servlet.ServletException;
import java.io.Closeable;
import java.io.File;
import java.io.IOException;
import java.util.function.Consumer;

public abstract class UndertowTestServer implements Closeable  {
    private Undertow server;

    protected static final int DEFAULT_PORT = 8972;

    public UndertowTestServer(int port, String dispatcherServletName, Consumer<DeploymentInfo> dispatcherSetupConsumer, Class<? extends ServletContextListener> guiceContextListenerClass, UndertowSlaLogParameters slaLogParameters) throws ServletException {
        DeploymentInfo deploymentInfo = Servlets.deployment();

        deploymentInfo.setDeploymentName(getClass().getSimpleName());
        deploymentInfo.setClassLoader(getClass().getClassLoader());
        deploymentInfo.setContextPath("/");

        // Dispatcher servlet
        dispatcherSetupConsumer.accept(deploymentInfo);

        // Guice listener and filter
        deploymentInfo.addListener(Servlets.listener(guiceContextListenerClass));
        deploymentInfo.addFilter(new FilterInfo("Guice Filter", com.google.inject.servlet.GuiceFilter.class));
        deploymentInfo.addFilterUrlMapping("Guice Filter", "/*", DispatcherType.REQUEST);

        // SLA-logging filter
        deploymentInfo.addFilter(
                new FilterInfo("SDMSlaLogFilter", SDMSLALoggingFilter.class)
                        .addInitParam("appName", slaLogParameters.getAppName())
                        .addInitParam("shortAppName", slaLogParameters.getShortAppName()));
        deploymentInfo.addFilterServletNameMapping("SDMSlaLogFilter", dispatcherServletName, DispatcherType.REQUEST);

        // Ensure that we can get the resources we need
        deploymentInfo.setResourceManager(new FileResourceManager(new File("src/main/webapp")));

        DeploymentManager manager = Servlets.defaultContainer().addDeployment(deploymentInfo);
        manager.deploy();

        HttpHandler handler = manager.start();

        this.server = Undertow
                .builder()
                .addHttpListener(port, "localhost")
                .setHandler(handler)
                .build();
    }

    public void start() {
        server.start();
    }

    @Override
    public void close() throws IOException {
        server.stop();
        server = null;
    }
}
