/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

package com.trifork.stamdata.authorization.models;


import com.google.inject.Inject;
import com.google.inject.Provider;
import com.trifork.stamdata.models.AuthorizationVO;
import com.trifork.stamdata.persistence.PersistenceModule;
import org.apache.log4j.Logger;

import java.io.IOException;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.time.LocalDate;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import static com.trifork.stamdata.Preconditions.checkNotNull;

/**
 * Data access object for authorizations.
 *
 * This class is thread safe.
 *
 * @author Thomas Børlum (thb@trifork.com)
 */
public class Authorization20240105Dao {

    Logger logger = Logger.getLogger(Authorization20240105Dao.class);

    private Provider<Connection> connectionProvider;

	@Inject
    Authorization20240105Dao(Provider<Connection> connectionProvider) throws IOException {

		this.connectionProvider = checkNotNull(connectionProvider);
	}

	@SuppressWarnings("unchecked")
	public List<AuthorizationVO> getAuthorizationsCpr(String cpr, Date currentTime) throws SQLException {
	    Connection connection = null;
        PreparedStatement preparedStatement = null;
        List<AuthorizationVO> authorizations = new ArrayList<AuthorizationVO>();

        try {
            connection = connectionProvider.get();

            String queryString = "SELECT Autorisationsnummer, " +
                    "                    CPR, " +
                    "                    Fornavn, " +
                    "                    Efternavn, " +
                    "                    UddannelsesKode, " +
                    "                    Uddannelsesland, " +
                    "                    Tidsberaenset, " +
                    "                    Tilsyn, " +
                    "                    Tilsynsbemaerkninger, " +
                    "                    Sanktionstype, " +
                    "                    Dispensation_75_aarsreglen, " +
                    "                    Selvstaendig_virke_dato, " +
                    "                    Selvstaendigt_virke, " +
                    "                    SelvstaendigvirkeUddLand, " +
                    "                    Speciale1, " +
                    "                    SpecialeDato1, " +
                    "                    SpecialeUddLand1, " +
                    "                    Speciale2, " +
                    "                    SpecialeDato2, " +
                    "                    SpecialeUddLand2, " +
                    "                    Speciale3, " +
                    "                    SpecialeDato3, " +
                    "                    SpecialeUddLand3, " +
                    "                    AutorisationGyldig, " +
                    "                    AutorisationsDato, " +
                    "                    CreatedDate, " +
                    "                    ModifiedDate, " +
                    "                    ValidFrom, " +
                    "                    ValidTo " +
                    "               from Autorisation5 "+
                    "where CPR=? AND ValidFrom <= ? AND (ValidTo > ? OR ValidTo IS NULL)";

            preparedStatement = connection.prepareStatement(queryString);
            int index=1;
            preparedStatement.setObject(index++, cpr);
            preparedStatement.setObject(index++, currentTime);
            preparedStatement.setObject(index++, currentTime);
            ResultSet resultSet = preparedStatement.executeQuery();
            while (resultSet.next()) {
                authorizations.add(mapAuthorization(resultSet));
            }
        } finally {
            PersistenceModule.close(preparedStatement);
            PersistenceModule.close(connection);
        }

		return authorizations;
	}

	private AuthorizationVO mapAuthorization(ResultSet resultSet) throws SQLException {
	    AuthorizationVO authorization = new AuthorizationVO();

	    authorization.setAuthorizationCode(resultSet.getString("Autorisationsnummer"));
	    authorization.setCpr(resultSet.getString("CPR"));
	    authorization.setFirstName(resultSet.getString("Fornavn"));
	    authorization.setLastName(resultSet.getString("Efternavn"));
	    authorization.setEducationCode(resultSet.getString("UddannelsesKode"));
        authorization.setUddannelsesland(resultSet.getString("Uddannelsesland"));
        authorization.setTidsbegraenset(resultSet.getString("Tidsberaenset"));
        authorization.setTilsyn(resultSet.getString("Tilsyn"));
        authorization.setTilsynsBemaerkninger(resultSet.getString("Tilsynsbemaerkninger"));
        authorization.setSanktionsType(resultSet.getString("Sanktionstype"));
        authorization.setDispensation75Aarsreglen(resultSet.getString("Dispensation_75_aarsreglen"));
        authorization.setSelvstaendigVirkeDato(getLocalDate(resultSet, "Selvstaendig_virke_dato"));
        authorization.setSelvstaendigtVirke(resultSet.getString("Selvstaendigt_virke"));
        authorization.setSelvstaendigvirkeUddLand(resultSet.getString("SelvstaendigvirkeUddLand"));

        authorization.setSpeciale1(resultSet.getString("Speciale1"));
        authorization.setSpecialeDato1(getLocalDate(resultSet, "SpecialeDato1"));
        authorization.setSpecialeUddLand1(resultSet.getString("SpecialeUddLand1"));
        authorization.setSpeciale2(resultSet.getString("Speciale2"));
        authorization.setSpecialeDato2(getLocalDate(resultSet, "SpecialeDato2"));
        authorization.setSpecialeUddLand2(resultSet.getString("SpecialeUddLand2"));
        authorization.setSpeciale3(resultSet.getString("Speciale3"));
        authorization.setSpecialeDato3(getLocalDate(resultSet, "SpecialeDato3"));
        authorization.setSpecialeUddLand3(resultSet.getString("SpecialeUddLand3"));

        authorization.setAutorisationGyldig(resultSet.getString("AutorisationGyldig"));
        authorization.setAutorisationsDato(getLocalDate(resultSet, "AutorisationsDato"));

	    authorization.setCreatedDate(resultSet.getDate("CreatedDate"));
	    authorization.setModifiedDate(resultSet.getDate("ModifiedDate"));
        authorization.setValidFrom(resultSet.getDate("ValidFrom"));
        authorization.setValidTo(resultSet.getDate("ValidTo"));

        return authorization;
    }

    @SuppressWarnings("unchecked")
	public List<AuthorizationVO> getAuthorizationsAuthorizationCode(String authorizationCode, Date currentTime) throws SQLException {
        Connection connection = null;
        PreparedStatement preparedStatement = null;
        List<AuthorizationVO> authorizations = new ArrayList<AuthorizationVO>();

        try {
            connection = connectionProvider.get();

            String queryString = "SELECT Autorisationsnummer, " +
                    "                    CPR, " +
                    "                    Fornavn, " +
                    "                    Efternavn, " +
                    "                    UddannelsesKode, " +
                    "                    Uddannelsesland, " +
                    "                    Tidsberaenset, " +
                    "                    Tilsyn, " +
                    "                    Tilsynsbemaerkninger, " +
                    "                    Sanktionstype, " +
                    "                    Dispensation_75_aarsreglen, " +
                    "                    Selvstaendig_virke_dato, " +
                    "                    Selvstaendigt_virke, " +
                    "                    SelvstaendigvirkeUddLand, " +
                    "                    Speciale1, " +
                    "                    SpecialeDato1, " +
                    "                    SpecialeUddLand1, " +
                    "                    Speciale2, " +
                    "                    SpecialeDato2, " +
                    "                    SpecialeUddLand2, " +
                    "                    Speciale3, " +
                    "                    SpecialeDato3, " +
                    "                    SpecialeUddLand3, " +
                    "                    AutorisationGyldig, " +
                    "                    AutorisationsDato, " +
                    "                    CreatedDate, " +
                    "                    ModifiedDate, " +
                    "                    ValidFrom, " +
                    "                    ValidTo " +
                    "               from Autorisation5 "+
                    "where Autorisationsnummer=? AND ValidFrom <= ? AND (ValidTo > ? OR ValidTo IS NULL)";

            preparedStatement = connection.prepareStatement(queryString);
            int index=1;
            preparedStatement.setObject(index++, authorizationCode);
            preparedStatement.setObject(index++, currentTime);
            preparedStatement.setObject(index++, currentTime);
            ResultSet resultSet = preparedStatement.executeQuery();
            while (resultSet.next()) {
                authorizations.add(mapAuthorization(resultSet));
            }
        } finally {
            PersistenceModule.close(preparedStatement);
            PersistenceModule.close(connection);
        }

        return authorizations;
	}

    private LocalDate getLocalDate(ResultSet resultSet, String columnName) throws SQLException {
        return resultSet.getDate(columnName) != null ? resultSet.getDate(columnName).toLocalDate() : null;
    }
}
