/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

package com.trifork.stamdata.authorization.webservice;

import com.google.inject.Inject;
import com.sun.xml.ws.developer.SchemaValidation;
import com.trifork.stamdata.SDMException;
import com.trifork.stamdata.authorization.SoapUtils20240105;
import com.trifork.stamdata.authorization.models.Authorization20240105Dao;
import com.trifork.stamdata.jaxws.GuiceInstanceResolver.GuiceWebservice;
import com.trifork.stamdata.models.AuthorizationVO;
import com.trifork.stamdata.util.SlaConsts;
import dk.medcom.dgws._2006._04.dgws_1_0.Header;
import dk.nsi._2024._01._05.stamdataauthorization.*;
import dk.nsi.stamdata.security.AccessControl;
import dk.nsi.stamdata.security.RequestContextProvider;
import dk.nsi.stamdata.security.RequestContextVO;
import dk.sds.nsp.audit.Audit;
import dk.sds.nsp.audit.AuditBuilder;
import dk.sds.nsp.audit.AuditBuilder.TypeOfInformation;
import dk.sdsd.nsp.slalog.api.SLALogItem;
import dk.sosi.seal.model.constants.FaultCodeValues;
import org.apache.log4j.Logger;
import org.oasis_open.docs.wss._2004._01.oasis_200401_wss_wssecurity_secext_1_0.Security;

import javax.jws.WebService;
import javax.xml.ws.Holder;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.time.LocalDate;
import java.time.format.DateTimeFormatter;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;


@WebService(endpointInterface="dk.nsi._2024._01._05.stamdataauthorization.AuthorizationCodePortType")
@GuiceWebservice
@SchemaValidation
public class AuthorizationCodeLookup20240105Impl implements AuthorizationCodePortType
{
	private final Authorization20240105Dao authorization20240105Dao;
    private final String WHITELIST_SERVICE_NAME = "SDMAuthorizationCode";
    private final SLALogItem slaLogItem;
    private final DateTimeFormatter pattern = DateTimeFormatter.ofPattern("yyyy-MM-dd");

    @Inject
    private RequestContextProvider requestContextProvider;

    @Inject
    private AccessControl accessControl;

	@Inject
    AuthorizationCodeLookup20240105Impl(Authorization20240105Dao authorization20240105Dao, SLALogItem slaLogItem)
	{
        this.authorization20240105Dao = authorization20240105Dao;
        this.slaLogItem = slaLogItem;
	}

	@Override
    public AuthorizationResponseType authorization(Holder<Security> wsseHeader, Holder<Header> medcomHeader, AuthorizationCodeRequestType parameters) throws DGWSFault {
        RequestContextVO requestContext = null;
        try {
            requestContext = requestContextProvider.getRequestContext(null, slaLogItem);
            accessControl.checkAccess(requestContext, WHITELIST_SERVICE_NAME);
        } catch (SDMException e) {
            throw new DGWSFault(e.getMessage(), FaultCodeValues.INVALID_IDCARD);
        }
        Header value = medcomHeader.value;
        if (value.getLinking() != null && value.getLinking().getMessageID() != null) {
            String messageID = value.getLinking().getMessageID();
            slaLogItem.setMessageId(messageID);
            slaLogItem.setSourceOperation("authorization20240105");
            slaLogItem.addCallParameter(SlaConsts.SLA_PARAM_CVR, requestContext.getServiceActor().getOrganizationId());
        }
        SoapUtils20240105.setHeadersToOutgoing(wsseHeader, medcomHeader);

        String authorizationCode = parameters.getAuthorizationCode();

        Logger logger = Logger.getLogger(AuthorizationCodeLookup20240105Impl.class);

        List<AuthorizationVO> authorizations = null;

        String cpr = "";
        Date now = Date.from(requestContext.getTime());

        try {
            authorizations = authorization20240105Dao.getAuthorizationsAuthorizationCode(authorizationCode,now);
        } catch (SQLException e) {
            logger.error(e.getMessage());
            throw new DGWSFault("Denne autorisationskode findes ikke.", FaultCodeValues.PROCESSING_PROBLEM);
        }

        AuthorizationResponseType response = new ObjectFactory().createAuthorizationResponseType();


        TypeOfInformation informationType = TypeOfInformation.SensitivePersonalInformation;
    	String component = "SDM";
    	String context = "SAES AuthorizationCodeLookup";
    	AuditBuilder auditBuilder = Audit.createAuditBuilder();
        auditBuilder.addAuditInformation(component, context, informationType, "authorization code", authorizationCode);

        if (!authorizations.isEmpty()) {

        	cpr = authorizations.get(0).getCpr();
        	try {
        	    authorizations = authorization20240105Dao.getAuthorizationsCpr(cpr,now);
            } catch (SQLException e) {
                logger.error(e.getMessage());
                throw new DGWSFault("Denne autorisationskode findes ikke.", FaultCodeValues.PROCESSING_PROBLEM);
            }

            response.setCpr(cpr);
            response.setFornavn(authorizations.get(0).getFirstName());
            response.setEfternavn(authorizations.get(0).getLastName());

        } else {
    		logger.info("Denne autorisationskode findes ikke");

    		throw new DGWSFault("Denne autorisationskode findes ikke.", FaultCodeValues.PROCESSING_PROBLEM);
    	}

        for (AuthorizationVO authorization : authorizations)
        {
            AuthorizationType authorizationType = new ObjectFactory().createAuthorizationType();
            authorizationType.setAutorisationsNummer(authorization.getAuthorizationCode());
            authorizationType.setUddannelsesKode(authorization.getEducationCode());
            authorizationType.setUddannelsesLand(authorization.getUddannelsesland());
            authorizationType.setTidsbegraenset(authorization.getTidsbegraenset());
            authorizationType.setTilsyn(authorization.getTilsyn());
            authorizationType.setTilsynsBemaerkninger(authorization.getTilsynsBemaerkninger());
            authorizationType.setSanktionsType(authorization.getSanktionsType());
            authorizationType.setDispensation75Aarsreglen(authorization.getDispensation75Aarsreglen());
            authorizationType.setSelvstaendigVirkeDato(getLocalDateString(authorization.getSelvstaendigVirkeDato()));
            authorizationType.setSelvstaendigtVirke(authorization.getSelvstaendigtVirke());
            authorizationType.setSelvstaendigVirkeUddLand(authorization.getSelvstaendigvirkeUddLand());
            authorizationType.setSpeciale1(authorization.getSpeciale1());
            authorizationType.setSpecialeDato1(getLocalDateString(authorization.getSpecialeDato1()));
            authorizationType.setSpecialeUddLand1(authorization.getSpecialeUddLand1());
            authorizationType.setSpeciale2(authorization.getSpeciale2());
            authorizationType.setSpecialeDato2(getLocalDateString(authorization.getSpecialeDato2()));
            authorizationType.setSpecialeUddLand2(authorization.getSpecialeUddLand2());
            authorizationType.setSpeciale3(authorization.getSpeciale3());
            authorizationType.setSpecialeDato3(getLocalDateString(authorization.getSpecialeDato3()));
            authorizationType.setSpecialeUddLand3(authorization.getSpecialeUddLand3());
            authorizationType.setAutorisationGyldig(authorization.getAutorisationGyldig());
            authorizationType.setAutorisationsDato(getLocalDateString(authorization.getAutorisationsDato()));

            // Map the name of the education if we know it.

            if (educations.containsKey(authorization.getEducationCode()))
            {
                authorizationType.setUddannelsesNavn(educations.get(authorization.getEducationCode()));
            }
            response.getAutorisation().add(authorizationType);
        }
        auditBuilder.addAuditInformation(component, context, informationType, "cpr", authorizations.get(0).getCpr());
        return response;
    }

    private String getLocalDateString(LocalDate localDate) {
        return localDate != null ? localDate.format(pattern) : null;
    }

    private static Map<String,String> educations = new HashMap<String, String>(); {

        educations.put("4498", "Optiker");
        educations.put("5151", "Fysioterapeut");
        educations.put("5153", "Ergoterapeut");
        educations.put("5155", "Fodterapeut");
        educations.put("5158", "Radiograf");
        educations.put("5159", "Bioanalytiker");
        educations.put("5166", "Sygeplejerske");
        educations.put("5175", "Jordemoder");
        educations.put("5265", "Kiropraktor");
        educations.put("5431", "Tandplejer");
        educations.put("5432", "Klinisk Tandtekniker");
        educations.put("5433", "Tandlæge");
        educations.put("5451", "Klinisk diætist");
        educations.put("7170", "Læge");
        educations.put("9495", "Bandagist");
    }
}
