/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package com.trifork.stamdata.authorization;

import static org.hamcrest.Matchers.*;
import static org.junit.Assert.*;

import java.io.ByteArrayOutputStream;
import java.net.URL;
import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import javax.inject.Inject;
import javax.inject.Named;
import javax.xml.namespace.QName;
import dk.nsi.stamdata.dgws.DGWSHeaderMock;
import dk.nsi.stamdata.dgws.SecurityWrapper;
import dk.nsi.stamdata.testing.*;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;

import com.google.common.collect.Lists;
import com.google.inject.Provider;
import com.trifork.stamdata.jaxws.SealNamespaceResolver;
import com.trifork.stamdata.models.AuthorizationVO;
import com.trifork.stamdata.persistence.PersistenceModule;

import dk.nsi.stamdata.jaxws.generated.AuthorizationCodePortType;
import dk.nsi.stamdata.jaxws.generated.AuthorizationCodeRequestType;
import dk.nsi.stamdata.jaxws.generated.AuthorizationCodeService;
import dk.nsi.stamdata.jaxws.generated.AuthorizationPortType;
import dk.nsi.stamdata.jaxws.generated.AuthorizationRequestType;
import dk.nsi.stamdata.jaxws.generated.AuthorizationResponseType;
import dk.nsi.stamdata.jaxws.generated.AuthorizationService;
import dk.nsi.stamdata.jaxws.generated.AuthorizationType;
import dk.nsi.stamdata.jaxws.generated.DGWSFault;
import dk.nsi.stamdata.jaxws.generated.ObjectFactory;
import dk.sosi.seal.model.constants.FaultCodeValues;


@RunWith(GuiceTestRunner.class)
public class TestAuthorizationServlet
{
    private UndertowJaxWsTestServer undertowTestServer;
    List<AuthorizationVO> authorizations = Lists.newArrayList();
    TestHelper helper = new TestHelper();

    static final String WHITELISTED_CVR = "33257872";
    static final String NON_WHITELISTED_CVR = "46837428";
    static final String SERVICE_NAME1 = "SDM";
    static final String SERVICE_NAME2 = "SDMAuthorizationCode";

    String cvr = WHITELISTED_CVR;

    @Inject
    private Provider<Connection> connectionProvider;

    @Inject
    Connection connection;

    @Inject
    @Named("run.undertow")
    private boolean runUndertow;

    @Inject
    @Named("saes.endpoint")
    private String saesEndpoint;

    StamdataTestHelper testHelper = new StamdataTestHelper();
    DatabaseTestInitializer dbInitializer;

    private AuthorizationRequestType request = new ObjectFactory().createAuthorizationRequestType();
    private AuthorizationCodeRequestType requestCode = new ObjectFactory().createAuthorizationCodeRequestType();
    private AuthorizationResponseType response = null;
    private AuthorizationPortType client;
    private AuthorizationCodePortType clientCode;

    @Before
    public void setUp() throws Exception
    {
        if(runUndertow) {
            dbInitializer = new DatabaseTestInitializer(connection);
            dbInitializer.initialize();

            undertowTestServer = new UndertowJaxWsTestServer(ComponentController.class, new UndertowSlaLogParameters("Stamdata Authorization Lookup", "stamdata-authorization-lookup"));
            undertowTestServer.start();
        }

        connection.setAutoCommit(false);
        Statement statement = connection.createStatement();
        statement.executeUpdate("INSERT IGNORE INTO whitelist_config SET component_name='"+ SERVICE_NAME1 +"', cvr='" + WHITELISTED_CVR + "'");
        statement.executeUpdate("INSERT IGNORE INTO whitelist_config SET component_name='"+ SERVICE_NAME2 +"', cvr='" + WHITELISTED_CVR + "'");
        connection.commit();
        authorizations.clear();
    }


    @After
    public void tearDown() throws Exception
    {
        SecurityContextTestHelper.resetContextOrganisation();
        if(runUndertow) {
            undertowTestServer.close();
        }
        authorizations.clear();
    }


    @Test
    public void shouldReturnTheExpectedAuthorizationWhenThereAreOthersThatDontMatchTheQuery() throws Exception
    {
        AuthorizationVO authorization1_1 = createAuthorization();
        authorization1_1.setCpr("1111122222");
        authorization1_1.setEducationCode("2131");
        authorization1_1.setFirstName("Peter");
        authorization1_1.setLastName("Andersen");
        authorization1_1.setAuthorizationCode("B1114");

        AuthorizationVO authorization1_2 = createAuthorization();
        authorization1_2.setCpr("1111122222");
        authorization1_2.setFirstName("Peter");
        authorization1_2.setLastName("Andersen");
        authorization1_2.setAuthorizationCode("A2114");
        authorization1_2.setEducationCode("5155");

        // Add another to make sure the right one is selected.

        AuthorizationVO authorization2 = createAuthorization();
        authorization2.setCpr("2222211111");

        request.setCpr(authorization1_1.getCpr());

        sendRequest();

        assertThat(response.getFirstName(), is(authorization1_1.getFirstName()));
        assertThat(response.getLastName(), is(authorization1_1.getLastName()));

        assertThat(response.getAuthorization().size(), is(2));

        AuthorizationType authorization = response.getAuthorization().get(0);

        assertTrue(checkAuthorization(authorization1_1, authorization) || checkAuthorization(authorization1_2, authorization));
        if (authorization.getEducationCode().equals("5155")) {
            assertThat(authorization.getEducationName(), is("Fodterapeut"));
        }

        authorization = response.getAuthorization().get(1);

        assertTrue(checkAuthorization(authorization1_1, authorization) || checkAuthorization(authorization1_2, authorization));
        if (authorization.getEducationCode().equals("5155")) {
            assertThat(authorization.getEducationName(), is("Fodterapeut"));
        }


    }


    public boolean checkAuthorization(AuthorizationVO authorization1, AuthorizationType authorization) {
        return authorization.getEducationCode().equals(authorization1.getEducationCode()) &&
               authorization.getAuthorizationCode().equals(authorization1.getAuthorizationCode());
    }

    @Test(expected = DGWSFault.class)
    public void shouldReturnFaultIfNotWhitelisted() throws Exception
    {
        AuthorizationVO authorization = createAuthorization();
        request.setCpr(authorization.getCpr());

        cvr = NON_WHITELISTED_CVR;
        SecurityContextTestHelper.setContextOrganisation(cvr);
        sendRequest();
    }

    @Test
    public void shouldReturnEnEmptyResponseIfNoAuthorizationsWereFound() throws Exception
    {
        createAuthorization();

        request.setCpr("0000000000");

        sendRequest();

        assertThat(response.getFirstName(), is(nullValue()));
    }

    @Test
    public void testAuthorizationCode() throws Exception
    {
    	AuthorizationVO authorization1_1 = createAuthorization();
        authorization1_1.setCpr("1111122222");
        authorization1_1.setEducationCode("2131");
        authorization1_1.setFirstName("Peter");
        authorization1_1.setLastName("Andersen");
        authorization1_1.setAuthorizationCode("B1114");

        AuthorizationVO authorization1_2 = createAuthorization();
        authorization1_2.setCpr("1111122222");
        authorization1_2.setFirstName("Peter");
        authorization1_2.setLastName("Andersen");
        authorization1_2.setAuthorizationCode("A2114");
        authorization1_2.setEducationCode("5155");

        // Add another to make sure the right one is selected.

        AuthorizationVO authorization2 = createAuthorization();
        authorization2.setCpr("2222211111");

        requestCode.setAuthorizationCode(authorization1_1.getAuthorizationCode());

        sendCodeRequest();

        assertThat(response.getFirstName(), is(authorization1_1.getFirstName()));
        assertThat(response.getLastName(), is(authorization1_1.getLastName()));
        assertThat(response.getCpr(), is(authorization1_1.getCpr()));

        assertThat(response.getAuthorization().size(), is(2));

        AuthorizationType authorization = response.getAuthorization().get(0);

        assertTrue(checkAuthorization(authorization1_1, authorization) || checkAuthorization(authorization1_2, authorization));
        if (authorization.getEducationCode().equals("5155")) {
            assertThat(authorization.getEducationName(), is("Fodterapeut"));
        }

        authorization = response.getAuthorization().get(1);

        assertTrue(checkAuthorization(authorization1_1, authorization) || checkAuthorization(authorization1_2, authorization));
        if (authorization.getEducationCode().equals("5155")) {
            assertThat(authorization.getEducationName(), is("Fodterapeut"));
        }
    }

    @Test
    public void testUnknownCpr() throws Exception
    {
        AuthorizationVO authorization1_1 = createAuthorization();
        authorization1_1.setCpr("1111122222");
        authorization1_1.setEducationCode("2131");
        authorization1_1.setFirstName("Peter");
        authorization1_1.setLastName("Andersen");
        authorization1_1.setAuthorizationCode("B1114");

        AuthorizationVO authorization1_2 = createAuthorization();
        authorization1_2.setCpr("1111122222");
        authorization1_2.setFirstName("Peter");
        authorization1_2.setLastName("Andersen");
        authorization1_2.setAuthorizationCode("A2114");
        authorization1_2.setEducationCode("5155");

        requestCode.setAuthorizationCode("Invalid");

        try {
			sendCodeRequest();
		} catch (DGWSFault e) {
			assertEquals(FaultCodeValues.PROCESSING_PROBLEM, e.getFaultInfo());
			assertEquals("Denne autorisationskode findes ikke.", e.getMessage());
			return;
		}
        fail();
    }

    @Test
    public void testXmlUtilException() throws Exception {

        String fileName = "src/test/resources/data/invalidXml.xml";
        String xml = RequestHelper.readFile(fileName);

        SecurityContextTestHelper.setContextOrganisation(WHITELISTED_CVR);
        String responseXml = RequestHelper.sendRequest(saesEndpoint, "http://nsi.dk/sdm/Gateway", xml);

        String faultstring = RequestHelper.getFaultstring(responseXml);

        assertEquals("Failed to copy a message", faultstring);
    }

    public AuthorizationVO createAuthorization()
    {
        AuthorizationVO auth = new AuthorizationVO("1234567890", "Ib", "Sørensen", "A1234", "2312");
        auth.setCreatedDate(Factories.DUMMY_DATE);
        auth.setModifiedDate(Factories.DUMMY_DATE);
        auth.setValidFrom(Factories.DUMMY_DATE);
        auth.setValidTo(Factories.YEAR_2100);
        authorizations.add(auth);
        return auth;
    }

    public void persistAuthorizations() throws SQLException
    {
        Connection connection = null;
        try {
            connection = connectionProvider.get();
            connection.createStatement().executeUpdate("DELETE FROM Autorisation2");
            connection.commit();
            for (AuthorizationVO authorization : authorizations) {
                testHelper.persist(connection, authorization);
            }
        } finally {
            PersistenceModule.close(connection);
        }
    }

    public void sendRequest() throws Exception
    {
        helper.enableAuditlogging(true);
        ByteArrayOutputStream outStream = helper.redirectSystemOutToStream();

        persistAuthorizations();

        final QName SERVICE_QNAME = new QName("http://nsi.dk/-/stamdata/3.0", "AuthorizationService");

        URL wsdlLocation = new URL(String.format("%s?wsdl", saesEndpoint));
        AuthorizationService serviceCatalog = new AuthorizationService(wsdlLocation, SERVICE_QNAME);

        // SEAL enforces that the XML prefixes are exactly
        // as it creates them. So we have to make sure we
        // don't change them.

        serviceCatalog.setHandlerResolver(new SealNamespaceResolver());

        client = serviceCatalog.getAuthorizationPort();

        SecurityWrapper headers = DGWSHeaderMock.getSecurityWrapper();

        response = client.authorization(headers.getSecurity(), headers.getMedcomHeader(), request);

        String[] logs = helper.getSystemLogs();
        List<String> authorizationCodes = new ArrayList<String>();

        assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SAES AuthorizationLookup\",\"type\":\"SensitivePersonalInformation\",\"key\":\"cpr\",\"value\":\""+request.getCpr()+"\"}", logs[0]);
        if (authorizations.size()>1) {

            for(int i = 0; i<authorizations.size(); i++) {
                if (authorizations.get(i).getCpr() == request.getCpr()) {
                    authorizationCodes.add("\""+authorizations.get(i).getAuthorizationCode()+"\"");
                }
            }
            Collections.sort(authorizationCodes);

            assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SAES AuthorizationLookup\",\"type\":\"SensitivePersonalInformation\",\"key\":\"authorization code\",\"values\":"+authorizationCodes.toString().replace(" ", "")+"}", logs[1]);
        } else if (authorizationCodes.isEmpty()){
            assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SAES AuthorizationLookup\",\"type\":\"SensitivePersonalInformation\",\"key\":\"authorization code\",\"values\":"+authorizationCodes+"}", logs[1]);
        } else {
            assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SAES AuthorizationLookup\",\"type\":\"SensitivePersonalInformation\",\"key\":\"authorization code\",\"value\":\""+authorizations.get(0).getAuthorizationCode()+"\"}", logs[1]);
        }
        helper.enableAuditlogging(false);
        helper.resetSystemOut();
    }

    public void sendCodeRequest() throws Exception
    {
        helper.enableAuditlogging(true);
        ByteArrayOutputStream outStream = helper.redirectSystemOutToStream();

        persistAuthorizations();

        final QName SERVICE_QNAME = new QName("http://nsi.dk/-/stamdata/3.0", "AuthorizationCodeService");

        URL wsdlLocation = new URL(String.format("%s?wsdl", saesEndpoint));
        AuthorizationCodeService serviceCatalog = new AuthorizationCodeService(wsdlLocation, SERVICE_QNAME);

        // SEAL enforces that the XML prefixes are exactly
        // as it creates them. So we have to make sure we
        // don't change them.

        serviceCatalog.setHandlerResolver(new SealNamespaceResolver());

        clientCode = serviceCatalog.getAuthorizationCodePort();

        SecurityWrapper headers = DGWSHeaderMock.getSecurityWrapper();

        response = clientCode.authorization(headers.getSecurity(), headers.getMedcomHeader(), requestCode);

        String[] logs = helper.getSystemLogs();

        assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SAES AuthorizationCodeLookup\",\"type\":\"SensitivePersonalInformation\",\"key\":\"authorization code\",\"value\":\""+requestCode.getAuthorizationCode()+"\"}", logs[0]);
        assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SAES AuthorizationCodeLookup\",\"type\":\"SensitivePersonalInformation\",\"key\":\"cpr\",\"value\":\""+authorizations.get(0).getCpr()+"\"}", logs[1]);

        helper.enableAuditlogging(false);
        helper.resetSystemOut();
    }
}
