/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata.replication.krs.webservice;

import com.google.inject.Inject;
import com.google.inject.Provider;
import com.trifork.stamdata.jaxws.SealNamespaceResolver;
import com.trifork.stamdata.persistence.PersistenceModule;
import dk.medcom.dgws._2006._04.dgws_1_0.Header;
import dk.nsi._2011._10._21.stamdatakrs.*;
import dk.nsi.stamdata.replication.TestTableCreator;
import dk.nsi.stamdata.replication.krs.ComponentController;
import dk.nsi.stamdata.replication.models.ClientDao;
import dk.nsi.stamdata.testing.DatabaseTestInitializer;
import dk.nsi.stamdata.testing.SecurityContextTestHelper;
import dk.nsi.stamdata.testing.UndertowSlaLogParameters;
import dk.nsi.stamdata.testing.UndertowJaxWsTestServer;
import dk.sosi.seal.model.AuthenticationLevel;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.oasis_open.docs.wss._2004._01.oasis_200401_wss_wssecurity_secext_1_0.Security;
import org.w3c.dom.Element;

import javax.inject.Named;
import javax.xml.namespace.QName;
import javax.xml.ws.Holder;
import java.net.URL;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.Map;

import static org.hamcrest.Matchers.is;
import static org.junit.Assert.assertThat;

@RunWith(GuiceTestRunner.class)
public class StamdataReplicationIdCardLevelAttackTest {
    public static final String WHITELISTED_CVR = "33257872";
    public static final String NON_WHITELISTED_CVR = "87654321";

    private UndertowJaxWsTestServer undertowTestServer;
    private StamdataReplication client;

    private ReplicationRequestType request;
    private ReplicationResponseType response;
    private Element anyAsElement;
    private DatabaseTestInitializer dbInitializer;

    @Inject
    private Provider<Connection> connectionProvider;
    @Inject
    private ClientDao clientDao;

    @Inject
    TestTableCreator testTableCreator;

    @Inject
    @Named("run.undertow")
    private boolean runUndertow;

    @Inject
    @Named("skrs.endpoint")
    private String skrsEndpoint;

    @Before
    public void setUp() throws Exception {

        dbInitializer = new DatabaseTestInitializer(connectionProvider.get());
        dbInitializer.initialize();

        if(runUndertow) {
            undertowTestServer = new UndertowJaxWsTestServer(8987, ComponentController.class, new UndertowSlaLogParameters("Stamdata Replication Module", "stamdata-batch-copy"));
            undertowTestServer.start();
        }

        URL wsdlLocation = new URL(String.format("%s?wsdl", skrsEndpoint));
        QName serviceName = new QName("http://nsi.dk/2011/10/21/StamdataKrs/", "StamdataReplicationService");
        StamdataReplicationService service = new StamdataReplicationService(wsdlLocation, serviceName);

        service.setHandlerResolver(new SealNamespaceResolver());
        client = service.getStamdataReplication();
        SecurityContextTestHelper.setContextUserSystemUser();
    }

    @After
    public void tearDown() throws Exception {
        if (runUndertow) {
            if (undertowTestServer != null) {
                undertowTestServer.close();
            }
            dbInitializer.getConnection().close();
        }
    }

    // Tests

    @Test
    public void levelThreeRequestDoesNotRaisesAnException() throws Exception {
        attemptCallWithLevel(3);
    }

    @Test(expected = ReplicationFault.class)
    public void levelFourRequestRaisesAnException() throws Exception {
        attemptCallWithLevel(4);
    }

    // Helper methods

    private void attemptCallWithLevel(int level) throws ReplicationFault, Exception
    {
        populateDatabase();

        if(level == 3) {
            SecurityContextTestHelper.setContextUserSystemUser();
        } else {
            SecurityContextTestHelper.setContextUserHealthcareProfessionalUser();
        }
        createCprPersonRegisterReplicationRequest();
        sendRequest();

        assertResponseContainsAtom();
    }

    private void populateDatabase() throws SQLException {
        Connection connection = null;

        try {
            connection = connectionProvider.get();

            connection.createStatement().executeUpdate("SET foreign_key_checks = 0");
            connection.createStatement().executeUpdate("DELETE FROM  Client_permissions");
            connection.createStatement().executeUpdate("DELETE FROM  Client");
            connection.createStatement().executeUpdate("SET foreign_key_checks = 1");
        } finally {
            PersistenceModule.close(connection);
        }

        testTableCreator.whiteList("cpr/person/v1",WHITELISTED_CVR);
    }

    private void createCprPersonRegisterReplicationRequest() {
        request = new ObjectFactory().createReplicationRequestType();
        request.setRegister("cpr");
        request.setDatatype("person");
        request.setVersion(1L);
        request.setOffset("0");
    }

    private void assertResponseContainsAtom() {
        assertThat(anyAsElement.getLocalName(), is("feed"));
        assertThat(anyAsElement.getNamespaceURI(), is("http://www.w3.org/2005/Atom"));
        assertThat(anyAsElement.getFirstChild().getFirstChild().getTextContent(), is("tag:nsi.dk,2011:cpr/person/v1"));
    }

    private static Map<Integer, AuthenticationLevel> mapFromIntegerLevelToEnum;
    static {
        mapFromIntegerLevelToEnum = new HashMap<Integer, AuthenticationLevel>();

        mapFromIntegerLevelToEnum.put(AuthenticationLevel.NO_AUTHENTICATION.getLevel(), AuthenticationLevel.NO_AUTHENTICATION);
        mapFromIntegerLevelToEnum.put(AuthenticationLevel.USERNAME_PASSWORD_AUTHENTICATION.getLevel(), AuthenticationLevel.USERNAME_PASSWORD_AUTHENTICATION);
        mapFromIntegerLevelToEnum.put(AuthenticationLevel.VOCES_TRUSTED_SYSTEM.getLevel(), AuthenticationLevel.VOCES_TRUSTED_SYSTEM);
        mapFromIntegerLevelToEnum.put(AuthenticationLevel.MOCES_TRUSTED_USER.getLevel(), AuthenticationLevel.MOCES_TRUSTED_USER);
    }

    private void sendRequest() throws ReplicationFault {
        Holder<Security> securityHeader;
        Holder<Header> medcomHeader;

        SecurityWrapper secutityHeadersNotWhitelisted = DGWSHeaderMock.getSecurityWrapper();
        securityHeader = secutityHeadersNotWhitelisted.getSecurity();
        medcomHeader = secutityHeadersNotWhitelisted.getMedcomHeader();

        response = client.replicate(securityHeader, medcomHeader, request);
        anyAsElement = (Element) response.getAny();
    }
}
