/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata.replication.krs.webservice;

import com.google.inject.Inject;
import com.google.inject.Provider;
import com.trifork.stamdata.jaxws.SealNamespaceResolver;
import com.trifork.stamdata.models.PersonVO;
import com.trifork.stamdata.models.YderVO;
import com.trifork.stamdata.persistence.PersistenceModule;
import dk.medcom.dgws._2006._04.dgws_1_0.Header;
import dk.nsi._2011._10._21.stamdatakrs.*;
import dk.nsi.stamdata.replication.TestTableCreator;
import dk.nsi.stamdata.replication.constants.ReplicationConstants;
import dk.nsi.stamdata.replication.krs.ComponentController;
import dk.nsi.stamdata.replication.models.ClientDao;
import dk.nsi.stamdata.testing.*;
import org.joda.time.DateTime;
import org.junit.After;
import org.junit.Before;
import org.junit.Ignore;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.oasis_open.docs.wss._2004._01.oasis_200401_wss_wssecurity_secext_1_0.Security;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import javax.inject.Named;
import javax.xml.namespace.NamespaceContext;
import javax.xml.namespace.QName;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.ws.Holder;
import javax.xml.ws.soap.SOAPFaultException;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;
import java.io.ByteArrayOutputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.math.BigInteger;
import java.net.URL;
import java.sql.Connection;
import java.sql.SQLException;
import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.Scanner;

import static org.junit.Assert.*;

@RunWith(GuiceTestRunner.class)
public class StamdataReplicationImplIntegrationTest
{
    private static final int SERVER_PORT = 8987;
    public static final String WHITELISTED_CVR = "33257872";
    public static final String NON_WHITELISTED_CVR = "87654321";
    private boolean isClientAuthority = false;

    private UndertowJaxWsTestServer undertowTestServer;
    TestHelper helper = new TestHelper();
    private StamdataReplication client;

    private ReplicationRequestType request;
    private ReplicationResponseType response;
    private Element anyAsElement;

    @Inject
    private Provider<Connection> connectionProvider;
    @Inject
    private ClientDao clientDao;

    @Inject TestTableCreator testTableCreator;

    @Inject
    @Named("run.undertow")
    private boolean runUndertow;

    @Inject
    @Named("skrs.endpoint")
    private String skrsEndpoint;

    private DateTime now, lastYear, nextYear;
	private List<String> permissions;

    StamdataTestHelper testHelper = new StamdataTestHelper();
	DatabaseTestInitializer dbInitializer;

	@Before
    public void setUp() throws Exception
    {
        if(runUndertow) {

            // Setup in-memory database
            dbInitializer = new DatabaseTestInitializer(connectionProvider.get());
            dbInitializer.initialize();

            // Start Undertow server
            undertowTestServer = new UndertowJaxWsTestServer(SERVER_PORT, ComponentController.class, new UndertowSlaLogParameters("Stamdata Replication Module", "stamdata-batch-copy"));
            undertowTestServer.start();

            // Set security context for development profile
            SecurityContextTestHelper.setContextUserSystemUser();
        }

        URL wsdlLocation = new URL(String.format("%s?wsdl", skrsEndpoint));
        QName serviceName = new QName("http://nsi.dk/2011/10/21/StamdataKrs/", "StamdataReplicationService");
        StamdataReplicationService service = new StamdataReplicationService(wsdlLocation, serviceName);

        service.setHandlerResolver(new SealNamespaceResolver());
        client = service.getStamdataReplication();

        now = DateTime.now();
        lastYear = now.minusYears(1);
        nextYear = now.plusYears(1);

        isClientAuthority = true;

        permissions = Arrays.asList(new String[] {
                "cpr/person/v1",
                "sikrede/sikrede/v1",
                "yderregister/yder/v3",
                "yderregister/person/v3",
                "yderregister/v2/yder/v3",
                "yderregister/v2/person/v3",
                "bemyndigelsesservice/bemyndigelse/v1",
                "sor/sygehusafdeling/v1",
                "vitamin/grunddata/v1",
                "vitamin/firmadata/v1",
                "vitamin/udgaaedenavne/v1",
                "vitamin/indholdsstoffer/v1",
                "cpr/barnrelation/v1",
                "tilskudsblanket/forhoejettakst/v1",
                "ddv/vaccinationplan/v2",
                "doseringsforslag/dosageproposal/v2",
                "nationalRoles/nationalRolePerson/v1"});
    }

    @After
    public void tearDown() throws Exception
    {
        if (runUndertow) {
            dbInitializer.getConnection().close();

            if (undertowTestServer != null) {
                undertowTestServer.close();
            }
        }
    }


    @Test // SDM-85 Ualmindelig ValidFrom-dato returneres som dato med ugyldig tidszone
    public void testReallyOldValidFromDate() throws Exception {
        Date validFrom=new SimpleDateFormat("yyyy-MM-dd").parse("0575-08-01");
        YderVO yder = testHelper.createYder("1234");
        yder.setValidFrom(validFrom);
        createReplicationRequest("yderregister", "yder", 3L);
        populateDatabaseAndSendRequest();

        assertResponseContainsValueOnXPath("//yder:yder/yder:ValidFrom", "0575-08-02T23:00:00.000Z");
        assertAuditLog();
    }

    @Test
    public void testRecordElements() throws Exception {
        testHelper.createYder("1234");
        testHelper.createYder("4321");
        createReplicationRequest("yderregister", "yder", 3L);
        populateDatabaseAndSendRequest();
        assertAuditLog();
    }

    @Test
    public void testRegisterVersionNullNoRegisterVersionInResponse() throws Exception {
        testHelper.createYder("1234");
        testHelper.createYder("4321");
        createReplicationRequest("yderregister", "yder", 3L, null);
        populateDatabaseAndSendRequest();
        assertAuditLog(false);
    }

    // Tests
    @Test
    public void testNotSoOldValidFromDate() throws Exception {
        // Efter 1894 skal vi ikke længere have utc
        //Instant validFrom = new Instant("1894-08-01");
        Date validFrom=new SimpleDateFormat("dd/MM-yyyy").parse("01/08-1894");
        YderVO yder = testHelper.createYder("1234");
        yder.setValidFrom(validFrom);
        createReplicationRequest("yderregister", "yder", 3L);
        populateDatabaseAndSendRequest();
        assertResponseContainsValueOnXPath("//yder:yder/yder:ValidFrom", "1894-08-01T00:00:00.000+01:00");
        assertAuditLog();
    }

    @Test(expected = SOAPFaultException.class)
    public void requestWithoutAnyReplicationTypeGivesSenderSoapFault() throws Exception
    {
        request = new ReplicationRequestType();
        populateDatabaseAndSendRequest();
    }

    @Test
    public void basicQueryDoesNotRaiseAnException() throws Exception
    {

        createSikredeSikredeRegisterReplicationRequest();
        populateDatabaseAndSendRequest();
        assertAuditLog();
    }

    @Ignore
    @Test(expected = ReplicationFault.class)
    public void cvrThatIsNotWhitelistedRaisesAnException() throws Exception
    {

        createSikredeSikredeRegisterReplicationRequest();
        isClientAuthority = false;
        populateDatabaseAndSendRequest();
    }

    @Test(expected = ReplicationFault.class)
    public void cvrThatIsWhitelistedForAnotherRegisterRaisesAnException() throws Exception
    {

        createSorHospitalRegisterReplicationRequest();
        populateDatabaseAndSendRequest();
    }

    @Test(expected = ReplicationFault.class)
    public void requestToAnUnexistingRegisterAndDatatypeRaisesAnException() throws Exception
    {

        createUnexistingRegisterAndDatatypeReplicationRequest();
        populateDatabaseAndSendRequest();
    }

    @Test(expected = ReplicationFault.class)
    public void requestToAnExistingRegisterButUnexistingDatatypeRaisesAnException() throws Exception
    {

        createExistingRegisterButUnexistingDatatypeReplicationRequest();
        populateDatabaseAndSendRequest();
    }

    @Test
    public void basicQueryOnEmptyDatabaseReturnsZeroResults() throws Exception
    {
        createSikredeSikredeRegisterReplicationRequest();
        populateDatabaseAndSendRequest();
        assertResponseContainsSikredeAtom();
        assertResponseContainsExactNumberOfRecords("sdm:sikrede", 0);
        assertAuditLog();
    }

    @Test
    public void testSikredeCopy() throws Exception
    {
        PersonVO person = testHelper.createPerson("1234567890");
        YderVO yder = testHelper.createYder("1234");
        testHelper.createSikrede(person, yder, "1", DateTime.now());
        createReplicationRequest("sikrede", "sikrede", 1L);
        populateDatabaseAndSendRequest();

        assertResponseContainsRecordAtom("sikrede", "sikrede");
        assertResponseContainsExactNumberOfRecords("sikrede:sikrede", 1);
        assertResponseContainsValueOnXPath("//sikrede:sikrede/sikrede:CPRnr", "1234567890");
        assertAuditLog();
    }

    @Test
    public void testNationalRolesCopy() throws Exception
    {
        testHelper.createNationalRole("CVR:25450442-RID:22019616", "nspSundAssistR3", DateTime.now().toDate());
        createReplicationRequest("nationalRoles", "nationalRolePerson", 1L);
        populateDatabaseAndSendRequest();

        assertResponseContainsRecordAtom("nationalRoles", "nationalRolePerson");
        assertResponseContainsExactNumberOfRecords("nationalRolePerson:nationalRolePerson", 1);
        assertResponseContainsValueOnXPath("//nationalRolePerson:nationalRolePerson/nationalRolePerson:externalName", "nspSundAssistR3");
        assertAuditLog();
    }

    @Test
    public void testYderregisterCopy() throws Exception
    {
        YderVO yder = testHelper.createYder("1234");
        yder.setHistIdYder("1234567890123456");

        createReplicationRequest("yderregister", "yder", 3L);

        populateDatabaseAndSendRequest();

        assertResponseContainsRecordAtom("yderregister", "yder", 3);
        assertResponseContainsExactNumberOfRecords("yder:yder", 1);
        assertResponseContainsValueOnXPath("//yder:yder/yder:HistIdYder", "1234567890123456");
        assertAuditLog();
    }

    @Test
    public void testYderregisterCopy_registerVersion2() throws Exception
    {
        YderVO yder = testHelper.createYder("1234");
        yder.setHistIdYder("1234567890123456");

        createReplicationRequest("yderregister", "yder", 3L, 2L);

        populateDatabaseAndSendRequest();

        assertResponseContainsRecordAtom("yderregister", "yder", 3, 2);
        assertResponseContainsExactNumberOfRecords("yder:yder", 1);
        assertResponseContainsValueOnXPath("//yder:yder/yder:HistIdYder", "1234567890123456");
        assertAuditLog(true);
    }

    @Test
    public void testYderregisterWhitelistningVersion() throws Exception
    {
         testYderregisterWhitelistning("yderregister/yder/v3");
    }

    @Test
    public void testYderregisterWhitelistningDatatype() throws Exception
    {
         testYderregisterWhitelistning("yderregister/yder/");
    }

    @Test
    public void testYderregisterWhitelistningRegister() throws Exception
    {
         testYderregisterWhitelistning("yderregister/");
    }

    @Test
    public void testYderregisterWhitelistningAll() throws Exception
    {
         testYderregisterWhitelistning("");
    }

    @Test
    public void testYderregisterWhitelistningVersionFail() throws Exception
    {
        try {
            testYderregisterWhitelistning("yderregister/yder/v1");
        } catch (ReplicationFault fault) {
            assertEquals("Dette cvr nummer (33257872) er ikke autoriseret til at hente denne type data (yderregister/yder/v3).",fault.getMessage());
            return;
        }
        fail();
    }

    @Test
    public void testYderregisterWhitelistningDatatypeFail() throws Exception
    {
        try {
            testYderregisterWhitelistning("yderregister/foo/");
        } catch (ReplicationFault fault) {
            assertEquals("Dette cvr nummer (33257872) er ikke autoriseret til at hente denne type data (yderregister/yder/v3).",fault.getMessage());
            return;
        }
        fail();
    }

    @Test
    public void testYderregisterWhitelistningRegisterFail() throws Exception
    {
        try {
            testYderregisterWhitelistning("foo");
        } catch (ReplicationFault fault) {
            assertEquals("Dette cvr nummer (33257872) er ikke autoriseret til at hente denne type data (yderregister/yder/v3).",fault.getMessage());
            return;
        }
        fail();
    }

    public void testYderregisterWhitelistning(String permission) throws Exception
    {
        permissions = Arrays.asList(new String[] {permission});
        YderVO yder = testHelper.createYder("1234");
        yder.setHistIdYder("1234567890123456");
        createReplicationRequest("yderregister", "yder", 3L);

        populateDatabaseAndSendRequest();

        assertResponseContainsRecordAtom("yderregister", "yder", 3);
        assertResponseContainsExactNumberOfRecords("yder:yder", 1);
        assertResponseContainsValueOnXPath("//yder:yder/yder:HistIdYder", "1234567890123456");
        assertAuditLog();
    }


    public String anyElementAsString() throws TransformerException
    {
        Transformer transformer = TransformerFactory.newInstance().newTransformer();
        ByteArrayOutputStream byteArrayOutputStream = new ByteArrayOutputStream();
        transformer.transform(new DOMSource(anyAsElement), new StreamResult(byteArrayOutputStream));
        return new String(byteArrayOutputStream.toByteArray());
    }

    @Test
    public void testThatWeAreAbleToHandleMoreThanTwentyRequestsForWsdlInSuccession() throws Exception
    {
        URL wsdlLocation = new URL(String.format("%s?wsdl", skrsEndpoint));
        QName serviceName = new QName("http://nsi.dk/2011/10/21/StamdataKrs/", "StamdataReplicationService");

        for (int i = 0; i < 20; i++)
        {
            StamdataReplicationService service = new StamdataReplicationService(wsdlLocation, serviceName);

            service.setHandlerResolver(new SealNamespaceResolver());
            client = service.getStamdataReplication();
        }
    }

    @Test
    public void testThatWeAreAbleToHandleMoreThanTwentyConnectionsInSuccession() throws Exception
    {
        createSikredeSikredeRegisterReplicationRequest();
        populateDatabaseAndSendRequest();
        for (int i = 0; i < 20; i++)
        {
            sendRequest();
            assertAuditLog();
        }
    }

    @Test
    public void testCPRBarnRelationCopy() throws Exception
    {
        request = new ObjectFactory().createReplicationRequestType();
        request.setRegister("cpr");
        request.setDatatype("barnrelation");
        request.setVersion(1L);
        request.setOffset("00000000000000000000");
        request.setMaxRecords(500L);

        populateDatabaseAndSendRequest();

        //printDocument(anyAsElement.getOwnerDocument(), System.out);

        assertResponseContainsRecordAtom("cpr", "barnrelation");
        assertAuditLog();
    }

	/**
	 * This is a regression test covering the change introduced by NSPSUPPORT-57 to the documentation
	 */
	@Test()
	public void testPermissionsAreNotCommaSeparated() throws Exception {
		createReplicationRequest("vitamin", "grunddata");

		// the documentation stated that this was the way to get the 4 vitamin permissions, so let's try it
		permissions = Arrays.asList("vitamin/grunddata/v1,vitamin/firmadata/v1,vitamin/udgaaedenavne/v1,vitamin/indholdsstoffer/v1");

		try {
			populateDatabaseAndSendRequest();
			fail("Expected SOAP fault saying that the client does not have the necessary permissions");
		} catch (ReplicationFault fault) {
			assertTrue("Fault message: " + fault.getMessage() , fault.getMessage().contains("Dette cvr nummer (33257872) er ikke autoriseret til at hente denne type data (vitamin/grunddata/v1)"));
		}
	}

    @Test
    public void testViewNotFound() throws Exception {
        try {
            createReplicationRequest("sikrede", "sikrede-plappe");
            populateDatabaseAndSendRequest();
        } catch (ReplicationFault fault) {
            assertTrue(fault.getMessage().contains("Datatype version ikke fundet: sikrede/sikrede-plappe/v1"));
        }
    }

    public void testXmlUtilException() throws Exception {

        String fileName = "src/test/resources/data/invalidXml.xml";
        String xml = RequestHelper.readFile(fileName);

        String responseXml = RequestHelper.sendRequest(skrsEndpoint, "http://nsi.dk/2011/10/21/StamdataKrs/replicate", xml);

        String faultstring = RequestHelper.getFaultstring(responseXml);
        assertEquals("Failed to copy a message", faultstring);
    }

    @Test
    public void testException() throws Exception {

        String fileName = "src/test/resources/data/validXml.xml";
        String xml = RequestHelper.readFile(fileName);

        String responseXml = RequestHelper.sendRequest(skrsEndpoint, "http://nsi.dk/2011/10/21/StamdataKrs/replicate", xml);

        String faultstring = RequestHelper.getFaultstring(responseXml);
        assertEquals("Dette cvr nummer (33257872) er ikke autoriseret til at hente denne type data (yderregister/yder/v3).", faultstring);
    }

    @Test
    public void replicate_registerVersionOmitted_defaultsTo1() throws Exception {
	    // Arrange
        request = new ObjectFactory().createReplicationRequestType();
        request.setRegister("cpr");
        request.setDatatype("barnrelation");
        request.setVersion(1L);
        request.setRegisterVersion(null);
        request.setOffset("00000000000000000000");
        request.setMaxRecords(500L);

        // Act
        sendRequest();

        // Assert
        assertResponseContainsRecordAtom("cpr", "barnrelation", 1);
        assertAuditLog();
    }

	// Helper methods

	// Pretty print XML document - good for debugging
    private static void printDocument(Document doc, OutputStream out) {
        try {
            TransformerFactory tf = TransformerFactory.newInstance();
            Transformer transformer = tf.newTransformer();
            transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "no");
            transformer.setOutputProperty(OutputKeys.METHOD, "xml");
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");
            transformer.setOutputProperty(OutputKeys.ENCODING, "UTF-8");
            transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "4");
            transformer.transform(new DOMSource(doc), new StreamResult(new OutputStreamWriter(out, "UTF-8")));
        } catch(Exception e) {
            // ignore - this is a test method
            e.printStackTrace();
        }
    }

    private void createReplicationRequest(String register, String datatype, long version, Long registerVersion) {
        request = new ObjectFactory().createReplicationRequestType();
        request.setRegister(register);
        request.setDatatype(datatype);
        request.setVersion(version);
        request.setRegisterVersion(registerVersion);
        request.setOffset("0");
    }

    private void createReplicationRequest(String register, String datatype, long version) {
        createReplicationRequest(register, datatype, version, null);
    }

    private void createReplicationRequest(String register, String datatype) {
    	createReplicationRequest(register, datatype, 1L);
    }

    private void createSikredeSikredeRegisterReplicationRequest()
    {
        request = new ObjectFactory().createReplicationRequestType();
        request.setRegister("sikrede");
        request.setDatatype("sikrede");
        request.setVersion(1L);
        request.setRegisterVersion(null);
        request.setOffset("0");
    }

    private void createSorHospitalRegisterReplicationRequest()
    {
        request = new ObjectFactory().createReplicationRequestType();
        request.setRegister("sor");
        request.setDatatype("sygehus");
        request.setVersion(1L);
        request.setRegisterVersion(ReplicationConstants.DEFAULT_REGISTER_VERSION);
        request.setOffset("0");
    }

    private void createUnexistingRegisterAndDatatypeReplicationRequest()
    {
        request = new ObjectFactory().createReplicationRequestType();
        request.setRegister("foo");
        request.setDatatype("bar");
        request.setVersion(1L);
        request.setRegisterVersion(ReplicationConstants.DEFAULT_REGISTER_VERSION);
        request.setOffset("0");
    }

    private void createExistingRegisterButUnexistingDatatypeReplicationRequest()
    {
        request = new ObjectFactory().createReplicationRequestType();
        request.setRegister("sor");
        request.setDatatype("bar");
        request.setVersion(1L);
        request.setRegisterVersion(ReplicationConstants.DEFAULT_REGISTER_VERSION);
        request.setOffset("0");
    }

    private void assertResponseContainsSikredeAtom()
    {
        assertEquals(anyAsElement.getLocalName(), "feed");
        assertEquals(anyAsElement.getNamespaceURI(), "http://www.w3.org/2005/Atom");
        assertEquals(anyAsElement.getFirstChild().getFirstChild().getTextContent(), "tag:nsi.dk,2011:sikrede/sikrede/v1");
    }

    private void assertResponseContainsRecordAtom(String register, String datatype, int version)
    {
    	assertEquals(anyAsElement.getLocalName(), "feed");
    	assertEquals(anyAsElement.getNamespaceURI(), "http://www.w3.org/2005/Atom");
    	assertEquals(anyAsElement.getFirstChild().getFirstChild().getTextContent(), "tag:nsi.dk,2011:" + register + "/" + datatype + "/v"+version);
    }

    private void assertResponseContainsRecordAtom(String register, String datatype, int version, int registerVersion)
    {
        assertEquals(anyAsElement.getLocalName(), "feed");
        assertEquals(anyAsElement.getNamespaceURI(), "http://www.w3.org/2005/Atom");
        assertEquals(anyAsElement.getFirstChild().getFirstChild().getTextContent(), "tag:nsi.dk,2011:" + register + "/v"+registerVersion + "/" + datatype + "/v"+version);
    }

    private void assertResponseContainsRecordAtom(String register, String datatype)
    {
    	assertResponseContainsRecordAtom(register, datatype, 1);
    }

    private void assertResponseContainsPersonWithSurNameMatchingGivenName(String givenName, String surName) throws XPathExpressionException
    {
        XPathExpression expression = createXpathExpression("//sdm:person[sdm:fornavn='" + givenName + "']/sdm:efternavn");
        String result = expression.evaluate(anyAsElement);
        assertEquals(result, surName);
    }

    private void assertResponseContainsExactNumberOfRecords(String tag, int n) throws XPathExpressionException
    {
        assertResponseContainsExactNumberOfRecords(tag, Integer.toString(n));
    }

    private void assertResponseContainsExactNumberOfRecords(String tag, String n) throws XPathExpressionException
    {
        XPathExpression expression = createXpathExpression("count(//" + tag + ")");
        String countAsString = expression.evaluate(anyAsElement);
        assertEquals(n, countAsString);
    }

    private void assertResponseContainsPersonWithCpr(String oracleCpr) throws XPathExpressionException
    {
        XPathExpression expression = createXpathExpression("//sdm:person/sdm:cpr");
        String result = expression.evaluate(anyAsElement);

        //printDocument(anyAsElement.getOwnerDocument(), System.out);

        assertEquals(result, oracleCpr);
    }

    private void assertResponseContainsValueOnXPath(String xPath, String oracle) throws XPathExpressionException
    {
        XPathExpression expression = createXpathExpression(xPath);
        String result = expression.evaluate(anyAsElement);
        assertEquals(oracle, result);
    }

    private String getOffsetFromAtomFirstEntry() throws XPathExpressionException
    {
        XPathExpression expression = createXpathExpression("//atom:entry/atom:id");
        String atomId = expression.evaluate(anyAsElement);
        String[] tokens = atomId.split("/");
        return tokens[0];
    }

    private String getOffsetFromAtomLastEntry() throws XPathExpressionException
    {
        XPathExpression expression = createXpathExpression("//atom:entry/atom:id");
        String atomId = expression.evaluate(anyAsElement);
        String[] tokens = atomId.split("/");
        return tokens[tokens.length - 1];
    }

    private XPathExpression createXpathExpression(String expression) throws XPathExpressionException
    {
        NamespaceContext context = new NamespaceContextMap(
                "krs", "http://nsi.dk/2011/10/21/StamdataKrs/",
                "atom", "http://www.w3.org/2005/Atom",
                "sdm", "http://nsi.dk/-/stamdata/3.0/cpr",
                "sikrede", "http://nsi.dk/-/stamdata/3.0/sikrede",
                "yder", "http://nsi.dk/-/stamdata/3.0/yderregister",
                "grunddata", "http://nsi.dk/-/stamdata/3.0/vitamin",
                "forhoejettakst", "http://nsi.dk/-/stamdata/3.0/tilskudsblanket",
                "bemyndigelse", "http://nsi.dk/-/stamdata/3.0/bemyndigelsesservice",
        		"dosageproposal", "http://nsi.dk/-/stamdata/3.0/doseringsforslag",
        		"doseringsforslag", "http://nsi.dk/-/stamdata/3.0/doseringsforslag",
                "nationalRolePerson", "http://nsi.dk/-/stamdata/3.0/nationalRoles");
        XPathFactory factory = XPathFactory.newInstance();
        XPath xpath = factory.newXPath();
        xpath.setNamespaceContext(context);
        return xpath.compile(expression);
    }

    private void populateDatabaseAndSendRequest() throws Exception
    {
        populateDatabase();
        sendRequest();
    }

    private void populateDatabase() throws SQLException {
        Connection connection = null;

        try {
            connection = connectionProvider.get();

            connection.createStatement().executeUpdate("DELETE FROM Client");
            connection.createStatement().executeUpdate("DELETE FROM Client_permissions");

            connection.commit();
        } finally {
            PersistenceModule.closeConnection(connection);
        }

        for (String permission : permissions) {
            testTableCreator.whiteList(permission,WHITELISTED_CVR);
        }

        try {
            connection = connectionProvider.get();
            testHelper.prepareTestdata(connection);
            connection.commit();
        } finally {
            PersistenceModule.closeConnection(connection);
        }
    }

    private void sendRequest() throws Exception
    {
        helper.enableAuditlogging(true);
        helper.redirectSystemOutToStream();

        Holder<Security> securityHeader;
        Holder<Header> medcomHeader;

        if (isClientAuthority)
        {
            SecurityWrapper secutityHeadersNotWhitelisted = DGWSHeaderMock.getSecurityWrapper();
            securityHeader = secutityHeadersNotWhitelisted.getSecurity();
            medcomHeader = secutityHeadersNotWhitelisted.getMedcomHeader();
        } else {
            SecurityWrapper securityHeaders = DGWSHeaderMock.getSecurityWrapper();
            securityHeader = securityHeaders.getSecurity();
            medcomHeader = securityHeaders.getMedcomHeader();
        }
        response = client.replicate(securityHeader, medcomHeader, request);
        anyAsElement = (Element) response.getAny();

        helper.resetSystemOut();
        helper.enableAuditlogging(false);
    }

    public void assertAuditLog() throws XPathExpressionException {
        if (!runUndertow) { return; }

        assertAuditLog(false);
    }

	public void assertAuditLog(boolean withRegVersion) throws XPathExpressionException {
        if (!runUndertow) { return; }

		String[] logs = helper.getSystemLogs();
        NodeList elements = anyAsElement.getElementsByTagName("atom:entry");

        assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SKRS Replicate\",\"type\":\"SensitivePersonalInformation\",\"key\":\"register\",\"value\":\"" + request.getRegister() + "\"}", logs[0]);
        assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SKRS Replicate\",\"type\":\"SensitivePersonalInformation\",\"key\":\"datatype\",\"value\":\"" + request.getDatatype() + "\"}", logs[1]);
        assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SKRS Replicate\",\"type\":\"SensitivePersonalInformation\",\"key\":\"version\",\"value\":\"" + request.getVersion() + "\"}", logs[2]);
        assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SKRS Replicate\",\"type\":\"SensitivePersonalInformation\",\"key\":\"registerVersion\",\"value\":\"" + (request.getRegisterVersion() != null ? request.getRegisterVersion() : ReplicationConstants.DEFAULT_REGISTER_VERSION) + "\"}", logs[3]);
        if (request.getMaxRecords() != null) {
        	assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SKRS Replicate\",\"type\":\"SensitivePersonalInformation\",\"key\":\"max records\",\"value\":\"" + request.getMaxRecords() + "\"}", logs[4]);
		} else {
			assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SKRS Replicate\",\"type\":\"SensitivePersonalInformation\",\"key\":\"max records\",\"value\":\"Undefined\"}", logs[4]);
		}
        assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SKRS Replicate\",\"type\":\"SensitivePersonalInformation\",\"key\":\"offset\",\"value\":\"" + request.getOffset() + "\"}", logs[5]);
        assertEquals(elements.getLength()+"", getValueFromLog(logs[6]).toString());
        if (elements.getLength() > 0) {
            Node first = elements.item(0);
            Node last = elements.item(elements.getLength()-1);
            int elementIndex = withRegVersion ? 4 : 3;
            String firstRequest = first.getFirstChild().getTextContent().split("/")[elementIndex];
            String lastRequest = last.getFirstChild().getTextContent().split("/")[elementIndex];
            assertEquals(getValueFromLog(logs[7]), new BigInteger(firstRequest));
            assertEquals(getValueFromLog(logs[8]), new BigInteger(lastRequest));
        }
	}

	public BigInteger getValueFromLog(String log) {
		Scanner scanner = new Scanner(log);

        scanner.useDelimiter("\"");
        scanner.findInLine("\"value\":");
        BigInteger value = scanner.nextBigInteger();
		return value;
	}
}
