/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata.replication.rfs.webservice;

import com.google.inject.Inject;
import com.google.inject.Provider;
import com.trifork.stamdata.jaxws.SealNamespaceResolver;
import com.trifork.stamdata.models.PersonVO;
import com.trifork.stamdata.models.YderVO;
import com.trifork.stamdata.persistence.PersistenceModule;
import dk.medcom.dgws._2006._04.dgws_1_0.Header;
import dk.nsi._2024._02._27.stamdatarfs.*;
import dk.nsi.stamdata.replication.TestTableCreator;
import dk.nsi.stamdata.replication.constants.ReplicationConstants;
import dk.nsi.stamdata.replication.models.ClientDao;
import dk.nsi.stamdata.replication.rfs.ComponentController;
import dk.nsi.stamdata.replication.webservice.DynamicViewXmlGenerator;
import dk.nsi.stamdata.replication.webservice.FaultCodes;
import dk.nsi.stamdata.replication.webservice.StamdataReplicationFault;
import dk.nsi.stamdata.testing.*;
import org.apache.commons.lang3.StringUtils;
import org.joda.time.DateTime;
import org.junit.*;
import org.junit.runner.RunWith;
import org.oasis_open.docs.wss._2004._01.oasis_200401_wss_wssecurity_secext_1_0.Security;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import javax.inject.Named;
import javax.xml.namespace.NamespaceContext;
import javax.xml.namespace.QName;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.ws.Holder;
import javax.xml.ws.WebServiceException;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;
import java.io.ByteArrayOutputStream;
import java.math.BigInteger;
import java.net.URL;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.*;

import static org.junit.Assert.assertEquals;

@RunWith(GuiceTestRunner.class)
public class StamdataReplicationRfs20240227IntegrationTest
{
    private static final int SERVER_PORT = 8987;
    public static final String WHITELISTED_CVR = "33257872";
    public static final String NON_WHITELISTED_CVR = "87654321";
    private boolean isClientAuthority = false;

    private UndertowJaxWsTestServer undertowTestServer;
    TestHelper helper = new TestHelper();
    private StamdataReplication client;

    private ReplicationRequestType request;
    private ReplicationResponseType response;
    private Element anyAsElement;

    @Inject
    private Provider<Connection> connectionProvider;
    @Inject
    private ClientDao clientDao;

    @Inject TestTableCreator testTableCreator;

    @Inject
    @Named("run.undertow")
    private boolean runUndertow;

    @Inject
    @Named("srfs.endpoint")
    private String srfsEndpoint;

    private DateTime now, lastYear, nextYear;
	private List<String> permissions;

    StamdataTestHelper testHelper = new StamdataTestHelper();
    private StamdataBatchCopyTestHelper stamdataBatchCopyTestHelper;
    private DatabaseTestInitializer dbInitializer;

	@Before
    public void setUp() throws Exception
    {
        if(runUndertow) {

            // Setup in-memory database
            dbInitializer = new DatabaseTestInitializer(connectionProvider.get());
            dbInitializer.initialize();

            // Start Undertow server
            undertowTestServer = new UndertowJaxWsTestServer(SERVER_PORT, ComponentController.class, new UndertowSlaLogParameters("Stamdata Replication Module", "stamdata-batch-copy"));
            undertowTestServer.start();

            // Set security context for development profile
            SecurityContextTestHelper.setContextUserSystemUser();
        }

//        URL wsdlLocation = new URL(String.format("%s?wsdl", srfsEndpoint));
        URL wsdlLocation = new URL(String.format("%s?wsdl", "http://localhost:8987/service/StamdataReplication-20240227"));
        QName serviceName = new QName("http://nsi.dk/2024/02/27/StamdataRfs/", "StamdataReplicationService");
        StamdataReplicationService service = new StamdataReplicationService(wsdlLocation, serviceName);

        service.setHandlerResolver(new SealNamespaceResolver());
        client = service.getStamdataReplication();

        now = DateTime.now();
        lastYear = now.minusYears(1);
        nextYear = now.plusYears(1);

        isClientAuthority = true;

	    permissions = Arrays.asList("cpr/person/v1",
                "sikrede/sikrede/v1",
                "yderregister/yder/v3",
                "yderregister/person/v3",
                "yderregister/v2/yder/v3",
                "yderregister/v2/person/v3",
                "bemyndigelsesservice/bemyndigelse/v1",
                "sor/sygehusafdeling/v1",
                "vitamin/grunddata/v1",
                "vitamin/firmadata/v1",
                "vitamin/udgaaedenavne/v1",
                "vitamin/indholdsstoffer/v1",
                "cpr/barnrelation/v1",
                "tilskudsblanket/forhoejettakst/v1",
                "ddv/vaccinationplan/v2",
                "doseringsforslag/dosageproposal/v2");

        stamdataBatchCopyTestHelper = new StamdataBatchCopyTestHelper();
    }

    @After
    public void tearDown() throws Exception
    {
    	if (undertowTestServer != null) {
    		undertowTestServer.close();
    	}
    }

    @Test
    @Ignore
    public void testSikredeCopy() throws Exception
    {
        PersonVO person = testHelper.createPerson("1234567890");
        YderVO yder = testHelper.createYder("1234");
        testHelper.createSikrede(person, yder, "1", DateTime.now());
        createReplicationRequest("sikrede", "sikrede", null);
        populateDatabaseAndSendRequest();

        assertResponseContainsRecordAtom("sikrede", "sikrede");
        assertResponseContainsExactNumberOfRecords("sikrede:sikrede", 1);
        assertResponseContainsValueOnXPath("//sikrede:sikrede/sikrede:CPRnr", "1234567890");
        assertAuditLog();
    }

    @Test
    public void testPersonLookup_singleId_singleVersion() throws Exception {
	    PersonVO person = testHelper.createPerson("2211657418");

        createReplicationRequest("cpr", "person", Arrays.asList("2211657418"));
        populateDatabaseAndSendRequest();

        assertResponseContainsRecordAtom("cpr", "person");
        assertResponseContainsExactNumberOfRecords("person:person", 1);
        assertResponseContainsValueOnXPath("//person:person/person:cpr", "2211657418");
        assertAuditLog();
    }

    @Test
    public void testPersonLookup_singleId_multipleVersions() throws Exception {
        PersonVO firstVersion = testHelper.createPerson("2211657418");
        firstVersion.setValidFrom(new Date(firstVersion.getValidFrom().getTime() + 1000));

        PersonVO secondVersion = testHelper.createPerson("2211657418");
        secondVersion.setValidFrom(new Date(secondVersion.getValidFrom().getTime() + 2000));
        secondVersion.setHusnummer("11");

        createReplicationRequest("cpr", "person", Arrays.asList("2211657418"));
        populateDatabaseAndSendRequest();

        assertResponseContainsRecordAtom("cpr", "person");
        assertResponseContainsExactNumberOfRecords("person:person", 1);
        assertResponseContainsValueOnXPath("//person:person/person:husnummer", "11");
        assertAuditLog();
    }

    @Test
    public void testPersonLookup_otherId_singleVersion() throws Exception {
        PersonVO person = testHelper.createPerson("1112579876");

        createReplicationRequest("cpr", "person", Arrays.asList("1112579876"));
        populateDatabaseAndSendRequest();

        assertResponseContainsRecordAtom("cpr", "person");
        assertResponseContainsExactNumberOfRecords("person:person", 1);
        assertResponseContainsValueOnXPath("//person:person/person:cpr", "1112579876");
        assertAuditLog();
    }

    @Test
    public void testPersonLookup_multipleIds() throws Exception {
        PersonVO person1 = testHelper.createPerson("1112579876");
        PersonVO person2 = testHelper.createPerson("2211657418");

        createReplicationRequest("cpr", "person", Arrays.asList("1112579876", "2211657418"));
        populateDatabaseAndSendRequest();

        assertResponseContainsRecordAtom("cpr", "person");
        assertResponseContainsExactNumberOfRecords("person:person", 2);
        assertResponseContainsValueOnXPath("//person:person/person:cpr[text()='1112579876']", "1112579876");
        assertResponseContainsValueOnXPath("//person:person/person:cpr[text()='2211657418']", "2211657418");
        assertAuditLog();
    }

    @Test
    public void testYder() throws Exception
    {
        YderVO yder = testHelper.createYder("123456");
        createReplicationRequest("yderregister", "yder", 3, Arrays.asList("123456"));
        populateDatabaseAndSendRequest();

        assertResponseContainsRecordAtom("yderregister", "yder", 3);
        assertResponseContainsExactNumberOfRecords("yder:yder", 1);
        assertResponseContainsValueOnXPath("//yder:yder/yder:YdernrYder", "123456");
        assertAuditLog();
    }

    @Test
    public void testPersonLookup_malformedId_emptyResult() throws Exception {
        PersonVO person1 = testHelper.createPerson("1112579876");
        PersonVO person2 = testHelper.createPerson("2211657418");

        createReplicationRequest("cpr", "person", Arrays.asList("foo", "bar"));
        populateDatabaseAndSendRequest();

        assertResponseContainsRecordAtom("cpr", "person");
        assertResponseContainsExactNumberOfRecords("person:person", 0);
        assertAuditLog();
    }

    @Test
    public void testPersonLookup_largeRequest() throws Exception {
	    List<String> cprNumbers = new ArrayList<>();
	    int maxRequestSize = 5000;
	    for(int i = 0; i < maxRequestSize; i++) {
	        String cpr = String.format("111257%s", StringUtils.leftPad(Integer.toString(i), 4, "0"));
	        cprNumbers.add(cpr);
        }

	    for(String cpr : cprNumbers) {
	        testHelper.createPerson(cpr);
        }

        createReplicationRequest("cpr", "person", cprNumbers);
        populateDatabaseAndSendRequest();

        assertResponseContainsRecordAtom("cpr", "person");
        assertResponseContainsExactNumberOfRecords("person:person", maxRequestSize);
        assertResponseContainsValueOnXPath("//person:person/person:cpr[text()='1112570000']", "1112570000");
        assertResponseContainsValueOnXPath("//person:person/person:cpr[text()='1112574999']", "1112574999");
        assertAuditLog();
    }

    @Test(expected = WebServiceException.class)
    public void testPersonLookup_tooLargeRequest() throws Exception {
        List<String> cprNumbers = new ArrayList<>();
        int maxRequestSize = 5001;
        for(int i = 0; i < maxRequestSize; i++) {
            String cpr = String.format("111257%s", StringUtils.leftPad(Integer.toString(i), 4, "0"));
            cprNumbers.add(cpr);
        }

        for(String cpr : cprNumbers) {
            testHelper.createPerson(cpr);
        }

        createReplicationRequest("cpr", "person", cprNumbers);
        populateDatabaseAndSendRequest();
    }

    @Test(expected = WebServiceException.class)
    public void testPersonLookup_emptyRequest() throws Exception {
        List<String> cprNumbers = new ArrayList<>();

        createReplicationRequest("cpr", "person", cprNumbers);
        populateDatabaseAndSendRequest();
    }

    @Test
    public void filterTest_shouldFailWithFilterThatDoesNotExist() {
        String cpr = "2211657418";
        String filterName = "nonExistingFilter";
        testHelper.createPerson(cpr);
        try {
            createReplicationRequest("cpr", "person", Arrays.asList(cpr), filterName);
            populateDatabaseAndSendRequest();
            Assert.fail("Expected an WebServiceException to be thrown");
        } catch (Exception e) {
            Assert.assertEquals("Invalid filter with name: " + filterName + " and CVR: " + WHITELISTED_CVR, e.getMessage());
        }
    }

    @Test
    public void filterTest_shouldPassWithExistingFilterAndValidDataType() throws Exception {
        String cpr = "2211657418";
        String filterName = "existingFilter";
        String register = "cpr";
        String datatype = "person";
        List<String> fieldsInFilter = Arrays.asList("cpr", "koen", "fornavn", "mellemnavn");
        testHelper.createPerson(cpr);

        createReplicationRequest(register, datatype, Arrays.asList(cpr), filterName);
        populateDatabaseAndSendRequest();

        assertResponseContainsRecordAtom(register, datatype);
        assertResponseContainsExactNumberOfRecords(datatype+":"+datatype, 1);
        assertResponseContainsValueOnXPath("//"+datatype+":"+datatype+"/"+datatype+":"+register, cpr);
        stamdataBatchCopyTestHelper.assertNodesInElementListAreInKeepList(anyAsElement, datatype, register, DynamicViewXmlGenerator.STAMDATA_NAMESPACE_URI_PREFIX, fieldsInFilter);
    }

    public String anyElementAsString() throws TransformerException
    {
        Transformer transformer = TransformerFactory.newInstance().newTransformer();
        ByteArrayOutputStream byteArrayOutputStream = new ByteArrayOutputStream();
        transformer.transform(new DOMSource(anyAsElement), new StreamResult(byteArrayOutputStream));
        return new String(byteArrayOutputStream.toByteArray());
    }

    private void createReplicationRequest(String register, String datatype, long version, Long registerVersion, List<String> idList, String filterName) {
        request = new ObjectFactory().createReplicationRequestType();
        request.setRegister(register);
        request.setDatatype(datatype);
        request.setVersion(version);
        request.setRegisterVersion(registerVersion);

        if(filterName != null) {
            request.setFilter(filterName);
        }
        if(idList != null) {
            request.setIdList(new IdList());
            request.getIdList().getId().addAll(idList);
        }
    }

    private void createReplicationRequest(String register, String datatype, long version, List<String> idList, String filterName) {
        createReplicationRequest(register, datatype, version, null, idList, filterName);
    }

    private void createReplicationRequest(String register, String datatype, long version, List<String> idList) {
        createReplicationRequest(register, datatype, version, null, idList, null);
    }

    private void createReplicationRequest(String register, String datatype, List<String> idList, String filterName) {
    	createReplicationRequest(register, datatype, 1L, idList, filterName);
    }

    private void createReplicationRequest(String register, String datatype, List<String> idList) {
        createReplicationRequest(register, datatype, 1L, idList, null);
    }

    private void assertResponseContainsRecordAtom(String register, String datatype, int version)
    {
    	assertEquals("feed", anyAsElement.getLocalName());
    	assertEquals("http://www.w3.org/2005/Atom", anyAsElement.getNamespaceURI());
    	assertEquals("tag:nsi.dk,2011:" + register + "/" + datatype + "/v"+version, anyAsElement.getFirstChild().getFirstChild().getTextContent());
    }

    private void assertResponseContainsRecordAtom(String register, String datatype)
    {
    	assertResponseContainsRecordAtom(register, datatype, 1);
    }

    private void assertResponseContainsExactNumberOfRecords(String tag, int n) throws XPathExpressionException
    {
        assertResponseContainsExactNumberOfRecords(tag, Integer.toString(n));
    }

    private void assertResponseContainsExactNumberOfRecords(String tag, String n) throws XPathExpressionException
    {
        XPathExpression expression = createXpathExpression("count(//" + tag + ")");
        String countAsString = expression.evaluate(anyAsElement);
        assertEquals(n, countAsString);
    }

    private void assertResponseContainsValueOnXPath(String xPath, String oracle) throws XPathExpressionException
    {
        XPathExpression expression = createXpathExpression(xPath);
        String result = expression.evaluate(anyAsElement);
        assertEquals(oracle, result);
    }

    private XPathExpression createXpathExpression(String expression) throws XPathExpressionException
    {
        NamespaceContext context = new NamespaceContextMap(
                "rfs", "http://nsi.dk/2011/10/21/StamdataRfs/",
                "atom", "http://www.w3.org/2005/Atom",
                "sdm", "http://nsi.dk/-/stamdata/3.0/cpr",
                "sikrede", "http://nsi.dk/-/stamdata/3.0/sikrede",
                "yder", "http://nsi.dk/-/stamdata/3.0/yderregister",
                "grunddata", "http://nsi.dk/-/stamdata/3.0/vitamin",
                "forhoejettakst", "http://nsi.dk/-/stamdata/3.0/tilskudsblanket",
                "bemyndigelse", "http://nsi.dk/-/stamdata/3.0/bemyndigelsesservice",
        		"dosageproposal", "http://nsi.dk/-/stamdata/3.0/doseringsforslag",
        		"doseringsforslag", "http://nsi.dk/-/stamdata/3.0/doseringsforslag",
                "person", "http://nsi.dk/-/stamdata/3.0/cpr");
        XPathFactory factory = XPathFactory.newInstance();
        XPath xpath = factory.newXPath();
        xpath.setNamespaceContext(context);
        return xpath.compile(expression);
    }

    private void populateDatabaseAndSendRequest() throws Exception
    {
        populateDatabase();
        sendRequest();
    }

    private void populateDatabase() throws SQLException {
        Connection connection = null;

        try {
            connection = connectionProvider.get();

            connection.createStatement().executeUpdate("DELETE FROM Client");
            connection.createStatement().executeUpdate("DELETE FROM Client_permissions");

            connection.commit();
        } finally {
            PersistenceModule.closeConnection(connection);
        }

        for (String permission : permissions) {
            testTableCreator.whiteList(permission,WHITELISTED_CVR);
        }

        try {
            connection = connectionProvider.get();
            testHelper.prepareTestdata(connection);
            connection.commit();
        } finally {
            PersistenceModule.closeConnection(connection);
        }
    }

    private void sendRequest() throws Exception
    {
        helper.enableAuditlogging(true);
        helper.redirectSystemOutToStream();

        Holder<Security> securityHeader;
        Holder<Header> medcomHeader;

        if (isClientAuthority)
        {
            SecurityWrapper secutityHeadersNotWhitelisted = DGWSHeaderMock.getSecurityWrapper();
            securityHeader = secutityHeadersNotWhitelisted.getSecurity();
            medcomHeader = secutityHeadersNotWhitelisted.getMedcomHeader();
        } else {
            SecurityWrapper securityHeaders = DGWSHeaderMock.getSecurityWrapper();
            securityHeader = securityHeaders.getSecurity();
            medcomHeader = securityHeaders.getMedcomHeader();
        }

        response = client.replicate(securityHeader, medcomHeader, request);
        anyAsElement = (Element) response.getAny();

        helper.resetSystemOut();
        helper.enableAuditlogging(false);
    }

    public void assertAuditLog() throws XPathExpressionException {
        if (!runUndertow) { return; }

        String[] logs = helper.getSystemLogs();
        NodeList elements = anyAsElement.getElementsByTagName("atom:entry");

        assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SRFS Replicate\",\"type\":\"SensitivePersonalInformation\",\"key\":\"register\",\"value\":\"" + request.getRegister() + "\"}", logs[0]);
        assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SRFS Replicate\",\"type\":\"SensitivePersonalInformation\",\"key\":\"datatype\",\"value\":\"" + request.getDatatype() + "\"}", logs[1]);
        assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SRFS Replicate\",\"type\":\"SensitivePersonalInformation\",\"key\":\"version\",\"value\":\"" + request.getVersion() + "\"}", logs[2]);
        assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SRFS Replicate\",\"type\":\"SensitivePersonalInformation\",\"key\":\"registerVersion\",\"value\":\"" + (request.getRegisterVersion() != null ? request.getRegisterVersion() : ReplicationConstants.DEFAULT_REGISTER_VERSION) + "\"}", logs[3]);
        if(request.getIdList() != null) {
            assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SRFS Replicate\",\"type\":\"SensitivePersonalInformation\",\"key\":\"idlist\",\"values\":[\"" + String.join("\",\"", request.getIdList().getId()) + "\"]}", logs[4]);
        }
        else {
            assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SRFS Replicate\",\"type\":\"SensitivePersonalInformation\",\"key\":\"idlist\",\"values\":[]}", logs[4]);
        }

        assertEquals(elements.getLength()+"", getValueFromLog(logs[5]).toString());
        if (elements.getLength() > 0) {
            Node first = elements.item(0);
            Node last = elements.item(elements.getLength()-1);
            int elementIndex = isDefaultRegisterVersion() ? 3 : 4;
            String firstRequest = first.getFirstChild().getTextContent().split("/")[elementIndex];
            String lastRequest = last.getFirstChild().getTextContent().split("/")[elementIndex];
            assertEquals(getValueFromLog(logs[6]), new BigInteger(firstRequest));
            assertEquals(getValueFromLog(logs[7]), new BigInteger(lastRequest));
        }
    }

	public BigInteger getValueFromLog(String log) {
		Scanner scanner = new Scanner(log);

        scanner.useDelimiter("\"");
        scanner.findInLine("\"value\":");
        BigInteger value = scanner.nextBigInteger();
		return value;
	}

    private boolean isDefaultRegisterVersion() {
        return request.getRegisterVersion() == null || request.getRegisterVersion() == 1;
    }
}
