/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata.replication.dao;

import com.google.inject.Inject;
import com.google.inject.Provider;
import com.trifork.stamdata.persistence.PersistenceModule;
import dk.nsi.stamdata.replication.vo.ViewMapVO;
import org.apache.commons.lang.StringUtils;

import java.sql.*;
import java.util.*;
import java.util.Date;
import java.util.stream.Collectors;

public class FilterManagementDAO {

    private static final String TABLE_FILTER_CONTENT = "Filter_content";
    private static final String FIELDS_FILTER_CONTENT = getFieldNames(FilterContentEnum.class);
    private static final Integer NUMBER_OF_FIELDS_IN_FILTER_CONTENT = FilterContentEnum.values().length;

    private static final String TABLE_FILTER_FIELDS = "Filter_fields";
    private static final String FIELDS_FILTER_FIELDS = getFieldNames(FilterFieldsEnum.class);
    private static final Integer NUMBER_OF_FIELDS_IN_FILTER_FIELDS = FilterFieldsEnum.values().length;

    private final Provider<Connection> connectionProvider;

    public static String getTableFilterContent() {
        return TABLE_FILTER_CONTENT;
    }

    public static String getFieldsFilterContent() {
        return FIELDS_FILTER_CONTENT;
    }

    public static Integer getNumberOfFieldsInFilterContent() {
        return NUMBER_OF_FIELDS_IN_FILTER_CONTENT;
    }

    public static String getTableFilterFields() {
        return TABLE_FILTER_FIELDS;
    }

    public static String getFieldsFilterFields() {
        return FIELDS_FILTER_FIELDS;
    }

    public static Integer getNumberOfFieldsInFilterFields() {
        return NUMBER_OF_FIELDS_IN_FILTER_FIELDS;
    }

    @Inject
    public FilterManagementDAO(Provider<Connection> connectionProvider) {
        this.connectionProvider = connectionProvider;
    }

    public void createFilter(String cvr, String filterName, ViewMapVO viewMapVOHolder, List<String> fieldList) throws SQLException {
        Connection connection = null;
        FilterContetnInfo filterContetnInfo = null;
        long filterContentID = 0L;
        try {
            // Get connection
            connection = connectionProvider.get();

            // If filtered exists get 'Id' and 'created' date
            filterContetnInfo = retrieveFilterFilterContetnInfo(cvr, filterName, connection);

            // Check if the filter exists, if found take created time and delete it after
            Timestamp createDate = null;
            if (filterContetnInfo != null) {
                filterContentID = filterContetnInfo.Id;
                createDate = filterContetnInfo.timestamp;
                deleteFilter(cvr, filterName, filterContentID, connection);
            } else {
                createDate = new Timestamp((new java.util.Date()).getTime());
            }

            // Insert the new filter intro Filter_Content
            filterContentID = insertFilterToFilterContent(cvr, filterName, createDate, viewMapVOHolder, connection);

            // Insert the fields intro Filder_Fields
            insertFieldsToFilterFields(filterContentID, createDate, fieldList, connection);
        } finally {
            PersistenceModule.close(connection);
        }
    }

    public FilterContetnInfo getFilterFilterId(String cvr, String filterName) throws SQLException
    {
        Connection connection = null;
        try {
            connection = connectionProvider.get();
            return retrieveFilterFilterContetnInfo(cvr, filterName, connection);
        } finally {
            PersistenceModule.close(connection);
        }
    }

    public FilterContetnInfo retrieveFilterFilterContetnInfo(String cvr, String filterName, Connection connection) throws SQLException
    {
        ResultSet resultSet = null;
        PreparedStatement preparedStatement = null;
        try {
            String queryString = "SELECT Id, " + FilterContentEnum.created.name() + " " +
                    "FROM " + TABLE_FILTER_CONTENT + " " +
                    "WHERE " + FilterContentEnum.cvr.name() + " = ? " +
                    "AND " + FilterContentEnum.name.name() + " = ? " +
                    "LIMIT 1 ";
            preparedStatement = connection.prepareStatement(queryString);
            preparedStatement.setString(1, cvr);
            preparedStatement.setString(2, filterName);
            resultSet = preparedStatement.executeQuery();

            if (resultSet.next()) {
                FilterContetnInfo filterContetnInfo = new FilterContetnInfo();
                filterContetnInfo.Id = resultSet.getLong("Id");
                filterContetnInfo.timestamp = resultSet.getTimestamp(FilterContentEnum.created.name());
                return filterContetnInfo;
            }
            return null;
        } finally {
            if (resultSet != null) resultSet.close();
            PersistenceModule.close(preparedStatement);
        }
    }

    public long insertFilterToFilterContent(String cvr, String filterName, Timestamp createDate, ViewMapVO viewMapVOHolder, Connection connection) throws SQLException
    {
        PreparedStatement preparedStatement = null;
        ResultSet generatedKeys = null;
        try {
            String queryString = "INSERT INTO " + TABLE_FILTER_CONTENT + " (" + FIELDS_FILTER_CONTENT + ") " +
                    "VALUES (" + generateSqlParameterPlaceholders(NUMBER_OF_FIELDS_IN_FILTER_CONTENT) + ") ";
            preparedStatement = connection.prepareStatement(queryString, Statement.RETURN_GENERATED_KEYS);

            // Set values for insert
            preparedStatement.setString(FilterContentEnum.getIndex(FilterContentEnum.cvr), cvr);
            preparedStatement.setString(FilterContentEnum.getIndex(FilterContentEnum.name), filterName);
            preparedStatement.setString(FilterContentEnum.getIndex(FilterContentEnum.register), viewMapVOHolder.getRegister());
            preparedStatement.setString(FilterContentEnum.getIndex(FilterContentEnum.datatype), viewMapVOHolder.getDatatype());
            preparedStatement.setLong(FilterContentEnum.getIndex(FilterContentEnum.version), viewMapVOHolder.getVersion());
            preparedStatement.setLong(FilterContentEnum.getIndex(FilterContentEnum.registerVersion), viewMapVOHolder.getRegisterVersion());
            preparedStatement.setTimestamp(FilterContentEnum.getIndex(FilterContentEnum.created), createDate);
            preparedStatement.setTimestamp(FilterContentEnum.getIndex(FilterContentEnum.modified), new Timestamp((new java.util.Date()).getTime()));
            preparedStatement.executeUpdate();

            // Retrieve the 'Id' (generated keys)
            generatedKeys = preparedStatement.getGeneratedKeys();
            if (generatedKeys.next()) {
                return generatedKeys.getLong(1);
            } else {
                throw new SQLException("No ID generated for the inserted filter.");
            }
        } finally {
            if (generatedKeys != null) generatedKeys.close();
            PersistenceModule.close(preparedStatement);
        }
    }

    public void insertFieldsToFilterFields(long filterContentID, Timestamp createDate, List<String> fieldList, Connection connection) throws SQLException {
        PreparedStatement preparedStatement = null;
        try {
            String queryString = "INSERT INTO " + TABLE_FILTER_FIELDS + " (" + FIELDS_FILTER_FIELDS + ") " +
                    "VALUES" + generateMultiInsertSqlPlaceholders(NUMBER_OF_FIELDS_IN_FILTER_FIELDS, fieldList.size()) + " ";
            preparedStatement = connection.prepareStatement(queryString);

            // Set values for insert
            for (int i = 0; i < fieldList.size(); i++) {
                // Adjust index for each insert, So the preparedStatement.setX will hit the next insert
                int parameterIndex = i * NUMBER_OF_FIELDS_IN_FILTER_FIELDS;
                preparedStatement.setLong(FilterFieldsEnum.getIndex(FilterFieldsEnum.filter_id) + parameterIndex, filterContentID);
                preparedStatement.setString(FilterFieldsEnum.getIndex(FilterFieldsEnum.field) + parameterIndex, fieldList.get(i));
                preparedStatement.setTimestamp(FilterFieldsEnum.getIndex(FilterFieldsEnum.created) + parameterIndex, createDate);
                preparedStatement.setDate(FilterFieldsEnum.getIndex(FilterFieldsEnum.modified) + parameterIndex, new java.sql.Date(new Date().getTime()));
            }
            preparedStatement.executeUpdate();
        } finally {
            PersistenceModule.close(preparedStatement);
        }
    }

    public void deleteFilter(String cvr, String filterName) throws SQLException {
        Connection connection = null;
        FilterContetnInfo filterContetnInfo = null;
        try {
            // Get connection
            connection = connectionProvider.get();

            // If filtered exists get 'Id' and 'created' date
            filterContetnInfo = retrieveFilterFilterContetnInfo(cvr, filterName, connection);

            // Check if the filter exists, if found take created time and delete it after
            if (filterContetnInfo != null) {
                long filterContentID = filterContetnInfo.Id;
                deleteFilter(cvr, filterName, filterContentID, connection);
            } else {
                throw new SQLException("No filter was found with cvr: " + cvr + " and name: " + filterName + ".");
            }
        } finally {
            PersistenceModule.close(connection);
        }
    }

    private void deleteFilter(String cvr, String filterName, long filterContentID, Connection connection) throws SQLException
    {
        PreparedStatement preparedStatementFields = null;
        PreparedStatement preparedStatementContent = null;
        try {
            // If a filter with the same cvr + name exists, it is deleted
            String deleteFieldsQueryString =
                    "DELETE FROM " + TABLE_FILTER_FIELDS + " " +
                    "WHERE " + FilterFieldsEnum.filter_id.name() + " = ?; ";
            preparedStatementFields = connection.prepareStatement(deleteFieldsQueryString);
            preparedStatementFields.setLong(1, filterContentID);
            int deleteFields = preparedStatementFields.executeUpdate();
            connection.commit();

            String deleteContentQueryString =
                    "DELETE FROM " + TABLE_FILTER_CONTENT + " " +
                            "WHERE " + FilterContentEnum.cvr.name() + " = ? " +
                            "AND " + FilterContentEnum.name.name() + " = ? ";

            preparedStatementContent = connection.prepareStatement(deleteContentQueryString);
            preparedStatementContent.setString(1, cvr);
            preparedStatementContent.setString(2, filterName);
            int deleteContent = preparedStatementContent.executeUpdate();
            connection.commit();
        } finally {
            PersistenceModule.close(preparedStatementFields);
            PersistenceModule.close(preparedStatementContent);
        }
    }

    public HashMap<String, Boolean> fieldsExistInSKRS(String datatype, List<String> fieldList) throws SQLException {
        HashMap<String, Boolean> result = new HashMap<>();
        PreparedStatement preparedStatement = null;
        Connection connection = null;
        ResultSet resultSet = null;
        try {
            // Get connection
            connection = connectionProvider.get();

            // Select from table where in fieldList
            String queryString = "SELECT v.datatype, c.feedColumnName FROM SKRSColumns as c " +
                    "INNER JOIN SKRSViewMapping as v ON v.idSKRSViewMapping = c.viewMap " +
                    "WHERE v.datatype = '" + datatype + "' " +
                    "AND c.feedColumnName in (" + generateSqlParameterPlaceholders(fieldList.size()) + ") ";
            preparedStatement = connection.prepareStatement(queryString);

            // Add fields name to preparedStatement
            for (int i = 0; i < fieldList.size(); i++) {
                preparedStatement.setString(i + 1, fieldList.get(i));
            }

            // Get the resultSet from query
            resultSet = preparedStatement.executeQuery();

            // Add fields found in resultSet to true in the end result
            while (resultSet.next()) {
                String field = resultSet.getString("feedColumnName");
                result.put(field, true);
            }

            // Mark fields that was not found as false in the end result
            for (String field : fieldList) {
                if (!result.containsKey(field)) {
                    result.put(field, false);
                }
            }
        } finally {
            PersistenceModule.close(preparedStatement);
            PersistenceModule.close(connection);
        }
        return result;
    }

    public List<String> getFieldsNotInSKRS(String datatype, List<String> fieldList) throws SQLException {
        HashMap<String, Boolean> existingFieldsMap = fieldsExistInSKRS(datatype, fieldList);
        List<String> fieldsNotInSKRS = existingFieldsMap.entrySet().stream()
                .filter(entry -> !entry.getValue())
                .map(Map.Entry::getKey)
                .collect(Collectors.toList());
        return fieldsNotInSKRS;
    }

    public List<String> getFilterFieldsFromDatabase(long filterId) throws SQLException {
        List<String> result = new ArrayList<>();
        PreparedStatement preparedStatement = null;
        Connection connection = null;
        ResultSet resultSet = null;
        try {
            // Get connection
            connection = connectionProvider.get();

            // Select from table where in fieldList
            String queryString = "SELECT * FROM " + TABLE_FILTER_FIELDS + " WHERE " + FilterFieldsEnum.filter_id.name() + " = ? ";
            preparedStatement = connection.prepareStatement(queryString);
            preparedStatement.setLong(1, filterId);
            resultSet = preparedStatement.executeQuery();

            // Add fields found in resultSet to true in the end result
            while (resultSet.next()) {
                String field = resultSet.getString(FilterFieldsEnum.field.name());
                result.add(field);
            }
        } finally {
            PersistenceModule.close(preparedStatement);
            if (resultSet != null) resultSet.close();
            PersistenceModule.close(connection);
        }
        return result;
    }

    public final class FilterContetnInfo {
        public long Id;
        public Timestamp timestamp;
    }

    public enum FilterContentEnum {
        cvr,
        name,
        register,
        datatype,
        version,
        registerVersion,
        created,
        modified;
        public static int getIndex(FilterContentEnum filterContentEnum) {
            return filterContentEnum.ordinal()+1;
        }
    }

    public enum FilterFieldsEnum {
        filter_id,
        field,
        created,
        modified;
        public static int getIndex(FilterFieldsEnum filterFieldsEnum) {
            return filterFieldsEnum.ordinal()+1;
        }
    }
    public static String getFieldNames(Class<? extends Enum<?>> enumClass) {
        Enum<?>[] fields = enumClass.getEnumConstants();
        String[] fieldNames = new String[fields.length];
        for (int i = 0; i < fields.length; i++) {
            fieldNames[i] = fields[i].toString();
        }
        return String.join(", ", fieldNames);
    }

    public static <T extends Enum<T>> int getIndex(Enum<T>[] values, T value) {
        for (int i = 0; i < values.length; i++) {
            if (values[i] == value) {
                return i;
            }
        }
        return -1; // If value not found
    }

    private String generateSqlParameterPlaceholders(int count) {
        String placeholders = StringUtils.repeat("?, ", count);
        placeholders = StringUtils.removeEnd(placeholders, ", ");
        return placeholders;
    }

    private String generateMultiInsertSqlPlaceholders(int parameterCount, int numberOfInserts) {
        String insertsString = StringUtils.repeat("(" + generateSqlParameterPlaceholders(parameterCount) + "), ", numberOfInserts);
        insertsString = StringUtils.removeEnd(insertsString, ", ");
        return insertsString;
    }
}
