/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata.replication.webservice;

import com.google.inject.Inject;
import com.google.inject.Provider;
import dk.nsi.stamdata.replication.dao.FilterManagementDAO;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import static com.trifork.stamdata.Preconditions.checkNotNull;

public class StamdataBachCopyFilter {

    private Provider<Connection> connectionProvider;

    @Inject
    public StamdataBachCopyFilter(Provider<Connection> connectionProvider) {
        this.connectionProvider = checkNotNull(connectionProvider, "connectionProvider");
    }

    public List<String> getFilterFieldsForReplicationRequest(String cvr, ReplicationRequest parameters) throws StamdataReplicationFault {
        List<String> fieldsToKeep = new ArrayList<>();

        // Check if the filter should be used
        if (!shouldFilterBeUsed(parameters)) return fieldsToKeep;

        // Get the filter content
        FilterManagementDAO filterManagementDAO = new FilterManagementDAO(connectionProvider);
        FilterManagementDAO.FilterContetnInfo filterContentInfo = null;
        try {
            filterContentInfo = filterManagementDAO.getFilterFilterId(cvr, parameters.getFilterName());
        } catch (SQLException e) {
            throw new StamdataReplicationFault("Error getting filter from database.", FaultCodes.INTERNAL_ERROR);
        }

        // Check if the filter is valid
        if (filterContentInfo == null) {
            throw new StamdataReplicationFault("Invalid filter with name: " + parameters.getFilterName() + " and CVR: " + cvr, FaultCodes.INVALID_FILTER);
        }

        // Check if the filter fields are valid
        try {
            fieldsToKeep = filterManagementDAO.getFilterFieldsFromDatabase(filterContentInfo.Id);
        } catch (SQLException e) {
            throw new StamdataReplicationFault("Error getting filter fields. Error: " + e, FaultCodes.INTERNAL_ERROR);
        }
        if (fieldsToKeep.isEmpty()) {
            throw new StamdataReplicationFault("Filter fields was Empty", FaultCodes.EMPTY_FILTER_Fields);
        }
        return fieldsToKeep;
    }

    public Document getFilteredDocument(ReplicationRequest parameters, Document document, List<String> fieldsToKeep) throws StamdataReplicationFault {
        // Check if the filter should be used
        if (fieldsToKeep.isEmpty()) return document;

        // Filter the fields
        try {
            document = removeUnwantedNodes(document, parameters, fieldsToKeep);
        } catch (Exception e) {
            throw new StamdataReplicationFault("Error occurred while filtering fields in the document.", FaultCodes.ERROR_FILTER_Fields);
        }
        return document;
    }

    private boolean shouldFilterBeUsed(ReplicationRequest parameters) {
        // Check if the request is a RFS request
        if (parameters.getRequestType() != RequestType.RFS) return false;
        // Check if there is a filter name
        if (parameters.getFilterName() == null) return false;

        return true;
    }

    public Document removeUnwantedNodes(Document document, ReplicationRequest parameters, List<String> fieldsToKeep) {
        // Get all <dataType:dataType> elements
        String stamdataNamespaceUri = DynamicViewXmlGenerator.STAMDATA_NAMESPACE_URI_PREFIX + parameters.getRegister();
        NodeList dataTypeNodes = document.getElementsByTagNameNS(stamdataNamespaceUri, parameters.getDatatype());

        for (int i = 0; i < dataTypeNodes.getLength(); i++) {
            // Get the <dataType:dataType> element
            Element personElement = (Element) dataTypeNodes.item(i);

            // Get all child elements of <dataType:dataType>
            NodeList childNodes = personElement.getChildNodes();

            // Iterate over the child nodes in reverse order to safely remove them
            for (int j = childNodes.getLength() - 1; j >= 0; j--) {
                Node childNode = childNodes.item(j);

                // Check if the child node is an Element
                if (childNode.getNodeType() != Node.ELEMENT_NODE) {
                    continue;
                }

                // If the child element's name is not in the list of elements to keep, remove it
                Element childElement = (Element) childNode;
                if (!fieldsToKeep.contains(childElement.getLocalName())) {
                    personElement.removeChild(childElement);
                }

            }
        }
        return document;
    }
}
