/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

package dk.nsi.stamdata.replication.webservice;

import com.google.inject.Inject;
import com.google.inject.Provider;
import com.trifork.stamdata.util.SlaConsts;
import dk.medcom.dgws._2006._04.dgws_1_0.Header;
import dk.nsi.stamdata.replication.constants.ReplicationConstants;
import dk.nsi.stamdata.replication.dynamic.DynamicRow;
import dk.nsi.stamdata.replication.dynamic.DynamicRowFetcher;
import dk.nsi.stamdata.replication.dynamic.DynamicViewMapper;
import dk.nsi.stamdata.replication.models.ClientDao;
import dk.nsi.stamdata.replication.models.ClientVO;
import dk.nsi.stamdata.replication.vo.ViewMapVO;
import dk.nsi.stamdata.security.RequestContextVO;
import dk.sds.nsp.audit.Audit;
import dk.sds.nsp.audit.AuditBuilder;
import dk.sdsd.nsp.slalog.api.SLALogItem;
import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;
import org.apache.log4j.MDC;
import org.joda.time.DateTime;
import org.joda.time.Instant;
import org.oasis_open.docs.wss._2004._01.oasis_200401_wss_wssecurity_secext_1_0.Security;
import org.oasis_open.docs.wss._2004._01.oasis_200401_wss_wssecurity_utility_1_0.Timestamp;
import org.w3c.dom.Document;

import javax.xml.transform.TransformerException;
import javax.xml.ws.Holder;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.Calendar;
import java.util.List;
import static java.lang.String.format;

public class StamdataReplicator {
    private static final Logger logger = Logger.getLogger(StamdataReplicator.class);

    private static final int MAX_RECORD_LIMIT = 2000;

    private final ClientDao clients;
    private final SLALogItem slaLogItem;
    private StamdataBachCopyFilter stamdataBachCopyFilter;

    private String cvr;

    @Inject
    private DynamicViewMapper dynamicViewMapper;

    @Inject
    private DynamicRowFetcher dynamicRowFetcher;

    @Inject
    private DynamicViewXmlGenerator dynamicViewXmlGenerator;

    private Provider<Connection> connectionProvider;

    @Inject
    StamdataReplicator(ClientDao clientDao, SLALogItem slaLogItem, Provider<Connection> connectionProvider)
    {
        this.clients = clientDao;
        this.slaLogItem = slaLogItem;
        this.connectionProvider = connectionProvider;
    }

    public Document replicate(Holder<Security> wsseHeader, Holder<Header> medcomHeader, ReplicationRequest parameters, RequestContextVO requestContext) throws StamdataReplicationFault {
        cvr = requestContext.getServiceActor().getOrganizationId();
        boolean includeRegisterVersion = true;
        if(parameters.getRegisterVersion() == null) {
            includeRegisterVersion = false;
            parameters.setRegisterVersion(ReplicationConstants.DEFAULT_REGISTER_VERSION);
        }

        Header value = medcomHeader.value;
        if (value.getLinking() != null && value.getLinking().getMessageID() != null) {
            String messageID = value.getLinking().getMessageID();
            slaLogItem.setMessageId(messageID);
            slaLogItem.setSourceOperation("replicate");
        }

        slaLogItem.addCallParameter(SlaConsts.SLA_PARAM_CVR, cvr);
        slaLogItem.addCallParameter(SlaConsts.SLA_PARAM_DATATYPE, parameters.getDatatype());
        slaLogItem.addCallParameter(SlaConsts.SLA_PARAM_REGISTER, parameters.getRegister());
        slaLogItem.addCallParameter(SlaConsts.SLA_PARAM_DATATYPE_VERSION, String.valueOf(parameters.getVersion()));
        slaLogItem.addCallParameter(SlaConsts.SLA_PARAM_REGISTER_VERSION, String.valueOf(parameters.getRegisterVersion()));
        slaLogItem.addCallParameter(SlaConsts.SLA_PARAM_FILTER, parameters.getFilterName());

        stamdataBachCopyFilter = new StamdataBachCopyFilter(connectionProvider);
        List<String> fieldsToKeep = stamdataBachCopyFilter.getFilterFieldsForReplicationRequest(cvr, parameters);
        auditLogInput(parameters, fieldsToKeep);


        // Replace the client request securityheader
        Security security = new Security();
        Timestamp ts = new Timestamp();
        ts.setCreated(Calendar.getInstance());
        security.setTimestamp(ts);
        wsseHeader.value = security;

        try
        {
            return handleRequestUsingDynamicViews(wsseHeader, medcomHeader, parameters, includeRegisterVersion, fieldsToKeep);
        }
        catch (StamdataReplicationFault e)
        {
            // Log an throw is normally an anti-pattern, but since
            // exceptions are part of JAX-WS's flow it is "OK" here.
            //
            logger.warn("The request could not be handled. This is likely the clients mistake.", e);

            throw e;
        }
        catch (Exception e)
        {
            throw new StamdataReplicationFault("An unhandled error occurred.", FaultCodes.INTERNAL_ERROR, e);
        }
    }

    private void auditLogInput(ReplicationRequest parameters, List<String> fieldsToKeep) {
        AuditBuilder.TypeOfInformation informationType = AuditBuilder.TypeOfInformation.SensitivePersonalInformation;
        String component = "SDM";
        String context = String.format("S%s Replicate", parameters.getRequestType().toString());

        AuditBuilder auditBuilder = Audit.createAuditBuilder();
        auditBuilder.addAuditInformation(component, context, informationType, "register", parameters.getRegister());
        auditBuilder.addAuditInformation(component, context, informationType, "datatype", parameters.getDatatype());
        auditBuilder.addAuditInformation(component, context, informationType, "version", parameters.getVersion());
        auditBuilder.addAuditInformation(component, context, informationType, "registerVersion", parameters.getRegisterVersion());

        if (parameters.getFilterName() != null) {
            auditBuilder.addAuditInformation(component, context, informationType, "filter", parameters.getFilterName());
            auditBuilder.addAuditInformation(component, context, informationType, "fieldsShowInResult", fieldsToKeep);
        }

        switch(parameters.getRequestType()) {
            case KRS:
                auditLogKrsInput(auditBuilder, component, context, informationType, parameters);
                break;
            case RFS:
                auditLogRfsInput(auditBuilder, component, context, informationType, parameters);
                break;
            default:
                throw new IllegalArgumentException(String.format("Unknown RequestType: %s", parameters.getRequestType()));
        }
    }

    private void auditLogKrsInput(AuditBuilder auditBuilder, String component, String context, AuditBuilder.TypeOfInformation informationType, ReplicationRequest parameters) {
        if (parameters.getMaxRecords() != null) {
            auditBuilder.addAuditInformation(component, context, informationType, "max records", parameters.getMaxRecords());
        } else {
            auditBuilder.addAuditInformation(component, context, informationType, "max records", "Undefined");
        }
        auditBuilder.addAuditInformation(component, context, informationType, "offset", parameters.getOffset());
    }

    private void auditLogRfsInput(AuditBuilder auditBuilder, String component, String context, AuditBuilder.TypeOfInformation informationType, ReplicationRequest parameters) {
        auditBuilder.addAuditInformation(component, context, informationType, "idlist", parameters.getIdList());
    }

    private Document handleRequestUsingDynamicViews(Holder<Security> wsseHeader, Holder<Header> medcomHeader,
                                                    ReplicationRequest parameters, boolean includeRegisterVersion, List<String> fieldsToKeep) throws StamdataReplicationFault {
        try {
            String viewPath = getViewPath(parameters);

            MDC.put("view", String.valueOf(viewPath));
            MDC.put("cvr", String.valueOf(cvr));

            // Start by checking if the view exist so we can respond with a better error than not authorized.
            if (!dynamicViewMapper.viewMapExists(parameters.getRegister(), parameters.getDatatype(), parameters.getVersion(), parameters.getRegisterVersion())) {
                throw new StamdataReplicationFault("Datatype version ikke fundet: " + viewPath, FaultCodes.UNKNOWN_VIEW);
            }

            ClientVO client = clients.findByCvr(cvr);
            if (client == null || !client.isAuthorizedFor(viewPath)) {
                throw new StamdataReplicationFault("Dette cvr nummer (" + String.valueOf(cvr) + ") er ikke autoriseret til at hente denne type data (" + String.valueOf(viewPath) + ").", FaultCodes.UNAUTHORIZED);
            }

            addParametersToMDC(parameters);

            ////////////////////
            // Do actual work here
            ViewMapVO view = dynamicViewMapper.getViewMapForView(parameters.getRegister(), parameters.getDatatype(), parameters.getVersion(), parameters.getRegisterVersion());

            Document feedDocument = createFeed(view, parameters, includeRegisterVersion);

            // Filter the feedDocument if needed
            feedDocument = stamdataBachCopyFilter.getFilteredDocument(parameters, feedDocument, fieldsToKeep);

            return feedDocument;
        } catch (StamdataReplicationFault e) {
            throw e;
        } catch (Exception e) {
            throw new StamdataReplicationFault("Could not complete the request do to an error.", FaultCodes.INTERNAL_ERROR, e);
        } finally {
            MDC.remove("view");
            MDC.remove("cvr");
            MDC.remove("offset");
            MDC.remove("limit");
        }
    }

    private void addParametersToMDC(ReplicationRequest parameters) throws StamdataReplicationFault {
        switch(parameters.getRequestType()) {
            case KRS:
                HistoryOffset offset = getOffset(parameters);
                int limit = getRecordLimit(parameters);

                MDC.put("offset", String.valueOf(offset));
                MDC.put("limit", String.valueOf(limit));
                break;
            case RFS:
                if(parameters.getIdList() != null) {
                    MDC.put("idlist", parameters.getIdList());
                }
                break;
            default:
                throw new IllegalArgumentException(String.format("Unknown RequestType: %s", parameters.getRequestType()));
        }
    }

    private int getRecordLimit(ReplicationRequest parameters)
    {
        if (parameters.getMaxRecords() == null) {
            return MAX_RECORD_LIMIT;
        } else {
            return Math.min(parameters.getMaxRecords().intValue(), MAX_RECORD_LIMIT);
        }
    }

    private Document createFeed(ViewMapVO view, ReplicationRequest parameters, boolean includeRegisterVersion) throws StamdataReplicationFault {
        try {
            List<DynamicRow> rows = null;
            switch(parameters.getRequestType()) {
                case KRS:
                    HistoryOffset offset = getOffset(parameters);
                    int limit = getRecordLimit(parameters);

                    long pid = Long.parseLong(offset.getRecordID());
                    Instant modifiedDate = new Instant(offset.getModifiedDate());

                    rows = dynamicRowFetcher.fetchRows(view, pid, modifiedDate, limit);
                    break;
                case RFS:
                    rows = dynamicRowFetcher.fetchRows(view, parameters.getIdList());
                    break;
                default:
                    throw new IllegalArgumentException(String.format("Unknown RequestType: %s", parameters.getRequestType()));
            }
            auditLogOutput(rows);
            return dynamicViewXmlGenerator.generateXml(view, rows, parameters.getRegister(), parameters.getDatatype(), DateTime.now(), includeRegisterVersion);
        } catch (SQLException e) {
            throw new RuntimeException("Database error", e);
        } catch (TransformerException e) {
            throw new RuntimeException("Transformer error", e);
        }
    }

    private void auditLogOutput(List<DynamicRow> rows) {
        AuditBuilder.TypeOfInformation informationType = AuditBuilder.TypeOfInformation.SensitivePersonalInformation;
        String component = "SDM";
        String context = "SKRS Replicate";

        AuditBuilder auditBuilder = Audit.createAuditBuilder();
        auditBuilder.addAuditInformation(component, context, informationType, "returned elements", rows.size());
        if (rows.size() > 0) {
            auditBuilder.addAuditInformation(component, context, informationType, "returned first element", DynamicViewXmlGenerator.mapRevisionId(rows.get(0)));
            auditBuilder.addAuditInformation(component, context, informationType, "returned last element", DynamicViewXmlGenerator.mapRevisionId(rows.get(rows.size()-1)));
        }
    }

    private String getViewPath(ReplicationRequest parameters)
    {
        if(parameters.getRegisterVersion() != null && parameters.getRegisterVersion() != ReplicationConstants.DEFAULT_REGISTER_VERSION) {
            return format("%s/v%d/%s/v%d", parameters.getRegister(), parameters.getRegisterVersion(), parameters.getDatatype(), parameters.getVersion());
        }
        return format("%s/%s/v%d", parameters.getRegister(), parameters.getDatatype(), parameters.getVersion());
    }

    private HistoryOffset getOffset(ReplicationRequest parameters) throws StamdataReplicationFault
    {
        try {
            return new HistoryOffset(parameters.getOffset());
        } catch (Exception e) {
            throw new StamdataReplicationFault("Invalid offset in the request.", FaultCodes.INVALID_OFFSET, e);
        }
    }
}
