/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata.cpr.dao;

import java.sql.*;
import java.time.LocalDateTime;
import java.util.*;
import java.util.Date;
import java.util.stream.Collectors;

import org.apache.commons.lang.StringUtils;

import com.google.inject.Inject;
import com.google.inject.Provider;
import com.trifork.stamdata.models.PersonVO;
import com.trifork.stamdata.persistence.PersistenceModule;

public class CprDAO {
	//alders beregning kopieret fra minlog2 (PersonStamDB.lookupAge)
    private final String SQL_PERSON = "CPR, Koen, Fornavn, FornavnMarkering, Mellemnavn, MellemnavnMarkering, Efternavn, EfternavnMarkering, CoNavn, Lokalitet, StandardAdresse, Vejnavn, Bygningsnummer, Husnummer, Etage, SideDoerNummer, " +
            "Bynavn, Postnummer, PostDistrikt, Status, NavneBeskyttelseStartDato, NavneBeskyttelseSletteDato, GaeldendeCPR, Foedselsdato, TIMESTAMPDIFF(YEAR, Foedselsdato, CURDATE()) Alder, Stilling, VejKode, KommuneKode, NavnTilAdressering, VejnavnTilAdressering, " +
            "FoedselsdatoMarkering, StatusDato, StatusMarkering, StartDato, StartDatoMarkering, SlutDato, SlutDatoMarkering, NavnStartDato, NavnStartDatoMarkering, AdresseringsNavn, Unreliable, " +
            "CreatedDate, ModifiedDate, ValidFrom, ValidTo " +
            "FROM v2_Person ";

    public enum WardRelation {
        BARN("barn"),
        BARN_UDEN("barn_uden"),
        MYNDLING("myndling");
        private final String value;

        WardRelation(String v) {
            value = v;
        }

        public String value() {
            return value;
        }

        public static WardRelation fromValue(String value) {
            for (WardRelation ward : WardRelation.values()) {
                if (ward.value.equals(value)) {
                    return ward;
                }
            }
            throw new IllegalArgumentException(value);
        }
    }

    public class Ward {
        String identifier;
        WardRelation relation;

        @Override
        public String toString() {
            return "Ward [identifier=" + identifier + ", relation=" + relation + "]";
        }

        public String getIdentifier() {
            return identifier;
        }

        public void setIdentifier(String identifier) {
            this.identifier = identifier;
        }

        public WardRelation getRelation() {
            return relation;
        }

        public void setRelation(WardRelation relation) {
            this.relation = relation;
        }
    }

    public enum CustodyRelation {
        MOR("mor"),
        MOR_UDEN("mor_uden"),
        FAR("far"),
        FAR_UDEN("far_uden"),
        ANDEN("anden"),
        VAERGE("vaerge");
        private final String value;

        CustodyRelation(String v) {
            value = v;
        }

        public String value() {
            return value;
        }

        public static CustodyRelation fromValue(String value) {
            for (CustodyRelation custody : CustodyRelation.values()) {
                if (custody.value.equals(value)) {
                    return custody;
                }
            }
            throw new IllegalArgumentException(value);
        }
    }

    public class Custody {
        String identifier;
        CustodyRelation relation;

        public String getIdentifier() {
            return identifier;
        }

        public void setIdentifier(String identifier) {
            this.identifier = identifier;
        }

        public CustodyRelation getRelation() {
            return relation;
        }

        public void setRelation(CustodyRelation relation) {
            this.relation = relation;
        }

        @Override
        public String toString() {
            return "Custody [identifier=" + identifier + ", relation=" + relation + "]";
        }
    }

    private final Provider<Connection> connectionProvider;

    @Inject
    public CprDAO(Provider<Connection> connectionProvider) {
        this.connectionProvider = connectionProvider;
    }

    public List<String> findRelationsAsGivingAuthority(String cpr) throws SQLException {
        return findRelationsAsGivingAuthority(cpr, new Date());
    }

    public List<String> findRelationsAsGivingAuthority(String cpr, Date currentTime) throws SQLException {
        PreparedStatement preparedStatement = null;
        Connection connection = connectionProvider.get();
        try {
            List<String> relations = new ArrayList<String>();

            // Query første forældre
            String barnMotherQuery = "select pf.MorCPR from PersonFamilie as pf join v2_ForaeldreMyndighedRelation as fmr on fmr.CPR = pf.CPR where pf.CPR = ? and fmr.TypeKode = '0003' and fmr.ValidFrom <= ? and (fmr.ValidTo > ? or fmr.ValidTo is null) and pf.ValidFrom <= ? and (pf.ValidTo > ? or pf.ValidTo is null)";
            preparedStatement = connection.prepareStatement(barnMotherQuery);
            preparedStatement.setObject(1, cpr);
            preparedStatement.setObject(2, currentTime);
            preparedStatement.setObject(3, currentTime);
            preparedStatement.setObject(4, currentTime);
            preparedStatement.setObject(5, currentTime);
            ResultSet resultSet = preparedStatement.executeQuery();
            while (resultSet.next()) {
                relations.add(resultSet.getString("MorCPR"));
            }

            // Query anden forældre
            String barnFatherQuery = "select pf.FarCPR from PersonFamilie as pf join v2_ForaeldreMyndighedRelation as fmr on fmr.CPR = pf.CPR where pf.CPR = ? and fmr.TypeKode = '0004' and fmr.ValidFrom <= ? and (fmr.ValidTo > ? or fmr.ValidTo is null)  and pf.ValidFrom <= ? and (pf.ValidTo > ? or pf.ValidTo is null)";
            preparedStatement = connection.prepareStatement(barnFatherQuery);
            preparedStatement.setObject(1, cpr);
            preparedStatement.setObject(2, currentTime);
            preparedStatement.setObject(3, currentTime);
            preparedStatement.setObject(4, currentTime);
            preparedStatement.setObject(5, currentTime);
            resultSet = preparedStatement.executeQuery();
            while (resultSet.next()) {
                relations.add(resultSet.getString("FarCPR"));
            }

            // Query forældremyndighed
            String autorityQuery = "select fmr.RelationCpr as CPR from v2_ForaeldreMyndighedRelation as fmr where fmr.CPR = ? and (fmr.TypeKode = '0005' OR fmr.TypeKode = '0006') and fmr.ValidFrom <= ? and (fmr.ValidTo > ? or fmr.ValidTo is null)";
            preparedStatement = connection.prepareStatement(autorityQuery);
            preparedStatement.setObject(1, cpr);
            preparedStatement.setObject(2, currentTime);
            preparedStatement.setObject(3, currentTime);
            resultSet = preparedStatement.executeQuery();
            while (resultSet.next()) {
                relations.add(resultSet.getString("CPR"));
            }

            return relations;
        } finally {
            PersistenceModule.close(preparedStatement);
            PersistenceModule.close(connection);
        }
    }
    
    public List<String> findRelationsAsHavingAuthority(String relationCpr, Date currentTime) throws SQLException {
        PreparedStatement preparedStatement = null;
        Connection connection = connectionProvider.get();
        try {
            List<String> relations = new ArrayList<String>();

            // Query første forældre
            String barnMotherQuery = "select pf.CPR from PersonFamilie as pf join v2_ForaeldreMyndighedRelation as fmr on fmr.CPR = pf.CPR where pf.MorCPR = ? and fmr.TypeKode = '0003' and fmr.ValidFrom <= ? and (fmr.ValidTo > ? or fmr.ValidTo is null) and pf.ValidFrom <= ? and (pf.ValidTo > ? or pf.ValidTo is null)";
            preparedStatement = connection.prepareStatement(barnMotherQuery);
            preparedStatement.setObject(1, relationCpr);
            preparedStatement.setObject(2, currentTime);
            preparedStatement.setObject(3, currentTime);
            preparedStatement.setObject(4, currentTime);
            preparedStatement.setObject(5, currentTime);
            ResultSet resultSet = preparedStatement.executeQuery();
            while (resultSet.next()) {
                relations.add(resultSet.getString("CPR"));
            }

            // Query anden forældre
            String barnFatherQuery = "select pf.CPR from PersonFamilie as pf join v2_ForaeldreMyndighedRelation as fmr on fmr.CPR = pf.CPR where pf.FarCPR = ? and fmr.TypeKode = '0004' and fmr.ValidFrom <= ? and (fmr.ValidTo > ? or fmr.ValidTo is null)  and pf.ValidFrom <= ? and (pf.ValidTo > ? or pf.ValidTo is null)";
            preparedStatement = connection.prepareStatement(barnFatherQuery);
            preparedStatement.setObject(1, relationCpr);
            preparedStatement.setObject(2, currentTime);
            preparedStatement.setObject(3, currentTime);
            preparedStatement.setObject(4, currentTime);
            preparedStatement.setObject(5, currentTime);
            resultSet = preparedStatement.executeQuery();
            while (resultSet.next()) {
                relations.add(resultSet.getString("CPR"));
            }

            // Query forældremyndighed
            String autorityQuery = "select fmr.CPR as CPR from v2_ForaeldreMyndighedRelation as fmr where fmr.RelationCpr = ? and (fmr.TypeKode = '0005' OR fmr.TypeKode = '0006') and fmr.ValidFrom <= ? and (fmr.ValidTo > ? or fmr.ValidTo is null)";
            preparedStatement = connection.prepareStatement(autorityQuery);
            preparedStatement.setObject(1, relationCpr);
            preparedStatement.setObject(2, currentTime);
            preparedStatement.setObject(3, currentTime);
            resultSet = preparedStatement.executeQuery();
            while (resultSet.next()) {
                relations.add(resultSet.getString("CPR"));
            }

            return relations;
        } finally {
            PersistenceModule.close(preparedStatement);
            PersistenceModule.close(connection);
        }
    }

    public String findPriorCpr(String cpr) throws SQLException {
        return findPriorCpr(cpr, new Date());
    }

    public String findPriorCpr(String cpr, Date currentTime) throws SQLException {
        PreparedStatement preparedStatement = null;
        Connection connection = null;
        try {
            String queryString = "select HistoriskCPR from HistoriskePersonnumre where CPR = ? and ValidFrom <= ? and (ValidTo > ? or ValidTo is null)";
            connection = connectionProvider.get();
            preparedStatement = connection.prepareStatement(queryString);
            preparedStatement.setObject(1, cpr);
            preparedStatement.setObject(2, currentTime);
            preparedStatement.setObject(3, currentTime);
            ResultSet resultSet = preparedStatement.executeQuery();
            if (resultSet.next()) {
                return resultSet.getString("HistoriskCPR");
            } else {
                return null;
            }
        } finally {
            PersistenceModule.close(preparedStatement);
            PersistenceModule.close(connection);
        }
    }

    public Date findMunicipalityCodeStartDate(String cpr) throws SQLException {
        return findMunicipalityCodeStartDate(cpr, new Date());
    }

    public Date findMunicipalityCodeStartDate(String cpr, Date currentTime) throws SQLException {
        PreparedStatement preparedStatement = null;
        Connection connection = null;
        try {
            String queryString = "select TilflytningKommuneDato from PersonKontakt where CPR = ? and ValidFrom <= ? and (ValidTo > ? or ValidTo is null)";
            connection = connectionProvider.get();
            preparedStatement = connection.prepareStatement(queryString);
            preparedStatement.setObject(1, cpr);
            preparedStatement.setObject(2, currentTime);
            preparedStatement.setObject(3, currentTime);
            ResultSet resultSet = preparedStatement.executeQuery();
            if (resultSet.next()) {
                return resultSet.getTimestamp("TilflytningKommuneDato");
            } else {
                return null;
            }
        } finally {
            PersistenceModule.close(preparedStatement);
            PersistenceModule.close(connection);
        }
    }

    public Collection<Custody> findCustodies(String cpr) throws SQLException {
        return findCustodies(cpr, new Date());
    }


    public List<String> findFMR(String cpr, Date currentTime) throws SQLException {
        PreparedStatement preparedStatement = null;
        Connection connection = null;
        List<String> fmrs = new ArrayList<String>();
        String query = "select TypeKode, RelationCpr from v2_ForaeldreMyndighedRelation " +
                " where CPR = ? and ValidFrom <= ? and (ValidTo > ? or ValidTo is null)";

        try {
            connection = connectionProvider.get();
            preparedStatement = connection.prepareStatement(query);
            preparedStatement.setObject(1, cpr);
            preparedStatement.setObject(2, currentTime);
            preparedStatement.setObject(3, currentTime);

            ResultSet resultSet = preparedStatement.executeQuery();

            while (resultSet.next()) {
                String type = resultSet.getString("Typekode");
                switch (type) {
                    case "0003":
                        fmrs.add(type);
                        break;
                    case "0004":
                        fmrs.add(type);
                        break;
                    case "0005":
                    case "0006":
                        fmrs.add(resultSet.getString("RelationCpr"));
                        break;
                }
            }

        } finally {
            PersistenceModule.close(preparedStatement);
            PersistenceModule.close(connection);
        }

        return fmrs;
    }

    public List<String> findFMRward(String cpr, Date currentTime) throws SQLException {
        PreparedStatement preparedStatement = null;
        Connection connection = null;
        List<String> fmrs = new ArrayList<String>();
        String query = "select Cpr from v2_ForaeldreMyndighedRelation " +
                " where RelationCPR = ? and (TypeKode = '0005' or TypeKode = '0006') and ValidFrom <= ? and (ValidTo > ? or ValidTo is null)";

        try {
            connection = connectionProvider.get();
            preparedStatement = connection.prepareStatement(query);
            preparedStatement.setObject(1, cpr);
            preparedStatement.setObject(2, currentTime);
            preparedStatement.setObject(3, currentTime);

            ResultSet resultSet = preparedStatement.executeQuery();

            while (resultSet.next()) {
                fmrs.add(resultSet.getString("Cpr"));
            }

        } finally {
            PersistenceModule.close(preparedStatement);
            PersistenceModule.close(connection);
        }

        return fmrs;
    }

    public List<String> findUVRward(String cpr, Date currentTime) throws SQLException {
        PreparedStatement preparedStatement = null;
        Connection connection = null;
        List<String> cprs = new ArrayList<String>();
        String query = "select RelationCpr from v2_UmyndiggoerelseVaergeRelation " +
                " where CPR = ? and ValidFrom <= ? and (ValidTo > ? or ValidTo is null)";

        try {
            connection = connectionProvider.get();
            preparedStatement = connection.prepareStatement(query);
            preparedStatement.setObject(1, cpr);
            preparedStatement.setObject(2, currentTime);
            preparedStatement.setObject(3, currentTime);

            ResultSet resultSet = preparedStatement.executeQuery();

            while (resultSet.next()) {
            	cprs.add(resultSet.getString("RelationCPR"));
            }

        } finally {
            PersistenceModule.close(preparedStatement);
            PersistenceModule.close(connection);
        }

        return cprs;
    }

    public List<String> findUVR(String cpr, Date currentTime) throws SQLException {
        PreparedStatement preparedStatement = null;
        Connection connection = null;
        List<String> cprs = new ArrayList<String>();
        String query = "select Cpr from v2_UmyndiggoerelseVaergeRelation " +
                " where RelationCPR = ? and ValidFrom <= ? and (ValidTo > ? or ValidTo is null)";

        try {
            connection = connectionProvider.get();
            preparedStatement = connection.prepareStatement(query);
            preparedStatement.setObject(1, cpr);
            preparedStatement.setObject(2, currentTime);
            preparedStatement.setObject(3, currentTime);

            ResultSet resultSet = preparedStatement.executeQuery();

            while (resultSet.next()) {
                cprs.add(resultSet.getString("CPR"));
            }

        } finally {
            PersistenceModule.close(preparedStatement);
            PersistenceModule.close(connection);
        }

        return cprs;
    }

    public Collection<Custody> findCustodies(String cpr, Date currentTime) throws SQLException {
        ArrayList<Custody> custodies = new ArrayList<Custody>();

        custodies.addAll(findMorFarCustodies(cpr, currentTime));

        List<String> fmrs = findFMR(cpr, currentTime);
        for (String fmr : fmrs) {
            if (!"0003".equals(fmr) && !"0004".equals(fmr)) {
                Custody custody = new Custody();
                custody.setIdentifier(fmr);
                custody.setRelation(CustodyRelation.ANDEN);
                custodies.add(custody);
            }
        }
        List<String> vaerger = findUVRward(cpr, currentTime);
        for (String vaerge : vaerger) {
            Custody custody = new Custody();
            custody.setIdentifier(vaerge);
            custody.setRelation(CustodyRelation.VAERGE);
            custodies.add(custody);
        }

        return custodies;
    }

    private List<Custody> findMorFarCustodies(String cpr, Date currentTime) throws SQLException {
        List<Custody> custodies = new ArrayList<>();

        PreparedStatement preparedStatement = null;
        Connection connection = null;

        String query = "select pf.CPR," +
                " pf.MorCPR," +
                " (case when pf.MorCPR is null or pf.MorCPR in ('', '0000000000') then 'INGEN_TYPE' else coalesce((select 'MOR' from v2_ForaeldreMyndighedRelation fmr where fmr.CPR = pf.CPR and fmr.TypeKode = '0003' and fmr.ValidFrom <= ? and (fmr.ValidTo > ? or fmr.ValidTo is null)), 'MOR_UDEN') end) MorCustodyRelationType," +
                " pf.FarCPR," +
                " (case when pf.FarCPR is null or pf.FarCPR in ('', '0000000000') then 'INGEN_TYPE' else coalesce((select 'FAR' from v2_ForaeldreMyndighedRelation fmr where fmr.CPR = pf.CPR and fmr.TypeKode = '0004' and fmr.ValidFrom <= ? and (fmr.ValidTo > ? or fmr.ValidTo is null)), 'FAR_UDEN') end) FarCustodyRelationType" +
                " from PersonFamilie pf" +
                " where pf.CPR = ? and pf.ValidFrom <= ? and (pf.ValidTo > ? or pf.ValidTo is null)";

        try {
            connection = connectionProvider.get();
            preparedStatement = connection.prepareStatement(query);
            preparedStatement.setObject(1, currentTime);
            preparedStatement.setObject(2, currentTime);
            preparedStatement.setObject(3, currentTime);
            preparedStatement.setObject(4, currentTime);
            preparedStatement.setObject(5, cpr);
            preparedStatement.setObject(6, currentTime);
            preparedStatement.setObject(7, currentTime);

            ResultSet resultSet = preparedStatement.executeQuery();
            if(!resultSet.next()) {
                return custodies;
            }

            List<String> blankCprs = Arrays.asList("", "0000000000");
            String morCpr = resultSet.getString("MorCPR");
            if(morCpr != null && !blankCprs.contains(morCpr)) {
                String morCustodyRelationType = resultSet.getString("MorCustodyRelationType");

                Custody custody = new Custody();
                custody.setIdentifier(morCpr);
                custody.setRelation(Enum.valueOf(CustodyRelation.class, morCustodyRelationType));
                custodies.add(custody);
            }

            String farCpr = resultSet.getString("FarCPR");
            if(farCpr != null && !blankCprs.contains(farCpr)) {
                String farCustodyRelationType = resultSet.getString("FarCustodyRelationType");

                Custody custody = new Custody();
                custody.setIdentifier(farCpr);
                custody.setRelation(Enum.valueOf(CustodyRelation.class, farCustodyRelationType));
                custodies.add(custody);
            }

            if(resultSet.next()) {
                throw new IllegalStateException(String.format("Multiple PersonFamilie rows retrieved for CPR %s!", cpr));
            }
        } finally {
            PersistenceModule.close(preparedStatement);
            PersistenceModule.close(connection);
        }

        return custodies;
    }

    public Collection<Ward> findWard(String cpr) throws SQLException {
        return findWard(cpr, new Date());
    }

    public Collection<Ward> findWard(String cpr, Date currentTime) throws SQLException {
        ArrayList<Ward> wards = new ArrayList<Ward>();

        wards.addAll(findMorFarWard(cpr, currentTime));

        List<String> fmrs = findFMRward(cpr, currentTime);
        for (String fmr : fmrs) {
            Ward ward = new Ward();
            ward.setIdentifier(fmr);
            ward.setRelation(WardRelation.BARN);
            wards.add(ward);
        }
        List<String> uvrs = findUVR(cpr, currentTime);
        for (String uvr : uvrs) {
            Ward ward = new Ward();
            ward.setIdentifier(uvr);
            ward.setRelation(WardRelation.MYNDLING);
            wards.add(ward);
        }

        return wards;
    }

    private List<Ward> findMorFarWard(String cpr, Date currentTime) throws SQLException {
        List<Ward> wards = new ArrayList<>();

        PreparedStatement preparedStatement = null;
        Connection connection = null;

        String query =
                " (select pf.CPR BarnCPR," +
                " (case when pf.CPR = '0000000000' then 'INGEN_TYPE' else coalesce((select 'BARN' from v2_ForaeldreMyndighedRelation fmr where fmr.CPR = pf.CPR and fmr.TypeKode = '0003' and fmr.ValidFrom <= ? and (fmr.ValidTo > ? or fmr.ValidTo is null) limit 1), 'BARN_UDEN') end) WardRelationType" +
                " from PersonFamilie pf" +
                " where pf.MorCPR = ? and pf.ValidFrom <= ? and (pf.ValidTo > ? or pf.ValidTo is null))" +
                " union " +
                " (select pf.CPR BarnCPR," +
                " (case when pf.CPR = '0000000000' then 'INGEN_TYPE' else coalesce((select 'BARN' from v2_ForaeldreMyndighedRelation fmr where fmr.CPR = pf.CPR and fmr.TypeKode = '0004' and fmr.ValidFrom <= ? and (fmr.ValidTo > ? or fmr.ValidTo is null) limit 1), 'BARN_UDEN') end) WardRelationType" +
                " from PersonFamilie pf" +
                " where pf.FarCPR = ? and pf.ValidFrom <= ? and (pf.ValidTo > ? or pf.ValidTo is null))";

        try {
            connection = connectionProvider.get();
            preparedStatement = connection.prepareStatement(query);
            preparedStatement.setObject(1, currentTime);
            preparedStatement.setObject(2, currentTime);
            preparedStatement.setObject(3, cpr);
            preparedStatement.setObject(4, currentTime);
            preparedStatement.setObject(5, currentTime);
            preparedStatement.setObject(6, currentTime);
            preparedStatement.setObject(7, currentTime);
            preparedStatement.setObject(8, cpr);
            preparedStatement.setObject(9, currentTime);
            preparedStatement.setObject(10, currentTime);

            ResultSet resultSet = preparedStatement.executeQuery();
            while(resultSet.next()) {
                String barnCpr = resultSet.getString("BarnCPR");
                String wardRelationType = resultSet.getString("WardRelationType");

                // BarnCPR is non-nullable in the database, so we just check that it does not consist of zeroes.
                if(!barnCpr.equals("0000000000")) {
                    Ward ward = new Ward();
                    ward.setIdentifier(barnCpr);
                    ward.setRelation(Enum.valueOf(WardRelation.class, wardRelationType));
                    wards.add(ward);
                }
            }
        } finally {
            PersistenceModule.close(preparedStatement);
            PersistenceModule.close(connection);
        }

        return wards;
    }

    public PersonVO findCpr(String cpr, Date currentTime) throws SQLException {
        PreparedStatement preparedStatement = null;
        Connection connection = null;
        try {
            String queryString = "SELECT " + SQL_PERSON + " WHERE CPR = ? AND ValidFrom <= ? AND (ValidTo > ? OR ValidTo IS NULL)";
            connection = connectionProvider.get();
            preparedStatement = connection.prepareStatement(queryString);
            preparedStatement.setObject(1, cpr);
            preparedStatement.setObject(2, currentTime);
            preparedStatement.setObject(3, currentTime);
            ResultSet resultSet = preparedStatement.executeQuery();
            if (resultSet.next()) {
                if (!resultSet.isLast()) {
                    throw new IllegalStateException("More that one person return for single cpr");
                }
                return mapPerson(resultSet);
            } else {
                return null;
            }
        } finally {
            PersistenceModule.close(preparedStatement);
            PersistenceModule.close(connection);
        }
    }

    public PersonVO mapPerson(ResultSet resultSet) throws SQLException {
        PersonVO person = new PersonVO();

        person.setCpr(resultSet.getString("CPR"));
        person.setKoen(resultSet.getString("Koen"));
        person.setFornavn(resultSet.getString("Fornavn"));
        person.setMellemnavn(resultSet.getString("Mellemnavn"));
        person.setEfternavn(resultSet.getString("Efternavn"));
        person.setCoNavn(resultSet.getString("CoNavn"));
        person.setLokalitet(resultSet.getString("Lokalitet"));
        person.setVejnavn(resultSet.getString("Vejnavn"));
        person.setBygningsnummer(resultSet.getString("Bygningsnummer"));
        person.setHusnummer(resultSet.getString("Husnummer"));
        person.setEtage(resultSet.getString("Etage"));
        person.setSideDoerNummer(resultSet.getString("SideDoerNummer"));
        person.setBynavn(resultSet.getString("Bynavn"));
        person.setPostnummer(resultSet.getString("Postnummer"));
        person.setPostdistrikt(resultSet.getString("PostDistrikt"));
        person.setStandardAdresse(resultSet.getString("StandardAdresse"));
        person.setStatus(resultSet.getString("Status"));
        person.setNavnebeskyttelsestartdato(resultSet.getDate("NavneBeskyttelseStartDato"));
        person.setNavnebeskyttelseslettedato(resultSet.getDate("NavneBeskyttelseSletteDato"));
        person.setGaeldendeCPR(resultSet.getString("GaeldendeCPR"));
        person.setFoedselsdato(resultSet.getDate("Foedselsdato"));

        person.setAlder(resultSet.getInt("Alder"));
    	if (resultSet.wasNull()) {
    		person.setAlder(null);
    	} 
        
        person.setFoedselsdatoMarkering(resultSet.getString("FoedselsdatoMarkering"));
        person.setStilling(resultSet.getString("Stilling"));
        person.setVejKode(resultSet.getString("VejKode"));
        person.setKommuneKode(resultSet.getString("KommuneKode"));
        person.setNavnTilAdressering(resultSet.getString("NavnTilAdressering"));
        person.setVejnavnTilAdressering(resultSet.getString("VejnavnTilAdressering"));
        person.setStatusDato(resultSet.getDate("StatusDato"));
        person.setUnreliable(resultSet.getString("Unreliable"));
        person.setCreatedDate(resultSet.getDate("CreatedDate"));
        person.setModifiedDate(resultSet.getDate("ModifiedDate"));
        person.setValidFrom(resultSet.getDate("ValidFrom"));
        person.setValidTo(resultSet.getDate("ValidTo"));
        return person;
    }

    public List<PersonVO> findBirthday(Date birthday, Date currentTime) throws SQLException {
        PreparedStatement preparedStatement = null;
        Connection connection = null;
        List<PersonVO> persons = new ArrayList<PersonVO>();
        try {
            String queryString = "SELECT " + SQL_PERSON + " WHERE Foedselsdato = ? AND ValidFrom <= ? AND (ValidTo > ? OR ValidTo IS NULL)";
            connection = connectionProvider.get();
            preparedStatement = connection.prepareStatement(queryString);
            preparedStatement.setObject(1, birthday);
            preparedStatement.setObject(2, currentTime);
            preparedStatement.setObject(3, currentTime);
            ResultSet resultSet = preparedStatement.executeQuery();
            while (resultSet.next()) {
                persons.add(mapPerson(resultSet));
            }
        } finally {
            PersistenceModule.close(preparedStatement);
            PersistenceModule.close(connection);
        }
        return persons;
    }

    public List<PersonVO> findName(String givenName, String middleName, String surName, Date currentTime) throws SQLException {
        PreparedStatement preparedStatement = null;
        Connection connection = null;
        List<PersonVO> persons = new ArrayList<PersonVO>();
        try {
            String SQL_MELLEMNAVN = "";
            if (StringUtils.isNotBlank(middleName)) {
                SQL_MELLEMNAVN = "AND Mellemnavn = ? ";
            }
            String queryString = "SELECT " + SQL_PERSON + " WHERE Fornavn = ? " + SQL_MELLEMNAVN + " AND Efternavn = ? AND ValidFrom <= ? AND (ValidTo > ? OR ValidTo IS NULL)";
            connection = connectionProvider.get();
            preparedStatement = connection.prepareStatement(queryString);
            int index = 1;
            preparedStatement.setObject(index++, givenName);
            if (StringUtils.isNotBlank(middleName)) {
                preparedStatement.setObject(index++, middleName);
            }
            preparedStatement.setObject(index++, surName);
            preparedStatement.setObject(index++, currentTime);
            preparedStatement.setObject(index++, currentTime);
            ResultSet resultSet = preparedStatement.executeQuery();
            while (resultSet.next()) {
                persons.add(mapPerson(resultSet));
            }
        } finally {
            PersistenceModule.close(preparedStatement);
            PersistenceModule.close(connection);
        }
        return persons;
    }

    public List<PersonVO> findDeceased(List<String> cprList, LocalDateTime currentTime) throws SQLException {
        List<PersonVO> persons = new ArrayList<>();
        if ((cprList == null || cprList.isEmpty()) || currentTime == null) {
            return persons;
        }

        PreparedStatement preparedStatement = null;
        Connection connection = null;
        try {
            // Finder alle personer med CPR i cprList, som er døde og hvor StatusDato er før currentTime
            // og hvor der ikke findes en nyere indgang i v2_Person for samme CPR med en anden status end '90'.
            // Hvis samme CPR har flere indgange i v2_Person med status '90', så vælges den nyeste.
            String sql = String.format(
                    "SELECT " + SQL_PERSON + " p1 " +
                    "WHERE CPR IN (%s) AND p1.Status = '90' AND p1.StatusDato < ? " +
                    "AND NOT EXISTS (" + // Finder en nyere indgang for samme CPR med en anden status end '90'
                    "    SELECT 1 FROM v2_Person p2 " +
                    "    WHERE p2.CPR = p1.CPR " +
                    "    AND p2.Status != '90' " +
                    "    AND p2.ValidFrom <= ? " +
                    "    AND (p2.ValidTo > p1.ValidTo OR p2.ValidTo IS NULL)" + // Sikker at vi kun kikker på indgange efter p1.ValidTo
                    ") " +
                    "AND p1.ValidTo = (" + // Sikker kun den nyeste indgang for samme CPR med status '90' vil blive returneret
                    "    SELECT MAX(p3.ValidTo) FROM v2_Person p3 " +
                    "    WHERE p3.CPR = p1.CPR AND p3.Status = '90'" +
                    ") ",
                    cprList.stream()
                            .map(v -> "?")
                            .collect(Collectors.joining(", "))
            );

            connection = connectionProvider.get();
            preparedStatement = connection.prepareStatement(sql);

            int index = 0;
            for( String cpr : cprList ) {
                preparedStatement.setString( ++index, cpr );
            }
            preparedStatement.setTimestamp(++index, java.sql.Timestamp.valueOf(currentTime)); // Sætter ? ved p1.StatusDato < ?
            preparedStatement.setTimestamp(++index, java.sql.Timestamp.valueOf(currentTime)); // Sætter ? ved p2.ValidFrom <= ?

            ResultSet resultSet = preparedStatement.executeQuery();
            while (resultSet.next()) {
                persons.add(mapPerson(resultSet));
            }

        } finally {
            PersistenceModule.close(preparedStatement);
            PersistenceModule.close(connection);
        }
        return persons;
    }
}
