/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata.cpr;

import static org.hamcrest.CoreMatchers.anyOf;
import static org.hamcrest.CoreMatchers.is;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertThat;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.sql.Connection;
import java.sql.SQLException;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.*;

import dk.nsi.stamdata.testing.DatabaseTestInitializer;

import org.joda.time.DateTime;
import org.joda.time.LocalDate;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;

import com.google.common.collect.Sets;
import com.google.inject.Inject;
import com.trifork.stamdata.models.ForaeldreMyndighedRelationVO.FMRTypeKode;
import com.trifork.stamdata.models.PersonVO;

import dk.nsi.stamdata.cpr.dao.CprDAO;
import dk.nsi.stamdata.cpr.dao.CprDAO.Custody;
import dk.nsi.stamdata.cpr.dao.CprDAO.CustodyRelation;
import dk.nsi.stamdata.cpr.dao.CprDAO.Ward;
import dk.nsi.stamdata.guice.GuiceTestRunner;
import dk.nsi.stamdata.testing.Factories;
import dk.nsi.stamdata.testing.StamdataTestHelper;

import javax.inject.Named;

@RunWith(GuiceTestRunner.class)
public class CprDaoTest {
    private static final DateTime YESTERDAY = new DateTime(Factories.YESTERDAY);
    private static final Date CURRENT_TIME = new Date();
    private static final String CPR  = "1234567890";
    private static final String CPR1 = "1234567891";
    private static final String CPR2 = "1234567892";
    private static final String CPR3 = "1234567893";
    private static final String CPR4 = "1234567894";
    private static final String CPR5 = "1234567895";
    private static final String CPR6 = "1234567896";

    @Inject
    private Connection connection;
    @Inject
    private CprDAO cprDao;

    @Inject
    @Named("run.undertow")
    private boolean runUndertow;

    StamdataTestHelper testHelper = new StamdataTestHelper();
    private DatabaseTestInitializer dbInitializer;
    
    @Before
    public void setUp() throws Exception {
        if (runUndertow) {
            dbInitializer = new DatabaseTestInitializer(connection);
            dbInitializer.initialize();
        }
        testHelper.clear();
    }

    @Test
    public void findCustodies() throws Exception {
        PersonVO person = testHelper.createPerson(CPR);
        PersonVO personMor = testHelper.createKvinde(CPR1);
        PersonVO personFar = testHelper.createMand(CPR2);
        testHelper.createForaeldreMyndighedRelation(person, FMRTypeKode.MOR);
        testHelper.createForaeldreMyndighedRelation(person, FMRTypeKode.FAR);
        testHelper.createPersonFamilie(person, personMor, personFar);
        testHelper.prepareTestdata(connection);

        Collection<Custody> custodies = cprDao.findCustodies(CPR);
        assertEquals(2,custodies.size());
        Iterator<Custody> iter = custodies.iterator();
        Custody custody = iter.next();
        assertEquals(CPR1, custody.getIdentifier());
        assertEquals(CustodyRelation.MOR, custody.getRelation());
        custody = iter.next();
        assertEquals(CPR2, custody.getIdentifier());
        assertEquals(CustodyRelation.FAR, custody.getRelation());
    }

    @Test
    public void findCustodiesMedmor() throws Exception {
        PersonVO person = testHelper.createPerson(CPR);
        PersonVO personMor = testHelper.createKvinde(CPR1);
        PersonVO personMedmor = testHelper.createKvinde(CPR3);
        testHelper.createForaeldreMyndighedRelation(person, FMRTypeKode.MOR);
        testHelper.createPersonFamilie(person, personMor, personMedmor);
        testHelper.prepareTestdata(connection);

        Collection<Custody> custodies = cprDao.findCustodies(CPR);
        assertEquals("Custodies " + Arrays.toString(custodies.toArray()), 2, custodies.size());
        Set<String> foundCpr = Sets.newHashSet();
        Iterator<Custody> iter = custodies.iterator();
        while (iter.hasNext()) {
            Custody custody = iter.next();
            foundCpr.add(custody.getIdentifier());
            switch (custody.getRelation()) {
                case MOR:
                    assertThat(custody.getIdentifier(), is(CPR1));
                    break;
                case FAR_UDEN:
                    assertThat(custody.getIdentifier(), is(CPR3));
                    break;
                default:
                    fail("Wrong relation: " + custody.toString());
                    break;
            }
        }
        assertEquals("Found cpr " + Arrays.toString(foundCpr.toArray()), 2, foundCpr.size());
    }

    @Test
    public void findCustodiesMedfar() throws Exception {
        PersonVO person = testHelper.createPerson(CPR);
        PersonVO personFar = testHelper.createMand(CPR2);
        PersonVO personMedfar = testHelper.createMand(CPR3);
        testHelper.createForaeldreMyndighedRelation(person, FMRTypeKode.FAR);
        testHelper.createPersonFamilie(person, personFar, personMedfar);
        testHelper.prepareTestdata(connection);

        Collection<Custody> custodies = cprDao.findCustodies(CPR);
        assertEquals("Custodies " + Arrays.toString(custodies.toArray()), 2, custodies.size());
        Set<String> foundCpr = Sets.newHashSet();
        Iterator<Custody> iter = custodies.iterator();
        while (iter.hasNext()) {
            Custody custody = iter.next();
            foundCpr.add(custody.getIdentifier());
            switch (custody.getRelation()) {
                case FAR:
                    assertThat(custody.getIdentifier(), is(CPR3));
                    break;
                case MOR_UDEN:
                    assertThat(custody.getIdentifier(), is(CPR2));
                    break;
                default:
                    fail("Wrong relation: " + custody.toString());
                    break;
            }
        }
        assertEquals("Found cpr " + Arrays.toString(foundCpr.toArray()), 2, foundCpr.size());
    }

    @Test
    public void findCustodiesVaerge() throws Exception {
        PersonVO person = testHelper.createPerson(CPR);
        PersonVO personMor = testHelper.createKvinde(CPR1);
        PersonVO personFar = testHelper.createMand(CPR2);
        PersonVO personVaerge = testHelper.createKvinde(CPR3);
        testHelper.createForaeldreMyndighedRelation(person, FMRTypeKode.MOR);
        testHelper.createPersonFamilie(person, personMor, personFar);
        testHelper.createVaerge(person, personVaerge);
        testHelper.prepareTestdata(connection);

        Collection<Custody> custodies = cprDao.findCustodies(CPR);
        assertEquals("Custodies " + Arrays.toString(custodies.toArray()), 3, custodies.size());
        Set<String> foundCpr = Sets.newHashSet();
        Iterator<Custody> iter = custodies.iterator();
        while (iter.hasNext()) {
            Custody custody = iter.next();
            foundCpr.add(custody.getIdentifier());
            switch (custody.getRelation()) {
                case MOR:
                    assertEquals(CPR1, custody.getIdentifier());
                    break;
                case VAERGE:
                    assertEquals(CPR3, custody.getIdentifier());
                    break;
                case FAR_UDEN:
                    assertEquals(CPR2, custody.getIdentifier());
                    break;
                default:
                    fail("Wrong relation: " + custody.toString());
                    break;
            }
        }
        assertEquals("Found cpr " + Arrays.toString(foundCpr.toArray()), 3, foundCpr.size());
    }

    @Test
    public void findCustodiesVaergeFar() throws Exception {
        PersonVO person = testHelper.createPerson(CPR);
        PersonVO personMor = testHelper.createKvinde(CPR1);
        PersonVO personFar = testHelper.createMand(CPR2);
        PersonVO personVaerge = testHelper.createKvinde(CPR3);
        testHelper.createForaeldreMyndighedRelation(person, FMRTypeKode.FAR);
        testHelper.createPersonFamilie(person, personMor, personFar);
        testHelper.createVaerge(person, personVaerge);
        testHelper.prepareTestdata(connection);

        Collection<Custody> custodies = cprDao.findCustodies(CPR);
        assertEquals("Custodies " + Arrays.toString(custodies.toArray()), 3, custodies.size());
        Set<String> foundCpr = Sets.newHashSet();
        Iterator<Custody> iter = custodies.iterator();
        while (iter.hasNext()) {
            Custody custody = iter.next();
            foundCpr.add(custody.getIdentifier());
            switch (custody.getRelation()) {
                case MOR_UDEN:
                    assertEquals(CPR1, custody.getIdentifier());
                    break;
                case FAR:
                    assertEquals(CPR2, custody.getIdentifier());
                    break;
                case VAERGE:
                    assertEquals(CPR3, custody.getIdentifier());
                    break;
                default:
                    fail("Wrong relation: " + custody.toString());
                    break;
            }
        }
        assertEquals("Found cpr " + Arrays.toString(foundCpr.toArray()), 3, foundCpr.size());
    }

    @Test
    public void findCustodiesAI() throws Exception {
        PersonVO person = testHelper.createPerson(CPR);
        PersonVO personMor = testHelper.createKvinde(CPR1);
        PersonVO personFar = testHelper.createMand(CPR2);
        testHelper.createForaeldreMyndighedRelation(person, FMRTypeKode.AI1, personMor);
        testHelper.createForaeldreMyndighedRelation(person, FMRTypeKode.AI1, personFar);
        testHelper.createPersonFamilie(person, null, null);
        testHelper.prepareTestdata(connection);

        Collection<Custody> custodies = cprDao.findCustodies(CPR);
        assertEquals("Custodies " + Arrays.toString(custodies.toArray()), 2, custodies.size());
        Set<String> foundCpr = Sets.newHashSet();
        Iterator<Custody> iter = custodies.iterator();
        while (iter.hasNext()) {
            Custody custody = iter.next();
            foundCpr.add(custody.getIdentifier());
            switch (custody.getRelation()) {
                case ANDEN:
                    assertThat(custody.getIdentifier(), anyOf(is(CPR1), is(CPR2)));
                    break;
                default:
                    fail("Wrong relation: " + custody.toString());
                    break;
            }
        }
        assertEquals("Found cpr " + Arrays.toString(foundCpr.toArray()), 2, foundCpr.size());
    }

    @Test
    public void findCustodiesAI2() throws Exception {
        PersonVO person = testHelper.createPerson(CPR);
        PersonVO personMor = testHelper.createKvinde(CPR1);
        PersonVO personMammy = testHelper.createKvinde(CPR2);
        testHelper.createForaeldreMyndighedRelation(person, FMRTypeKode.AI1, personMor);
        testHelper.createForaeldreMyndighedRelation(person, FMRTypeKode.AI2, personMammy);
        testHelper.createPersonFamilie(person, null, null);
        testHelper.prepareTestdata(connection);

        Collection<Custody> custodies = cprDao.findCustodies(CPR);
        assertEquals("Custodies " + Arrays.toString(custodies.toArray()), 2, custodies.size());
        Set<String> foundCpr = Sets.newHashSet();
        Iterator<Custody> iter = custodies.iterator();
        while (iter.hasNext()) {
            Custody custody = iter.next();
            foundCpr.add(custody.getIdentifier());
            switch (custody.getRelation()) {
                case ANDEN:
                    assertThat(custody.getIdentifier(), anyOf(is(CPR1), is(CPR2)));
                    break;
                default:
                    fail("Wrong relation: " + custody.toString());
                    break;
            }
        }
        assertEquals("Found cpr " + Arrays.toString(foundCpr.toArray()), 2, foundCpr.size());
    }

    @Test
    public void findMunicipalityCodeStartDate() throws SQLException {
        testHelper.createPerson(CPR);
        testHelper.prepareTestdata(connection);
        StamdataTestHelper.assertDates(Factories.YEAR_2000, cprDao.findMunicipalityCodeStartDate(CPR));
    }

    @Test
    public void findPriorCpr() throws SQLException {
        testHelper.createPerson(CPR);
        testHelper.prepareTestdata(connection);
        assertEquals(StamdataTestHelper.PRIOR_CPR, cprDao.findPriorCpr(CPR));
    }

    @Test
    public void findRelations() throws Exception {
        PersonVO person = testHelper.createPerson(CPR);
        PersonVO personMor = testHelper.createKvinde(CPR1);
        PersonVO personFar = testHelper.createMand(CPR2);
        testHelper.createForaeldreMyndighedRelation(person, FMRTypeKode.MOR);
        testHelper.createForaeldreMyndighedRelation(person, FMRTypeKode.FAR);
        testHelper.createPersonFamilie(person, personMor, personFar);
        testHelper.prepareTestdata(connection);

        Collection<String> relations = cprDao.findRelationsAsGivingAuthority(CPR);
        assertEquals(2, relations.size());
        Iterator<String> iter = relations.iterator();
        String relations_cpr = iter.next();
        assertEquals(CPR1, relations_cpr);
        relations_cpr = iter.next();
        assertEquals(CPR2, relations_cpr);
    }

    @Test
    public void findRelationsVaerge() throws Exception {
        PersonVO person = testHelper.createPerson(CPR);
        PersonVO person1 = testHelper.createKvinde(CPR1);
        PersonVO person2 = testHelper.createKvinde(CPR2);
        testHelper.createPersonFamilie(person, null, null);
        testHelper.createForaeldreMyndighedRelation(person, FMRTypeKode.AI1, person1);
        testHelper.createForaeldreMyndighedRelation(person, FMRTypeKode.AI2, person2);
        testHelper.prepareTestdata(connection);

        Collection<String> relations = cprDao.findRelationsAsGivingAuthority(CPR);
        assertEquals(2, relations.size());
        Iterator<String> iter = relations.iterator();
        String relations_cpr = iter.next();
        assertEquals(CPR1, relations_cpr);
        relations_cpr = iter.next();
        assertEquals(CPR2, relations_cpr);
    }

    @Test
    public void findWardMor() throws Exception {
        PersonVO person = testHelper.createKvinde(CPR);
        PersonVO barn1 = testHelper.createPerson(CPR1);
        PersonVO barn2 = testHelper.createPerson(CPR2);
        PersonVO barn3 = testHelper.createPerson(CPR3);
        PersonVO barn4 = testHelper.createPerson(CPR4);
        PersonVO personMedmor = testHelper.createKvinde(CPR5);

        testHelper.createPersonFamilie(barn1, person, personMedmor);
        testHelper.createForaeldreMyndighedRelation(barn1, FMRTypeKode.MOR);
        testHelper.createForaeldreMyndighedRelation(barn1, FMRTypeKode.FAR);

        testHelper.createPersonFamilie(barn2, person, null);
        testHelper.createForaeldreMyndighedRelation(barn2, FMRTypeKode.FAR);

        testHelper.createPersonFamilie(barn3, null, null);
        testHelper.createForaeldreMyndighedRelation(barn3, FMRTypeKode.AI1, person);

        testHelper.createPersonFamilie(barn4, null, null);
        testHelper.createVaerge(barn4, person);
        testHelper.prepareTestdata(connection);

        Collection<Ward> wards = cprDao.findWard(CPR);
        assertEquals("Wards " + Arrays.toString(wards.toArray()), 4, wards.size());
        Set<String> foundCpr = Sets.newHashSet();
        Iterator<Ward> iter = wards.iterator();
        while (iter.hasNext()) {
            Ward ward = iter.next();
            foundCpr.add(ward.getIdentifier());
            switch (ward.getRelation()) {
                case BARN:
                    assertThat(ward.getIdentifier(), anyOf(is(CPR1), is(CPR3)));
                    break;
                case BARN_UDEN:
                    assertThat(ward.getIdentifier(), is(CPR2));
                    break;
                case MYNDLING:
                    assertThat(ward.getIdentifier(), is(CPR4));
                    break;
                default:
                    fail("Wrong relation: " + ward.toString());
                    break;
            }
        }
        assertEquals("Found cpr " + Arrays.toString(foundCpr.toArray()), 4, foundCpr.size());
    }

    @Test
    public void findWardFar() throws Exception {
        PersonVO person = testHelper.createMand(CPR);
        PersonVO barn1 = testHelper.createPerson(CPR1);
        PersonVO barn2 = testHelper.createPerson(CPR2);
        PersonVO barn3 = testHelper.createPerson(CPR3);
        PersonVO barn4 = testHelper.createPerson(CPR4);

        testHelper.createPersonFamilie(barn1, null, person);
        testHelper.createForaeldreMyndighedRelation(barn1, FMRTypeKode.MOR);
        testHelper.createForaeldreMyndighedRelation(barn1, FMRTypeKode.FAR);

        testHelper.createPersonFamilie(barn2, null, person);
        testHelper.createForaeldreMyndighedRelation(barn2, FMRTypeKode.FAR);

        testHelper.createPersonFamilie(barn3, null, null);
        testHelper.createForaeldreMyndighedRelation(barn3, FMRTypeKode.AI1, person);

        testHelper.createPersonFamilie(barn4, null, null);
        testHelper.createVaerge(barn4, person);
        testHelper.prepareTestdata(connection);

        Collection<Ward> wards = cprDao.findWard(CPR);
        assertEquals("Wards " + Arrays.toString(wards.toArray()), 4, wards.size());
        Set<String> foundCpr = Sets.newHashSet();
        Iterator<Ward> iter = wards.iterator();
        while (iter.hasNext()) {
            Ward ward = iter.next();
            foundCpr.add(ward.getIdentifier());
            switch (ward.getRelation()) {
                case BARN:
                    assertThat(ward.getIdentifier(), anyOf(is(CPR1), is(CPR2), is(CPR3)));
                    break;
                case MYNDLING:
                    assertThat(ward.getIdentifier(), is(CPR4));
                    break;
                default:
                    fail("Wrong relation: " + ward.toString());
                    break;
            }
        }
        assertEquals("Found cpr " + Arrays.toString(foundCpr.toArray()), 4, foundCpr.size());
    }
    
    @Test
    public void testPersonInformation() throws Exception {
        testHelper.createPerson(CPR);
        testHelper.prepareTestdata(connection);

        PersonVO result = cprDao.findCpr(CPR, new Date());
        assertNotNull(result);
        assertEquals("Standard adressevej 321", result.getStandardAdresse());
        assertEquals("BirkeBy", result.getBynavn());
        assertEquals("6666", result.getPostnummer());
    }
    
    @Test
    public void testCprNotExists() throws Exception {
    	PersonVO result = cprDao.findCpr("NOT_FOUND", new Date());
        assertNull(result);
    }
    
    @Test
    public void testPersonInformationStatus() throws Exception {
        testHelper.createPersonWithStatus(CPR, "40");
        testHelper.prepareTestdata(connection);

        PersonVO result = cprDao.findCpr(CPR, new Date());
        assertNotNull(result);
        assertEquals("40", result.getStatus());
    }

    @Test
    public void testPersonGetAge() throws Exception {
        Integer personAge = 20;
    	Date yesterdayAdjustedSomeYears = DateTime.now().minusDays(1).withHourOfDay(0).withMinuteOfHour(0).withSecondOfMinute(0).withMillisOfSecond(0).minusYears(personAge).toDate();
        PersonVO person = testHelper.createPersonWithBirthday(yesterdayAdjustedSomeYears);
        testHelper.prepareTestdata(connection);

        PersonVO result = cprDao.findCpr(person.getCpr(), new Date());
        assertNotNull(result);
        assertEquals(personAge, result.getAlder());
    }


    @Test
    public void testPersonGetAgeNegative() throws Exception {
        Integer personAge = -20;
    	Date yesterdayAdjustedSomeYears = DateTime.now().minusDays(1).withHourOfDay(0).withMinuteOfHour(0).withSecondOfMinute(0).withMillisOfSecond(0).minusYears(personAge).toDate();
        PersonVO person = testHelper.createPersonWithBirthday(yesterdayAdjustedSomeYears);
        testHelper.prepareTestdata(connection);

        PersonVO result = cprDao.findCpr(person.getCpr(), new Date());
        assertNotNull(result);

        // There is a little difference between in-memory db and mysql.. maybe something with date/time setup
        if (runUndertow) {
            assertEquals(personAge, result.getAlder());
        } else {
            assertEquals(personAge.intValue()+1, result.getAlder().intValue());
        }
    }
    
    @Test
    public void testPersonGetAgeCprNotExists() throws Exception {
        PersonVO result = cprDao.findCpr("findes ikke", new Date());
        assertNull(result);
    }

    @Test
    public void testFindUVR() throws Exception {
        PersonVO person = testHelper.createPerson(CPR);
        PersonVO personVaerge = testHelper.createKvinde(CPR3);
        testHelper.createVaerge(person, personVaerge);
        testHelper.prepareTestdata(connection);
    	
        List<String> result = cprDao.findUVR(CPR3, new Date());
        
        assertNotNull(result);
        assertEquals(CPR, result.get(0));
    }
    
    @Test
    public void testForaeldremyndighedMor() throws Exception {
        PersonVO barn1 = testHelper.createPerson(CPR1);
        PersonVO barn2 = testHelper.createPerson(CPR2);
        PersonVO mor = testHelper.createKvinde(CPR);
        PersonVO andenPerson = testHelper.createPerson(CPR3);
        testHelper.createForaeldreMyndighedRelation(barn1, FMRTypeKode.MOR, mor);
        testHelper.createForaeldreMyndighedRelation(barn1, FMRTypeKode.AI2, andenPerson);
        testHelper.createForaeldreMyndighedRelation(barn2, FMRTypeKode.MOR, mor);
        
        testHelper.createPersonFamilie(barn1, mor, null);
        testHelper.createPersonFamilie(barn2, mor, null);

        testHelper.prepareTestdata(connection);
    	
        List<String> result = cprDao.findRelationsAsHavingAuthority(mor.getCpr(), new Date());
        
        assertNotNull(result);
        assertTrue(result.contains(barn1.getCpr()));
        assertTrue(result.contains(barn2.getCpr()));
    }    
    
    @Test
    public void testForaeldremyndighedFar() throws Exception {
        PersonVO barn1 = testHelper.createPerson(CPR1);
        PersonVO barn2 = testHelper.createPerson(CPR2);
        PersonVO far = testHelper.createKvinde(CPR);
        PersonVO andenPerson = testHelper.createPerson(CPR3);
        PersonVO mor = testHelper.createKvinde(CPR4);
        testHelper.createForaeldreMyndighedRelation(barn1, FMRTypeKode.FAR, far);
        testHelper.createForaeldreMyndighedRelation(barn1, FMRTypeKode.AI2, andenPerson);
        testHelper.createForaeldreMyndighedRelation(barn2, FMRTypeKode.FAR, far);
        testHelper.createForaeldreMyndighedRelation(barn2, FMRTypeKode.MOR, mor);
        
        testHelper.createPersonFamilie(barn1, null, far);
        testHelper.createPersonFamilie(barn2, mor, far);

        testHelper.prepareTestdata(connection);
    	
        List<String> result = cprDao.findRelationsAsHavingAuthority(far.getCpr(), new Date());
        
        assertNotNull(result);
        assertTrue(result.contains(barn1.getCpr()));
        assertTrue(result.contains(barn2.getCpr()));
    }    
    
    @Test
    public void testForaeldremyndighedAnden() throws Exception {
        PersonVO barn1 = testHelper.createPerson("barn");
        PersonVO personRelation1 = testHelper.createPerson(CPR1);
        PersonVO personRelation2 = testHelper.createPerson(CPR2);
        testHelper.createForaeldreMyndighedRelation(barn1, FMRTypeKode.AI1, personRelation1);
        testHelper.createForaeldreMyndighedRelation(barn1, FMRTypeKode.AI2, personRelation2);

        testHelper.prepareTestdata(connection);
    	
        List<String> result = cprDao.findRelationsAsHavingAuthority(personRelation1.getCpr(), new Date());
        
        assertNotNull(result);
        assertTrue(result.contains(barn1.getCpr()));

    }    

    @Test
    public void testDeceasedEmptyInputGivesEmptyResult() throws SQLException {
        List<PersonVO> resultFromNullInput = cprDao.findDeceased(null, LocalDateTime.now());
        List<PersonVO> resultFromEmptyListInput = cprDao.findDeceased(new ArrayList<>(), LocalDateTime.now());

        assertNotNull(resultFromNullInput);
        assertNotNull(resultFromEmptyListInput);
    }

    @Test
    public void testFindDeceasedFilteringStatus() throws SQLException {
        testHelper.createPersonWithStatus(CPR1, "01");
        testHelper.createPersonWithStatus(CPR2, "90");
        testHelper.prepareTestdata(connection);

        List<PersonVO> result = cprDao.findDeceased(Arrays.asList(CPR1, CPR2), LocalDateTime.now());

        assertNotNull(result);
        assertEquals(1, result.size());
        assertEquals(CPR2, result.get(0).getCpr());
    }

    @Test
    public void testFindDeceasedFilteringStatusDato() throws SQLException {
        testHelper.createPersonWithStatus(CPR1, "90").setStatusDato(Factories.YEAR_2000);
        testHelper.createPersonWithStatus(CPR2, "90").setStatusDato(Factories.YESTERDAY);
        testHelper.prepareTestdata(connection);

        List<PersonVO> result = cprDao.findDeceased(Arrays.asList(CPR1, CPR2), LocalDateTime.now().minusDays(2));

        assertNotNull(result);
        assertEquals(1, result.size());
        assertEquals(CPR1, result.get(0).getCpr());
    }

    @Test
    public void testFindDeceasedReturnMultipleFound() throws SQLException {
        testHelper.createPersonWithStatus(CPR1, "90");
        testHelper.createPersonWithStatus(CPR2, "90");
        testHelper.prepareTestdata(connection);

        List<PersonVO> result = cprDao.findDeceased(Arrays.asList(CPR1, CPR2, CPR3), LocalDateTime.now());

        List<String> expectedCprList = Arrays.asList(CPR1, CPR2);
        assertNotNull(result);
        assertEquals(expectedCprList.size(), result.size());
        assertTrue(result.stream().allMatch(personVO -> expectedCprList.contains(personVO.getCpr())));
    }

    @Test
    public void testFindDeceased_WithPersonWhoIsDeadButIsAliveInTheFuture() throws SQLException {
        createDeceasedTestData(CPR1,"01", "90", "01");
        testHelper.prepareTestdata(connection);

        List<PersonVO> result = cprDao.findDeceased(Arrays.asList(CPR1, CPR2, CPR3), LocalDateTime.now());

        assertNotNull(result);
        assertEquals(0, result.size());
    }

    @Test
    public void testFindDeceased_WithPersonWhoIsAliveButDeadInTheFuture() throws SQLException {
        createDeceasedTestData(CPR1,"90", "01", "90");
        testHelper.prepareTestdata(connection);

        List<PersonVO> result = cprDao.findDeceased(Arrays.asList(CPR1, CPR2, CPR3), LocalDateTime.now());

        List<String> expectedCprList = Collections.singletonList(CPR1);
        assertNotNull(result);
        assertEquals(expectedCprList.size(), result.size());
        assertTrue(result.stream().allMatch(personVO -> expectedCprList.contains(personVO.getCpr())));
    }

    @Test
    public void testFindDeceased_WithNoDeceasedPersons() throws SQLException {
        createDeceasedTestData(CPR1, "01", "01", "01");
        testHelper.prepareTestdata(connection);

        List<PersonVO> result = cprDao.findDeceased(Arrays.asList(CPR1, CPR2, CPR3), LocalDateTime.now());

        assertNotNull(result);
        assertEquals(0, result.size());
    }

    @Test
    public void testFindDeceased_backInTime_WithPersonWhoIsDeadButIsAliveInTheFuture() throws SQLException {
        createDeceasedTestData(CPR1,"01", "90", "90");
        testHelper.prepareTestdata(connection);

        List<PersonVO> result0 = cprDao.findDeceased(Arrays.asList(CPR1, CPR2, CPR3), LocalDateTime.now().minusDays(99)); // Does not exist in the database at that time
        List<PersonVO> result1 = cprDao.findDeceased(Arrays.asList(CPR1, CPR2, CPR3), LocalDateTime.now().minusDays(10)); // 01 Alive at that time
        List<PersonVO> result2 = cprDao.findDeceased(Arrays.asList(CPR1, CPR2, CPR3), LocalDateTime.now().minusDays(9));  // 01 Alive at that time
        List<PersonVO> result3 = cprDao.findDeceased(Arrays.asList(CPR1, CPR2, CPR3), LocalDateTime.now().minusDays(8));  // 90 Dead at that time

        // result0 should be empty, since the person does not exist in the database at that time
        assertNotNull(result0);
        assertEquals(0, result0.size());

        // result1 should be empty, since the person is alive at that time
        assertNotNull(result1);
        assertEquals(0, result1.size());

        // result2 should be empty, since the person is alive at that time
        assertNotNull(result2);
        assertEquals(0, result2.size());

        // result3 should contain the person, since the person is dead at that time
        List<String> expectedCprList = Collections.singletonList(CPR1);
        assertNotNull(result3);
        assertEquals(expectedCprList.size(), result3.size());
    }


    @Test
    public void testFindDeceased_WithMultipleDeceasedStatus_ReturnsNewsValidTo() throws SQLException {
        // createDeceasedTestData returns statusDato for den nyeste person (dateAfterToDayStatus) den bruger vi til at sammenligne med hvad der er returneret fra databasen
        // Vi kan ikke gøre det på med validTo eller validFrom, da PersonVO som bliver returneret fra databasen ikke indeholder validTo og validFrom
        DateTime newsPersonStatusDate = createDeceasedTestData(CPR1,"90", "90", "90");
        testHelper.prepareTestdata(connection);

        List<PersonVO> result = cprDao.findDeceased(Arrays.asList(CPR1, CPR2, CPR3), LocalDateTime.now());

        List<String> expectedCprList = Collections.singletonList(CPR1);
        assertNotNull(result);
        assertEquals(expectedCprList.size(), result.size());
        assertTrue(result.stream().allMatch(personVO -> expectedCprList.contains(personVO.getCpr())));

        // Valider at vi får returnerede den nyeste indgang i databasen, den med den nyeste statusDato og derved også nyeste validTo dato på den måde test daten er sat op
        LocalDate dateFromDatabase = new DateTime(result.get(0).getStatusDato()).toLocalDate();
        assertEquals(newsPersonStatusDate.toLocalDate(), dateFromDatabase);
    }

    @Test
    public void testFindDeceased_WithDeceasedStatus() throws SQLException {
        createDeceasedTestData(CPR1,"01", "90", "90");
        createDeceasedTestData(CPR2,"01", "90", null);
        testHelper.prepareTestdata(connection);

        List<PersonVO> result = cprDao.findDeceased(Arrays.asList(CPR1, CPR2, CPR3), LocalDateTime.now());

        List<String> expectedCprList = Arrays.asList(CPR1, CPR2);
        assertNotNull(result);
        assertEquals(expectedCprList.size(), result.size());
        assertTrue(result.stream().allMatch(personVO -> expectedCprList.contains(personVO.getCpr())));
    }

    @Test
    public void testFindDeceased_WithMixedDeceasedAndNonDeceasedPersons() throws SQLException {
        createDeceasedTestData(CPR1, "01", "01", "01"); // CPR1 is not deceased
        createDeceasedTestData(CPR2, "01", "01", null); // CPR2 is not deceased
        createDeceasedTestData(CPR3, "01", "90", "01"); // CPR3 is not deceased
        createDeceasedTestData(CPR4, "01", "90", null); // CPR4 is deceased
        createDeceasedTestData(CPR5, "90", "01", "01"); // CPR5 is not deceased
        createDeceasedTestData(CPR6, "90", null, null); // CPR6 is deceased
        testHelper.prepareTestdata(connection);

        List<PersonVO> result = cprDao.findDeceased(Arrays.asList(CPR1, CPR2, CPR3, CPR4, CPR5, CPR6), LocalDateTime.now());

        List<String> expectedCprList = Arrays.asList(CPR4, CPR6);
        assertNotNull(result);
        assertEquals(expectedCprList.size(), result.size());
        assertTrue(result.stream().allMatch(personVO -> expectedCprList.contains(personVO.getCpr())));
    }

    @Test
    public void testFindDeceased_WithMultipleStatusChanges() throws SQLException  {
        createDeceasedTestPerson(CPR1, "01", new DateTime("2018-09-05T12:30:00"), new DateTime("1910-10-21T00:00:00"), new DateTime("2018-12-07T00:00:00"));
        createDeceasedTestPerson(CPR1, "90", new DateTime("2018-12-06T14:39:00"), new DateTime("2018-12-07T00:00:00"), new DateTime("2020-03-12T00:00:00"));
        createDeceasedTestPerson(CPR1, "01", null, new DateTime("2020-03-12T00:00:00"), new DateTime("2999-12-31T00:00:00"));
        testHelper.prepareTestdata(connection);

        List<PersonVO> result = cprDao.findDeceased(Collections.singletonList(CPR1), LocalDateTime.now());

        assertNotNull(result);
        assertEquals(0, result.size());
    }

    @Test
    public void testFindDeceased_backInTime_WithMultipleStatusChanges() throws SQLException  {
        createDeceasedTestPerson(CPR1, "01", new DateTime("2018-09-05T12:30:00"), new DateTime("1910-10-21T00:00:00"), new DateTime("2018-12-07T00:00:00"));
        createDeceasedTestPerson(CPR1, "90", new DateTime("2018-12-06T14:39:00"), new DateTime("2018-12-07T00:00:00"), new DateTime("2020-03-12T00:00:00"));
        createDeceasedTestPerson(CPR1, "01", null, new DateTime("2020-03-12T00:00:00"), new DateTime("2999-12-31T00:00:00"));
        testHelper.prepareTestdata(connection);

        String findDeceasedBeforeStatus90 = "2018-11-06T00:00:00"; // Er før status 90 og i den første perioden med status 01
        LocalDateTime localDateTime = LocalDateTime.parse(findDeceasedBeforeStatus90, DateTimeFormatter.ISO_LOCAL_DATE_TIME);
        List<PersonVO> result = cprDao.findDeceased(Collections.singletonList(CPR1), localDateTime);

        assertNotNull(result);
        assertEquals(0, result.size());
    }

    public DateTime createDeceasedTestData(String cpr, String dateBeforeToDayStatus, String toDayStatus, String dateAfterToDayStatus) throws SQLException {
        DateTime toDay = DateTime.now();

        DateTime statusDato = toDay.minusDays(11);
        DateTime validFrom = toDay.minusDays(11);
        DateTime validTo = toDay.minusDays(10);
        createDeceasedTestPerson(cpr, dateBeforeToDayStatus, statusDato, validFrom, validTo);

        if (toDayStatus != null) {
            statusDato = validTo;
            validFrom = validTo;
            validTo = validTo.plusDays(2);
            createDeceasedTestPerson(cpr, toDayStatus, statusDato, validFrom, validTo);
        }

        if (dateAfterToDayStatus != null) {
            statusDato = validTo;
            validFrom = validTo;
            validTo = validTo.plusDays(2);
            createDeceasedTestPerson(cpr, dateAfterToDayStatus, statusDato, validFrom, validTo);
        }

        // Returner statusDato for at kunne sammenligne med hvad der er returneret fra databasen
        // Da PersonVO som bliver returnerede fra databasen ikke indeholder validTo og validFrom
        return statusDato;
    }

    private void createDeceasedTestPerson(String cpr, String status, DateTime statusDato, DateTime validFrom, DateTime validTo) {
        PersonVO person = testHelper.createPersonWithStatus(cpr, status);
        if (statusDato != null) {
            person.setStatusDato(statusDato.toDate());
        }
        person.setValidFrom(validFrom.toDate());
        person.setValidTo(validTo.toDate());
    }
}
