/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata.cpr.medcom;

import static com.trifork.stamdata.Preconditions.checkNotNull;

import java.io.StringWriter;
import java.sql.SQLException;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;

import javax.annotation.Resource;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.ws.WebServiceContext;
import javax.xml.ws.handler.MessageContext;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;
import org.xml.sax.SAXParseException;

import com.trifork.stamdata.Nullable;
import com.trifork.stamdata.models.PersonVO;
import com.trifork.stamdata.models.YderVO;

import dk.nsi.stamdata.cpr.SoapUtils;
import dk.nsi.stamdata.cpr.dao.CprDAO;
import dk.nsi.stamdata.cpr.dao.SikredeDAO;
import dk.nsi.stamdata.cpr.dao.YderDAO;
import dk.nsi.stamdata.cpr.util.UnreliableHandler;
import dk.sds.nsp.audit.Audit;
import dk.sds.nsp.audit.AuditBuilder;
import dk.sds.nsp.audit.AuditBuilder.TypeOfInformation;

abstract public class DetGodeCPROpslagCommon {

    protected final UnreliableHandler unreliableHandler;
    protected final SikredeDAO sikredeDAO;
    protected final CprDAO cprDao;
    protected final YderDAO yderDao;

    @Resource
    protected WebServiceContext wsContext;

    private static final Logger logger = Logger.getLogger(DetGodeCPROpslagCommon.class);

    public DetGodeCPROpslagCommon(CprDAO cprDao, SikredeDAO sikredeDAO, YderDAO yderDao,
                                  UnreliableHandler unreliableHandler) {

        this.cprDao = cprDao;
        this.sikredeDAO = sikredeDAO;
        this.yderDao = yderDao;
        this.unreliableHandler = unreliableHandler;
    }

    protected YderVO getYder(String ydernummer, Date currentDate) throws SQLException {
        return yderDao.findYder(ydernummer, currentDate);
    }

    // HELPERS
    protected PersonVO fetchPersonWithPnr(String cpr, Date currentTime) {
        checkNotNull(cpr, "pnr");
        PersonVO person = null;
        try {
            person = cprDao.findCpr(cpr, currentTime);
        } catch (Exception e) {
            logger.error(e.getMessage(),e);
        }
        unreliableHandler.defaultUnreliableFields(person);

        return person;
    }

    // HELPERS
    protected void checkInputParameters(@Nullable String pnr) {
        if (StringUtils.isBlank(pnr)) {
            throw SoapUtils.newSOAPSenderFault("PersonCivilRegistrationIdentifier was not set in request, but is required.");
        }
    }

    protected void logAccess(String requestedCPR, String clientCVR, String context) {
        logger.info("type=auditlog, service=stamdata-cpr, client_cvr="+clientCVR+", requested_cpr="+requestedCPR);

        TypeOfInformation informationType = AuditBuilder.TypeOfInformation.SensitivePersonalInformation;
    	String component = "SDM";
    	String key = "cpr";

        AuditBuilder auditBuilder = Audit.createAuditBuilder();
        auditBuilder.addAuditInformation(component, context, informationType, key, requestedCPR);
    }

    protected void checkParseError(Object input, Class inputType) {
        MessageContext messageContext = wsContext.getMessageContext();
        SAXParseException parseException = (SAXParseException)messageContext.get(ValidationHandler.PARSE_ERROR);
        if (parseException != null) {
            throw SoapUtils.newSOAPSenderFault(parseException.getMessage());
        }
        StringWriter writer = new StringWriter();
        try {
            JAXBContext context = getJAXBContext(inputType);
            Marshaller m = context.createMarshaller();
            m.marshal(input, writer);
            messageContext.put(ValidationHandler.REQUEST_PAYLOAD, writer.toString());
        } catch (JAXBException e) {
            logger.warn("Unable to marshal request", e);
        }
        messageContext.put(ValidationHandler.REQUEST_DONE, true);
    }

    static Map<Class,JAXBContext> jaxbcontexts = new HashMap<Class,JAXBContext>();

    public JAXBContext getJAXBContext(Class inputType) throws JAXBException {
        JAXBContext jaxbcontext = jaxbcontexts.get(inputType);
        if (jaxbcontext == null) {
            jaxbcontext = JAXBContext.newInstance(inputType);
            jaxbcontexts.put(inputType, jaxbcontext);
        }
        return jaxbcontext;
    }
}
