/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata.cpr.medcom.v104;

import java.sql.SQLException;
import java.text.ParseException;
import java.util.Date;

import javax.jws.WebParam;
import javax.jws.WebService;
import javax.xml.ws.Holder;

import com.trifork.stamdata.SDMException;
import dk.nsi.stamdata.security.AccessControl;
import dk.nsi.stamdata.security.RequestContextProvider;
import dk.nsi.stamdata.security.RequestContextVO;
import org.apache.log4j.Logger;

import com.google.inject.Inject;
import com.sun.xml.ws.developer.SchemaValidation;
import com.trifork.stamdata.jaxws.GuiceInstanceResolver.GuiceWebservice;
import com.trifork.stamdata.models.PersonVO;
import com.trifork.stamdata.models.YderVO;

import dk.nsi.stamdata.cpr.SoapUtils;
import dk.nsi.stamdata.cpr.dao.CprDAO;
import dk.nsi.stamdata.cpr.dao.SikredeDAO;
import dk.nsi.stamdata.cpr.dao.YderDAO;
import dk.nsi.stamdata.cpr.mapping.v104.PersonMapper;
import dk.nsi.stamdata.cpr.medcom.DetGodeCPROpslagCommon;
import dk.nsi.stamdata.cpr.medcom.FaultMessages;
import dk.nsi.stamdata.cpr.medcom.ValidationHandler;
import dk.nsi.stamdata.cpr.models.HealthCareInformation;
import dk.nsi.stamdata.cpr.util.UnreliableHandler;
import dk.nsi.stamdata.jaxws.generated.Header;
import dk.nsi.stamdata.jaxws.generated.Security;
import dk.sdsd.nsp.slalog.api.SLALogItem;
import dk.sosi.seal.model.constants.FaultCodeValues;
import oio.medcom.cprservice._1_0_4.DGWSFault;
import oio.medcom.cprservice._1_0_4.DetGodeCPROpslag;
import oio.medcom.cprservice._1_0_4.GetPersonInformationIn;
import oio.medcom.cprservice._1_0_4.GetPersonInformationOut;
import oio.medcom.cprservice._1_0_4.GetPersonWithHealthCareInformationIn;
import oio.medcom.cprservice._1_0_4.GetPersonWithHealthCareInformationOut;
import oio.medcom.cprservice._1_0_4.PersonInformationStructureType;
import oio.medcom.cprservice._1_0_4.PersonWithHealthCareInformationStructureType;

@WebService(endpointInterface="oio.medcom.cprservice._1_0_4.DetGodeCPROpslag")
@GuiceWebservice
@SchemaValidation(handler = ValidationHandler.class)
public class DetGodeCPROpslagImpl extends DetGodeCPROpslagCommon implements DetGodeCPROpslag
{
    private static final String NS_DET_GODE_CPR_OPSLAG_104 = "urn:oio:medcom:cprservice:1.0.4";
    private static final String NS_DGWS_1_0 = "http://www.medcom.dk/dgws/2006/04/dgws-1.0.xsd";
    private static final String NS_WS_SECURITY = "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd";
    private static final String WHITE_LIST_SERVICE_NAME = "SDM";

    private final RequestContextProvider requestContextProvider;
    private final PersonMapper personMapper;
    private final SLALogItem slaLogItem;
    private static final Logger logger = Logger.getLogger(DetGodeCPROpslagImpl.class);

    @Inject
    private AccessControl accessControl;

    /**
     * Dummy
     */
    public DetGodeCPROpslagImpl() {
        super(null, null, null, null);
        slaLogItem = null;
        requestContextProvider = null;
        personMapper = null;
    }

    @Inject
    DetGodeCPROpslagImpl(RequestContextProvider requestContextProvider, CprDAO cprDao, PersonMapper personMapper, SikredeDAO sikredeDAO, YderDAO yderDao,
                         SLALogItem slaLogItem, UnreliableHandler unreliableHandler)
    {
        super(cprDao, sikredeDAO, yderDao, unreliableHandler);
        this.personMapper = personMapper;
        this.requestContextProvider = requestContextProvider;
        personMapper.setCprDao(cprDao);
        this.slaLogItem = slaLogItem;
    }

    @Override
    public GetPersonWithHealthCareInformationOut getPersonWithHealthCareInformation(
            @WebParam(name = "Security", targetNamespace = NS_WS_SECURITY, mode = WebParam.Mode.INOUT, partName = "wsseHeader") Holder<Security> wsseHeader,
            @WebParam(name = "Header", targetNamespace = NS_DGWS_1_0, mode = WebParam.Mode.INOUT, partName = "medcomHeader") Holder<Header> medcomHeader,
            @WebParam(name = "getPersonWithHealthCareInformationIn", targetNamespace = NS_DET_GODE_CPR_OPSLAG_104, partName = "parameters") GetPersonWithHealthCareInformationIn input)
            throws DGWSFault
    {
        RequestContextVO requestContext = null;
        try {
            requestContext = requestContextProvider.getRequestContext(input.getPersonCivilRegistrationIdentifier(), slaLogItem);
            accessControl.checkAccess(requestContext, WHITE_LIST_SERVICE_NAME);
        } catch (SDMException e) {
            throw SoapUtils.newDGWSFault(wsseHeader, medcomHeader, e.getMessage(), FaultCodeValues.INVALID_IDCARD, DGWSFault.class);
        }
        checkParseError(input, GetPersonWithHealthCareInformationIn.class);
        SoapUtils.updateSlaLog(medcomHeader, "getPersonWithHealthCareInformation", slaLogItem, requestContext.getServiceActor().getOrganizationId());
        SoapUtils.setHeadersToOutgoing(wsseHeader, medcomHeader);

        Date now = Date.from(requestContext.getTime());

        // 1. Check the white list to see if the client is authorized.
        String pnr = input.getPersonCivilRegistrationIdentifier();
        logAccess(pnr, requestContext.getServiceActor().getOrganizationId(), "SCES v1.0.4 getPersonWithHealthcareInformation");

        // 2. Validate the input parameters.
        checkInputParameters(pnr);

        // 2. Fetch the person from the database.
        PersonVO person = fetchPersonWithPnr(pnr,now);

        if (person == null) {
            returnNoData(wsseHeader, medcomHeader);
        }

        HealthCareInformation healthCareInformation = null;
        try {
            healthCareInformation = sikredeDAO.findHealthCareInformationWithFrontierInformation(pnr, now);
        } catch (SQLException e) {
            returnProcessingProblem(wsseHeader, medcomHeader, e);
        } catch (ParseException e) {
            returnProcessingProblem(wsseHeader, medcomHeader, e);
		}

        YderVO yder = null;
        if (healthCareInformation != null) {
            try {
                yder = getYder(healthCareInformation.getProviderNo(), now);
            } catch (SQLException e) {
                returnProcessingProblem(wsseHeader, medcomHeader, e);
            }
        }
        YderVO yderGL = null;
        if (healthCareInformation != null && healthCareInformation.getSYdernrGl() != null) {
            try {
				yderGL = getYder(healthCareInformation.getSYdernrGl(), now);
            } catch (SQLException e) {
                returnProcessingProblem(wsseHeader, medcomHeader, e);
            }
        }


        // If the relation is not found we have to use fake data to satisfy the
        // output format.
        GetPersonWithHealthCareInformationOut output = new GetPersonWithHealthCareInformationOut();
        PersonWithHealthCareInformationStructureType personWithHealthCareInformation = null;
        personWithHealthCareInformation = personMapper.map(person, healthCareInformation, yder, yderGL);

        output.setPersonWithHealthCareInformationStructure(personWithHealthCareInformation);

        return output;
    }

    @Override
    public GetPersonInformationOut getPersonInformation(
            @WebParam(name = "Security", targetNamespace = NS_WS_SECURITY, header = true, mode = WebParam.Mode.INOUT, partName = "wsseHeader") Holder<Security> wsseHeader,
            @WebParam(name = "Header", targetNamespace = NS_DGWS_1_0, header = true, mode = WebParam.Mode.INOUT, partName = "medcomHeader") Holder<Header> medcomHeader,
            @WebParam(name = "getPersonInformationIn", targetNamespace = NS_DET_GODE_CPR_OPSLAG_104, partName = "parameters") GetPersonInformationIn input) throws DGWSFault {
        RequestContextVO requestContext = null;
        try {
            requestContext = requestContextProvider.getRequestContext(input.getPersonCivilRegistrationIdentifier(), slaLogItem);
        } catch (SDMException e) {
            throw SoapUtils.newDGWSFault(wsseHeader, medcomHeader, e.getMessage(), FaultCodeValues.INVALID_IDCARD, DGWSFault.class);
        }
        try { accessControl.checkAccess(requestContext, WHITE_LIST_SERVICE_NAME); }
        catch (SDMException e) {
            throw SoapUtils.newDGWSFault(wsseHeader, medcomHeader, e.getMessage(), FaultCodeValues.NOT_AUTHORIZED, DGWSFault.class);
        }
        checkParseError(input, GetPersonInformationIn.class);
        SoapUtils.updateSlaLog(medcomHeader, "getPersonInformation", slaLogItem, requestContext.getServiceActor().getOrganizationId());
        SoapUtils.setHeadersToOutgoing(wsseHeader, medcomHeader);

        Date now = Date.from(requestContext.getTime());

        // 1. Check the white list to see if the client is authorized.
        String pnr = input.getPersonCivilRegistrationIdentifier();
        logAccess(pnr, requestContext.getServiceActor().getOrganizationId(), "SCES v1.0.4 getPersonInformation");

        // 2. Validate the input parameters.
        checkInputParameters(pnr);

        // 3. Fetch the person from the database.
        PersonVO person = null;

        person = fetchPersonWithPnr(pnr,now);

        if (person == null) {
            returnNoData(wsseHeader, medcomHeader);
        }

        // We now have the requested person. Use it to fill in
        // the response.
        GetPersonInformationOut output = new GetPersonInformationOut();

        PersonInformationStructureType personInformation = personMapper.map(person, PersonMapper.CPRProtectionLevel.DoNotCensorCPR);

        output.setPersonInformationStructure(personInformation);

        return output;
    }

    private void returnProcessingProblem(Holder<Security> wsseHeader, Holder<Header> medcomHeader, Exception e) throws DGWSFault {
        logger.error(e.getMessage(), e);
        throw SoapUtils.newDGWSFault(wsseHeader, medcomHeader, FaultMessages.INTERNAL_SERVER_ERROR, FaultCodeValues.PROCESSING_PROBLEM, DGWSFault.class);
    }

    private void returnNoData(Holder<Security> wsseHeader, Holder<Header> medcomHeader) throws DGWSFault {
        throw SoapUtils.newDGWSFault(wsseHeader, medcomHeader, FaultMessages.NO_DATA_FOUND_FAULT_MSG, FaultMessages.NO_DATA_FOUND_FAULT_MSG, DGWSFault.class);
    }

}
