/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata.cpr.pvit;

import java.sql.SQLException;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.XMLGregorianCalendar;

import dk.nsi._2014._03._01.stamdatacpr.PersonLookupResponseType;
import dk.nsi.stamdata.cpr.mapping.v102.PersonMapper;
import oio.medcom.cprservice._1_0.PersonInformationStructureType;
import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import com.google.common.collect.Maps;
import com.trifork.stamdata.models.PersonVO;

import dk.nsi._2011._09._23.stamdatacpr.NamePersonQueryType;
import dk.nsi.stamdata.cpr.dao.CprDAO;
import dk.nsi.stamdata.cpr.util.UnreliableHandler;

public abstract class StamdataPersonResponseFinderCommon {

    private final static Logger logger = Logger.getLogger(StamdataPersonResponseFinderCommon.class);
    private CprDAO cprDao;
    private final UnreliableHandler unreliableHandler;
    private final PersonMapper personMapper;
    private String auditlogText;

    public StamdataPersonResponseFinderCommon(CprDAO cprDao, PersonMapper personMapper, UnreliableHandler unreliableHandler)
    {
        this.cprDao = cprDao;
        this.personMapper = personMapper;
        this.unreliableHandler = unreliableHandler;
    }

    public abstract PersonInformationStructureType mapPerson(PersonVO person, PersonMapper.CPRProtectionLevel cprProtection);

    protected PersonMapper getPersonMapper() {
        return personMapper;
    }

    public PersonLookupResponseType answerCprRequest(String cvr, String cpr, PersonMapper.CPRProtectionLevel cprProtection) throws SQLException, DatatypeConfigurationException {
        PersonLookupResponseType response = new PersonLookupResponseType();
        List<PersonInformationStructureType> personInformationStructure = response.getPersonInformationStructure();
        PersonVO person = findPersonForCprRequest(cvr, cpr, new Date());
        if (person != null) {
            personInformationStructure.add(mapPerson(person, cprProtection));
        }
        return response;
    }

    public PersonLookupResponseType answerCivilRegistrationNumberListPersonRequest(String cvr, List<String> civilRegistrationNumberList, PersonMapper.CPRProtectionLevel cprProtection, String cprForFiltering) throws SQLException {
        PersonLookupResponseType response = new PersonLookupResponseType();
        List<PersonInformationStructureType> personInformationStructure = response.getPersonInformationStructure();
        for (PersonVO person : findPersonsForCivilRegistrationNumberListPersonRequest(cvr, civilRegistrationNumberList, new Date())) {
        	if (includePerson(cprForFiltering, person.getCpr())) {
        		personInformationStructure.add(mapPerson(person, cprProtection));
        	}
        }
        return response;
    }

    public PersonLookupResponseType answerBirthDatePersonRequest(String cvr, XMLGregorianCalendar birthDate, PersonMapper.CPRProtectionLevel cprProtection, String cprForFiltering) throws SQLException, DatatypeConfigurationException {
        PersonLookupResponseType response = new PersonLookupResponseType();
        List<PersonInformationStructureType> personInformationStructure = response.getPersonInformationStructure();
        List<PersonVO> persons = findPersonsForBirthDatePersonRequest(cvr, birthDate, new Date());
        for (PersonVO person : persons) {
        	if (includePerson(cprForFiltering, person.getCpr())) {
        		personInformationStructure.add(mapPerson(person, cprProtection));	
        	}
        }
        logger.info("type=auditlog, filtering_based_on_cpr="+ (persons.size() - personInformationStructure.size()));
        return response;
    }

    public PersonLookupResponseType answerNamePersonRequest(String cvr, NamePersonQueryType namePerson, PersonMapper.CPRProtectionLevel cprProtection, String cprForFiltering) throws SQLException, DatatypeConfigurationException {
        PersonLookupResponseType response = new PersonLookupResponseType();
        List<PersonInformationStructureType> personInformationStructure = response.getPersonInformationStructure();

        List<PersonVO> persons = findPersonsForNamePersonRequest(cvr, namePerson, new Date());
        for (PersonVO person : persons) {
        	if (includePerson(cprForFiltering, person.getCpr())) {
        		personInformationStructure.add(mapPerson(person, cprProtection));
        	}
        }
        logger.info("type=auditlog, filtering_based_on_cpr="+ (persons.size() - personInformationStructure.size()));
        return response;
    }

    protected PersonVO findPersonForCprRequest(String cvr, String cpr, Date currentTime) throws SQLException {
        PersonVO person = cprDao.findCpr(cpr, currentTime);
        if (person != null) {
            unreliableHandler.defaultUnreliableFields(person);
        }
        boolean wasFound = (person != null);
        logger.info("type=auditlog, client_cvr="+cvr+", requested_cpr="+cpr+", record_was_returned="+wasFound);
        return person;
    }

    protected List<PersonVO> findPersonsForCivilRegistrationNumberListPersonRequest(String cvr, List<String> civilRegistrationNumberList, Date currentTime) throws SQLException {
        List<PersonVO> persons = new LinkedList<PersonVO>();
        for (String cpr : civilRegistrationNumberList) {
            PersonVO person = findPersonForCprRequest(cvr, cpr, currentTime);
            if (person != null) {
                persons.add(person);
            }
        }
        return persons;
    }

    protected List<PersonVO> findPersonsForBirthDatePersonRequest(String cvr, XMLGregorianCalendar birthDate, Date currentTime) throws SQLException
    {
        List<PersonVO> persons = cprDao.findBirthday(birthDate.toGregorianCalendar().getTime(),currentTime);
        logger.info("type=auditlog, client_cvr="+cvr+", search_birthday_param=" + birthDate.toGregorianCalendar());
        for (PersonVO person : persons) {
            unreliableHandler.defaultUnreliableFields(person);
            logger.info("type=auditlog, client_cvr=" + cvr + ", cpr_of_returned_person=" + person.getCpr());
        }
        return persons;
    }

    protected List<PersonVO> findPersonsForNamePersonRequest(String cvr, NamePersonQueryType namePerson, Date currentTime) throws SQLException
    {
        Map<String, Object> columnValuePairs = Maps.newHashMap();

        columnValuePairs.put("Fornavn", namePerson.getPersonGivenName());
        columnValuePairs.put("Efternavn", namePerson.getPersonSurnameName());

        if (!StringUtils.isBlank(namePerson.getPersonMiddleName())) {
            columnValuePairs.put("Mellemnavn", namePerson.getPersonMiddleName());
        }
        logger.info("type=auditlog, client_cvr="+cvr+", requested_name="+namePerson.getPersonGivenName()+" "+namePerson.getPersonMiddleName()+" "+namePerson.getPersonSurnameName());

        List<PersonVO> persons = cprDao.findName(namePerson.getPersonGivenName(),namePerson.getPersonMiddleName(),namePerson.getPersonSurnameName(),currentTime);
        for (PersonVO person : persons) {
            logger.info("type=auditlog, client_cvr="+cvr+", cvr_of_returned_person="+person.getCpr());
            unreliableHandler.defaultUnreliableFields(person);
        }

        return persons;
    }

	public void setAuditlogText(String auditlogText) {
		this.auditlogText = auditlogText;
	}

	public String getAuditlogText() {
		return this.auditlogText;
	}
	
    private boolean includePerson(String cprForFiltering, String actualCpr) {
    	// hvis specifik cpr er medsendt, så skal det fundne cpr matche
    	if (cprForFiltering != null && !cprForFiltering.equals(actualCpr)) {
    		return false;
    	} else {
    		return true;
    	}
    }

}
