/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata.cpr.pvit.v120;

import java.sql.SQLException;

import javax.inject.Inject;
import javax.jws.WebService;
import javax.xml.datatype.DatatypeConfigurationException;

import org.apache.log4j.Logger;

import com.sun.xml.ws.developer.SchemaValidation;
import com.trifork.stamdata.SDMException;
import com.trifork.stamdata.jaxws.GuiceInstanceResolver;

import dk.nsi._2011._09._23.stamdatacpr.PersonLookupRequestType;
import dk.nsi._2014._03._01.stamdatacpr.PersonLookupResponseType;
import dk.nsi._2020._10._01.stamdatacpr.PersonSearchRequestType;
import dk.nsi._2020._10._01.stamdatacpr.StamdataPersonLookup;
import dk.nsi.stamdata.cpr.SoapUtils;
import dk.nsi.stamdata.cpr.medcom.FaultMessages;
import dk.nsi.stamdata.cpr.pvit.StamdataLookupCommon;
import dk.nsi.stamdata.cpr.pvit.v101.StamdataPersonResponseFinder;
import dk.nsi.stamdata.cpr.pvit.v110.StamdataPersonWithConsentResponseFinder;
import dk.nsi.stamdata.security.AccessControl;
import dk.nsi.stamdata.security.RequestContextProvider;
import dk.nsi.stamdata.security.RequestContextVO;
import dk.sdsd.nsp.slalog.api.SLALogItem;
import dk.sosi.seal.model.constants.FaultCodeValues;

@WebService(endpointInterface="dk.nsi._2020._10._01.stamdatacpr.StamdataPersonLookup")
@GuiceInstanceResolver.GuiceWebservice
@SchemaValidation
public class StamdataPersonLookupImpl extends StamdataLookupCommon implements StamdataPersonLookup {
    private final static Logger logger = Logger.getLogger(StamdataPersonLookupImpl.class);
    private static final String WHITELIST_PERSON_DETAILS = "SDM";
    private static final String WHITELIST_WITH_CONSENT = "SDMwithConsent";
    private static final String WHITELIST_SEARCH = "SDMwithCPR";
    private final StamdataPersonResponseFinder stamdataPersonResponseFinder;
    private final StamdataPersonWithConsentResponseFinder stamdataPersonWithConsentResponseFinder;
    private final StamdataSearchPersonResponseFinder stamdataSearchPersonResponseFinder;
    private final RequestContextProvider requestContextProvider;

    @Inject
    private AccessControl accessControl;

    /**
     * Constructor for this implementation as Guice request scoped bean
     */
    @Inject
    StamdataPersonLookupImpl(RequestContextProvider requestContextProvider, StamdataPersonResponseFinder stamdataPersonResponseFinder, StamdataPersonWithConsentResponseFinder stamdataPersonWithConsentResponseFinder, StamdataSearchPersonResponseFinder stamdataSearchPersonResponseFinder, SLALogItem slaLogItem) {
        super(slaLogItem);
        this.stamdataPersonResponseFinder = stamdataPersonResponseFinder;
        this.stamdataPersonWithConsentResponseFinder = stamdataPersonWithConsentResponseFinder;
        this.stamdataSearchPersonResponseFinder = stamdataSearchPersonResponseFinder;
        this.requestContextProvider = requestContextProvider;
    }

    @Override
    public PersonLookupResponseType getPersonDetails(PersonLookupRequestType request) {
        RequestContextVO requestContext = null;
        try {
            requestContext = requestContextProvider.getRequestContext(request.getCivilRegistrationNumberPersonQuery(), slaLogItem);
            accessControl.checkAccess(requestContext, WHITELIST_PERSON_DETAILS);
        } catch (SDMException e) {
        	throw SoapUtils.newSOAPFaultException(e.getMessage(), FaultCodeValues.INVALID_IDCARD);
        }
        SoapUtils.updateSlaLog(requestContext.getMessageId(), "getPersonDetails", slaLogItem, requestContext.getServiceActor().getOrganizationId());

        try {
        	stamdataPersonResponseFinder.setAuditlogText("SCES v1.2.0 getPersonDetails");
        	PersonLookupResponseType response = StamdataLookupCommon.personDetails(request, requestContext.getServiceActor().getOrganizationId(), stamdataPersonResponseFinder, accessControl.getCprForFilteringBasedOnAccess(requestContext));
            return response;
        } catch (SQLException | DatatypeConfigurationException e) {
            logger.error(e.getMessage(), e);
            throw SoapUtils.newSOAPFaultException(FaultMessages.INTERNAL_SERVER_ERROR, FaultCodeValues.PROCESSING_PROBLEM);
        } catch (SDMException e) {
        	throw SoapUtils.newSOAPFaultException(e.getMessage(), FaultCodeValues.INVALID_IDCARD);
		}
    }

    @Override
    public PersonLookupResponseType getPersonDetailsWithConsent(PersonLookupRequestType request) {
        RequestContextVO requestContext = null;
        try {
            requestContext = requestContextProvider.getRequestContext(request.getCivilRegistrationNumberPersonQuery(), slaLogItem);
            accessControl.checkAccess(requestContext, WHITELIST_WITH_CONSENT);
        } catch (SDMException e) {
            throw SoapUtils.newSOAPFaultException(e.getMessage(), FaultCodeValues.INVALID_IDCARD);
        }
        SoapUtils.updateSlaLog(requestContext.getMessageId(), "getPersonDetailsWithConsent", slaLogItem, requestContext.getServiceActor().getOrganizationId());

        try {
        	stamdataPersonWithConsentResponseFinder.setAuditlogText("SCES v1.2.0 getPersonDetailsWithConsent");
        	PersonLookupResponseType response = StamdataLookupCommon.personDetails(request, requestContext.getServiceActor().getOrganizationId(), stamdataPersonWithConsentResponseFinder, accessControl.getCprForFilteringBasedOnAccess(requestContext));
            return response;
        } catch (SQLException | DatatypeConfigurationException e) {
            logger.error(e.getMessage(), e);
            throw SoapUtils.newSOAPFaultException(FaultMessages.INTERNAL_SERVER_ERROR, FaultCodeValues.PROCESSING_PROBLEM);
        } catch (SDMException e) {
        	throw SoapUtils.newSOAPFaultException(e.getMessage(), FaultCodeValues.INVALID_IDCARD);
        }
    }

    @Override
    public PersonLookupResponseType searchPersonDetails(PersonSearchRequestType request) {
        RequestContextVO requestContext = null;
        try {
            requestContext = requestContextProvider.getRequestContext(null, slaLogItem);
            accessControl.checkAccess(requestContext, WHITELIST_SEARCH);
        } catch (SDMException e) {
            throw SoapUtils.newSOAPFaultException(e.getMessage(), FaultCodeValues.INVALID_IDCARD);
        }
        SoapUtils.updateSlaLog(requestContext.getMessageId(), "searchPersonDetails", slaLogItem, requestContext.getServiceActor().getOrganizationId());

        try {
            stamdataSearchPersonResponseFinder.setAuditlogText("SCES v1.2.0 searchPersonDetails");
            PersonLookupResponseType response =  StamdataLookupCommon.personDetails(request, requestContext.getServiceActor().getOrganizationId(), stamdataSearchPersonResponseFinder, accessControl.getCprForFilteringBasedOnAccess(requestContext));
            return response;
        } catch (SQLException | DatatypeConfigurationException e) {
            logger.error(e.getMessage(), e);
            throw SoapUtils.newSOAPFaultException(FaultMessages.INTERNAL_SERVER_ERROR, FaultCodeValues.PROCESSING_PROBLEM);
        } catch (SDMException e) {
        	throw SoapUtils.newSOAPFaultException(e.getMessage(), FaultCodeValues.INVALID_IDCARD);
        }
    }
}
