/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata.cpr.medcom;

import com.google.inject.Inject;
import com.google.inject.Provider;
import com.trifork.stamdata.jaxws.SealNamespaceResolver;
import com.trifork.stamdata.models.ForaeldreMyndighedRelationVO.FMRTypeKode;
import com.trifork.stamdata.models.PersonVO;
import com.trifork.stamdata.models.SikredeVO;
import com.trifork.stamdata.models.YderVO;
import dk.nsi.stamdata.cpr.AbstractPersonLookupTest;
import dk.nsi.stamdata.cpr.SoapUtils;
import dk.nsi.stamdata.dgws.DGWSHeaderMock;
import dk.nsi.stamdata.dgws.SecurityWrapper;
import dk.nsi.stamdata.guice.GuiceTestRunner;
import dk.nsi.stamdata.security.WhitelistService;
import dk.nsi.stamdata.testing.*;
import oio.medcom.cprservice._1_0_4_1.*;
import org.apache.log4j.*;
import org.joda.time.DateTime;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;

import javax.xml.namespace.QName;
import javax.xml.soap.SOAPFault;
import javax.xml.ws.soap.SOAPFaultException;
import java.io.ByteArrayOutputStream;
import java.net.URL;
import java.sql.Connection;
import java.sql.Statement;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import static junit.framework.Assert.assertNull;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.fail;

@RunWith(GuiceTestRunner.class)
public class DetGodeCPROpslag1041Test extends AbstractPersonLookupTest {

    public static final QName DET_GODE_CPR_OPSLAG_1041_SERVICE = new QName("urn:oio:medcom:cprservice:1.0.4.1", "DetGodeCPROpslagService");
    public static final String WHITELISTED_CVR = "33257872";

    @Inject
    private Provider<Connection> connectionProvider;

    private DetGodeCPROpslag client;

    private GetPersonInformationIn request = new GetPersonInformationIn();
    private GetPersonInformationOut response;
    private GetPersonWithHealthCareInformationOut healthCareResponse;

    private GetPersonWithHealthCareInformationIn healthCareRequest = new GetPersonWithHealthCareInformationIn();

    private StamdataTestHelper testHelper = new StamdataTestHelper();

    TestHelper helper = new TestHelper();
    private DatabaseTestInitializer dbInitializer;

    private Comparator<PersonCustodyStructureType> personCustodyStructureComparator = new Comparator<PersonCustodyStructureType>() {
        public int compare(PersonCustodyStructureType person, PersonCustodyStructureType person1) {
            return person.getPersonCustodyIdentifier().compareTo(person1.getPersonCustodyIdentifier());
        }
    };

    @Inject
    Connection connection;

    @Before
    public void setUpDB() throws Exception
    {
        connection.setAutoCommit(false);
        dbInitializer = new DatabaseTestInitializer(connection);
        dbInitializer.initialize();
        Statement statement = connection.createStatement();
        statement.executeUpdate("INSERT IGNORE INTO whitelist_config SET component_name='"+ WhitelistService.DEFAULT_SERVICE_NAME +"', cvr='"+WHITELISTED_CVR+"'");
        connection.commit();
    }

    @Before
    public void setUp() throws Exception {
        URL wsdlLocation = new URL( String.format("%s/service/DetGodeCPROpslag-1.0.4.1?wsdl", getScesEndpoint()));
        DetGodeCPROpslagService serviceCatalog =
                new DetGodeCPROpslagService(wsdlLocation, DET_GODE_CPR_OPSLAG_1041_SERVICE);
        serviceCatalog.setHandlerResolver(new SealNamespaceResolver());
        client = serviceCatalog.getDetGodeCPROpslag();
    }

    private void sendPersonRequest(String cpr) throws Exception {
    	helper.enableAuditlogging(true);
    	ByteArrayOutputStream outStream = helper.redirectSystemOutToStream();
        request.setPersonCivilRegistrationIdentifier(cpr);
        try {
            performRequest();
        } finally {
            helper.enableAuditlogging(false);
            helper.resetSystemOut();
        }
        assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SCES v1.0.4.1 getPersonInformation\",\"type\":\"SensitivePersonalInformation\",\"key\":\"cpr\",\"value\":\"" + request.getPersonCivilRegistrationIdentifier() + "\"}" + System.lineSeparator(), outStream.toString());
    }

    private void sendPersonRequestLevel4(String cpr) throws Exception {
        request.setPersonCivilRegistrationIdentifier(cpr);
        performRequestLevel4();
    }

    private void performRequestLevel4() throws Exception {
        SecurityWrapper securityHeaders = DGWSHeaderMock.getSecurityWrapper();
        response = client.getPersonInformation(securityHeaders.getSecurity(), securityHeaders.getMedcomHeader(), request);
    }

    private void performRequest() throws Exception {
        SecurityWrapper securityHeaders = DGWSHeaderMock.getSecurityWrapper();
        response = client.getPersonInformation(securityHeaders.getSecurity(), securityHeaders.getMedcomHeader(), request);
    }

    private void sendHealthCareRequest(String cpr) throws Exception {
        healthCareRequest.setPersonCivilRegistrationIdentifier(cpr);
        sendHealthCareRequest();
    }

    private void sendHealthCareRequest() throws Exception {
        helper.enableAuditlogging(true);
        ByteArrayOutputStream outStream = helper.redirectSystemOutToStream();

        testHelper.prepareTestdata(connection);
        SecurityWrapper securityHeaders = DGWSHeaderMock.getSecurityWrapper();
        healthCareResponse = client.getPersonWithHealthCareInformation(securityHeaders.getSecurity(), securityHeaders.getMedcomHeader(), healthCareRequest);

        assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SCES v1.0.4.1 getPersonWithHealthcareInformation\",\"type\":\"SensitivePersonalInformation\",\"key\":\"cpr\",\"value\":\""+healthCareRequest.getPersonCivilRegistrationIdentifier()+"\"}"+System.lineSeparator(), outStream.toString());
        helper.enableAuditlogging(false);
        helper.resetSystemOut();
    }

    @Test
    public void requestForPersonInformation() throws Exception {
        PersonVO person = testHelper.createPerson();
        PersonVO personSpouse = testHelper.createKvinde(StamdataTestHelper.CPR_SPOUSE);
        PersonVO personFar = testHelper.createMand(StamdataTestHelper.CPR_FAR);
        PersonVO personMor = testHelper.createKvinde(StamdataTestHelper.CPR_MOR);
        PersonVO personBarn1 = testHelper.createPerson(StamdataTestHelper.CPR_BARN1);
        PersonVO personBarn2 = testHelper.createPerson(StamdataTestHelper.CPR_BARN2);
        PersonVO personBarn3 = testHelper.createPerson(StamdataTestHelper.CPR_BARN3);
        PersonVO personBarn4 = testHelper.createPerson(StamdataTestHelper.CPR_BARN4);
        testHelper.createSpouse(person, personSpouse);
        testHelper.createForaeldreMyndighedRelation(person, FMRTypeKode.FAR);
        testHelper.createForaeldreMyndighedRelation(person, FMRTypeKode.MOR);
        testHelper.createPersonFamilie(person, personMor, personFar);

        testHelper.createForaeldreMyndighedRelation(personBarn1, FMRTypeKode.FAR);
        testHelper.createPersonFamilie(personBarn1, null, person);

        testHelper.createForaeldreMyndighedRelation(personBarn2, FMRTypeKode.MOR);
        testHelper.createPersonFamilie(personBarn2, null, person);

        testHelper.createForaeldreMyndighedRelation(personBarn3, FMRTypeKode.AI1, person);
        testHelper.createVaerge(personBarn4, person);

        testHelper.prepareTestdata(connection);
        sendPersonRequest(person.getCpr());

        AssertResponse1041.assertPersonInformationStructure(response.getPersonInformationStructure());
    }

    @Test
    public void requestForPersonInformationWithLevel4() throws Exception {
        PersonVO person = testHelper.createPerson();
        PersonVO personSpouse = testHelper.createKvinde(StamdataTestHelper.CPR_SPOUSE);
        PersonVO personFar = testHelper.createMand(StamdataTestHelper.CPR_FAR);
        PersonVO personMor = testHelper.createKvinde(StamdataTestHelper.CPR_MOR);
        PersonVO personBarn1 = testHelper.createPerson(StamdataTestHelper.CPR_BARN1);
        PersonVO personBarn2 = testHelper.createPerson(StamdataTestHelper.CPR_BARN2);
        PersonVO personBarn3 = testHelper.createPerson(StamdataTestHelper.CPR_BARN3);
        PersonVO personBarn4 = testHelper.createPerson(StamdataTestHelper.CPR_BARN4);
        testHelper.createSpouse(person, personSpouse);
        testHelper.createForaeldreMyndighedRelation(person, FMRTypeKode.FAR);
        testHelper.createForaeldreMyndighedRelation(person, FMRTypeKode.MOR);
        testHelper.createPersonFamilie(person, personMor, personFar);

        testHelper.createForaeldreMyndighedRelation(personBarn1, FMRTypeKode.FAR);
        testHelper.createPersonFamilie(personBarn1, null, person);

        testHelper.createForaeldreMyndighedRelation(personBarn2, FMRTypeKode.MOR);
        testHelper.createPersonFamilie(personBarn2, null, person);

        testHelper.createForaeldreMyndighedRelation(personBarn3, FMRTypeKode.AI1, person);
        testHelper.createVaerge(personBarn4, person);

        testHelper.prepareTestdata(connection);
        sendPersonRequestLevel4(person.getCpr());

        AssertResponse1041.assertPersonInformationStructure(response.getPersonInformationStructure());
    }

    @Test
    public void requestForMissingPerson() throws Exception {
        final ByteArrayOutputStream outContent = getLogsFrom(SoapUtils.class);

        try {
            sendPersonRequest(StamdataTestHelper.CPR_MISSING);
            fail();
        } catch (DGWSFault e) {
            assertEquals("Ingen data fundet",e.getMessage());
        }
    }

    @Test
    public void requestForMissingPersonSoapFault() {
        try {
            request.setPersonCivilRegistrationIdentifier(StamdataTestHelper.CPR_MISSING);
            SecurityWrapper securityHeaders = DGWSHeaderMock.getSecurityWrapper();
            client.getPersonInformation(securityHeaders.getSecurity(), securityHeaders.getMedcomHeader(), request);
            fail();
        } catch (DGWSFault e) {
            assertEquals("Ingen data fundet",e.getMessage());
        }
    }

    public ByteArrayOutputStream getLogsFrom(Class logClass) {
        final String APPENDER_NAME = "log4jRuleAppender";
        final Layout LAYOUT = new SimpleLayout();

        Logger logger;
        final ByteArrayOutputStream outContent = new ByteArrayOutputStream();

        logger = Logger.getLogger(logClass);
        Appender appender = new WriterAppender(LAYOUT, outContent);
        appender.setName(APPENDER_NAME);
        logger.addAppender(appender);
        return outContent;
    }

    @Test
    public void requestForPersonInformationWithProtection() throws Exception {
        PersonVO person = testHelper.createPerson();
        person.setNavnebeskyttelsestartdato(testHelper.formatSimpleDate("2000-01-01"));
        person.setNavnebeskyttelseslettedato(testHelper.formatSimpleDate("2050-01-01"));
        PersonVO personSpouse = testHelper.createKvinde(StamdataTestHelper.CPR_SPOUSE);
        PersonVO personFar = testHelper.createMand(StamdataTestHelper.CPR_FAR);
        PersonVO personMor = testHelper.createKvinde(StamdataTestHelper.CPR_MOR);
        PersonVO personBarn1 = testHelper.createPerson(StamdataTestHelper.CPR_BARN1);
        PersonVO personBarn2 = testHelper.createPerson(StamdataTestHelper.CPR_BARN2);
        PersonVO personBarn3 = testHelper.createPerson(StamdataTestHelper.CPR_BARN3);
        PersonVO personBarn4 = testHelper.createPerson(StamdataTestHelper.CPR_BARN4);
        testHelper.createSpouse(person, personSpouse);
        testHelper.createForaeldreMyndighedRelation(person, FMRTypeKode.FAR);
        testHelper.createForaeldreMyndighedRelation(person, FMRTypeKode.MOR);
        testHelper.createPersonFamilie(person, personMor, personFar);

        testHelper.createForaeldreMyndighedRelation(personBarn1, FMRTypeKode.FAR);
        testHelper.createPersonFamilie(personBarn1, null, person);

        testHelper.createForaeldreMyndighedRelation(personBarn2, FMRTypeKode.MOR);
        testHelper.createPersonFamilie(personBarn2, null, person);

        testHelper.createForaeldreMyndighedRelation(personBarn3, FMRTypeKode.AI1, person);
        testHelper.createVaerge(personBarn4, person);

        testHelper.prepareTestdata(connection);
        sendPersonRequest(person.getCpr());

        AssertResponse1041.assertPersonInformationStructureProtected(response.getPersonInformationStructure());
    }

    @Test
    public void requestForPersonWithHealthCareInformation() throws Exception {
        PersonVO person1 = testHelper.createPerson();
        PersonVO person2 = testHelper.createPerson();
        PersonVO personSpouse = testHelper.createKvinde(StamdataTestHelper.CPR_SPOUSE);
        PersonVO personFar = testHelper.createMand(StamdataTestHelper.CPR_FAR);
        PersonVO personMor = testHelper.createKvinde(StamdataTestHelper.CPR_MOR);
        PersonVO personBarn1 = testHelper.createPerson(StamdataTestHelper.CPR_BARN1);
        PersonVO personBarn2 = testHelper.createPerson(StamdataTestHelper.CPR_BARN2);
        PersonVO personBarn3 = testHelper.createPerson(StamdataTestHelper.CPR_BARN3);
        PersonVO personBarn4 = testHelper.createPerson(StamdataTestHelper.CPR_BARN4);
        testHelper.createSpouse(person1, personSpouse);
        testHelper.createForaeldreMyndighedRelation(person1, FMRTypeKode.FAR);
        testHelper.createForaeldreMyndighedRelation(person1, FMRTypeKode.MOR);
        testHelper.createPersonFamilie(person1, personMor, personFar);

        testHelper.createForaeldreMyndighedRelation(personBarn1, FMRTypeKode.FAR);
        testHelper.createPersonFamilie(personBarn1, null, person1);

        testHelper.createForaeldreMyndighedRelation(personBarn2, FMRTypeKode.MOR);
        testHelper.createPersonFamilie(personBarn2, null, person1);

        testHelper.createForaeldreMyndighedRelation(personBarn3, FMRTypeKode.AI1, person1);
        testHelper.createVaerge(personBarn4, person1);
        YderVO yderRecord1 = testHelper.createYder("1234");
        YderVO yderRecord2 = testHelper.createYder("4321");
        SikredeVO sikredeRecord1 = testHelper.createSikredeWithFrontierInfo(person1, yderRecord1, "1", new DateTime(Factories.YESTERDAY),new DateTime(Factories.YEAR_2000),true);
        sikredeRecord1.setSYdernrGl("4321");
        testHelper.createSikrede(person2, yderRecord2, "4", new DateTime(2011, 10, 10, 0, 0));
        // Having multiple persons ensures that we are selecting the
        // right one.

        sendHealthCareRequest(person1.getCpr());

        PersonWithHealthCareInformationStructureType personWithHealthCareInformationStructure = healthCareResponse.getPersonWithHealthCareInformationStructure();
        AssertResponse1041.assertPersonWithHealthCareInformationStructure(personWithHealthCareInformationStructure);
    }

    @Test
    public void requestForPersonWithHealthCareInformationWithProtection() throws Exception {
        PersonVO person1 = testHelper.createPerson();
        person1.setNavnebeskyttelsestartdato(testHelper.formatSimpleDate("2000-01-01"));
        person1.setNavnebeskyttelseslettedato(testHelper.formatSimpleDate("2050-01-01"));
        PersonVO person2 = testHelper.createPerson();
        PersonVO personSpouse = testHelper.createKvinde(StamdataTestHelper.CPR_SPOUSE);
        PersonVO personFar = testHelper.createMand(StamdataTestHelper.CPR_FAR);
        PersonVO personMor = testHelper.createKvinde(StamdataTestHelper.CPR_MOR);
        PersonVO personBarn1 = testHelper.createPerson(StamdataTestHelper.CPR_BARN1);
        PersonVO personBarn2 = testHelper.createPerson(StamdataTestHelper.CPR_BARN2);
        PersonVO personBarn3 = testHelper.createPerson(StamdataTestHelper.CPR_BARN3);
        PersonVO personBarn4 = testHelper.createPerson(StamdataTestHelper.CPR_BARN4);
        testHelper.createSpouse(person1, personSpouse);
        testHelper.createForaeldreMyndighedRelation(person1, FMRTypeKode.FAR);
        testHelper.createForaeldreMyndighedRelation(person1, FMRTypeKode.MOR);
        testHelper.createPersonFamilie(person1, personMor, personFar);

        testHelper.createForaeldreMyndighedRelation(personBarn1, FMRTypeKode.FAR);
        testHelper.createPersonFamilie(personBarn1, null, person1);

        testHelper.createForaeldreMyndighedRelation(personBarn2, FMRTypeKode.MOR);
        testHelper.createPersonFamilie(personBarn2, null, person1);

        testHelper.createForaeldreMyndighedRelation(personBarn3, FMRTypeKode.AI1, person1);
        testHelper.createVaerge(personBarn4, person1);
        YderVO yderRecord1 = testHelper.createYder("1234");
        YderVO yderRecord2 = testHelper.createYder("4321");
        SikredeVO sikredeRecord1 = testHelper.createSikredeWithFrontierInfo(person1, yderRecord1, "1", new DateTime(Factories.YESTERDAY),new DateTime(Factories.YEAR_2000),true);
        sikredeRecord1.setSYdernrGl("4321");
        testHelper.createSikrede(person2, yderRecord2, "4", new DateTime(2011, 10, 10, 0, 0));
        // Having multiple persons ensures that we are selecting the
        // right one.

        sendHealthCareRequest(person1.getCpr());

        PersonWithHealthCareInformationStructureType personWithHealthCareInformationStructure = healthCareResponse.getPersonWithHealthCareInformationStructure();
        AssertResponse1041.assertPersonWithHealthCareInformationStructureProtected(personWithHealthCareInformationStructure);
    }

    @Test
    public void requestWithGroupCode0() throws Exception {
        PersonVO person1 = testHelper.createPerson();
        PersonVO person2 = testHelper.createPerson();
        YderVO yderRecord1 = testHelper.createYder("1234");
        YderVO yderRecord2 = testHelper.createYder("4321");
        SikredeVO sikredeRecord1 = testHelper.createSikredeWithFrontierInfo(person1, yderRecord1, "0", new DateTime(Factories.YESTERDAY),new DateTime(Factories.YEAR_2000),true);
        sikredeRecord1.setSYdernrGl("4321");
        testHelper.createSikrede(person2, yderRecord2, "4", new DateTime(2011, 10, 10, 0, 0));
        // Having multiple persons ensures that we are selecting the
        // right one.

        sendHealthCareRequest(person1.getCpr());

        PersonWithHealthCareInformationStructureType personWithHealthCareInformationStructure = healthCareResponse.getPersonWithHealthCareInformationStructure();
        PersonHealthCareInformationStructureType personHealthCareInformationStructure = personWithHealthCareInformationStructure.getPersonHealthCareInformationStructure();
        AssertResponse1041.assertPersonHealthCareInformationStructure(personHealthCareInformationStructure, PublicHealthInsuranceGroupIdentifierType.SYGESIKRINGSGRUPPE_0);
    }

    @Test
    public void requestWithEmptySikredeTable() throws Exception {
        PersonVO person1 = testHelper.createPerson();

        sendHealthCareRequest(person1.getCpr());

        PersonWithHealthCareInformationStructureType personWithHealthCareInformationStructure = healthCareResponse.getPersonWithHealthCareInformationStructure();
        PublicHealthInsuranceGroupIdentifierType publicHealthInsuranceGroupIdentifier =
                personWithHealthCareInformationStructure.getPersonHealthCareInformationStructure().getPersonPublicHealthInsurance().getPublicHealthInsuranceGroupIdentifier();
        // Fallback PublicHealthInsuranceGroupIdentifier should be sygesikringsgruppe_0
        assertEquals(PublicHealthInsuranceGroupIdentifierType.SYGESIKRINGSGRUPPE_0, publicHealthInsuranceGroupIdentifier);
    }

    @Test
    public void requestForSeperatedPerson() throws Exception {
        PersonVO person = testHelper.createPerson();
        PersonVO personSpouse = testHelper.createPerson(StamdataTestHelper.CPR_SPOUSE);
        testHelper.createSeperated(person, personSpouse);

        testHelper.prepareTestdata(connection);
        sendPersonRequest(person.getCpr());
        PersonInformationStructureType personInformationStructure = response.getPersonInformationStructure();

        assertNull(personInformationStructure.getPersonSpouseStructure());
    }

    @Test
    public void requestForUnmarriedPerson() throws Exception {
        PersonVO person = testHelper.createPerson();
        testHelper.createUnmarried(person);

        testHelper.prepareTestdata(connection);
        sendPersonRequest(person.getCpr());
        PersonInformationStructureType personInformationStructure = response.getPersonInformationStructure();

        assertNull(personInformationStructure.getPersonSpouseStructure());
    }

    @Test
    public void testSignatureInvalidModelBuildException() throws Exception {

        String fileName = "src/test/resources/data/validXml.xml";
        String xml = RequestHelper.readFile(fileName);
        String replication_endpoint = String.format("%s/service/DetGodeCPROpslag-1.0.4", getScesEndpoint());

        SecurityContextTestHelper.setIdCardExpired();
        String responseXml = RequestHelper.sendRequest(replication_endpoint, "http://rep.oio.dk/medcom.sundcom.dk/xml/wsdl/2007/06/28/getPersonInformation", xml);

        String faultstring = RequestHelper.getFaultstring(responseXml);
        assertEquals("Cannot find dispatch method for {urn:oio:medcom:cprservice:1.0.2}getPersonInformationIn", faultstring);
    }

    @Test
    public void requestForCustodyAndenIndehaver() throws Exception {
        PersonVO person = testHelper.createPerson(StamdataTestHelper.CPR_BARN1);
        PersonVO personFar = testHelper.createMand(StamdataTestHelper.CPR_FAR);
        PersonVO personMor = testHelper.createKvinde(StamdataTestHelper.CPR_MOR);

        testHelper.createPersonFamilie(person, null, personFar);
        testHelper.createForaeldreMyndighedRelation(person, FMRTypeKode.FAR, personFar);

        testHelper.createForaeldreMyndighedRelation(person, FMRTypeKode.AI1, personMor);

        testHelper.prepareTestdata(connection);
        sendPersonRequest(StamdataTestHelper.CPR_BARN1);

        PersonInformationStructureType personInformationStructure = response.getPersonInformationStructure();
        List<PersonCustodyStructureType> personCustodyStructure = personInformationStructure.getPersonCustodyStructure();
        assertEquals(2,personCustodyStructure.size());
        PersonCustodyStructureType personCustody = personCustodyStructure.get(0);
        AssertResponse1041.assertpersonCustody(personCustody, StamdataTestHelper.CPR_FAR, PersonCustodyRelationType.FAR);
        personCustody = personCustodyStructure.get(1);
        AssertResponse1041.assertpersonCustody(personCustody, StamdataTestHelper.CPR_MOR, PersonCustodyRelationType.ANDEN);

        sendPersonRequest(StamdataTestHelper.CPR_FAR);

        personInformationStructure = response.getPersonInformationStructure();
        List<PersonWardStructureType> personWardStructure = personInformationStructure.getPersonWardStructure();
        assertEquals(1,personWardStructure.size());
        PersonWardStructureType ward = personWardStructure.get(0);
        assertEquals(PersonWardRelationType.BARN,ward.getPersonWardRelation());
        assertEquals(StamdataTestHelper.CPR_BARN1, ward.getPersonWardIdentifier());

        sendPersonRequest(StamdataTestHelper.CPR_MOR);

        personInformationStructure = response.getPersonInformationStructure();
        personWardStructure = personInformationStructure.getPersonWardStructure();
        assertEquals(1,personWardStructure.size());
        ward = personWardStructure.get(0);
        assertEquals(PersonWardRelationType.BARN,ward.getPersonWardRelation());
        assertEquals(StamdataTestHelper.CPR_BARN1, ward.getPersonWardIdentifier());
    }

    @Test
    public void requestForCustodyMedmor() throws Exception {
        PersonVO person = testHelper.createPerson(StamdataTestHelper.CPR_BARN1);
        PersonVO personMor = testHelper.createKvinde(StamdataTestHelper.CPR_MOR);
        PersonVO personMedmor = testHelper.createKvinde(StamdataTestHelper.CPR_MEDMOR);

        testHelper.createPersonFamilie(person, personMor, null);
        testHelper.createForaeldreMyndighedRelation(person, FMRTypeKode.MOR, personMor);
        testHelper.createForaeldreMyndighedRelation(person, FMRTypeKode.AI1, personMedmor);

        testHelper.prepareTestdata(connection);
        sendPersonRequest(StamdataTestHelper.CPR_BARN1);

        PersonInformationStructureType personInformationStructure = response.getPersonInformationStructure();
        List<PersonCustodyStructureType> personCustodyStructure = personInformationStructure.getPersonCustodyStructure();
        Collections.sort(personCustodyStructure, personCustodyStructureComparator);
        assertEquals(2,personCustodyStructure.size());
        PersonCustodyStructureType personCustody = personCustodyStructure.get(0);
        AssertResponse1041.assertpersonCustody(personCustody, StamdataTestHelper.CPR_MOR, PersonCustodyRelationType.MOR);
        personCustody = personCustodyStructure.get(1);
        AssertResponse1041.assertpersonCustody(personCustody, StamdataTestHelper.CPR_MEDMOR, PersonCustodyRelationType.ANDEN);

        sendPersonRequest(StamdataTestHelper.CPR_MOR);

        personInformationStructure = response.getPersonInformationStructure();
        List<PersonWardStructureType> personWardStructure = personInformationStructure.getPersonWardStructure();
        assertEquals(1,personWardStructure.size());
        PersonWardStructureType ward = personWardStructure.get(0);
        assertEquals(PersonWardRelationType.BARN,ward.getPersonWardRelation());
        assertEquals(StamdataTestHelper.CPR_BARN1, ward.getPersonWardIdentifier());

        sendPersonRequest(StamdataTestHelper.CPR_MEDMOR);

        personInformationStructure = response.getPersonInformationStructure();
        personWardStructure = personInformationStructure.getPersonWardStructure();
        assertEquals(1,personWardStructure.size());
        ward = personWardStructure.get(0);
        assertEquals(PersonWardRelationType.BARN,ward.getPersonWardRelation());
        assertEquals(StamdataTestHelper.CPR_BARN1, ward.getPersonWardIdentifier());
    }

    @Test
    public void requestForCustodyMedfar() throws Exception {
        PersonVO person = testHelper.createPerson(StamdataTestHelper.CPR_BARN1);
        PersonVO personMor = testHelper.createKvinde(StamdataTestHelper.CPR_MOR);
        PersonVO personMedfar = testHelper.createMand(StamdataTestHelper.CPR_MEDFAR);

        testHelper.createPersonFamilie(person, personMor, null);
        testHelper.createForaeldreMyndighedRelation(person, FMRTypeKode.MOR, personMor);
        testHelper.createForaeldreMyndighedRelation(person, FMRTypeKode.AI1, personMedfar);

        testHelper.prepareTestdata(connection);
        sendPersonRequest(StamdataTestHelper.CPR_BARN1);

        PersonInformationStructureType personInformationStructure = response.getPersonInformationStructure();
        List<PersonCustodyStructureType> personCustodyStructure = personInformationStructure.getPersonCustodyStructure();
        Collections.sort(personCustodyStructure, personCustodyStructureComparator);
        assertEquals(2,personCustodyStructure.size());
        PersonCustodyStructureType personCustody = personCustodyStructure.get(0);
        AssertResponse1041.assertpersonCustody(personCustody, StamdataTestHelper.CPR_MOR, PersonCustodyRelationType.MOR);
        personCustody = personCustodyStructure.get(1);
        AssertResponse1041.assertpersonCustody(personCustody, StamdataTestHelper.CPR_MEDFAR, PersonCustodyRelationType.ANDEN);

        sendPersonRequest(StamdataTestHelper.CPR_MOR);

        personInformationStructure = response.getPersonInformationStructure();
        List<PersonWardStructureType>  personWardStructure = personInformationStructure.getPersonWardStructure();
        assertEquals(1,personWardStructure.size());
        PersonWardStructureType ward = personWardStructure.get(0);
        assertEquals(PersonWardRelationType.BARN,ward.getPersonWardRelation());
        assertEquals(StamdataTestHelper.CPR_BARN1, ward.getPersonWardIdentifier());

        sendPersonRequest(StamdataTestHelper.CPR_MEDFAR);

        personInformationStructure = response.getPersonInformationStructure();
        personWardStructure = personInformationStructure.getPersonWardStructure();
        assertEquals(1,personWardStructure.size());
        ward = personWardStructure.get(0);
        assertEquals(PersonWardRelationType.BARN,ward.getPersonWardRelation());
        assertEquals(StamdataTestHelper.CPR_BARN1, ward.getPersonWardIdentifier());
    }

    @Test
    public void requestForVaerge() throws Exception {
        PersonVO person = testHelper.createPerson(StamdataTestHelper.CPR_BARN1);
        PersonVO personVaerge = testHelper.createPerson(StamdataTestHelper.CPR_VAERGE);
        testHelper.createVaerge(person, personVaerge);

        testHelper.prepareTestdata(connection);
        sendPersonRequest(person.getCpr());
        PersonInformationStructureType personInformationStructure = response.getPersonInformationStructure();
        List<PersonCustodyStructureType> personCustodyStructure = personInformationStructure.getPersonCustodyStructure();

        assertEquals(1,personCustodyStructure.size());
        PersonCustodyStructureType personCustody = personCustodyStructure.get(0);
        AssertResponse1041.assertpersonCustody(personCustody, StamdataTestHelper.CPR_VAERGE, PersonCustodyRelationType.VAERGE);

        sendPersonRequest(personVaerge.getCpr());

        personInformationStructure = response.getPersonInformationStructure();
        List<PersonWardStructureType> personWardStructure = personInformationStructure.getPersonWardStructure();
        assertEquals(1,personWardStructure.size());
        PersonWardStructureType ward = personWardStructure.get(0);
        assertEquals(PersonWardRelationType.MYNDLING,ward.getPersonWardRelation());
        assertEquals(StamdataTestHelper.CPR_BARN1, ward.getPersonWardIdentifier());
    }

    @Test
    public void requestForVaergeUdenCPR() throws Exception {
        PersonVO person = testHelper.createPerson(StamdataTestHelper.CPR_BARN1);
        PersonVO personVaerge = testHelper.createPerson(StamdataTestHelper.CPR_VAERGE);
        testHelper.createVaerge(person, null);

        testHelper.prepareTestdata(connection);
        sendPersonRequest(person.getCpr());
        PersonInformationStructureType personInformationStructure = response.getPersonInformationStructure();
        List<PersonCustodyStructureType> personCustodyStructure = personInformationStructure.getPersonCustodyStructure();

        assertEquals(0,personCustodyStructure.size());

        sendPersonRequest(personVaerge.getCpr());

        personInformationStructure = response.getPersonInformationStructure();
        List<PersonWardStructureType> personWardStructure = personInformationStructure.getPersonWardStructure();
        assertEquals(0,personWardStructure.size());
    }

    @Test
    public void requestForPersonWithoutStatus() throws Exception {
        PersonVO person = testHelper.createPerson();
        person.setStatus(null);
        testHelper.prepareTestdata(connection);
        try {
            sendPersonRequest(person.getCpr());
        } catch (Exception e) {
            SOAPFault fault = ((SOAPFaultException)e).getFault();
            Assert.assertTrue(fault.getFirstChild().getFirstChild().toString().contains("Client"));
        }
    }
    
    @Test
    public void requestFrontierPersonWithSBSStatsborgerskabNullAndSBSStatsborgerskabKodeAvailable() throws Exception {
        PersonVO person1 = testHelper.createPerson();

        YderVO yder1 = testHelper.createYder("1234");

        DateTime fromDate = new DateTime(2011, 10, 10, 0, 0);

        SikredeVO sikredeVO = testHelper.createSikredeWithFrontierInfo(person1, yder1, "9", fromDate, fromDate,true);
        sikredeVO.setSBSStatsborgerskab(null);

        sendHealthCareRequest(person1.getCpr());
        
        Assert.assertNotNull(healthCareResponse.getPersonWithHealthCareInformationStructure().getFrontierInformationStructure().getFrontierCitizenshipStructure());
        Assert.assertEquals("GB",healthCareResponse.getPersonWithHealthCareInformationStructure().getFrontierInformationStructure().getFrontierCitizenshipStructure().getFrontierCitizenshipCode());
        Assert.assertEquals("United Kingdom of Great Britain and Northern Ireland (the)",healthCareResponse.getPersonWithHealthCareInformationStructure().getFrontierInformationStructure().getFrontierCitizenshipStructure().getFrontierCitizenship());

    } //Andre kombinationer af SBSStatsborgerskab ikke-udfyldt findes i DetGodeCPROpslagIntegrationTest som generel test, da løsningen er lavet end. Ovenstående test er lavet for at tjekke at denne version har ændringen.

}
