/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata.cpr.pvit;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNull;

import java.io.ByteArrayOutputStream;
import java.net.URL;
import java.sql.Connection;
import java.sql.Statement;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.List;

import javax.inject.Inject;
import javax.xml.datatype.DatatypeFactory;
import javax.xml.datatype.XMLGregorianCalendar;
import javax.xml.namespace.QName;

import org.junit.Before;
import org.junit.Ignore;
import org.junit.Test;

import com.google.common.collect.Lists;
import com.trifork.stamdata.models.PersonVO;

import dk.nsi._2011._09._23.stamdatacpr.NamePersonQueryType;
import dk.nsi._2011._09._23.stamdatacpr.PersonLookupRequestType;
import dk.nsi._2014._03._01.stamdatacpr.PersonLookupResponseType;
import dk.nsi._2020._10._01.stamdatacpr.PersonSearchRequestType;
import dk.nsi._2020._10._01.stamdatacpr.StamdataPersonLookup;
import dk.nsi._2020._10._01.stamdatacpr.StamdataPersonLookupService;
import dk.nsi.stamdata.cpr.AbstractPersonLookupTest;
import dk.nsi.stamdata.security.WhitelistService;
import dk.nsi.stamdata.testing.DatabaseTestInitializer;
import dk.nsi.stamdata.testing.Factories;
import dk.nsi.stamdata.testing.StamdataTestHelper;
import dk.nsi.stamdata.testing.TestHelper;
import dk.oio.rep.cpr_dk.xml.schemas.core._2006._01._17.RegularCPRPersonType;
import dk.oio.rep.itst_dk.xml.schemas._2006._01._17.PersonNameStructureType;
import oio.medcom.cprservice._1_0.PersonAddressStructureType;
import oio.medcom.cprservice._1_0.PersonInformationStructureType;

@Ignore("Skal kun køres vha. de klasser der nedarver herfra")
public class StamdataPersonLookup120 extends AbstractPersonLookupTest {
    public static final QName PVIT_SERVICE_QNAME = new QName("http://nsi.dk/2020/10/01/StamdataCpr/", "StamdataPersonLookupService");

    protected List<PersonVO> persons = Lists.newArrayList();
    protected PersonLookupRequestType getRequest = new PersonLookupRequestType();
    protected PersonSearchRequestType searchRequest = new PersonSearchRequestType();
    private PersonLookupRequestType requestWithConsent = new PersonLookupRequestType();
    protected PersonLookupResponseType response;

    TestHelper helper = new TestHelper();
    protected DatabaseTestInitializer dbInitializer;

    public static final String WHITELISTED_CVR = "33257872";

    protected static StamdataPersonLookup client;
    private static StamdataPersonLookupService serviceCatalog;

    SimpleDateFormat dateFormatter = new SimpleDateFormat("yyyy-MM-dd");
    SimpleDateFormat birthdateFormatter = new SimpleDateFormat("yyyy-MM-ddXXX");

    protected enum RequestType { GET_PERSON, GET_PERSON_WITH_CONSENT, SEARCH_PERSON }

    @Inject
    Connection connection;

    @Before
    public void setup() throws Exception {
        dbInitializer = new DatabaseTestInitializer(connection);
        dbInitializer.initialize();
        setUpDB();
        setUpClient();
        setContextOrganisation();
        setContextUserCitizenUser();
    }
    protected void setContextOrganisation() {
    	//empty default
    }
    protected void setContextUserCitizenUser() {
    	//empty default
    }
	protected void setContextUserCitizenUser(String cpr) {
		//empty default
	}


    public void setUpDB() throws Exception
    {
        connection.setAutoCommit(false);
        Statement statement = connection.createStatement();
        statement.executeUpdate("INSERT IGNORE INTO whitelist_config SET component_name='"+ WhitelistService.DEFAULT_SERVICE_NAME +"', cvr='"+WHITELISTED_CVR+"'");
        statement.executeUpdate("INSERT IGNORE INTO whitelist_config SET component_name='SDMwithConsent', cvr='"+WHITELISTED_CVR+"'");
        statement.executeUpdate("INSERT IGNORE INTO whitelist_config SET component_name='SDMwithCPR', cvr='"+WHITELISTED_CVR+"'");
        connection.commit();
    }

    public void setUpClient() throws Exception {
        URL wsdlLocation = new URL(String.format("%s/service/StamdataPersonLookup-1.2.0?wsdl", getScesEndpoint()));
        serviceCatalog = new StamdataPersonLookupService(wsdlLocation, PVIT_SERVICE_QNAME);

        client = serviceCatalog.getStamdataPersonLookup();
    }

    @Test
    public void requestCPRprotectedWithoutConsent1() throws Exception {

        PersonVO person = addPerson("0103952595");
        setContextUserCitizenUser(person.getCpr());
        getRequest.setCivilRegistrationNumberPersonQuery("0103952595");
        prepareDatabaseAndSendRequest();
        assertResponseWithProtectedAddress();
        assertCpr(person);

        String[] logs = helper.getSystemLogs();
        assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SCES v1.2.0 getPersonDetails\",\"type\":\"SensitivePersonalInformation\",\"key\":\"cpr\",\"value\":\"" + getRequest.getCivilRegistrationNumberPersonQuery() + "\"}", logs[0]);
    }
    
    
    @Test
    public void requestCPRprotectedWithoutConsent3() throws Exception {
    	
    	PersonVO person = addPerson("0103952595");
    	setContextUserCitizenUser(person.getCpr());
        GregorianCalendar cal = new GregorianCalendar();
        cal.setTime(persons.get(0).getFoedselsdato());
        XMLGregorianCalendar birthday = DatatypeFactory.newInstance().newXMLGregorianCalendar(cal);
        getRequest.setBirthDatePersonQuery(birthday);
        prepareDatabaseAndSendRequest();
        assertResponseWithProtectedAddress();
        assertCpr("0000000000",null);

        String[] logs = helper.getSystemLogs();
        assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SCES v1.2.0 getPersonDetails\",\"type\":\"SensitivePersonalInformation\",\"key\":\"birthdate\",\"value\":\""+birthdateFormatter.format(getRequest.getBirthDatePersonQuery().toGregorianCalendar().getTime())+"\"}", logs[0]);
    }

    @Test
    public void requestCPRprotectedWithoutConsent4() throws Exception {
    	
    	PersonVO person = addPerson("0103952595");
    	setContextUserCitizenUser(person.getCpr());
        NamePersonQueryType name = new NamePersonQueryType();
        name.setPersonGivenName("Peter");
        name.setPersonMiddleName("Sigurd");
        name.setPersonSurnameName("Andersen");
        getRequest.setNamePersonQuery(name);
        prepareDatabaseAndSendRequest();
        assertResponseWithProtectedAddress();
        assertCpr("0000000000",null);

        String[] logs = helper.getSystemLogs();
        assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SCES v1.2.0 getPersonDetails\",\"type\":\"SensitivePersonalInformation\",\"key\":\"name\",\"value\":\""+getRequest.getNamePersonQuery().getPersonGivenName()+" "+getRequest.getNamePersonQuery().getPersonMiddleName()+" "+getRequest.getNamePersonQuery().getPersonSurnameName()+"\"}", logs[0]);
    }

    @Test
    public void requestCPRProtectionNotStartedWithConsent() throws Exception {
    	
    	PersonVO person = addPersonProtectionNotStarted("0103952595");
    	setContextUserCitizenUser(person.getCpr());
        getRequest.setCivilRegistrationNumberPersonQuery("0103952595");
        prepareDatabaseAndSendRequest(RequestType.GET_PERSON_WITH_CONSENT);
        assertResponseNotProtectedWithFullAddress();

        String[] logs = helper.getSystemLogs();
        assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SCES v1.2.0 getPersonDetailsWithConsent\",\"type\":\"SensitivePersonalInformation\",\"key\":\"cpr\",\"value\":\""+getRequest.getCivilRegistrationNumberPersonQuery()+"\"}", logs[0]);
    }

    @Test
    public void requestCPRProtectionNotStartedWithoutConsent() throws Exception {
    	
    	PersonVO person = addPersonProtectionNotStarted("0103952595");
    	setContextUserCitizenUser(person.getCpr());
        getRequest.setCivilRegistrationNumberPersonQuery("0103952595");
        prepareDatabaseAndSendRequest();
        assertResponseNotProtectedWithFullAddress();

        String[] logs = helper.getSystemLogs();
        assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SCES v1.2.0 getPersonDetails\",\"type\":\"SensitivePersonalInformation\",\"key\":\"cpr\",\"value\":\""+getRequest.getCivilRegistrationNumberPersonQuery()+"\"}", logs[0]);
    }

    @Test
    public void requestCPRprotectedWithConsent1() throws Exception { 
    	
        PersonVO person = addPerson("0103952595");
        setContextUserCitizenUser(person.getCpr());
        getRequest.setCivilRegistrationNumberPersonQuery("0103952595");
        prepareDatabaseAndSendRequest(RequestType.GET_PERSON_WITH_CONSENT);
        assertResponseWithFullAddress(true);
        assertCpr(person);

        String[] logs = helper.getSystemLogs();
        assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SCES v1.2.0 getPersonDetailsWithConsent\",\"type\":\"SensitivePersonalInformation\",\"key\":\"cpr\",\"value\":\""+getRequest.getCivilRegistrationNumberPersonQuery()+"\"}", logs[0]);
    }


    @Test
    public void requestCPRprotectedWithConsent3() throws Exception {
    	
    	PersonVO person = addPerson("0103952595");
    	setContextUserCitizenUser(person.getCpr());
        GregorianCalendar cal = new GregorianCalendar();
        cal.setTime(persons.get(0).getFoedselsdato());
        XMLGregorianCalendar birthday = DatatypeFactory.newInstance().newXMLGregorianCalendar(cal);
        getRequest.setBirthDatePersonQuery(birthday);
        prepareDatabaseAndSendRequest(RequestType.GET_PERSON_WITH_CONSENT);
        assertResponseWithFullAddress();
        assertCpr("0000000000",null);

        String[] logs = helper.getSystemLogs();
        assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SCES v1.2.0 getPersonDetailsWithConsent\",\"type\":\"SensitivePersonalInformation\",\"key\":\"birthdate\",\"value\":\""+birthdateFormatter.format(getRequest.getBirthDatePersonQuery().toGregorianCalendar().getTime())+"\"}", logs[0]);
    }

    @Test
    public void requestCPRprotectedWithConsent4() throws Exception {
    	
    	PersonVO person = addPerson("0103952595");
    	setContextUserCitizenUser(person.getCpr());
        NamePersonQueryType name = new NamePersonQueryType();
        name.setPersonGivenName("Peter");
        name.setPersonMiddleName("Sigurd");
        name.setPersonSurnameName("Andersen");
        getRequest.setNamePersonQuery(name);
        prepareDatabaseAndSendRequest(RequestType.GET_PERSON_WITH_CONSENT);
        assertResponseWithFullAddress();
        assertCpr("0000000000",null);

        String[] logs = helper.getSystemLogs();
        assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SCES v1.2.0 getPersonDetailsWithConsent\",\"type\":\"SensitivePersonalInformation\",\"key\":\"name\",\"value\":\""+getRequest.getNamePersonQuery().getPersonGivenName()+" "+getRequest.getNamePersonQuery().getPersonMiddleName()+" "+getRequest.getNamePersonQuery().getPersonSurnameName()+"\"}", logs[0]);
    }

    
    
    @Test
    @Ignore
    public void requestCPROutdatedProtectedWithConsent() throws Exception {
    	
    	PersonVO person = addPersonWithOutdatedProtection("0103952595");
    	setContextUserCitizenUser(person.getCpr());
        getRequest.setCivilRegistrationNumberPersonQuery("0103952595");
        prepareDatabaseAndSendRequest(RequestType.GET_PERSON_WITH_CONSENT);
        assertResponseWithFullAddress(false);

        String[] logs = helper.getSystemLogs();
        assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SCES v1.2.0 getPersonDetailsWithConsent\",\"type\":\"SensitivePersonalInformation\",\"key\":\"cpr\",\"value\":\"" + getRequest.getCivilRegistrationNumberPersonQuery() + "\"}", logs[0]);
    }
    
    @Test
    public void requestSearchCpr_name() throws Exception {
    	
    	String cpr = "0103952595";
        PersonVO person = addPerson(cpr);
        setContextUserCitizenUser(person.getCpr());        

        NamePersonQueryType namePersonQuery = new NamePersonQueryType();
        namePersonQuery.setPersonGivenName("Peter");
        namePersonQuery.setPersonMiddleName("Sigurd");
        namePersonQuery.setPersonSurnameName("Andersen");

        searchRequest.setNamePersonQuery(namePersonQuery);

        prepareDatabaseAndSendRequest(RequestType.SEARCH_PERSON);
        assertResponseWithFullAddress(true);
        assertCpr(person);

        String[] logs = helper.getSystemLogs();
        String query = "Peter Sigurd Andersen";
        assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SCES v1.2.0 searchPersonDetails\",\"type\":\"SensitivePersonalInformation\",\"key\":\"name\",\"value\":\"" + query + "\"}", logs[0]);
    }

    @Test
    public void requestSearchCpr_birthDate() throws Exception {

    	String cpr = "0103952595";
    	PersonVO person = addPerson(cpr);
    	setContextUserCitizenUser(person.getCpr());

        GregorianCalendar cal = new GregorianCalendar();
        cal.setTime(persons.get(0).getFoedselsdato());
        XMLGregorianCalendar birthday = DatatypeFactory.newInstance().newXMLGregorianCalendar(cal);

        searchRequest.setBirthDatePersonQuery(birthday);

        prepareDatabaseAndSendRequest(RequestType.SEARCH_PERSON);
        assertResponseWithFullAddress(true);
        assertCpr(person);

        String[] logs = helper.getSystemLogs();
        assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SCES v1.2.0 searchPersonDetails\",\"type\":\"SensitivePersonalInformation\",\"key\":\"birthdate\",\"value\":\"" + birthdateFormatter.format(searchRequest.getBirthDatePersonQuery().toGregorianCalendar().getTime()) + "\"}", logs[0]);
    }


    protected void assertCpr(PersonVO person) {
        assertCpr(person.getCpr(),person.getGaeldendeCPR());
    }
    private void assertCpr(String cpr, String gaeldendeCpr) {
        assertString(cpr,response.getPersonInformationStructure().get(0).getRegularCPRPerson().getSimpleCPRPerson().getPersonCivilRegistrationIdentifier());
        assertString(gaeldendeCpr,response.getPersonInformationStructure().get(0).getCurrentPersonCivilRegistrationIdentifier());
    }

    protected PersonVO addPerson(String cpr) throws ParseException {
        PersonVO person = Factories.createPersonWithCPR(cpr);
        person.setNavnebeskyttelsestartdato(dateFormatter.parse("2000-01-01"));
        person.setNavnebeskyttelseslettedato(dateFormatter.parse("2050-01-01"));
        persons.add(person);
        return person;
    }

    private PersonVO addPersonWithOutdatedProtection(String cpr) throws ParseException {
        PersonVO person = Factories.createPersonWithCPR(cpr);
        person.setNavnebeskyttelsestartdato(dateFormatter.parse("2000-01-01"));
        person.setNavnebeskyttelseslettedato(dateFormatter.parse("2010-01-01"));
        persons.add(person);
        return person;
    }
    
    private PersonVO addPersonProtectionNotStarted(String cpr) throws ParseException {
        PersonVO person = Factories.createPersonWithCPR(cpr);
        person.setNavnebeskyttelsestartdato(dateFormatter.parse("2999-01-01"));
        persons.add(person);
        return person;
    }

    private void assertResponseWithProtectedAddress() {
        assertResponseWithProtectedAddress(0, true);
    }
    
    protected void assertResponseWithProtectedAddress(int index) {
        assertResponseWithProtectedAddress(index, true);
    }

    private void assertResponseWithProtectedAddress(int index, boolean piprotected) {
        PersonInformationStructureType personInformationStructure = assertPersonInformantionProtected(index, piprotected);
        PersonAddressStructureType personAddressStructure = assertPersonInformantionWithProtectedName(personInformationStructure);
        assertString(null,personAddressStructure.getAddressComplete().getAddressPostal().getCountryIdentificationCode());
        assertString("ADRESSEBESKYTTET",personAddressStructure.getAddressComplete().getAddressPostal().getDistrictName());
        assertString(null,personAddressStructure.getAddressComplete().getAddressPostal().getDistrictSubdivisionIdentifier());
        assertString(null,personAddressStructure.getAddressComplete().getAddressPostal().getFloorIdentifier());
        assertString(null,personAddressStructure.getAddressComplete().getAddressPostal().getMailDeliverySublocationIdentifier());
        assertString("0000",personAddressStructure.getAddressComplete().getAddressPostal().getPostCodeIdentifier());
        assertString(null,personAddressStructure.getAddressComplete().getAddressPostal().getPostOfficeBoxIdentifier());
        assertString(null,personAddressStructure.getAddressComplete().getAddressPostal().getStreetBuildingIdentifier());
        assertString("ADRESSEBESKYTTET",personAddressStructure.getAddressComplete().getAddressPostal().getStreetName());
        assertString(null,personAddressStructure.getAddressComplete().getAddressPostal().getStreetNameForAddressingName());
        assertString(null,personAddressStructure.getAddressComplete().getAddressPostal().getSuiteIdentifier());
        assertString("0000",personAddressStructure.getCountyCode());
        assertString(null,personAddressStructure.getCareOfName());
    }

    private void assertString(String expected, Object actual) {
        if (expected == null) {
            assertNull((actual == null) ? "" : "Excepted null but got "+actual.toString(), actual);
        } else {
            assertEquals(expected, actual);
        }
    }
    
    protected void assertResponseSize(int size) {
    	assertEquals(size, response.getPersonInformationStructure().size());
    }
    
    private void assertResponseWithFullAddress() {
        assertResponseWithFullAddress(true);
    }

    protected void assertResponseWithFullAddress(boolean piprotected) {
        assertResponseWithFullAddress(0, piprotected);
    }
    
    protected void assertResponseWithFullAddress(int index) {
        assertResponseWithFullAddress(index, true);
    }
    
    private void assertResponseNotProtectedWithFullAddress() {
        assertPersonInformantionNotProtected(0, false);
    }

    private void assertResponseWithFullAddress(int index, boolean piprotected) {
        PersonInformationStructureType personInformationStructure = assertPersonInformantionProtected(index, piprotected);
        PersonAddressStructureType personAddressStructure = assertPersonInformantionWithFullName(personInformationStructure);
        assertString("DK",personAddressStructure.getAddressComplete().getAddressPostal().getCountryIdentificationCode().getValue());
        assertString("Überwald",personAddressStructure.getAddressComplete().getAddressPostal().getDistrictName());
        assertString("BirkeBy",personAddressStructure.getAddressComplete().getAddressPostal().getDistrictSubdivisionIdentifier());
        assertString("12",personAddressStructure.getAddressComplete().getAddressPostal().getFloorIdentifier());
        assertString(null,personAddressStructure.getAddressComplete().getAddressPostal().getMailDeliverySublocationIdentifier());
        assertString("6666",personAddressStructure.getAddressComplete().getAddressPostal().getPostCodeIdentifier());
        assertString(null,personAddressStructure.getAddressComplete().getAddressPostal().getPostOfficeBoxIdentifier());
        assertString("10",personAddressStructure.getAddressComplete().getAddressPostal().getStreetBuildingIdentifier());
        assertString("Ørstedgade",personAddressStructure.getAddressComplete().getAddressPostal().getStreetName());
        assertString("Østergd.",personAddressStructure.getAddressComplete().getAddressPostal().getStreetNameForAddressingName());
        assertString("tv",personAddressStructure.getAddressComplete().getAddressPostal().getSuiteIdentifier());
        assertString("1083",personAddressStructure.getCountyCode());
        assertString("Søren Petersen",personAddressStructure.getCareOfName());
    }

    private PersonInformationStructureType assertPersonInformantionProtected(int index, boolean piprotected) {
        PersonInformationStructureType personInfoStructure = response.getPersonInformationStructure().get(index);
        boolean personInformationProtectionIndicator = personInfoStructure.getRegularCPRPerson().isPersonInformationProtectionIndicator();
        assertEquals(piprotected, personInformationProtectionIndicator);
        PersonAddressStructureType personAddressStructure = personInfoStructure.getPersonAddressStructure();
        Date protectionStartDate = personAddressStructure.getPersonInformationProtectionStartDate().toGregorianCalendar().getTime();
        assertEquals("2000-01-01",dateFormatter.format(protectionStartDate));
        return personInfoStructure;
    }
    
    private PersonInformationStructureType assertPersonInformantionNotProtected(int index, boolean piprotected) {
        PersonInformationStructureType personInfoStructure = response.getPersonInformationStructure().get(index);
        boolean personInformationProtectionIndicator = personInfoStructure.getRegularCPRPerson().isPersonInformationProtectionIndicator();
        assertEquals(piprotected, personInformationProtectionIndicator);
        PersonAddressStructureType personAddressStructure = personInfoStructure.getPersonAddressStructure();
        Date protectionStartDate = personAddressStructure.getPersonInformationProtectionStartDate().toGregorianCalendar().getTime();
        assertEquals("2999-01-01",dateFormatter.format(protectionStartDate));
        return personInfoStructure;
    }

    private PersonAddressStructureType assertPersonInformantionWithFullName(PersonInformationStructureType personInfoStructure) {
        RegularCPRPersonType regularCPRPerson = personInfoStructure.getRegularCPRPerson();
        PersonNameStructureType personNameStructure = regularCPRPerson.getSimpleCPRPerson().getPersonNameStructure();
        assertString("Peter",personNameStructure.getPersonGivenName());
        assertString("Sigurd",personNameStructure.getPersonMiddleName());
        assertString("Andersen",personNameStructure.getPersonSurnameName());
        assertString("Peter,Andersen",regularCPRPerson.getPersonNameForAddressingName());
        PersonAddressStructureType personAddressStructure = personInfoStructure.getPersonAddressStructure();
        Date protectionStartDate = personAddressStructure.getPersonInformationProtectionStartDate().toGregorianCalendar().getTime();
        assertEquals("2000-01-01",dateFormatter.format(protectionStartDate));
        return personAddressStructure;
    }

    private PersonAddressStructureType assertPersonInformantionWithProtectedName(PersonInformationStructureType personInfoStructure) {
        RegularCPRPersonType regularCPRPerson = personInfoStructure.getRegularCPRPerson();
        PersonNameStructureType personNameStructure = regularCPRPerson.getSimpleCPRPerson().getPersonNameStructure();
        assertString("ADRESSEBESKYTTET",personNameStructure.getPersonGivenName());
        assertString(null,personNameStructure.getPersonMiddleName());
        assertString("ADRESSEBESKYTTET",personNameStructure.getPersonSurnameName());
        assertString("ADRESSEBESKYTTET",regularCPRPerson.getPersonNameForAddressingName());
        PersonAddressStructureType personAddressStructure = personInfoStructure.getPersonAddressStructure();
        Date protectionStartDate = personAddressStructure.getPersonInformationProtectionStartDate().toGregorianCalendar().getTime();
        assertEquals("2000-01-01",dateFormatter.format(protectionStartDate));
        return personAddressStructure;
    }

    protected void prepareDatabaseAndSendRequest() throws Exception {
        prepareDatabaseAndSendRequest(RequestType.GET_PERSON);
    }

    protected void prepareDatabaseAndSendRequest(RequestType requestType) throws Exception {
        connection.createStatement().executeUpdate("DELETE FROM v2_Person");
        for (PersonVO person : persons) {
            StamdataTestHelper.persist(connection, person);
        }
        connection.commit();

        sendRequest(requestType);
    }

    private void sendRequest(RequestType requestType) throws Exception {
        helper.enableAuditlogging(true);
        ByteArrayOutputStream outStream = helper.redirectSystemOutToStream();

        switch(requestType) {
            case GET_PERSON:
                response = client.getPersonDetails(getRequest);
                break;
            case GET_PERSON_WITH_CONSENT:
                response = client.getPersonDetailsWithConsent(getRequest);
                break;
            case SEARCH_PERSON:
                response = client.searchPersonDetails(searchRequest);
                break;
            default:
                throw new IllegalArgumentException("Unsupported RequestType: " + requestType);
        }

        helper.enableAuditlogging(false);
        helper.resetSystemOut();
    }
}
