/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata.idws;

import dk.nsi.common.OrganisationIdentifier;
import dk.nsi.common.OrganisationIdentifierType;
import dk.nsi.hsuid.*;
import dk.nsi.hsuid._2016._08.hsuid_1_1.HsuidHeader;
import org.apache.log4j.Logger;

import java.util.ArrayList;
import java.util.List;

/**
 * A header factory used for test purposes by the consent services integration tests
 */
public final class StamdataCprServiceHeaderFactory {

    private static final Logger logger = Logger.getLogger(StamdataCprServiceHeaderFactory.class);

    private StamdataCprServiceHeaderFactory() {
    }
    
    /** 
     * @return A valid {@link HsuidHeader} with default information, where the user is a citizen
     */
    public static HsuidHeader getCitizenHsuidHeader(String cpr) {
        return getValidCitizenHsuidHeader("issuer", cpr, "System owner", "System name", "0.0", "Responsible org");
    }

    /** 
     * @return A valid {@link HsuidHeader} filled in with the supplied information, where the user is a citizen
     */
    public static HsuidHeader getValidCitizenHsuidHeader(String issuer, String cpr, String systemOwner, String systemName, String systemVersion,
            String responsibleOrg) {
        List<Attribute> attributes = new ArrayList<Attribute>(6);
        addBasicAttributes(UserTypeAttribute.LegalValues.CITIZEN, cpr, systemOwner, systemName, systemVersion, responsibleOrg, attributes);

            return HealthcareServiceUserIdentificationHeaderUtil.createHealthcareServiceUserIdentificationHeader(issuer, attributes);
    }
    
    /** 
     * @return A valid {@link HsuidHeader} with default information, where the user is a healthcare professional
     */
    public static HsuidHeader getHealthcareProfessionalHsuidHeader(String cpr) {
        return getValidHealthcareProfessionalHsuidHeader(UserTypeAttribute.LegalValues.HEALTHCAREPROFESSIONAL, "issuer", cpr, "System owner", "System name", "0.0", "Responsible org",
                new OrganisationIdentifier("12345678"), null, cpr, "-");
    }

    /** 
     * @return A valid {@link HsuidHeader} with default information, where the user is a healthcare professional working
     * on behalf of another healthcare professional
     */
    public static HsuidHeader getHealthcareProfessionalHsuidHeader(String userCpr, String responsibleUserCpr, String responsibleUserAuth) {
        return getValidHealthcareProfessionalHsuidHeader(UserTypeAttribute.LegalValues.HEALTHCAREPROFESSIONAL, "issuer", userCpr, "System owner", "System name", "0.0", "Responsible org",
                new OrganisationIdentifier("12345678"), null, responsibleUserCpr, responsibleUserAuth);
    }

    /** 
     * @return A valid {@link HsuidHeader} filled in with the supplied information, where the user is a healthcare professional
     */
    public static HsuidHeader getValidHealthcareProfessionalHsuidHeader(UserTypeAttribute.LegalValues userType, String issuer, String cpr, String systemOwner, String systemName,
            String systemVersion, String responsibleOrg, OrganisationIdentifier orgUsingId1, OrganisationIdentifier orgUsingId2,
            String responsibleUser, String responsibleUserAuthorization) {
        List<Attribute> attributes;
        if (orgUsingId2 == null) {
            attributes = new ArrayList<Attribute>(9);
        } else {
            attributes = new ArrayList<Attribute>(10);
        }
        addBasicAttributes(userType, cpr, systemOwner, systemName, systemVersion, responsibleOrg, attributes);
        attributes.add(getOrganisationIdentifierAttribute(orgUsingId1));
        if (orgUsingId2 != null) {
            attributes.add(getOrganisationIdentifierAttribute(orgUsingId2));
        }
        attributes.add(new ResponsibleUserCivilRegistrationNumberAttribute(responsibleUser));
        attributes.add(new ResponsibleUserAuthorizationCodeAttribute(responsibleUserAuthorization));

        return HealthcareServiceUserIdentificationHeaderUtil.createHealthcareServiceUserIdentificationHeader(issuer, attributes);
    }

    private static OrganisationIdentifierAttribute getOrganisationIdentifierAttribute(OrganisationIdentifier org) {
        OrganisationIdentifierAttribute.LegalFormatNames type;

        OrganisationIdentifierType identifierType = org.getIdentifierType();
        if (identifierType.equals(OrganisationIdentifierType.SOR_CODE)) {
            type = OrganisationIdentifierAttribute.LegalFormatNames.SOR;
        } else if (identifierType.equals(OrganisationIdentifierType.SHAK_CODE)) {
            type = OrganisationIdentifierAttribute.LegalFormatNames.SHAK;
        } else if (identifierType.equals(OrganisationIdentifierType.DEA_NUMBER)) {
            type = OrganisationIdentifierAttribute.LegalFormatNames.GP_NUMBER;
        } else {
            throw new IllegalArgumentException("The organisation is not valid: " + org);
        }
        return new OrganisationIdentifierAttribute(org.getIdentifier(), type);
    }

    private static void addBasicAttributes(UserTypeAttribute.LegalValues userType, String cpr, String systemOwner, String systemName, String systemVersion, String responsibleOrg,
            List<Attribute> attributes) {
        attributes.add(UserTypeAttribute.get(userType));
        attributes.add(new ActingUserCivilRegistrationNumberAttribute(cpr));
        attributes.add(new SystemVendorNameAttribute(systemOwner));
        attributes.add(new SystemNameAttribute(systemName));
        attributes.add(new SystemVersionAttribute(systemVersion));
        attributes.add(new OperationsOrganisationNameAttribute(responsibleOrg));
    }
}
