/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

package dk.nsi.filter.management.sfm.webservice;

import com.google.inject.Provider;
import com.google.inject.name.Named;
import com.sun.xml.ws.developer.SchemaValidation;
import com.trifork.stamdata.SDMException;
import com.trifork.stamdata.jaxws.GuiceInstanceResolver;
import com.trifork.stamdata.util.FilterFaultMessageEnum;
import com.trifork.stamdata.util.FilterManagementProperties;
import dk.medcom.dgws._2006._04.dgws_1_0.Header;
import dk.nsi._2024._04._03.stamdatasfm.*;
import dk.nsi.filter.management.PersonInformation.PersonInformationClient;
import dk.nsi.filter.management.PersonInformation.PersonInformationClientException;
import dk.nsi.filter.management.ServiceErrorCallback;
import dk.nsi.stamdata.replication.dao.FilterManagementDAO;
import dk.nsi.stamdata.replication.dynamic.DynamicViewMapper;
import dk.nsi.stamdata.replication.vo.ViewMapVO;
import dk.nsi.stamdata.replication.webservice.FaultCodes;
import dk.nsi.stamdata.security.AccessControl;
import dk.nsi.stamdata.security.RequestContextProvider;
import dk.nsi.stamdata.security.RequestContextVO;
import org.apache.commons.lang.StringUtils;
import org.apache.http.client.HttpClient;
import org.apache.http.impl.client.HttpClientBuilder;
import org.apache.log4j.Logger;
import org.oasis_open.docs.wss._2004._01.oasis_200401_wss_wssecurity_secext_1_0.Security;

import com.google.inject.Inject;

import javax.jws.WebService;
import javax.xml.ws.Holder;
import java.sql.Connection;
import java.sql.SQLException;
import java.time.Duration;
import java.util.List;

@WebService(endpointInterface="dk.nsi._2024._04._03.stamdatasfm.StamdataFilterPortType")
@GuiceInstanceResolver.GuiceWebservice
@SchemaValidation
public class FilterManagementImpl implements StamdataFilterPortType {

    private static final Logger logger = Logger.getLogger(FilterManagementImpl.class);
    public static final int FILTER_NAME_MIN_SIZE = FilterManagementProperties.getFilterNameMinSize();
    public static final int FILTER_NAME_MAX_SIZE = FilterManagementProperties.getFilterNameMaxSize();

    private PersonInformationClient personInformationClient;

    private ViewMapVO viewMapVOHolder;

    private ValidateTextField validateTextField;
    private FilterManagementDAO filterManagementDAO;

    @Inject
    private DynamicViewMapper dynamicViewMapper;

    @Inject
    private RequestContextProvider requestContextProvider;

    @Inject
    private AccessControl accessControl;
    @Inject
    private Provider<Connection> connectionProvider;

    @Inject
    FilterManagementImpl(@Named("service.endpointPersonInformation") String url) {
        HttpClient httpClient = HttpClientBuilder.create().build();
        ServiceErrorCallback serviceErrorCallback = new ServiceErrorCallback(Duration.ofMinutes(10));
        this.personInformationClient = new PersonInformationClient(httpClient, url, serviceErrorCallback);
        validateTextField = new ValidateTextField();
    }

    @Override
    public FilterResponseType writeFilter(Holder<Security> wsseHeader, Holder<Header> medcomHeader, FilterNewRequestType parameters) throws FilterFaultMessage {
        logger.info("Start writeFilter");
        RequestContextVO requestContext = null;
        try {
            requestContext = requestContextProvider.getRequestContext(null, null);
            accessControl.checkAccess(requestContext, false, null);
        } catch (SDMException e) {
            logger.error(e.getMessage(), e);
            throw new FilterFaultMessage(e.getMessage(), null);
        }
        FilterResponseType response = null;
        String filterName = null;
        List<String> fieldList = null;

        // Construct the output container.
        response = new ObjectFactory().createFilterResponseType();

        // Check if the name is valid
        filterName = parameters.getName();
        validateFilterName(filterName);

        // Check if the ViewPath is valid and create viewMapVOHolder
        String viewPath = parameters.getViewPath();
        validateViewPath(viewPath);

        // Check if the different fields from DataTypeFieldList is valid
        filterManagementDAO = new FilterManagementDAO(connectionProvider);
        fieldList = parameters.getDataTypeFieldList();
        validateDataTypeFieldList(viewMapVOHolder.getDatatype(), fieldList);

        // Get cvr and create new filter
        String cvr = requestContext.getServiceActor().getOrganizationId();
        try {
            filterManagementDAO.createFilter(cvr, filterName, viewMapVOHolder, fieldList);
        } catch (SQLException e) {
            logger.error(e.getMessage(), e);
            throw new FilterFaultMessage(FilterFaultMessageEnum.ERROR_CREATING_FILTER.getErrorMessage(), null);
        }

        // Set the status to "OK"
        response.setStatus("OK");

        return response;
    }

    @Override
    public FilterResponseType deleteFilter(Holder<Security> wsseHeader, Holder<Header> medcomHeader, FilterDeleteRequestType parameters) throws FilterFaultMessage {
        logger.info("Start deleteFilter");
        RequestContextVO requestContext = null;
        try {
            requestContext = requestContextProvider.getRequestContext(null, null);
            accessControl.checkAccess(requestContext, false, null);
        } catch (SDMException e) {
            logger.error(e.getMessage(), e);
            throw new FilterFaultMessage(e.getMessage(), null);
        }

        // Construct the output container.
        FilterResponseType response = new ObjectFactory().createFilterResponseType();

        // Check if the name is valid
        String filterName = parameters.getName();
        validateFilterName(filterName);

        // Get cvr and create new filter
        String cvr = requestContext.getServiceActor().getOrganizationId();
        try {
            filterManagementDAO = new FilterManagementDAO(connectionProvider);
            filterManagementDAO.deleteFilter(cvr, filterName);
        } catch (SQLException e) {
            logger.error(e.getMessage(), e);
            throw new FilterFaultMessage(FilterFaultMessageEnum.ERROR_DELETING_FILTER.getErrorMessage(), null);
        }

        // Set the status to "OK"
        response.setStatus("OK");

        return response;
    }

    public void validateFilterName(String filterName) throws FilterFaultMessage {
        if (filterName == null) {
            throw new FilterFaultMessage(FilterFaultMessageEnum.NULL_FILTER_NAME.getErrorMessage(), null);
        }
        if (filterName.length() < FILTER_NAME_MIN_SIZE) {
            throw new FilterFaultMessage(FilterFaultMessageEnum.TOO_SHORT_FILTER_NAME.getErrorMessage(), null);
        }
        if (filterName.length() > FILTER_NAME_MAX_SIZE) {
            throw new FilterFaultMessage(FilterFaultMessageEnum.TOO_LARGE_FILTER_NAME.getErrorMessage(), null);
        }
        try {
            if (validateTextField.containsCpr(filterName) && personInformationClient.personExists(filterName)) {
                throw new FilterFaultMessage(FilterFaultMessageEnum.CPR_NUMBER_FILTER_NAME.getErrorMessage(), null);
            }
        } catch (PersonInformationClientException e) {
            logger.error(e.getMessage(), e);
            throw new FilterFaultMessage(FilterFaultMessageEnum.CPR_NUMBER_FILTER_NAME.getErrorMessage(), null);
        }
    }

    public void validateViewPath(String viewPath) throws FilterFaultMessage {
        if (viewPath == null) {
            throw new FilterFaultMessage(FilterFaultMessageEnum.NULL_VIEW_PATH.getErrorMessage(), null);
        }

        viewMapVOHolder = getViewMapv0(viewPath);
        try {
            if (!dynamicViewMapper.viewMapExists(viewMapVOHolder.getRegister(), viewMapVOHolder.getDatatype(), viewMapVOHolder.getVersion(), viewMapVOHolder.getRegisterVersion())) {
                throw new FilterFaultMessage(FilterFaultMessageEnum.NO_VIEW_MAP.getErrorMessage(), null);
            }
        } catch (SQLException e) {
            logger.error(e.getMessage(), e);
            throw new FilterFaultMessage(FilterFaultMessageEnum.ERROR_VALIDATING_IEW_MAP_SQL.getErrorMessage(), FaultCodes.UNKNOWN_VIEW);
        }
    }

    private void validateDataTypeFieldList(String datatype, List<String> fieldList) throws FilterFaultMessage {
        if (fieldList == null) {
            throw new FilterFaultMessage(FilterFaultMessageEnum.NULL_FIELD_LIST.getErrorMessage(), null);
        }
        if (fieldList.isEmpty()) {
            throw new FilterFaultMessage(FilterFaultMessageEnum.EMPTY_FIELD_LIST.getErrorMessage(), null);
        }
        try {
            List<String> fieldsNotInSKRS = filterManagementDAO.getFieldsNotInSKRS(datatype, fieldList);
            if (fieldsNotInSKRS.size() > 0) {
                String errorMessage = FilterFaultMessageEnum.FIELDS_NOT_IN_DATABASE.getErrorMessage() + String.join(", ", fieldsNotInSKRS) + ".";
                throw new FilterFaultMessage(errorMessage, null);
            }
        } catch (SQLException e) {
            logger.error(e.getMessage(), e);
            throw new FilterFaultMessage(FilterFaultMessageEnum.ERROR_VALIDATING_FIELDS_SQL.getErrorMessage(), null);
        }
    }


    public ViewMapVO getViewMapv0(String viewPath) {
        // viewPath format: register/version/datatype/Registerversion
        // eks with Registerversion:    vitamin/1/grunddata/1
        // eks without Registerversion: vitamin/1/grunddata/
        String[] viewPathList = viewPath.split("/");
        String register = viewPathList[0];
        long version = Long.parseLong(viewPathList[1]);
        String dataType = viewPathList[2];
        long registerVersion = StringUtils.isEmpty(viewPathList[1]) ? Long.parseLong(viewPathList[3]) : 1L;

        ViewMapVO viewMap = new ViewMapVO();
        viewMap.setCreatedDate(null);
        viewMap.setDatatype(dataType);
        viewMap.setRegister(register);
        viewMap.setTableName(null);
        viewMap.setVersion(version);
        viewMap.setRegisterVersion(registerVersion);

        return viewMap;
    }
}
