/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.filter.management;

import com.google.inject.Inject;
import com.google.inject.Provider;
import com.trifork.stamdata.persistence.PersistenceModule;
import dk.nsi.stamdata.replication.dao.FilterManagementDAO;
import dk.nsi.stamdata.testing.DatabaseTestInitializer;
import org.apache.commons.dbutils.QueryRunner;
import org.junit.Assert;

import java.sql.*;
import java.util.ArrayList;
import java.util.List;

public class TestTableCreator {
    public final static String TEST_TABLE_CONTENT = "Filter_content";
    public final static String TEST_TABLE_FIELDS = "Filter_fields";

    @Inject
    private Provider<Connection> connectionProvider;

    private DatabaseTestInitializer dbInitializer;

    public void createTestTableFilterContent(boolean runUndertow) throws SQLException {
        if (runUndertow) {
            this.dbInitializer = new DatabaseTestInitializer(connectionProvider.get());
            dbInitializer.initialize();
        }
        String sqlContentTable =
                "CREATE TABLE IF NOT EXISTS " + TEST_TABLE_CONTENT + " (\n" +
                "  id       BIGINT NOT NULL AUTO_INCREMENT PRIMARY KEY,\n" +
                "  cvr      VARCHAR(8) NOT NULL,\n" +
                "  name     VARCHAR(60) NOT NULL,\n" +
                "  register VARCHAR(255) NOT NULL,\n" +
                "  datatype VARCHAR(255) NOT NULL,\n" +
                "  version  INT NOT NULL,\n" +
                "  registerVersion INT NOT NULL DEFAULT 1,\n" +
                "  created  TIMESTAMP NOT NULL,\n" +
                "  modified TIMESTAMP NOT NULL,\n" +
                "  CONSTRAINT filter_uk unique KEY (cvr, name)\n" +
                ") ENGINE=InnoDB";
        Connection conn = null;
        QueryRunner run = new QueryRunner();
        try {
            conn = connectionProvider.get();
            run.update(conn, sqlContentTable);
            run.update(conn, "DELETE FROM " + TEST_TABLE_CONTENT);
        } finally {
            PersistenceModule.close(conn);
        }
    }

    public void createTestTableFilterFields(boolean runUndertow) throws SQLException {
        if (runUndertow) {
            this.dbInitializer = new DatabaseTestInitializer(connectionProvider.get());
            dbInitializer.initialize();
        }
        String sqlFieldsTable =
                "CREATE TABLE IF NOT EXISTS " + TEST_TABLE_FIELDS + " (\n" +
                        "  id        BIGINT NOT NULL AUTO_INCREMENT PRIMARY KEY,\n" +
                        "  filter_id BIGINT NOT NULL,\n" +
                        "  field     VARCHAR(255) NOT NULL,\n" +
                        "  created   TIMESTAMP NOT NULL,\n" +
                        "  modified  TIMESTAMP NOT NULL,\n" +
                        "FOREIGN KEY (filter_id) REFERENCES Filter_content(id) ON DELETE CASCADE\n" +
                        ") ENGINE=InnoDB";
        Connection conn = null;
        QueryRunner run = new QueryRunner();
        try {
            conn = connectionProvider.get();
            run.update(conn, sqlFieldsTable);
            run.update(conn, "DELETE FROM " + TEST_TABLE_FIELDS);
        } finally {
            PersistenceModule.close(conn);
        }
    }

    public void compareFilterFromTestAndDatabase(String cvr, String filterName, String viewPathFromTest, List<String> fieldList) throws SQLException
    {
        Connection connection = null;
        try {
            // Get connection
            connection = connectionProvider.get();

            // Get data from database
            FilterContentInfo filterContentInfo = getFilterContent(cvr, filterName, connection);
            List<FilterFieldsInfo> filterFieldsInfoList = getFilterFieldsByFilterId(filterContentInfo.id, connection);

            // Compare Filter content data
            String viewPathFromDatabase = getViewPath(filterContentInfo.register, filterContentInfo.version, filterContentInfo.datatype, filterContentInfo.registerVersion);
            Assert.assertEquals(viewPathFromTest, viewPathFromDatabase);
            Assert.assertEquals(cvr, filterContentInfo.cvr);
            Assert.assertEquals(filterName, filterContentInfo.name);

            // Compare filter fields data
            Assert.assertEquals(fieldList.size(), filterFieldsInfoList.size());

            for (int i = 0; i < fieldList.size(); i++) {
                String fieldFromTest = fieldList.get(i);
                String fieldFromDatabase = filterFieldsInfoList.get(i).field;
                Assert.assertEquals(fieldFromTest, fieldFromDatabase);
            }

        } finally {
            PersistenceModule.close(connection);
        }
    }

    public void checkFilterIsNotInDatabase(String cvr, String filterName) throws SQLException
    {
        Connection connection = null;
        try {
            // Get connection
            connection = connectionProvider.get();

            // Get data from database
            FilterContentInfo filterContentInfo = getFilterContent(cvr, filterName, connection);

            // Compare Filter content data
            Assert.assertNull(filterContentInfo);

        } finally {
            PersistenceModule.close(connection);
        }
    }

    public FilterContentInfo getFilterContent(String cvr, String filterName, Connection connection) throws SQLException
    {
        ResultSet resultSet = null;
        PreparedStatement preparedStatement = null;
        try {

            String queryString = "SELECT * " +
                    "FROM " + FilterManagementDAO.getTableFilterContent() + " " +
                    "WHERE " + FilterManagementDAO.FilterContentEnum.cvr.name() + " = ? " +
                    "AND " + FilterManagementDAO.FilterContentEnum.name.name() + " = ? " +
                    "LIMIT 1 ";
            preparedStatement = connection.prepareStatement(queryString);
            preparedStatement.setString(1, cvr);
            preparedStatement.setString(2, filterName);
            resultSet = preparedStatement.executeQuery();

            FilterContentInfo filterContentInfo = null;
            if (resultSet.next()) {
                filterContentInfo = new FilterContentInfo(resultSet);
            }
            return filterContentInfo;
        } finally {
            if (resultSet != null) resultSet.close();
            PersistenceModule.close(preparedStatement);
        }
    }

    public List<FilterFieldsInfo> getFilterFieldsByFilterId(long filterId, Connection connection) throws SQLException {
        List<FilterFieldsInfo> filterFieldsList = new ArrayList<>();
        PreparedStatement preparedStatement = null;
        ResultSet resultSet = null;

        try {
            String queryString = "SELECT * FROM " + FilterManagementDAO.getTableFilterFields() + " " +
                    "WHERE " + FilterManagementDAO.FilterFieldsEnum.filter_id + " = ?";
            preparedStatement = connection.prepareStatement(queryString);
            preparedStatement.setLong(1, filterId);
            resultSet = preparedStatement.executeQuery();

            while (resultSet.next()) {
                FilterFieldsInfo filterFields = new FilterFieldsInfo(resultSet);
                filterFieldsList.add(filterFields);
            }
            return filterFieldsList;
        } finally {
            if (resultSet != null) {
                resultSet.close();
            }
            if (preparedStatement != null) {
                preparedStatement.close();
            }
        }
    }

    public String getViewPath(String register, long version, String dataType) {
        return getViewPath(register, version, dataType, 0L);
    }

    public String getViewPath(String register, long version, String dataType, long registerverion) {
        // viewPath format: register/version/datatype/Registerversion
        // eks with Registerversion:    vitamin/1/grunddata/1
        // eks without Registerversion: vitamin/1/grunddata/
        if (registerverion == 0F) {
            return String.format("%s/%d/%s/", register, version, dataType);
        } else {
            return String.format("%s/%d/%s/%d", register, version, dataType, registerverion);
        }
    }

    public class FilterContentInfo {
        private long id;
        private String cvr;
        private String name;
        private String register;
        private String datatype;
        private long version;
        private long registerVersion;
        private Timestamp created;
        private Timestamp modified;

        public FilterContentInfo(ResultSet resultSet) throws SQLException {
            this.id = resultSet.getLong("id");
            this.cvr = resultSet.getString("cvr");
            this.name = resultSet.getString("name");
            this.register = resultSet.getString("register");
            this.datatype = resultSet.getString("datatype");
            this.version = resultSet.getLong("version");
            this.registerVersion = resultSet.getLong("registerVersion");
            this.created = resultSet.getTimestamp("created");
            this.modified = resultSet.getTimestamp("modified");
        }
    }
    public class FilterFieldsInfo {
        private long id;
        private long filterId;
        private String field;
        private Timestamp created;
        private Timestamp modified;

        public FilterFieldsInfo(ResultSet resultSet) throws SQLException {
            this.id = resultSet.getLong("id");
            this.filterId = resultSet.getLong("filter_id");
            this.field = resultSet.getString("field");
            this.created = resultSet.getTimestamp("created");
            this.modified = resultSet.getTimestamp("modified");
        }
    }
}
