/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.filter.management.sfm.webservice;


import com.google.inject.Inject;
import com.google.inject.Provider;
import com.trifork.stamdata.jaxws.SealNamespaceResolver;
import dk.medcom.dgws._2006._04.dgws_1_0.Header;
import dk.nsi._2024._04._03.stamdatasfm.*;
import dk.nsi.filter.management.TestTableCreator;
import dk.nsi.filter.management.sfm.ComponentController;
import dk.nsi.stamdata.testing.*;
import org.junit.*;
import org.junit.runner.RunWith;
import org.oasis_open.docs.wss._2004._01.oasis_200401_wss_wssecurity_secext_1_0.Security;

import javax.inject.Named;
import javax.xml.namespace.QName;
import javax.xml.ws.Holder;
import java.net.URL;
import java.sql.Connection;
import java.util.Arrays;
import java.util.List;

@RunWith(GuiceTestRunner.class)
public class FilterManagementImplIntegrationTest
{
    private static final int SERVER_PORT = 8987;
    public static final String CVR = "33257872";
    private boolean isClientAuthority = false;

    private UndertowJaxWsTestServer undertowTestServer;
    private StamdataFilterPortType client;
    private FilterNewRequestType filterNewRequest;
    private FilterDeleteRequestType filterDeleteRequest;
    private FilterResponseType response;
    private String responseStatus;
    DatabaseTestInitializer dbInitializer;
    TestHelper helper = new TestHelper();

    @Inject
    private TestTableCreator testTableCreator;
    @Inject
    private Provider<Connection> connectionProvider;
    @Inject
    @Named("run.undertow")
    private boolean runUndertow;
    @Inject
    @Named("sfm.endpoint")
    private String sfmEndpoint;


	@Before
    public void setUp() throws Exception
    {
        if(runUndertow) {
            // Setup in-memory database
            dbInitializer = new DatabaseTestInitializer(connectionProvider.get());
            dbInitializer.initialize();

            // Start Undertow server
            undertowTestServer = new UndertowJaxWsTestServer(SERVER_PORT, ComponentController.class, new UndertowSlaLogParameters("Filter managment Module", "Stamdata-Filter-Management-Service"));
            undertowTestServer.start();

            // Set security context for development profile
            SecurityContextTestHelper.setContextUserSystemUser();
        }

        URL wsdlLocation = new URL(String.format("%s?wsdl", sfmEndpoint));
        QName serviceName = new QName("http://nsi.dk/2024/04/03/StamdataSfm", "StamdataFilterService");
        StamdataFilterService service = new StamdataFilterService(wsdlLocation, serviceName);

        service.setHandlerResolver(new SealNamespaceResolver());
        client = service.getStamdataFilterPort();
        testTableCreator.createTestTableFilterContent(runUndertow);
        testTableCreator.createTestTableFilterFields(runUndertow);
    }

    @After
    public void tearDown() throws Exception
    {
        if (runUndertow) {
            dbInitializer.getConnection().close();

            if (undertowTestServer != null) {
                undertowTestServer.close();
            }
        }
    }

    @Test
    public void createFilter() throws Exception {
        // Given
        String filterName = "Filter_createFilter";
        String filterViewPath = testTableCreator.getViewPath("vitamin", 1L, "grunddata", 1L);
        List<String> fieldList = Arrays.asList( "drugID", "varetype", "varedeltype", "alfabetSekvensplads");

        createNewFilterRequest(filterName, filterViewPath, fieldList);
        sendRequest(false);

        testTableCreator.compareFilterFromTestAndDatabase(CVR, filterName, filterViewPath, fieldList);
    }

    @Test
    public void createFilter_WhenFilterWithSameNameExist() throws Exception {
        // Given
        String filterName = "Filter_createFilter_WhenFilterWithSameNameExist";
        String filterViewPath = testTableCreator.getViewPath("vitamin", 1L, "grunddata", 1L);
        List<String> fieldList = Arrays.asList( "drugID", "varetype", "varedeltype", "alfabetSekvensplads");

        // When
        createNewFilterRequest(filterName, filterViewPath, fieldList);
        sendRequest(false);

        filterViewPath = testTableCreator.getViewPath("vitamin", 1L, "grunddata", 1L);
        fieldList = Arrays.asList( "drugID", "alfabetSekvensplads");
        createNewFilterRequest(filterName, filterViewPath, fieldList);
        sendRequest(false);

        // Then
        testTableCreator.compareFilterFromTestAndDatabase(CVR, filterName, filterViewPath, fieldList);
    }

    @Test
    public void createFilter_WhenNameIsTooSmall() throws Exception {
        // Given
        String filterName = "12";
        String filterViewPath = testTableCreator.getViewPath("vitamin", 1L, "grunddata", 1L);
        List<String> fieldList = Arrays.asList( "drugID", "varetype", "varedeltype", "alfabetSekvensplads");

        Exception exception = null;
        try {
            createNewFilterRequest(filterName, filterViewPath, fieldList);
            sendRequest(false);
        } catch (Exception e) {
            exception = e;
        }

        Assert.assertNotNull(exception);
        testTableCreator.checkFilterIsNotInDatabase(CVR, filterName);
    }

    @Test
    public void createFilter_WhenNameTooBig() throws Exception {
        //Giver
        String filterName = "123456789_123456789_123456789_123456789_123456789_123456789_1"; //
        String filterViewPath = testTableCreator.getViewPath("vitamin", 1L, "grunddata", 1L);
        List<String> fieldList = Arrays.asList( "drugID", "varetype", "varedeltype", "alfabetSekvensplads");

        Exception exception = null;
        try {
            createNewFilterRequest(filterName, filterViewPath, fieldList);
            sendRequest(false);
        } catch (Exception e) {
            exception = e;
        }

        Assert.assertNotNull(exception);
        testTableCreator.checkFilterIsNotInDatabase(CVR, filterName);
    }

    @Test
    public void createFilter_WhenNameIsACprNr() throws Exception {
        //Giver
        String filterName = "0304784567";
        String filterViewPath = testTableCreator.getViewPath("vitamin", 1L, "grunddata", 1L);
        List<String> fieldList = Arrays.asList( "drugID", "varetype", "varedeltype", "alfabetSekvensplads");

        Exception exception = null;
        try {
            createNewFilterRequest(filterName, filterViewPath, fieldList);
            sendRequest(false);
        } catch (Exception e) {
            exception = e;
        }

        Assert.assertNotNull(exception);
        testTableCreator.checkFilterIsNotInDatabase(CVR, filterName);
    }

    @Test
    public void createFilter_WhenNameIsACprNrThatDoesNotExist() throws Exception {
        //Giver
        String filterName = "0123456789";
        String filterViewPath = testTableCreator.getViewPath("vitamin", 1L, "grunddata", 1L);
        List<String> fieldList = Arrays.asList( "drugID", "varetype", "varedeltype", "alfabetSekvensplads");

        // When
        createNewFilterRequest(filterName, filterViewPath, fieldList);
        sendRequest(false);

        // Then
        testTableCreator.compareFilterFromTestAndDatabase(CVR, filterName, filterViewPath, fieldList);
    }

    @Test
    public void createFilter_WhenViewPathNotInSKRS() throws Exception {
        // Given
        String filterName = "Filter_createFilter_WhenViewPathNotInSKRS";
        String filterViewPath = testTableCreator.getViewPath("DenneFindesHeltSikkerIkkeISKRS", 1L, "grunddata", 1L);
        List<String> fieldList = Arrays.asList( "drugID", "varetype", "varedeltype", "alfabetSekvensplads");

        Exception exception = null;
        try {
            createNewFilterRequest(filterName, filterViewPath, fieldList);
            sendRequest(false);
        } catch (Exception e) {
            exception = e;
        }

        Assert.assertNotNull(exception);
        testTableCreator.checkFilterIsNotInDatabase(CVR, filterName);
    }

    @Test
    public void deleteFilter_WhenFilterExist() throws Exception {
        // Given
        String filterName = "Filter_deleteFilter_WhenFilterExist";
        String filterViewPath = testTableCreator.getViewPath("vitamin", 1L, "grunddata", 1L);
        List<String> fieldList = Arrays.asList( "drugID", "varetype", "varedeltype", "alfabetSekvensplads");

        // When and then
        createNewFilterRequest(filterName, filterViewPath, fieldList);
        sendRequest(false);
        testTableCreator.compareFilterFromTestAndDatabase(CVR, filterName, filterViewPath, fieldList);

        createDeleteFilterRequest(filterName);
        sendRequest(true);
        testTableCreator.checkFilterIsNotInDatabase(CVR, filterName);
    }

    @Test
    public void deleteFilter_WhenNoFilterWithThatNameExist() throws Exception {
        // Given
        String filterName = "Filter_deleteFilter_WhenNoFilterWithThatNameExist";

        Exception exception = null;
        try {
            createDeleteFilterRequest(filterName);
            sendRequest(true);
        } catch (Exception e) {
            exception = e;
        }

        Assert.assertNotNull(exception);
        testTableCreator.checkFilterIsNotInDatabase(CVR, filterName);
    }

    private void createNewFilterRequest(String filterName, String filterViewPath, List<String> dataTypeFieldList) {
        filterNewRequest = new ObjectFactory().createFilterNewRequestType();
        filterNewRequest.setName(filterName);
        filterNewRequest.setViewPath(filterViewPath);
        filterNewRequest.getDataTypeFieldList().addAll(dataTypeFieldList);
    }

    private void createDeleteFilterRequest(String filterName) {
        filterDeleteRequest = new ObjectFactory().createFilterDeleteRequestType();
        filterDeleteRequest.setName(filterName);
    }

    private void sendRequest(boolean isDelete) throws Exception
    {
        helper.enableAuditlogging(true);
        helper.redirectSystemOutToStream();

        Holder<Security> securityHeader;
        Holder<Header> medcomHeader;

        if (isClientAuthority)
        {
            SecurityWrapper secutityHeadersNotWhitelisted = DGWSHeaderMock.getSecurityWrapper();
            securityHeader = secutityHeadersNotWhitelisted.getSecurity();
            medcomHeader = secutityHeadersNotWhitelisted.getMedcomHeader();
        } else {
            SecurityWrapper securityHeaders = DGWSHeaderMock.getSecurityWrapper();
            securityHeader = securityHeaders.getSecurity();
            medcomHeader = securityHeaders.getMedcomHeader();
        }

        if (isDelete) {
            response = client.deleteFilter(securityHeader, medcomHeader, filterDeleteRequest);
        } else {
            response = client.writeFilter(securityHeader, medcomHeader, filterNewRequest);
        }
        responseStatus = response.getStatus();

        helper.resetSystemOut();
        helper.enableAuditlogging(false);
    }
}
