/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata;

import FunctionalExt.Either;
import dk.nsp.test.idp.NspSoapClient;
import dk.nsp.test.idp.model.Identity;
import org.apache.commons.io.IOUtils;
import org.w3c.dom.Document;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.soap.MessageFactory;
import javax.xml.soap.SOAPBody;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPFault;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.net.URI;
import java.nio.charset.StandardCharsets;

import static FunctionalExt.Either.left;
import static FunctionalExt.Either.right;

public final class NspSoap {
    private NspSoap(){}

    public static class NspSoapResponseData {
        public final String
                content,
                statusMessage;
        public final int status;
        public final boolean isFault;

        public NspSoapResponseData(String content, int status, String statusMessage, boolean isFault) {
            this.content = content;
            this.status = status;
            this.statusMessage = statusMessage;
            this.isFault = isFault;
        }

        public static NspSoapResponseData from(NspSoapClient.NspSoapResponse res) {
            try {
                return new NspSoapResponseData(IOUtils.toString(res.getResponse(), StandardCharsets.UTF_8),
                        res.getStatus(),
                        res.getStatusMessage(),
                        res.isFault());
            } catch (IOException e) { throw new RuntimeException(e); }
        }
    }

    public static <T> Either<Either<RuntimeException, SOAPFault>, NspSoapResponseData> executeNspRequest(URI uri,
                                                                                                         String soapAction,
                                                                                                         Identity identity,
                                                                                                         JAXBElement<T> req,
                                                                                                         String... hs) {
        return toXml(req)
                .mapLeft(Either::<RuntimeException, SOAPFault>left)
                .flatMap(r -> executeNspRequest(uri, soapAction, identity, r, hs));
    }

    public static <T> Either<Either<RuntimeException, SOAPFault>, NspSoapResponseData> executeNspRequest(URI uri,
                                                                                                         String soapAction,
                                                                                                         Identity identity,
                                                                                                         T req,
                                                                                                         String... hs) {
        return toXml(req)
                .mapLeft(Either::<RuntimeException, SOAPFault>left)
                .flatMap(r -> executeNspRequest(uri, soapAction, identity, r, hs));
    }

    public static Either<Either<RuntimeException, SOAPFault>, NspSoapResponseData> executeNspRequest(URI uri,
                                                                                                     String soapAction,
                                                                                                     Identity identity,
                                                                                                     String req,
                                                                                                     String... hs) {
        try (NspSoapClient c = new NspSoapClient(System.out::println)) {
            try (NspSoapClient.NspSoapResponse res = c.request(uri, soapAction).as(identity).execute(req, hs))
            { return validate(NspSoapResponseData.from(res)); }
        } catch (Exception e) { throw new RuntimeException(e); }
    }

    private static Either<Either<RuntimeException, SOAPFault>, NspSoapResponseData> validate(NspSoapResponseData res) {
        try {
            SOAPBody sb = MessageFactory
                    .newInstance()
                    .createMessage(null, new ByteArrayInputStream(res.content.getBytes()))
                    .getSOAPBody();
            return sb.hasFault()
                    ? left(right(sb.getFault()))
                    : right(res);
        } catch (Exception e) { return left(left(new RuntimeException(e))); }
    }

    public static <T> Either<RuntimeException, String> toXml(JAXBElement<T> v) {
        final ByteArrayOutputStream writer = new ByteArrayOutputStream();
        try {
            JAXBContext.newInstance(v.getValue().getClass())
                    .createMarshaller()
                    .marshal(v, writer);
        } catch (JAXBException e) { return left(new RuntimeException(e)); }
        return right(new String(writer.toByteArray(), StandardCharsets.UTF_8));
    }

    public static <T> Either<RuntimeException, String> toXml(T v) {
        try { return toXml(JAXBContext.newInstance(v.getClass()), v); }
        catch (JAXBException e) { return left(new RuntimeException(e)); }
    }

    public static <T> Either<RuntimeException, String> toXml(JAXBContext c, T v) {
        final ByteArrayOutputStream writer = new ByteArrayOutputStream();
        try { c.createMarshaller().marshal(v, writer); }
        catch (JAXBException e) { return left(new RuntimeException(e)); }
        return right(new String(writer.toByteArray(), StandardCharsets.UTF_8));
    }

    public static <T> Either<RuntimeException, T> parse(NspSoapResponseData response, Class<T> c) {
        try {
            final Document message = MessageFactory
                    .newInstance()
                    .createMessage(null, new ByteArrayInputStream(response.content.getBytes()))
                    .getSOAPBody()
                    .extractContentAsDocument();
            return right(JAXBContext.newInstance(c)
                    .createUnmarshaller()
                    .unmarshal(message, c)
                    .getValue());
        } catch (JAXBException | SOAPException | IOException e) { return left(new RuntimeException(e)); }
    }
}