/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata.authorization;

import authorization20240105.AuthorizationResponseType;
import dk.nsi.stamdata.client.AuthorisationClient;
import dk.nsi.stamdata.configuration.NspServiceException;
import dk.nsi.stamdata.configuration.SDMContextConfiguration;
import dk.nsp.test.idp.OrganizationIdentities;
import dk.nsp.test.idp.model.Identity;
import dk.sosi.seal.model.AuthenticationLevel;
import org.junit.*;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit4.rules.SpringClassRule;
import org.springframework.test.context.junit4.rules.SpringMethodRule;

import javax.xml.bind.JAXBException;


@RunWith(Parameterized.class)
@ContextConfiguration(classes = SDMContextConfiguration.class )
public class Authorization20240105IntegrationTest {

    @ClassRule
    public static final SpringClassRule scr = new SpringClassRule();

    @Rule
    public final SpringMethodRule smr = new SpringMethodRule();

    @Parameterized.Parameter(0)
    public AuthenticationLevel authenticationLevel;

    @Value("${saes.endpoint}")
    private String saesEndpoint; // = "http://test1-cnsp.ekstern-test.nspop.dk:8080/stamdata-authorization-lookup-ws";

    private Identity identity = OrganizationIdentities.sundhedsdatastyrelsen();

    private AuthorisationClient authorisationClient;

    @Parameterized.Parameters(name = "{0}")
    public static AuthenticationLevel[] data() {
        return new AuthenticationLevel[]{AuthenticationLevel.VOCES_TRUSTED_SYSTEM};
    }

    @Before
    public void setup() {
        authorisationClient = new AuthorisationClient();
    }

    @Test
    public void getPersonWithAuthorization() throws NspServiceException , JAXBException{

        // Given
        String cpr = "0304784599";

        // When
        AuthorizationResponseType reponse = authorisationClient.callAuthorise20240105(cpr, identity, saesEndpoint);

        // Then
        Assert.assertEquals(1, reponse.getAutorisation().size());
        Assert.assertEquals("87RTY", reponse.getAutorisation().get(0).getAutorisationsNummer());
        Assert.assertEquals("7170", reponse.getAutorisation().get(0).getUddannelsesKode());
        Assert.assertEquals("Ole", reponse.getFornavn());
        Assert.assertEquals("Olsen", reponse.getEfternavn());
        Assert.assertEquals("0304784599", reponse.getCpr());
        Assert.assertEquals("Uddannelsesland", reponse.getAutorisation().get(0).getUddannelsesLand());
        Assert.assertEquals("Tidsbegraenset", reponse.getAutorisation().get(0).getTidsbegraenset());
        Assert.assertEquals("Tilsyn", reponse.getAutorisation().get(0).getTilsyn());
        Assert.assertEquals("Tilsynsbemaerkninger", reponse.getAutorisation().get(0).getTilsynsBemaerkninger());
        Assert.assertEquals("Sanktionstype", reponse.getAutorisation().get(0).getSanktionsType());
        Assert.assertEquals("Dispensation_75_aarsreglen", reponse.getAutorisation().get(0).getDispensation75Aarsreglen());
        Assert.assertEquals("2024-01-05", reponse.getAutorisation().get(0).getSelvstaendigVirkeDato());
        Assert.assertEquals("Selvstaendigt_virke", reponse.getAutorisation().get(0).getSelvstaendigtVirke());
        Assert.assertEquals("SelvstaendigvirkeUddLand", reponse.getAutorisation().get(0).getSelvstaendigVirkeUddLand());
        Assert.assertEquals("Speciale1", reponse.getAutorisation().get(0).getSpeciale1());
        Assert.assertEquals("2024-01-01", reponse.getAutorisation().get(0).getSpecialeDato1());
        Assert.assertEquals("SpecialeUddLand1", reponse.getAutorisation().get(0).getSpecialeUddLand1());
        Assert.assertEquals("Speciale2", reponse.getAutorisation().get(0).getSpeciale2());
        Assert.assertEquals("2024-01-02", reponse.getAutorisation().get(0).getSpecialeDato2());
        Assert.assertEquals("SpecialeUddLand2", reponse.getAutorisation().get(0).getSpecialeUddLand2());
        Assert.assertEquals("Speciale3", reponse.getAutorisation().get(0).getSpeciale3());
        Assert.assertEquals("2024-01-03", reponse.getAutorisation().get(0).getSpecialeDato3());
        Assert.assertEquals("SpecialeUddLand3", reponse.getAutorisation().get(0).getSpecialeUddLand3());
        Assert.assertEquals("Y", reponse.getAutorisation().get(0).getAutorisationGyldig());
        Assert.assertEquals("2024-01-05", reponse.getAutorisation().get(0).getAutorisationsDato());
    }

    @Test
    public void getPersonWithoutAuthorization() throws JAXBException, NspServiceException{

        // Given
        String cpr = "0000000000";

        // When
        AuthorizationResponseType reponse = authorisationClient.callAuthorise20240105(cpr, identity, saesEndpoint);

        // Then
        Assert.assertEquals(reponse.getAutorisation().size(), 0);
    }

    @Test
    public void testUserAuthorizationRetrieval_WhereDatesDatabaseIsNull() throws JAXBException, NspServiceException{

        // Given
        String cpr = "0404199999";

        // When
        AuthorizationResponseType reponse = authorisationClient.callAuthorise20240105(cpr, identity, saesEndpoint);

        // Then
        Assert.assertEquals(reponse.getAutorisation().size(), 1);
    }

    @Test
    public void getPersonWithAuthorizationCode() throws JAXBException, NspServiceException{

        // Given
        String authorizationCode = "87RTY";

        // When
        AuthorizationResponseType reponse = authorisationClient.callAuthoriseCode20240105(authorizationCode, identity, saesEndpoint);

        // Then
        Assert.assertEquals(reponse.getAutorisation().size(), 1);
        Assert.assertEquals(reponse.getAutorisation().get(0).getAutorisationsNummer(), "87RTY");
        Assert.assertEquals(reponse.getAutorisation().get(0).getUddannelsesKode(), "7170");
        Assert.assertEquals("Ole", reponse.getFornavn());
        Assert.assertEquals("Olsen", reponse.getEfternavn());
        Assert.assertEquals("0304784599", reponse.getCpr());
        Assert.assertEquals("Uddannelsesland", reponse.getAutorisation().get(0).getUddannelsesLand());
        Assert.assertEquals("Tidsbegraenset", reponse.getAutorisation().get(0).getTidsbegraenset());
        Assert.assertEquals("Tilsyn", reponse.getAutorisation().get(0).getTilsyn());
        Assert.assertEquals("Tilsynsbemaerkninger", reponse.getAutorisation().get(0).getTilsynsBemaerkninger());
        Assert.assertEquals("Sanktionstype", reponse.getAutorisation().get(0).getSanktionsType());
        Assert.assertEquals("Dispensation_75_aarsreglen", reponse.getAutorisation().get(0).getDispensation75Aarsreglen());
        Assert.assertEquals("2024-01-05", reponse.getAutorisation().get(0).getSelvstaendigVirkeDato());
        Assert.assertEquals("Selvstaendigt_virke", reponse.getAutorisation().get(0).getSelvstaendigtVirke());
        Assert.assertEquals("SelvstaendigvirkeUddLand", reponse.getAutorisation().get(0).getSelvstaendigVirkeUddLand());
        Assert.assertEquals("Speciale1", reponse.getAutorisation().get(0).getSpeciale1());
        Assert.assertEquals("2024-01-01", reponse.getAutorisation().get(0).getSpecialeDato1());
        Assert.assertEquals("SpecialeUddLand1", reponse.getAutorisation().get(0).getSpecialeUddLand1());
        Assert.assertEquals("Speciale2", reponse.getAutorisation().get(0).getSpeciale2());
        Assert.assertEquals("2024-01-02", reponse.getAutorisation().get(0).getSpecialeDato2());
        Assert.assertEquals("SpecialeUddLand2", reponse.getAutorisation().get(0).getSpecialeUddLand2());
        Assert.assertEquals("Speciale3", reponse.getAutorisation().get(0).getSpeciale3());
        Assert.assertEquals("2024-01-03", reponse.getAutorisation().get(0).getSpecialeDato3());
        Assert.assertEquals("SpecialeUddLand3", reponse.getAutorisation().get(0).getSpecialeUddLand3());
        Assert.assertEquals("Y", reponse.getAutorisation().get(0).getAutorisationGyldig());
        Assert.assertEquals("2024-01-05", reponse.getAutorisation().get(0).getAutorisationsDato());
    }

    @Test
    public void getPersonWithAuthorizationCode_WhereDatesDatabaseIsNull() throws JAXBException, NspServiceException{

        // Given
        String authorizationCode = "SSSSS";

        // When
        AuthorizationResponseType reponse = authorisationClient.callAuthoriseCode20240105(authorizationCode, identity, saesEndpoint);

        // Then
        Assert.assertEquals(reponse.getAutorisation().size(), 1);
        Assert.assertEquals(reponse.getAutorisation().get(0).getAutorisationsNummer(), "SSSSS");
        Assert.assertEquals(reponse.getAutorisation().get(0).getUddannelsesKode(), "7170");
        Assert.assertEquals("DatoNull", reponse.getFornavn());
        Assert.assertEquals("notHere", reponse.getEfternavn());
        Assert.assertEquals("0404199999", reponse.getCpr());
        Assert.assertEquals("", reponse.getAutorisation().get(0).getUddannelsesLand());
        Assert.assertEquals("Nej", reponse.getAutorisation().get(0).getTidsbegraenset());
        Assert.assertEquals("Nej", reponse.getAutorisation().get(0).getTilsyn());
        Assert.assertEquals("", reponse.getAutorisation().get(0).getTilsynsBemaerkninger());
        Assert.assertEquals("", reponse.getAutorisation().get(0).getSanktionsType());
        Assert.assertEquals("Nej", reponse.getAutorisation().get(0).getDispensation75Aarsreglen());
        Assert.assertNull(reponse.getAutorisation().get(0).getSelvstaendigVirkeDato());
        Assert.assertEquals("Nej", reponse.getAutorisation().get(0).getSelvstaendigtVirke());
        Assert.assertEquals("", reponse.getAutorisation().get(0).getSelvstaendigVirkeUddLand());
        Assert.assertEquals("", reponse.getAutorisation().get(0).getSpeciale1());
        Assert.assertNull(reponse.getAutorisation().get(0).getSpecialeDato1());
        Assert.assertEquals("", reponse.getAutorisation().get(0).getSpecialeUddLand1());
        Assert.assertEquals("", reponse.getAutorisation().get(0).getSpeciale2());
        Assert.assertNull(reponse.getAutorisation().get(0).getSpecialeDato2());
        Assert.assertEquals("", reponse.getAutorisation().get(0).getSpecialeUddLand2());
        Assert.assertEquals("", reponse.getAutorisation().get(0).getSpeciale3());
        Assert.assertNull(reponse.getAutorisation().get(0).getSpecialeDato3());
        Assert.assertEquals("", reponse.getAutorisation().get(0).getSpecialeUddLand3());
        Assert.assertEquals("Y", reponse.getAutorisation().get(0).getAutorisationGyldig());
        Assert.assertNull(reponse.getAutorisation().get(0).getAutorisationsDato());
    }
}
