/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata.client;

import dk.nsi.stamdata.configuration.NspServiceException;
import dk.nsp.test.idp.NspSoapClient;
import dk.nsp.test.idp.model.Identity;
import org.apache.commons.io.IOUtils;
import org.apache.log4j.Logger;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.soap.MessageFactory;
import javax.xml.soap.SOAPMessage;
import java.io.ByteArrayInputStream;
import java.io.StringReader;
import java.io.StringWriter;
import java.net.URI;
import java.nio.charset.StandardCharsets;

public class AbstractSDMClient {

    private Logger log = Logger.getLogger(getClass());

    public <REQUEST,RESPONSE> RESPONSE callNoXmlRootElement(Identity identity, REQUEST request, JAXBContext jaxbContext, String action, URI uri, Class c) throws JAXBException, NspServiceException {
        Marshaller jaxbMarshaller = jaxbContext.createMarshaller();
        StringWriter sw = new StringWriter();
        jaxbMarshaller.marshal(request, sw);

        log.debug("This is the request we are sending to SDM (" + uri + "):");
        log.debug(sw.toString());

        try(NspSoapClient client = new NspSoapClient(this::log)) {
            NspSoapClient.NspSoapRequest soapRequest = null;
            JAXBElement<RESPONSE> jaxbElement;

            if (identity != null) {
                soapRequest = client.request(uri, action).as(identity);
            } else {
                soapRequest = client.request(uri, action).withoutSecurity();
            }

            try (NspSoapClient.NspSoapResponse soapResponse = soapRequest.execute(sw.toString())) {
                String responseXml = IOUtils.toString(soapResponse.getResponse(), StandardCharsets.UTF_8);

                log.debug("This is the response we get from SDM:");
                log.debug(responseXml.toString());

                SOAPMessage message = MessageFactory.newInstance().createMessage(null, new ByteArrayInputStream(responseXml.getBytes()));
                jaxbElement = (JAXBElement<RESPONSE>) jaxbContext.createBinder().unmarshal(message.getSOAPBody().extractContentAsDocument(), c);


                if (soapResponse.isFault()) {
                    throw new NspServiceException(responseXml);
                }
                return (RESPONSE) jaxbElement.getValue();
            }
        } catch (NspServiceException e){
            throw e;
        } catch (Exception e){
            throw new RuntimeException(e);
        }
    }

    public <REQUEST,RESPONSE> RESPONSE call(Identity identity, REQUEST request, JAXBContext jaxbContext, String action, URI uri) throws JAXBException, NspServiceException {
        Marshaller jaxbMarshaller = jaxbContext.createMarshaller();
        StringWriter sw = new StringWriter();
        jaxbMarshaller.marshal(request, sw);

        log.debug("This is the request we are sending to SDM (" + uri + "):");
        log.debug(sw.toString());

        try(NspSoapClient client = new NspSoapClient(this::log)) {
            NspSoapClient.NspSoapRequest soapRequest = null;

            if (identity != null) {
                soapRequest = client.request(uri, action).as(identity);
            } else {
                soapRequest = client.request(uri, action).withoutSecurity();
            }

            try (NspSoapClient.NspSoapResponse soapResponse = soapRequest.execute(sw.toString())) {
                String responseXml = IOUtils.toString(soapResponse.getResponse(), StandardCharsets.UTF_8);

                log.debug("This is the response we get from SDM:");
                log.debug(responseXml.toString());

                org.xmlsoap.schemas.soap.envelope.Envelope responseEnvelope = (org.xmlsoap.schemas.soap.envelope.Envelope) jaxbContext.createUnmarshaller().unmarshal(new StringReader(responseXml));

                if (soapResponse.isFault()) {
                    throw new NspServiceException(responseXml);
                }
                return (RESPONSE) responseEnvelope.getBody().getAny().get(0);
            }
        } catch (NspServiceException e){
            throw e;
        } catch (Exception e){
            throw new RuntimeException(e);
        }
    }


    private void log(String xml) {
        if (Boolean.getBoolean(this.getClass().getSimpleName() + ".log")) {
            String method = "Unknown";
            for (StackTraceElement e : new Exception().fillInStackTrace().getStackTrace()) {
                if (e.getClassName().equals(getClass().getName())) {
                    method = e.getMethodName();
                }
            }
            System.out.println("[" + method + "]");
            System.out.println(xml.replace("\n", ""));
        }
    }
}
