/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata.filterManagement;

import com.trifork.stamdata.util.FilterFaultMessageEnum;
import filterManagement.FilterResponseType;
import dk.nsi.stamdata.client.FilterManagementClient;
import dk.nsi.stamdata.configuration.NspServiceException;
import dk.nsi.stamdata.configuration.SDMContextConfiguration;
import dk.nsp.test.idp.OrganizationIdentities;
import dk.nsp.test.idp.model.Identity;
import dk.sosi.seal.model.AuthenticationLevel;
import org.junit.*;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit4.rules.SpringClassRule;
import org.springframework.test.context.junit4.rules.SpringMethodRule;

import javax.xml.bind.JAXBException;
import java.util.Arrays;
import java.util.List;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

@RunWith(Parameterized.class)
@ContextConfiguration(classes = SDMContextConfiguration.class )
public class FilterManagementIntegrationTest {
    private FilterManagementClient client;
    private Identity identity = OrganizationIdentities.sundhedsdatastyrelsen();

    @ClassRule
    public static final SpringClassRule scr = new SpringClassRule();

    @Rule
    public final SpringMethodRule smr = new SpringMethodRule();

    @Parameterized.Parameter(0)
    public AuthenticationLevel authenticationLevel;

    @Value("${sfm.endpoint}")
    private String sfmEndpoint;

    @Parameterized.Parameters(name = "{0}")
    public static AuthenticationLevel[] data() {
        return new AuthenticationLevel[]{AuthenticationLevel.VOCES_TRUSTED_SYSTEM};
    }

    @Before
    public void setup()
    {
        client = new FilterManagementClient();
    }

    @Test
    public void createFilter() throws JAXBException, NspServiceException {
        // Given
        String filterName = "Filter_createFilter";
        String filterViewPath = getViewPath("vitamin", 1L, "grunddata", 1L);
        List<String> fieldList = Arrays.asList( "drugID", "varetype", "varedeltype", "alfabetSekvensplads");

        // When
        FilterResponseType response = callFilterManagementNew(filterName, filterViewPath, fieldList);

        // Then
        Assert.assertEquals(response.getStatus(), "OK");
    }

    @Test
    public void createFilter_WhenFilterWithSameNameExist() throws JAXBException, NspServiceException {
        // Given
        String filterName = "Filter_createFilter_WhenFilterWithSameNameExist";
        String filterViewPath = getViewPath("vitamin", 1L, "grunddata", 1L);
        List<String> fieldList = Arrays.asList( "drugID", "varetype", "varedeltype", "alfabetSekvensplads");

        // When
        FilterResponseType response1 = callFilterManagementNew(filterName, filterViewPath, fieldList);

        filterViewPath = getViewPath("vitamin", 1L, "grunddata", 1L);
        fieldList = Arrays.asList( "drugID", "alfabetSekvensplads");
        FilterResponseType response2 = callFilterManagementNew(filterName, filterViewPath, fieldList);

        // Then
        Assert.assertEquals(response1.getStatus(), "OK");
        Assert.assertEquals(response2.getStatus(), "OK");
    }

    @Test
    public void createFilter_WhenNameIsTooSmall() throws JAXBException, NspServiceException {
        // Given
        String filterName = "12";
        String filterViewPath = getViewPath("vitamin", 1L, "grunddata", 1L);
        List<String> fieldList = Arrays.asList( "drugID", "varetype", "varedeltype", "alfabetSekvensplads");

        testFilterManagementWithError(filterName, filterViewPath, fieldList, FilterFaultMessageEnum.TOO_SHORT_FILTER_NAME);
    }

    @Test
    public void createFilter_WhenNameTooBig() throws JAXBException, NspServiceException {
        //Giver
        String filterName = "123456789_123456789_123456789_123456789_123456789_123456789_1"; //
        String filterViewPath = getViewPath("vitamin", 1L, "grunddata", 1L);
        List<String> fieldList = Arrays.asList( "drugID", "varetype", "varedeltype", "alfabetSekvensplads");

        testFilterManagementWithError(filterName, filterViewPath, fieldList, FilterFaultMessageEnum.TOO_LARGE_FILTER_NAME);
    }

    @Test
    public void createFilter_WhenNameIsACprNr() throws JAXBException, NspServiceException {
        //Giver
        String filterName = "0304784567";
        String filterViewPath = getViewPath("vitamin", 1L, "grunddata", 1L);
        List<String> fieldList = Arrays.asList( "drugID", "varetype", "varedeltype", "alfabetSekvensplads");

        testFilterManagementWithError(filterName, filterViewPath, fieldList, FilterFaultMessageEnum.CPR_NUMBER_FILTER_NAME);
    }

    @Test
    public void createFilter_WhenNameIsACprNrThatDoesNotExist() throws JAXBException, NspServiceException {
        //Giver
        String filterName = "0123456789";
        String filterViewPath = getViewPath("vitamin", 1L, "grunddata", 1L);
        List<String> fieldList = Arrays.asList( "drugID", "varetype", "varedeltype", "alfabetSekvensplads");

        // When
        FilterResponseType response = callFilterManagementNew(filterName, filterViewPath, fieldList);

        // Then
        Assert.assertEquals(response.getStatus(), "OK");
    }

    @Test
    public void createFilter_WhenViewPathNotInSKRS() throws JAXBException, NspServiceException {
        // Given
        String filterName = "Filter_createFilter_WhenViewPathNotInSKRS";
        String filterViewPath = getViewPath("DenneFindesHeltSikkerIkkeISKRS", 1L, "grunddata", 1L);
        List<String> fieldList = Arrays.asList( "drugID", "varetype", "varedeltype", "alfabetSekvensplads");

        testFilterManagementWithError(filterName, filterViewPath, fieldList, FilterFaultMessageEnum.NO_VIEW_MAP);
    }

    @Test
    public void createFilter_WhenFieldNotInSKRS() throws JAXBException, NspServiceException {
        // Given
        String filterName = "Filter_createFilter_WhenFieldNotInSKRS";
        String filterViewPath = getViewPath("vitamin", 1L, "grunddata", 1L);
        String fieldNotInSKRS = "EtFeltSomIkkeFindes";
        List<String> fieldList = Arrays.asList( fieldNotInSKRS, "varetype");

        testFilterManagementWithError(filterName, filterViewPath, fieldList, FilterFaultMessageEnum.FIELDS_NOT_IN_DATABASE, fieldNotInSKRS + ".");
    }

    @Test
    public void deleteFilter_WhenFilterExist() throws JAXBException, NspServiceException {
        // Given
        String filterName = "Filter_deleteFilter_WhenFilterExist";
        String filterViewPath = getViewPath("vitamin", 1L, "grunddata", 1L);
        List<String> fieldList = Arrays.asList( "drugID", "varetype", "varedeltype", "alfabetSekvensplads");

        // When
        FilterResponseType response1 = callFilterManagementNew(filterName, filterViewPath, fieldList);
        FilterResponseType response2 = callFilterManagementDelete(filterName);


        // Then
        Assert.assertEquals(response1.getStatus(), "OK");
        Assert.assertEquals(response2.getStatus(), "OK");
    }

    @Test
    public void deleteFilter_WhenNoFilterWithThatNameExist() throws JAXBException, NspServiceException {
        // Given
        String filterName = "Filter_deleteFilter_WhenNoFilterWithThatNameExist";

        testFilterManagementWithError(filterName, FilterFaultMessageEnum.ERROR_DELETING_FILTER);
    }

    public FilterResponseType callFilterManagementNew(String filterName, String filterViewPath, List<String> dataTypeFieldList) throws JAXBException, NspServiceException
    {
        return client.callFilterManagementNew(filterName, filterViewPath, dataTypeFieldList, identity, sfmEndpoint);
    }

    public FilterResponseType callFilterManagementDelete(String filterName) throws JAXBException, NspServiceException
    {
        return client.callFilterManagementDelete(filterName, identity, sfmEndpoint);
    }

    public static String getViewPath(String register, long version, String dataType) {
        return getViewPath(register, version, dataType, 0L);
    }

    public static String getViewPath(String register, long version, String dataType, long registerverion) {
        // viewPath format: register/version/datatype/Registerversion
        // eks with Registerversion:    vitamin/1/grunddata/1
        // eks without Registerversion: vitamin/1/grunddata/
        if (registerverion == 0L) {
            return String.format("%s/%d/%s/", register, version, dataType);
        } else {
            return String.format("%s/%d/%s/%d", register, version, dataType, registerverion);
        }
    }
    public void testFilterManagementWithError(String filterName, String filterViewPath, List<String> fieldList, FilterFaultMessageEnum filterFaultMessageEnum) throws JAXBException, NspServiceException {
        testFilterManagementWithError(filterName, filterViewPath, fieldList, filterFaultMessageEnum, "");
    }

    public void testFilterManagementWithError(String filterName, String filterViewPath, List<String> fieldList, FilterFaultMessageEnum filterFaultMessageEnum, String addStringToError) throws JAXBException, NspServiceException {
        // When and Then
        String filterFaultMessage = "";
        try {
            FilterResponseType response = callFilterManagementNew(filterName, filterViewPath, fieldList);
        } catch (NspServiceException e) {
            filterFaultMessage = extractFilterFaultErrorMessage(e);
        }
        String errorMessage = filterFaultMessageEnum.getErrorMessage() + addStringToError;
        // Verify the message of the thrown exception
        Assert.assertEquals(errorMessage, filterFaultMessage);
    }

    public void testFilterManagementWithError(String filterName, FilterFaultMessageEnum filterFaultMessageEnum) throws JAXBException, NspServiceException {
        // When and Then
        String filterFaultMessage = "";
        try {
            FilterResponseType response = callFilterManagementDelete(filterName);
        } catch (NspServiceException e) {
            filterFaultMessage = extractFilterFaultErrorMessage(e);
        }
        String errorMessage = filterFaultMessageEnum.getErrorMessage();
        // Verify the message of the thrown exception
        Assert.assertEquals(errorMessage, filterFaultMessage);
    }

    public static String extractFilterFaultErrorMessage(NspServiceException e) {
        try {
            DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
            DocumentBuilder builder = factory.newDocumentBuilder();
            Document doc = builder.parse(new org.xml.sax.InputSource(new java.io.StringReader(e.getMessage())));

            // In Body -> Fault -> detail this have a child ns3:FilterFaultResponseType (that is our error)
            NodeList filterFaultResponseList = doc.getElementsByTagName("ns3:FilterFaultResponseType");
            if (filterFaultResponseList.getLength() <= 0) return null;

            // s3:FilterFaultResponseType next sibling node is the exception itself
            Node filterFaultResponse = filterFaultResponseList.item(0);
            Node nextSibling = filterFaultResponse.getNextSibling();
            while (nextSibling != null && nextSibling.getNodeType() != Node.ELEMENT_NODE) {
                nextSibling = nextSibling.getNextSibling();
            }

            // If we found the exception get the message, this is our custom thrown message
            if (nextSibling != null && "ns2:exception".equals(nextSibling.getNodeName())) {
                NodeList messageNodes = ((Element) nextSibling).getElementsByTagName("message");
                if (messageNodes.getLength() > 0) {
                    Element messageElement = (Element) messageNodes.item(0);
                    return messageElement.getTextContent();
                }
            }
        } catch (Exception ec) {
            ec.printStackTrace();
        }
        return null;
    }
}
