/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata.lookup;

import dk.nsi._2011._09._23.stamdatacpr.PersonLookupRequestType;
import dk.nsi._2014._03._01.stamdatacpr.PersonLookupResponseType;
import dk.nsi.stamdata.client.LookupClient;
import dk.nsi.stamdata.configuration.NspServiceException;
import dk.nsi.stamdata.configuration.SDMContextConfiguration;
import dk.nsp.test.idp.builder.Builders;
import dk.nsp.test.idp.model.CitizenIdentity;
import dk.nsp.test.idp.vault.RemoteCredentialVault;
import oio.medcom.cprservice._1_0.PersonInformationStructureType;
import org.junit.Before;
import org.junit.ClassRule;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit4.rules.SpringClassRule;
import org.springframework.test.context.junit4.rules.SpringMethodRule;

import javax.xml.bind.JAXBException;

import static org.junit.Assert.assertTrue;

import java.net.URI;
import java.util.List;


@ContextConfiguration(classes = SDMContextConfiguration.class )
public class StamdataPersonLookup120IDWSTest {

    @ClassRule
    public static final SpringClassRule scr = new SpringClassRule();

    @Rule
    public final SpringMethodRule smr = new SpringMethodRule();

    @Value("${sces.endpoint}")
    private String scesEndpoint;

    private LookupClient lookupClient;

    private static CitizenIdentity identity(String audience) {
    	return Builders.citizenIdentity()
    			.representing(Builders.citizen()
    					.identifiedBy("0304784567"))
    			.usedBy(Builders.serviceConsumer()
    					.named("Service Consumer Test")
    					.identifiedBy(new RemoteCredentialVault(URI.create("https://www.nspop.dk/download/attachments/190481050/NSP_Test_Service_Consumer_sds.p12"), "Test1234".toCharArray(), true)))
    			.verifiedBy(Builders.identityProvider()
    					.named("https://idp.test.nspop.dk")
    					.withAudience("https://bootstrap.sts.nspop.dk/")
    					.identifiedBy(new RemoteCredentialVault(URI.create("https://www.nspop.dk/download/attachments/190481050/NSP_Test_Identity_Provider_sds.p12"), "Test1234".toCharArray(), true)))
    			.using(Builders.securityTokenService()
    					.at(URI.create("http://test1.ekstern-test.nspop.dk:8080/sts/services/Bst2Idws")))
    			.consumedBy(Builders.serviceProvider()
    					.withAudience(audience))
    			.build();
    }
    
    private static CitizenIdentity identityForOtherPerson(String audience) {
    	return Builders.citizenIdentity()
    			.representing(Builders.citizen()
    					.identifiedBy("0304784568"))
    			.usedBy(Builders.serviceConsumer()
    					.named("Service Consumer Test")
    					.identifiedBy(new RemoteCredentialVault(URI.create("https://www.nspop.dk/download/attachments/190481050/NSP_Test_Service_Consumer_sds.p12"), "Test1234".toCharArray(), true)))
    			.verifiedBy(Builders.identityProvider()
    					.named("https://idp.test.nspop.dk")
    					.withAudience("https://bootstrap.sts.nspop.dk/")
    					.identifiedBy(new RemoteCredentialVault(URI.create("https://www.nspop.dk/download/attachments/190481050/NSP_Test_Identity_Provider_sds.p12"), "Test1234".toCharArray(), true)))
    			.using(Builders.securityTokenService()
    					.at(URI.create("http://test1.ekstern-test.nspop.dk:8080/sts/services/Bst2Idws")))
    			.consumedBy(Builders.serviceProvider()
    					.withAudience(audience))
    			.build();
    }

    @Before
    public void setup() {
        lookupClient = new LookupClient();
    }

    @Test
    public void getPersonDetailsSearchName_Success() throws JAXBException, NspServiceException {

        // Given
    	String expectedName = "Egon";
        //create request
        String birthdate = "04/03/1978"; // month/day/year

        // When
        PersonLookupResponseType reponse = lookupClient.callGetPersonDetails120(birthdate, identity("https://fsk"), scesEndpoint+"/service/StamdataPersonLookup-1.2.0");

        // Then
        assertTrue(reponse.getPersonInformationStructure().size()>0);
        assertTrue(findName(reponse.getPersonInformationStructure(), expectedName));
    }
    
    @Test
    public void getPersonDetailsSearchCpr_Success() throws JAXBException, NspServiceException {
    	
        // Given
        PersonLookupRequestType request = new PersonLookupRequestType();
        request.setCivilRegistrationNumberPersonQuery("0304784567");

        // When
        PersonLookupResponseType reponse = lookupClient.callGetPersonDetailsWithConsent110(request, identity("https://fsk"), scesEndpoint+"/service/StamdataPersonLookup-1.2.0");

        // Then
        assertTrue(reponse.getPersonInformationStructure().size()==1);
    }
    
    @Test
    public void getPersonDetailsSearchName_OtherPerson_NoResult() throws JAXBException, NspServiceException {
    	
        // Given
    	String expectedName = "Egon";
        //create request
        String birthdate = "04/03/1978"; // month/day/year

        // When
        PersonLookupResponseType reponse = lookupClient.callGetPersonDetails120(birthdate, identityForOtherPerson("https://fsk"), scesEndpoint+"/service/StamdataPersonLookup-1.2.0");

        // Then
        assertTrue(reponse.getPersonInformationStructure().size()==0);

    }

	@Rule
	public ExpectedException exceptionRule = ExpectedException.none();
    @Test
    public void getPersonDetailsSearchCpr_OtherPerson_Fail() throws JAXBException, NspServiceException {
    	
    	exceptionRule.expect(NspServiceException.class);
	    exceptionRule.expectMessage("Adgang til CPR-nummeret ikke tilladt");
    	
        // Given
        PersonLookupRequestType request = new PersonLookupRequestType();
        request.setCivilRegistrationNumberPersonQuery("0103952595");

        // When
        PersonLookupResponseType reponse = lookupClient.callGetPersonDetailsWithConsent110(request, identityForOtherPerson("https://fsk"), scesEndpoint+"/service/StamdataPersonLookup-1.2.0");

        // Then
        // exception
    }
    
    private boolean findName(List<PersonInformationStructureType> personList, String expectedGivenName) {
        for (PersonInformationStructureType elm : personList) {
            if (elm.getRegularCPRPerson().getSimpleCPRPerson().getPersonNameStructure().getPersonGivenName().equals(expectedGivenName)) {
                return true;
            }
        }
        return false;
    }
    
}
