/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata.personalinformation.v1;

import dk.nsi.stamdata.configuration.SDMContextConfiguration;
import dk.nsi.stamdata.personalinformation.IntegrationTestConfiguration;
import dk.sosi.seal.model.AuthenticationLevel;
import org.junit.Assert;
import org.junit.Before;
import org.junit.ClassRule;
import org.junit.Rule;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.openapitools.model.Age;
import org.openapitools.model.Custody;
import org.openapitools.model.Person;
import org.openapitools.model.Status;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit4.rules.SpringClassRule;
import org.springframework.test.context.junit4.rules.SpringMethodRule;

import java.time.LocalDate;
import java.time.Period;
import java.time.format.DateTimeFormatter;
import java.util.Map;

@RunWith(Parameterized.class)
@ContextConfiguration(classes = SDMContextConfiguration.class )
public class PersonalInformationTest {

    @ClassRule
    public static final SpringClassRule scr = new SpringClassRule();

    @Rule
    public final SpringMethodRule smr = new SpringMethodRule();

    @Parameterized.Parameter(0)
    public AuthenticationLevel authenticationLevel;

    @Value("${personinformation.endpoint}")
    private String personinformationEndpoint;

    private IntegrationTestConfiguration integrationTestConfiguration;

    @Parameterized.Parameters(name = "{0}")
    public static AuthenticationLevel[] data() {
        return new AuthenticationLevel[]{AuthenticationLevel.VOCES_TRUSTED_SYSTEM};
    }

    @Before
    public void setup() {
        integrationTestConfiguration = new IntegrationTestConfiguration(personinformationEndpoint);
    }

    @Test
    public void getPersonInformation() {
        // Given
        String cpr = "0304784567";

        // When
        Person reponse = integrationTestConfiguration.callPersonInformationService("v1", cpr);

        // Then
        Assert.assertEquals("0304784567", reponse.getCpr());
    }

    @Test
    public void getPersonInformationStatus() {
        // Given
        String cpr = "0304784567";

        // When
        Status reponse = integrationTestConfiguration.callPersonInformationStatusService(new Status(), "v1", cpr);

        // Then
        Assert.assertEquals("1", reponse.getStatus().toString());
    }


    @Test
    public void getPersonInformationAge() {
        // Given
        String cpr = "0304784567";

        // When
        Age reponse = integrationTestConfiguration.callPersonInformationAgeService("v1", cpr);

        // Then

        // convert cpr date to age
        DateTimeFormatter formatter = DateTimeFormatter.ofPattern("dd/MM/yyyy");
        String birthDate = "03/04/1978";
        int age = Period.between(LocalDate.parse(birthDate, formatter), LocalDate.now()).getYears();

        Assert.assertEquals(age, reponse.getAge().intValue());
    }
    @Test
    public void getPersonInformationBirthday() {
        // Given
        String cpr = "0304784567";

        // When
        Map<String,String> reponse = integrationTestConfiguration.callPersonInformationBirthdayService("v1", cpr);

        // Then
        Assert.assertEquals("1978-04-03", reponse.get("birthday"));
    }
    @Test
    public void getPersonInformationName() {
        // Given
        String cpr = "0304784567";

        // When
        Map<String,String> reponse = integrationTestConfiguration.callPersonInformationNameService("v1", cpr);

        // Then
        Assert.assertEquals("Egon", reponse.get("givenName"));
        Assert.assertEquals("", reponse.get("middleName"));
        Assert.assertEquals("Olsen", reponse.get("familyName"));
        Assert.assertEquals("Egon Olsen", reponse.get("nameForAddressing"));
    }
    @Test
    public void getPersonInformationCustody() {
        // Given
        String cpr = "1508902650";

        // When
        Custody reponse = integrationTestConfiguration.callPersonInformationCustodyService("v1", cpr);

        // Then
        Assert.assertEquals("0810503130", reponse.getWardCustody().get(0).getRelationCpr());
    }
    @Test
    public void getPersonInformationAddress() {
        // Given
        String cpr = "0304784567";

        // When
        Map<String,String> reponse = integrationTestConfiguration.callPersonInformationAddressService("v1", cpr);

        // Then
        Assert.assertEquals(3, reponse.size());
        Assert.assertEquals("Bogenhøjvej 230", reponse.get("address"));
        Assert.assertEquals("6666", reponse.get("zipCode"));
        Assert.assertEquals("u00dcberwald", reponse.get("cityName"));
    }
    @Test
    public void getPersonInformationAddressDeceased() {
        // Given
        String cpr = "1234567896"; // Cpr er sat til at have status 90, dvs. død

        // When
        Map<String,String> reponse = integrationTestConfiguration.callPersonInformationAddressService("v1", cpr);

        // Then
        Assert.assertEquals(0, reponse.size());
    }
}
