/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata.personinformation.service.v1;

import com.google.inject.Inject;
import com.trifork.stamdata.models.PersonVO;

import dk.nsi.stamdata.cpr.dao.CprDAO;

import dk.nsi.stamdata.personinformation.service.PersonApiServiceUtil;
import org.apache.log4j.Logger;
import org.openapitools.model.*;

import javax.ws.rs.*;
import javax.ws.rs.core.Response;
import java.sql.SQLException;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.stream.Collectors;

public class PersonApiServiceImpl implements PersonApiService {
    protected final CprDAO cprDAO;
    private static final Logger logger = Logger.getLogger(PersonApiServiceImpl.class);

    @Inject
    public PersonApiServiceImpl(CprDAO cprDAO) {
        this.cprDAO = cprDAO;
    }

    @Override
    public Response getPerson(@PathParam("cpr") String cpr) throws SQLException {
        if(cprDAO.findCpr(cpr, new Date()) != null) {
            Person person = new Person();
            person.setCpr(cpr);

            return Response
                    .ok()
                    .entity(person)
                    .build();
        }

        return Response
                .status(Response.Status.NOT_FOUND)
                .entity("{}")
                .header("Content-Type", "application/json")
                .build();
    }
    
    @Override
    public Response getPersonStatus(@PathParam("cpr") String cpr) throws SQLException {
    	PersonVO personVO = cprDAO.findCpr(cpr, new Date());
    	if(personVO != null) {
    		Integer personStatus = PersonApiServiceUtil.statusToInteger(personVO.getStatus());
    		if (personStatus >= 0) {
    			Status status = new Status();
    			status.setStatus(personStatus);
    			return Response
    					.ok()
    					.entity(status)
    					.build();
    		}
    	}
        return Response
                .status(Response.Status.NOT_FOUND)
                .entity("{}")
                .header("Content-Type", "application/json")
                .build();
    }

    @Override
    public Response getPersonAge(@PathParam("cpr") String cpr) throws SQLException {
    	PersonVO personVO = cprDAO.findCpr(cpr, new Date());
    	if(personVO != null && personVO.getAlder() != null && personVO.getAlder() > -1) {
    		Age age = new Age();
    		age.setAge(personVO.getAlder());
    		return Response
    				.ok()
    				.entity(age)
    				.build();
    	}
        return Response
                .status(Response.Status.NOT_FOUND)
                .entity("{}")
                .header("Content-Type", "application/json")
                .build();
    }
    
    @Override
    public Response getPersonBirthday(@PathParam("cpr") String cpr) throws SQLException {
    	PersonVO personVO = cprDAO.findCpr(cpr, new Date());
    	if(personVO != null && personVO.getFoedselsdato() != null)  {
    		Birthday birthday = new Birthday();
    		birthday.setBirthday(convertToLocalDateViaInstant(personVO.getFoedselsdato()));
    		return Response
    				.ok()
    				.entity(birthday)
    				.build();
    	}
        return Response
                .status(Response.Status.NOT_FOUND)
                .entity("{}")
                .header("Content-Type", "application/json")
                .build();
    }
    
    @Override
    public Response getPersonCustody(@PathParam("cpr") String cpr) throws SQLException {
    	List<CustodyWardCustody> custodyWardCustodyList = getAndBuildWardCustody(cpr);
    	List<CustodyParentalCustody> custodyParentalCustodyList = getAndbuildParentalCustody(cpr);
    	
    	if (!custodyWardCustodyList.isEmpty() || !custodyParentalCustodyList.isEmpty()) {
    		Custody custody = new Custody();
    		custody.setWardCustody(custodyWardCustodyList);
    		custody.setParentalCustody(custodyParentalCustodyList);
    		return Response
    				.ok()
    				.entity(custody)
    				.build();
    	}
        return Response
                .status(Response.Status.NOT_FOUND)
                .entity("{}")
                .header("Content-Type", "application/json")
                .build();
    }
    
    @Override
    public Response getPersonName(@PathParam("cpr") String cpr) throws SQLException {
    	PersonVO personVO = cprDAO.findCpr(cpr, new Date());
    	if(personVO != null && (personVO.getFornavn() != null || personVO.getMellemnavn() != null || personVO.getEfternavn() != null || personVO.getNavnTilAdressering() != null))  {
    		Name name = new Name();
    		name.setGivenName(personVO.getFornavn());
    		name.setMiddleName(personVO.getMellemnavn());
    		name.setFamilyName(personVO.getEfternavn());
    		name.setNameForAddressing(personVO.getNavnTilAdressering());
    		return Response
    				.ok()
    				.entity(name)
    				.build();
    	}
        return Response
                .status(Response.Status.NOT_FOUND)
                .entity("{}")
                .header("Content-Type", "application/json")
                .build();
    }

	@Override
	public Response getDeceased(@QueryParam("cpr") List<String> cprList, @QueryParam("dato") String dato) throws SQLException {
		LocalDateTime statusDato;
		try {
			statusDato = LocalDateTime.from(DateTimeFormatter.ISO_DATE_TIME.parse(dato));
		} catch (java.time.format.DateTimeParseException dtpe) {
			return Response
					.status(Response.Status.BAD_REQUEST)
					.entity("[]")
					.header("Content-Type", "application/json")
					.build();
		}

		List<PersonVO> personVOList = cprDAO.findDeceased(cprList, statusDato);

		List<Person> result = personVOList.stream()
				.map(personVO -> {
					Person person = new Person();
					person.setCpr(personVO.getCpr());

					return person;
				})
				.collect(Collectors.toList());

		return Response
				.ok()
				.entity(result)
				.build();
	}

	 @Override
	 public Response getPersonPostalAddress(@PathParam("cpr") String cpr) throws SQLException {
	     PersonVO personVO = cprDAO.findCpr(cpr, new Date());
	     if(isValidPersonForAddress(personVO)) {
			 Address address = new Address();
			 address.setAddress(personVO.getStandardAdresse());
			 address.setCityName(personVO.getPostdistrikt());
			 address.setZipCode(personVO.getPostnummer());
	         return Response
	                 .ok()
	                 .entity(address)
	                 .build();
	     }
	     return Response
	             .status(Response.Status.NOT_FOUND)
	             .entity("{}")
	             .header("Content-Type", "application/json")
	             .build();
	 }

	private boolean isValidPersonForAddress(PersonVO personVO) {
		List<String> validStatusCodes = Arrays.asList("01", "03", "05", "07");
		return personVO != null && personVO.getStandardAdresse() != null && personVO.getPostdistrikt() != null
				&& personVO.getPostnummer() != null && personVO.getStatus() != null
				&& validStatusCodes.contains(personVO.getStatus());
	}

	private  List<CustodyWardCustody> getAndBuildWardCustody(String cpr) throws SQLException {
    	List<CustodyWardCustody> custodyWardCustodyList = new ArrayList<>();
    	
    	List<String> umyndiggoerelseVaergeRelationList = cprDAO.findUVR(cpr, new Date());
    	for (String umyndiggoerelseVaergeRelation : umyndiggoerelseVaergeRelationList) {
    		CustodyWardCustody personInformationGuardianFor = new CustodyWardCustody();
    		personInformationGuardianFor.setRelationCpr(umyndiggoerelseVaergeRelation);
    		custodyWardCustodyList.add(personInformationGuardianFor);
    	}
    
    	return custodyWardCustodyList;

    }

    private  List<CustodyParentalCustody> getAndbuildParentalCustody(String cpr) throws SQLException {
    	List<CustodyParentalCustody> custodyParentalCustodyList = new ArrayList<>();
    	
    	List<String> cprList = cprDAO.findRelationsAsHavingAuthority (cpr, new Date());
    	for (String relationCpr : cprList) {
    		
    		CustodyParentalCustody custodyParentalCustody = new CustodyParentalCustody();
    		custodyParentalCustody.setRelationCpr(relationCpr);

    		PersonVO personVO = cprDAO.findCpr(relationCpr, new Date()); //da ingen forventning om mange børn/opslag, hentes cpr numre enkeltvis 
    		if(personVO != null && personVO.getAlder() != null && personVO.getAlder() > -1) {
    			custodyParentalCustody.setAge(personVO.getAlder());
    		}
    		custodyParentalCustodyList.add(custodyParentalCustody);
    	}
    	
    	return custodyParentalCustodyList;

    }
    
    private LocalDate convertToLocalDateViaInstant(Date dateToConvert) {
    	logger.debug("Dato er: " + ((dateToConvert != null) ? dateToConvert.toString() : null));

    	if (dateToConvert != null) {
    		java.util.Date safeDate = new Date(dateToConvert.getTime());
            return safeDate.toInstant()
                    .atZone(ZoneId.systemDefault())
                    .toLocalDate();
    	} else {
    		logger.info("No date value available");
    		return null;
    	}
    }
}
