/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata.integrationtest.service._2024._08._01;

import com.google.inject.Inject;
import com.trifork.stamdata.models.ForaeldreMyndighedRelationVO.FMRTypeKode;
import com.trifork.stamdata.models.PersonVO;
import dk.nsi.stamdata.guice.GuiceTestRunner;
import dk.nsi.stamdata.personinformation.ComponentController;
import dk.nsi.stamdata.personinformation.service._2024._08._01.PersonApiService;
import dk.nsi.stamdata.testing.DatabaseTestInitializer;
import dk.nsi.stamdata.testing.StamdataTestHelper;
import dk.nsi.stamdata.testing.UndertowRestEasyTestServer;
import dk.nsi.stamdata.testing.UndertowSlaLogParameters;
import org.jboss.resteasy.client.jaxrs.ResteasyClient;
import org.jboss.resteasy.client.jaxrs.ResteasyClientBuilder;
import org.jboss.resteasy.client.jaxrs.ResteasyWebTarget;
import org.joda.time.DateTime;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.openapitools.model.Address;
import org.openapitools.model.Age;
import org.openapitools.model.Custody;
import org.openapitools.model.Name;
import org.openapitools.model.Person;
import org.openapitools.model.StatusCode;

import javax.inject.Named;
import javax.ws.rs.core.GenericType;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.UriBuilder;
import java.io.IOException;
import java.sql.Connection;
import java.time.LocalDateTime;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.List;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;

@RunWith(GuiceTestRunner.class)
public class PersonInformationIT {
	
    private final StamdataTestHelper testHelper = new StamdataTestHelper();

    @Inject
    private Connection connection;
    private PersonApiService proxy;

    private UndertowRestEasyTestServer undertowServer;

    @Inject
    @Named("run.undertow")
    private boolean runUndertow;

    @Inject
    @Named("personinformation.endpoint")
    private String personInformationEndpoint;

    @Before
    public void setUpDB() throws Exception {
        if (runUndertow) {
            DatabaseTestInitializer dbInitializer = new DatabaseTestInitializer(connection);
            dbInitializer.initialize();
        }
        connection.setAutoCommit(false);
    }

    @Before
    public void setUp() throws Exception {
        if(runUndertow) {
            undertowServer = new UndertowRestEasyTestServer(8100, ComponentController.class, new UndertowSlaLogParameters("Stamdata Person Information service", "stamdata-personinformation"));
            undertowServer.start();
        }

        ResteasyClient client = new ResteasyClientBuilder().build();
        ResteasyWebTarget target = client.target(UriBuilder.fromPath(personInformationEndpoint));
        proxy = target.proxy(PersonApiService.class);
    }

    @After
    public void tearDown() throws IOException {
        if(runUndertow) {
            undertowServer.close();
        }
    }

    @Test
    public void requestForPerson() throws Exception {
        PersonVO testPerson = testHelper.createPerson();
        testHelper.prepareTestdata(connection);

        Response response = proxy.getPerson(testPerson.getCpr());

        assertNotNull(response);
        assertEquals(200, response.getStatus());
        Person person = response.readEntity(Person.class);
        assertEquals(testPerson.getCpr(), person.getCpr());
    }

    @Test
    public void requestForMissingPerson() throws Exception {
        Response response = proxy.getPerson("NOT_FOUND");

        assertEquals(404, response.getStatus());
        assertEquals("application/json", response.getHeaderString("content-type"));
        assertEquals("{}", response.readEntity(String.class));
    }
    
    @Test
    public void requestForPersonsStatus() throws Exception {
        PersonVO testPerson = testHelper.createPerson();
        testHelper.prepareTestdata(connection);
        String personsStatus = testPerson.getStatus();

        Response response = proxy.getPersonStatus(testPerson.getCpr());

        assertNotNull(response);
        assertEquals(200, response.getStatus());
        StatusCode statusCode = response.readEntity(StatusCode.class);
        assertEquals(personsStatus, statusCode.getStatusCode());
    }
    
    @Test
    public void requestForMissingPersonsStatus() throws Exception {
        Response response = proxy.getPersonStatus("NOT_FOUND");

        assertEquals(404, response.getStatus());
        assertEquals("application/json", response.getHeaderString("content-type"));
        assertEquals("{}", response.readEntity(String.class));
    }
    
    @Test
    public void requestForPersonsAge() throws Exception {
        Integer personAge = 20;
    	Date yesterdayMinusPersonAge = DateTime.now().minusDays(1).withHourOfDay(0).withMinuteOfHour(0).withSecondOfMinute(0).withMillisOfSecond(0).minusYears(personAge).toDate();
        PersonVO testPerson = testHelper.createPersonWithBirthday(yesterdayMinusPersonAge);
        testHelper.prepareTestdata(connection);

        Response response = proxy.getPersonAge(testPerson.getCpr());

        assertNotNull(response);
        assertEquals(200, response.getStatus());
        Age ageFound = response.readEntity(Age.class);
        assertEquals(personAge, ageFound.getAge());
    }
    
    @Test
    public void requestForMissingPersonsAge() throws Exception {
        Response response = proxy.getPersonAge("NOT_FOUND");

        assertEquals(404, response.getStatus());
        assertEquals("application/json", response.getHeaderString("content-type"));
        assertEquals("{}", response.readEntity(String.class));
    }

    @Test
    public void requestForPersonsBirthday() throws Exception {
        Date birthday = new GregorianCalendar(2002, Calendar.SEPTEMBER, 15).getTime();
        PersonVO testPerson = testHelper.createPersonWithBirthday(birthday);
        testHelper.prepareTestdata(connection);

        Response response = proxy.getPersonBirthday(testPerson.getCpr());

        assertNotNull(response);
        assertEquals(200, response.getStatus());
        assertEquals("{\"birthday\":\"2002-09-15\"}", response.readEntity(String.class));
        
    }
    
    @Test
    public void requestForMissingPersonsBirthday() throws Exception {
        Response response = proxy.getPersonBirthday("NOT_FOUND");

        assertEquals(404, response.getStatus());
        assertEquals("application/json", response.getHeaderString("content-type"));
        assertEquals("{}", response.readEntity(String.class));
    }
    
    @Test
    public void requestForPersonsCustody() throws Exception {

        PersonVO testPerson = testHelper.createPerson("person");
        PersonVO testPersonHarVaerge = testHelper.createPerson("harVaerge");
        PersonVO testPersonDatter = testHelper.createKvinde("datter");
        testHelper.createForaeldreMyndighedRelation(testPersonDatter, FMRTypeKode.AI1, testPerson);
        testHelper.createVaerge(testPersonHarVaerge, testPerson);
        testHelper.prepareTestdata(connection);

        Response response = proxy.getPersonCustody(testPerson.getCpr());

        assertNotNull(response);
        assertEquals(200, response.getStatus());
        Custody custodyFound = response.readEntity(Custody.class);
        assertNotNull(custodyFound.getParentalCustody());
        assertEquals(1, custodyFound.getParentalCustody().size());
        assertEquals(testPersonDatter.getCpr(), custodyFound.getParentalCustody().get(0).getRelationCpr());
        assertEquals(testPersonDatter.getAlder(), custodyFound.getParentalCustody().get(0).getAge());
        assertNotNull(custodyFound.getWardCustody());
        assertEquals(1, custodyFound.getWardCustody().size());
        assertEquals(testPersonHarVaerge.getCpr(), custodyFound.getWardCustody().get(0).getRelationCpr());

    }
    
    @Test
    public void requestForMissingPersonsCustody() throws Exception {
        Response response = proxy.getPersonCustody("NOT_FOUND");

        assertEquals(404, response.getStatus());
        assertEquals("application/json", response.getHeaderString("content-type"));
        assertEquals("{}", response.readEntity(String.class));
    }
    
    @Test
    public void requestForPersonsName() throws Exception {
    	
        PersonVO testPerson = testHelper.createPerson();
        testHelper.prepareTestdata(connection);
        String fornavn = testPerson.getFornavn();
        String mellemnavn = testPerson.getMellemnavn();
        String efternavn = testPerson.getEfternavn();
        String navnTilAdr = testPerson.getNavnTilAdressering();

        Response response = proxy.getPersonName(testPerson.getCpr());

        assertNotNull(response);
        assertEquals(200, response.getStatus());
        Name name = response.readEntity(Name.class);
        assertEquals(fornavn, name.getGivenName());
        assertEquals(mellemnavn, name.getMiddleName());
        assertEquals(efternavn, name.getFamilyName());
        assertEquals(navnTilAdr, name.getNameForAddressing());

    }
    
    @Test
    public void requestForMissingPersonsName() throws Exception {
        Response response = proxy.getPersonName("NOT_FOUND");

        assertEquals(404, response.getStatus());
        assertEquals("application/json", response.getHeaderString("content-type"));
        assertEquals("{}", response.readEntity(String.class));
    }

    @Test
    public void requestForDeceaced() throws Exception {
        PersonVO testPerson = testHelper.createPerson();
        testPerson.setStatus("90"); // deceased
        testHelper.prepareTestdata(connection);

        Response response = proxy.getDeceased(Arrays.asList(testPerson.getCpr()), LocalDateTime.now().toString());

        assertNotNull(response);
        assertEquals(200, response.getStatus());

        List<Person> personList = response.readEntity(new GenericType<List<Person>>() {});
        assertEquals(1, personList.size());
        assertEquals(testPerson.getCpr(), personList.get(0).getCpr());
    }

    @Test
    public void requestForPersonsAddress() throws Exception {

        PersonVO testPerson = testHelper.createPerson();
        testHelper.prepareTestdata(connection);
        String adresse = testPerson.getStandardAdresse();
        String cityName = testPerson.getPostdistrikt();
        String zipCode = testPerson.getPostnummer();

        Response response = proxy.getPersonPostalAddress(testPerson.getCpr());

        assertNotNull(response);
        assertEquals(200, response.getStatus());
        Address address = response.readEntity(Address.class);
        assertEquals(adresse, address.getAddress());
        assertEquals(cityName, address.getCityName());
        assertEquals(zipCode, address.getZipCode());

    }

    @Test
    public void requestForPersonsAddressForDeceased() throws Exception {

        PersonVO testPerson = testHelper.createPerson();
        testPerson.setStatus("90");
        testHelper.prepareTestdata(connection);
        Response response = proxy.getPersonPostalAddress(testPerson.getCpr());

        assertNotNull(response);
        assertEquals(404, response.getStatus());
        assertEquals("application/json", response.getHeaderString("content-type"));
        assertEquals("{}", response.readEntity(String.class));
    }

    @Test
    public void requestForPersonsAddressForStatus01() throws Exception {
        PersonVO testPerson = testHelper.createPerson();
        testPerson.setStatus("01");
        testHelper.prepareTestdata(connection);
        Response response01 = proxy.getPersonPostalAddress(testPerson.getCpr());

        assertNotNull(response01);
        assertEquals(200, response01.getStatus());
    }

    @Test
    public void requestForPersonsAddressForStatus03() throws Exception {
        PersonVO testPerson = testHelper.createPerson();
        testPerson.setStatus("03");
        testHelper.prepareTestdata(connection);
        Response response03 = proxy.getPersonPostalAddress(testPerson.getCpr());

        assertNotNull(response03);
        assertEquals(200, response03.getStatus());
    }

    @Test
    public void requestForPersonsAddressForStatus05() throws Exception {
        PersonVO testPerson = testHelper.createPerson();
        testPerson.setStatus("05");
        testHelper.prepareTestdata(connection);
        Response response05 = proxy.getPersonPostalAddress(testPerson.getCpr());

        assertNotNull(response05);
        assertEquals(200, response05.getStatus());
    }

    @Test
    public void requestForPersonsAddressForStatus07() throws Exception {
        PersonVO testPerson = testHelper.createPerson();
        testPerson.setStatus("07");
        testHelper.prepareTestdata(connection);
        Response response07 = proxy.getPersonPostalAddress(testPerson.getCpr());

        assertNotNull(response07);
        assertEquals(200, response07.getStatus());
    }

    @Test
    public void requestForPersonsAddressForIllegalStatus() throws Exception {
        PersonVO testPerson = testHelper.createPerson();
        testPerson.setStatus("99");
        testHelper.prepareTestdata(connection);
        Response response = proxy.getPersonPostalAddress(testPerson.getCpr());

        assertNotNull(response);
        assertEquals(404, response.getStatus());
        assertEquals("application/json", response.getHeaderString("content-type"));
        assertEquals("{}", response.readEntity(String.class));
    }

    @Test
    public void requestForMissingPersonsAddress() throws Exception {
        Response response = proxy.getPersonPostalAddress("NOT_FOUND");

        assertEquals(404, response.getStatus());
        assertEquals("application/json", response.getHeaderString("content-type"));
        assertEquals("{}", response.readEntity(String.class));
    }
}
