/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata.personinformation.service._2024._08._01;


import com.trifork.stamdata.models.PersonVO;
import dk.nsi.stamdata.cpr.dao.CprDAO;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Mockito;
import org.openapitools.model.Address;
import org.openapitools.model.Age;
import org.openapitools.model.Birthday;
import org.openapitools.model.Custody;
import org.openapitools.model.Name;
import org.openapitools.model.Person;
import org.openapitools.model.StatusCode;

import javax.ws.rs.core.Response;
import java.sql.SQLException;
import java.text.SimpleDateFormat;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.List;
import java.util.Locale;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

public class PersonApiServiceImplTest {
    private PersonApiServiceImpl personApiServiceImpl;
    private CprDAO cprDAO;

    @Before
    public void setup() {
        cprDAO = Mockito.mock(CprDAO.class);

        personApiServiceImpl = new PersonApiServiceImpl(cprDAO);
    }

    @Test
    public void testPersonDoesNotExists() throws SQLException {
        String input = "DOES_NOT_EXIST";
        Mockito.when(cprDAO.findCpr(Mockito.eq(input), Mockito.any() )).thenReturn(null);

        Response response = personApiServiceImpl.getPerson(input);
        assertNotNull(response);
        assertEquals(404, response.getStatus());
        assertEquals("application/json", response.getHeaderString("content-type"));
        assertEquals("{}", response.getEntity());
    }

    @Test
    public void testPersonExists() throws SQLException {
        String input = "DOES_EXIST";
        PersonVO person = new PersonVO();
        Mockito.when(cprDAO.findCpr(Mockito.eq(input), Mockito.any() )).thenReturn(person);

        Response response = personApiServiceImpl.getPerson(input);

        assertNotNull(response.getEntity());
        assertEquals(input, ((Person)(response.getEntity())).getCpr());
    }
    
    @Test
    public void testPersonStatusPersonExistsWithStatus() throws SQLException {
        String input = "DOES_EXIST";
        String personStatusCode = "03";
        PersonVO person = new PersonVO();
        person.setStatus("03");
        Mockito.when(cprDAO.findCpr(Mockito.eq(input), Mockito.any() )).thenReturn(person);

        Response response = personApiServiceImpl.getPersonStatus(input);

        assertNotNull(response.getEntity());
        assertEquals(personStatusCode, ((StatusCode)(response.getEntity())).getStatusCode());
    }
    
    @Test
    public void testPersonStatusPersonDoesNotExistsWithStatus() throws SQLException {
        String input = "DOES_NOT_EXIST";
        Mockito.when(cprDAO.findCpr(Mockito.eq(input), Mockito.any() )).thenReturn(null);

        Response response = personApiServiceImpl.getPersonStatus(input);
        assertNotNull(response);
        assertEquals(404, response.getStatus());
        assertEquals("application/json", response.getHeaderString("content-type"));
        assertEquals("{}", response.getEntity());
    }
    
    @Test
    public void testPersonStatisPersonExistsWithNotNumericStatus() throws SQLException {
        String input = "DOES_EXIST";
        PersonVO person = new PersonVO();
        person.setStatus("03Alfa");
        Mockito.when(cprDAO.findCpr(Mockito.eq(input), Mockito.any() )).thenReturn(person);

        Response response = personApiServiceImpl.getPersonStatus(input);

        assertNotNull(response.getEntity());
        assertEquals(404, response.getStatus());
        assertEquals("application/json", response.getHeaderString("content-type"));
        assertEquals("{}", response.getEntity());
    }
    
    @Test
    public void testPersonStatusPersonExistsWithNoStatus() throws SQLException {
        String input = "DOES_EXIST";
        PersonVO person = new PersonVO();
        person.setStatus(null);
        Mockito.when(cprDAO.findCpr(Mockito.eq(input), Mockito.any() )).thenReturn(person);

        Response response = personApiServiceImpl.getPersonStatus(input);

        assertNotNull(response.getEntity());
        assertEquals(404, response.getStatus());
        assertEquals("application/json", response.getHeaderString("content-type"));
        assertEquals("{}", response.getEntity());
    }
    
    @Test
    public void testPersonAgePersonExistsWithAge() throws SQLException {
        String input = "DOES_EXIST";
        Integer personAge = 25;
        PersonVO person = new PersonVO();
        person.setAlder(personAge);
        Mockito.when(cprDAO.findCpr(Mockito.eq(input), Mockito.any() )).thenReturn(person);

        Response response = personApiServiceImpl.getPersonAge(input);

        assertNotNull(response.getEntity());
        assertEquals(personAge, ((Age)(response.getEntity())).getAge());
    }
    
    @Test
    public void testPersonAgePersonExistsWithAgeFutureDate() throws SQLException {
        String input = "DOES_EXIST";
        Integer personAge = -5;
        PersonVO person = new PersonVO();
        person.setAlder(personAge);
        Mockito.when(cprDAO.findCpr(Mockito.eq(input), Mockito.any() )).thenReturn(person);

        Response response = personApiServiceImpl.getPersonAge(input);

        assertNotNull(response);
        assertEquals(404, response.getStatus());
        assertEquals("application/json", response.getHeaderString("content-type"));
        assertEquals("{}", response.getEntity());
    }
    
    @Test
    public void testPersonAgePersonDoesNotExistsWithAge() throws SQLException {
        String input = "DOES_NOT_EXIST";
        Mockito.when(cprDAO.findCpr(Mockito.eq(input), Mockito.any() )).thenReturn(null);

        Response response = personApiServiceImpl.getPersonAge(input);
        assertNotNull(response);
        assertEquals(404, response.getStatus());
        assertEquals("application/json", response.getHeaderString("content-type"));
        assertEquals("{}", response.getEntity());
    }

    @Test
    public void testPersonBirthDayPersonExistsWithBirthday() throws SQLException {
        String input = "DOES_EXIST";
        GregorianCalendar foedselsdato =  new GregorianCalendar(2014, Calendar.FEBRUARY, 11);
        PersonVO person = new PersonVO();
        person.setFoedselsdato(foedselsdato.getTime());
        Mockito.when(cprDAO.findCpr(Mockito.eq(input), Mockito.any() )).thenReturn(person);

        Response response = personApiServiceImpl.getPersonBirthday(input);

        assertNotNull(response.getEntity());
        Birthday birthday = (Birthday)(response.getEntity());
        
        assertEquals(foedselsdato.get(Calendar.YEAR), birthday.getBirthday().getYear());
        assertEquals(foedselsdato.get(Calendar.MONTH), birthday.getBirthday().getMonthValue()-1);
        String dayWeekTextExpected = new SimpleDateFormat("EEEE", Locale.ENGLISH).format(foedselsdato.getTime()).toUpperCase();
        assertEquals(dayWeekTextExpected, birthday.getBirthday().getDayOfWeek().toString());
    }
    
    @Test
    public void testPersonBirthdayPersonDoesNotExistsWithBirthday() throws SQLException {
        String input = "DOES_NOT_EXIST";
        Mockito.when(cprDAO.findCpr(Mockito.eq(input), Mockito.any() )).thenReturn(null);

        Response response = personApiServiceImpl.getPersonBirthday(input);
        assertNotNull(response);
        assertEquals(404, response.getStatus());
        assertEquals("application/json", response.getHeaderString("content-type"));
        assertEquals("{}", response.getEntity());
    }

    
    @Test
    public void testPersonCustodyWardCustodyExists() throws SQLException {
    	String input = "DOES_EXIST";

    	String relationCpr1 = "Rel1";
    	String relationCpr2 = "Rel2";

    	List<String> custodyWardCustodyList = new ArrayList<>();
    	custodyWardCustodyList.add(relationCpr1);
    	custodyWardCustodyList.add(relationCpr2);
    	Mockito.when(cprDAO.findUVR(Mockito.eq(input), Mockito.any() )).thenReturn(custodyWardCustodyList);

    	List<String> custodyParentalCustodyList = new ArrayList<>();
    	Mockito.when(cprDAO.findRelationsAsHavingAuthority(Mockito.eq(input), Mockito.any() )).thenReturn(custodyParentalCustodyList);


    	Response response = personApiServiceImpl.getPersonCustody(input);

    	assertNotNull(response.getEntity());
    	Custody custody = (Custody)(response.getEntity());

    	assertEquals(2, custody.getWardCustody().size());
    	assertEquals(relationCpr1, custody.getWardCustody().get(0).getRelationCpr());
    	assertEquals(relationCpr2, custody.getWardCustody().get(1).getRelationCpr());

    }
    
    @Test
    public void testPersonCustodyParentalCustodyExists() throws SQLException {
        String input = "DOES_EXIST";
    	String relationCprBarn1 = "Rel1";
    	String relationCprBarn2 = "Rel2";
    	
    	List<String> custodyWardCustodyList = new ArrayList<>();
    	Mockito.when(cprDAO.findUVR(Mockito.eq(input), Mockito.any() )).thenReturn(custodyWardCustodyList);

        List<String> custodyParentalCustodyList = new ArrayList<>();
        custodyParentalCustodyList.add(relationCprBarn1);
        custodyParentalCustodyList.add(relationCprBarn2);
		Mockito.when(cprDAO.findRelationsAsHavingAuthority(Mockito.eq(input), Mockito.any() )).thenReturn(custodyParentalCustodyList);
		
        Integer childAge = 7;
        PersonVO child1 = new PersonVO();
        child1.setCpr(relationCprBarn1);
        child1.setAlder(childAge);
        Mockito.when(cprDAO.findCpr(Mockito.eq(relationCprBarn1), Mockito.any() )).thenReturn(child1);
        PersonVO child2 = new PersonVO();
        child2.setCpr(relationCprBarn2);
        child2.setAlder(childAge);
        Mockito.when(cprDAO.findCpr(Mockito.eq(relationCprBarn2), Mockito.any() )).thenReturn(child2);
		
        Response response = personApiServiceImpl.getPersonCustody(input);

        assertNotNull(response.getEntity());
    	Custody custody = (Custody)(response.getEntity());

    	assertEquals(2, custody.getParentalCustody().size());
    	assertEquals(relationCprBarn1, custody.getParentalCustody().get(0).getRelationCpr());
    	assertEquals(childAge, custody.getParentalCustody().get(0).getAge());
    	assertEquals(relationCprBarn2, custody.getParentalCustody().get(1).getRelationCpr());
    	assertEquals(childAge, custody.getParentalCustody().get(1).getAge());

    }
    
    @Test
    public void testPersonAgeCustodyDoesNotExists() throws SQLException {
        String input = "DOES_NOT_EXIST";
        
        List<String> custodyWardCustodyList = new ArrayList<>();
		Mockito.when(cprDAO.findUVR(Mockito.eq(input), Mockito.any() )).thenReturn(custodyWardCustodyList);
		
		List<String> custodyParentalCustodyList = new ArrayList<>();
		Mockito.when(cprDAO.findRelationsAsHavingAuthority(Mockito.eq(input), Mockito.any() )).thenReturn(custodyParentalCustodyList);


        Response response = personApiServiceImpl.getPersonCustody(input);
        assertNotNull(response);
        assertEquals(404, response.getStatus());
        assertEquals("application/json", response.getHeaderString("content-type"));
        assertEquals("{}", response.getEntity());
    }

    @Test
    public void testPersonNamePersonExistsWithGivenName() throws SQLException {
        String input = "DOES_EXIST";
        String fornavn = "Fornavn";
        PersonVO person = new PersonVO();
        person.setFornavn(fornavn);
        Mockito.when(cprDAO.findCpr(Mockito.eq(input), Mockito.any() )).thenReturn(person);

        Response response = personApiServiceImpl.getPersonName(input);

        assertNotNull(response.getEntity());
        Name name = (Name)(response.getEntity());
        
        assertEquals(fornavn, name.getGivenName());
        
    }
    
    @Test
    public void testPersonNamePersonExistsWithMiddleName() throws SQLException {
        String input = "DOES_EXIST";
        String mellemnavn = "Mellemnavn";
        PersonVO person = new PersonVO();
        person.setMellemnavn(mellemnavn);
        Mockito.when(cprDAO.findCpr(Mockito.eq(input), Mockito.any() )).thenReturn(person);

        Response response = personApiServiceImpl.getPersonName(input);

        assertNotNull(response.getEntity());
        Name name = (Name)(response.getEntity());
        
        assertEquals(mellemnavn, name.getMiddleName());
        
    }
    
    @Test
    public void testPersonNamePersonExistsWithFamilyName() throws SQLException {
        String input = "DOES_EXIST";
        String efternavn = "Efternavn";
        PersonVO person = new PersonVO();
        person.setEfternavn(efternavn);
        Mockito.when(cprDAO.findCpr(Mockito.eq(input), Mockito.any() )).thenReturn(person);

        Response response = personApiServiceImpl.getPersonName(input);

        assertNotNull(response.getEntity());
        Name name = (Name)(response.getEntity());
        
        assertEquals(efternavn, name.getFamilyName());
        
    }
    
    @Test
    public void testPersonNamePersonExistsWithNameForAddressing() throws SQLException {
        String input = "DOES_EXIST";
        String navnTilAdressering = "NavnTilAdressering";
        PersonVO person = new PersonVO();
        person.setNavnTilAdressering(navnTilAdressering);
        Mockito.when(cprDAO.findCpr(Mockito.eq(input), Mockito.any() )).thenReturn(person);

        Response response = personApiServiceImpl.getPersonName(input);

        assertNotNull(response.getEntity());
        Name name = (Name)(response.getEntity());
        
        assertEquals(navnTilAdressering, name.getNameForAddressing());
        
    }
    
    @Test
    public void testPersonNamePersonDoesNotExists() throws SQLException {
        String input = "DOES_NOT_EXIST";
        Mockito.when(cprDAO.findCpr(Mockito.eq(input), Mockito.any() )).thenReturn(null);

        Response response = personApiServiceImpl.getPersonName(input);
        assertNotNull(response);
        assertEquals(404, response.getStatus());
        assertEquals("application/json", response.getHeaderString("content-type"));
        assertEquals("{}", response.getEntity());
    }
    
    @Test
    public void testPersonNamePersonDoesNotExistsWithNames() throws SQLException {
        String input = "NAME_DOES_NOT_EXIST";
        PersonVO person = new PersonVO();
        Mockito.when(cprDAO.findCpr(Mockito.eq(input), Mockito.any() )).thenReturn(person);

        Response response = personApiServiceImpl.getPersonName(input);
        assertNotNull(response);
        assertEquals(404, response.getStatus());
        assertEquals("application/json", response.getHeaderString("content-type"));
        assertEquals("{}", response.getEntity());
    }


    @Test(expected = SQLException.class)
    public void testExceptionRethrown() throws SQLException {
        String input = "SOME_ERROR";
        Mockito.when(cprDAO.findCpr(Mockito.eq(input), Mockito.any() )).thenThrow(new SQLException());

        personApiServiceImpl.getPerson(input);
    }

    @Test
    public void testFindDeceased() throws SQLException {
        PersonVO person = new PersonVO();
        person.setCpr("cpr");

        List<String> input = Arrays.asList(person.getCpr());
        Mockito.when(cprDAO.findDeceased(Mockito.eq(input), Mockito.any())).thenReturn(Arrays.asList(person));

        Response response = personApiServiceImpl.getDeceased(input, LocalDateTime.now().toString());

        assertNotNull(response);
        assertNotNull(response.getEntity());
        List<Person> result = (List<Person>) response.getEntity();

        assertEquals(input.size(), result.size());
        assertTrue(result.stream().map(Person::getCpr).allMatch(input::contains));
    }

    @Test
    public void testPersonAddressPersonExistsAndStatus01() throws SQLException {
        String input = "DOES_EXIST";
        String address = "TestAddress";
        String city = "TestCity";
        String zipCode = "TestZipCode";
        PersonVO person = new PersonVO();
        person.setStandardAdresse(address);
        person.setPostdistrikt(city);
        person.setPostnummer(zipCode);
        person.setStatus("01");
        Mockito.when(cprDAO.findCpr(Mockito.eq(input), Mockito.any() )).thenReturn(person);

        Response response = personApiServiceImpl.getPersonPostalAddress(input);

        assertNotNull(response.getEntity());
        Address responseEntity = (Address) (response.getEntity());

        assertEquals(address, responseEntity.getAddress());
        assertEquals(city, responseEntity.getCityName());
        assertEquals(zipCode, responseEntity.getZipCode());
    }

    @Test
    public void testPersonAddressPersonExistsAndStatus03() throws SQLException {
        String input = "DOES_EXIST";
        String address = "TestAddress";
        String city = "TestCity";
        String zipCode = "TestZipCode";
        PersonVO person = new PersonVO();
        person.setStandardAdresse(address);
        person.setPostdistrikt(city);
        person.setPostnummer(zipCode);
        person.setStatus("03");
        Mockito.when(cprDAO.findCpr(Mockito.eq(input), Mockito.any() )).thenReturn(person);

        Response response = personApiServiceImpl.getPersonPostalAddress(input);

        assertNotNull(response.getEntity());
        Address responseEntity = (Address) (response.getEntity());

        assertEquals(address, responseEntity.getAddress());
        assertEquals(city, responseEntity.getCityName());
        assertEquals(zipCode, responseEntity.getZipCode());
    }

    @Test
    public void testPersonAddressPersonExistsDeceased() throws SQLException {
        String input = "DOES_EXIST";
        String address = "TestAddress";
        String city = "TestCity";
        String zipCode = "TestZipCode";
        PersonVO person = new PersonVO();
        person.setStandardAdresse(address);
        person.setPostdistrikt(city);
        person.setPostnummer(zipCode);
        person.setStatus("90");
        Mockito.when(cprDAO.findCpr(Mockito.eq(input), Mockito.any() )).thenReturn(person);

        Response response = personApiServiceImpl.getPersonPostalAddress(input);
        assertNotNull(response);
        assertEquals(404, response.getStatus());
        assertEquals("application/json", response.getHeaderString("content-type"));
        assertEquals("{}", response.getEntity());
    }

    @Test
    public void requestForPersonsAddressForIllegalStatus() throws SQLException {
        String input = "DOES_EXIST";
        String address = "TestAddress";
        String city = "TestCity";
        String zipCode = "TestZipCode";
        PersonVO person = new PersonVO();
        person.setStandardAdresse(address);
        person.setPostdistrikt(city);
        person.setPostnummer(zipCode);
        person.setStatus("99");
        Mockito.when(cprDAO.findCpr(Mockito.eq(input), Mockito.any() )).thenReturn(person);

        Response response = personApiServiceImpl.getPersonPostalAddress(input);
        assertNotNull(response);
        assertEquals(404, response.getStatus());
        assertEquals("application/json", response.getHeaderString("content-type"));
        assertEquals("{}", response.getEntity());
    }

    @Test
    public void testPersonAddressPersonDoesNotExists() throws SQLException {
        String input = "DOES_NOT_EXIST";
        Mockito.when(cprDAO.findCpr(Mockito.eq(input), Mockito.any() )).thenReturn(null);

        Response response = personApiServiceImpl.getPersonPostalAddress(input);
        assertNotNull(response);
        assertEquals(404, response.getStatus());
        assertEquals("application/json", response.getHeaderString("content-type"));
        assertEquals("{}", response.getEntity());
    }
}
